
{*******************************************************}
{                                                       }
{       SAB Object Framework                            }
{       SAB Classes                                     }
{                                                       }
{       Copyright (c) 2017-23 TDP                       }
{                                                       }
{****************************************************tk*}

(*******************************************************}

  Class hierarchy:

    TObjectList<TSabObject>
      TSabList

    TObject
      TSabObject
        TSabStation
        TSabConfiguration
        TSabSafeBox
          TSabSafeBoxGroup
        TSabRecord
          TSabAccountCommon
            TSabAccount
            TSabSecAccount
          TSabFile
            TSabCertificate
            TSabExternalFile  (TSabExternalAttributes)
          TSabCreditCard

{*******************************************************)

unit SAB.Classes;

{ DEFINE EXTERNAL_FILE_HAS_PASSWORD}
{$DEFINE FILE_STREAM_EX}

interface

uses
  System.SysUtils, System.Classes, System.Contnrs,
  System.Generics.Collections, System.JSON, System.NetEncoding, Winapi.Windows,
  SAB.Types, SAB.Uri, SAB.Consts;

const
  MAX_SAB_FILENAME_LENGTH = 255;

  CertificateExtensions: array[0..13] of string = (
    '.CA-BUNDLE',
    '.CER',
    '.CERT',
    '.CRL',
    '.CRT',
    '.CSR',
    '.DER',
    '.P12',
    '.P7B',
    '.P7R',
    '.P7S',
    '.PEM',
    '.PFX',
    '.SPC'
  );

const
  EXTFILE_SIGNATURE = $13ECBDFA;  // FABDEC13
  EXTFILE_PREPOSTFIX = SizeOf(UInt32) { + SizeOf(UInt32) } + 16;

{$IFDEF FILE_STREAM_EX}
const
  fmOpenSync    = $10000;  // shr 16
  fmOpenDirect  = $20000;  // shr 17
  fmOpenNoATime = $40000;  // shr 18

  fmEncrypted  = $100000;  // shr 20
{$ENDIF}

type
  TSabObject = class;
  TSabList = class;
  SabException = System.SysUtils.Exception;

  TSabListSortCompare = function(List: TSabList; Index1, Index2: Integer): Integer;

{$IFDEF FILE_STREAM_EX}
  { TFileStreamEx }

  // FILE_FLAG_NO_BUFFERING + FILE_FLAG_WRITE_THROUGH
  // FILE_FLAG_SEQUENTIAL_SCAN + FILE_FLAG_RANDOM_ACCESS
  TFileStreamEx = class(THandleStream)
  protected
    FFileName: String;
  public
    constructor Create(const AFileName: string; Mode: DWORD);
    destructor Destroy; override;

    property FileName: String read FFileName;
  end;
{$ENDIF}

  { TSearchResults }

  TSearchResult = class(TObject)
  strict private
    FId: TSabId;
    FName: string;
    FDescription: string;
    FType: string;
    FOwnerId: Integer;
    FTemplate: string;
    FParent: string;

    FSabObjType: TSabObjectType;
  public
    constructor Create(const AId: TSabId; const AName, ADescription, AType: string; OwnerId: Integer; ATemplate, AParent: string);
    class function CreateObject(const JSON: TJSONObject): TSearchResult;

    property ObjType: TSabObjectType read FSabObjType;

    property Id: TSabId read FId;
    property Name: string read FName;
    property Description: string read FDescription;
    property ObjTypeStr: string read FType;
    property OwnerId: Integer read FOwnerId;
    property Template: string read FTemplate;
    property SafeBox: string read FParent;
  end;

  TSearchResultList = class(TObjectList)
  private
    function GetResultItem(Index: Integer): TSearchResult;
  public
    property Items[Index: Integer]: TSearchResult read GetResultItem; default;
  end;

  { TDeleted Items }
  TDeletedItem = class(TObject)
  strict private
    FId: TSabId;
    FName: string;
    FOwner: string;
    FDomain: string;
    FOwnerId: Integer;
    FType: string;
    FDescription: string;
    FWhiteBox: Boolean;
    FLogin: string;
    FFromSite: string;
    FTemplate: string;
    FParent: string;
    FDate: TDateTime;
    FDeletedParent: Boolean;
    FDeletedParentType: string;
    FDeletedParentId: TSabId;
    FDeletedParentName: string;
    FIcon: HICON;

    FSabObjType: TSabObjectType;
  public
    constructor Create(const AId: TSabId; const AName, AOwner, ADomain, AType, ADescription, ALogin, AFromSite, ATemplate, AParent: string;
      AOwnerId: Integer; ADate: TDateTime;
      ADeletedParent: Boolean; const ADeletedParentType: string; ADeletedParentId: TSabId; const ADeletedParentName: string;
      AIcon: HICON);
    class function CreateObject(const JSON: TJSONObject): TDeletedItem;

    property ObjType: TSabObjectType read FSabObjType;

    property Id: TSabId read FId;                                // id
    property Name: string read FName;                            // name
    property Owner: string read FOwner;
    property Domain: string read FDomain;
    property OwnerId: Integer read FOwnerId;
    property TypeOf: string read FType;                          // type
    property Description: string read FDescription;              // description
    property WhiteBox: Boolean read FWhiteBox;
    property Login: string read FLogin;
    property FromSite: string read FFromSite;
    property Template: string read FTemplate;
    property Parent: string read FParent;                        // SafeBox
    property DateOf: TDateTime read FDate;                       // date
    property DeletedParent: Boolean read FDeletedParent;
    property DeletedParentType: string read FDeletedParentType;
    property DeletedParentId: TSabId read FDeletedParentId;
    property DeletedParentName: string read FDeletedParentName;
    property Icon: HICON read FIcon;
  end;

  TDeletedItemList = class(TObjectList)
  private
    function GetResultItem(Index: Integer): TDeletedItem;
  public
    function FindBy(const AName: string; out AIndex: Integer): Boolean;

    property Items[Index: Integer]: TDeletedItem read GetResultItem; default;
  end;

  TSabList = class(System.Generics.Collections.TObjectList<TSabObject>)
  private
    FSorted: Boolean;
    FDuplicates: TDuplicates;

    function CompareStrings(const S1, S2: string): Integer;
    procedure QuickSort(L, R: Integer; SCompare: TSabListSortCompare);
    function GetSize: Integer;
    function GetNumberOfMarked: Integer;
  public
    constructor Create;

    procedure Clear;
    procedure SortByName;

    function FindBy(const Id: TSabId): TSabObject;
    function Find(const S: string; var Index: Integer): Boolean;
    function FindSafeBox(Id: TSabId; out Name: string): Boolean; overload;
    function FindSafeBox(const Name: string; out Id: TSabId): Boolean; overload;
    procedure Populate(const List: TStringList; const WhatObjects: TSabObjectSet = AllSabRecords; const AllObjects: Boolean = False);

    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
    property Sorted: Boolean read FSorted;

    property Size: Integer read GetSize;
    property NumberOfMarked: Integer read GetNumberOfMarked;
  end;

  TSabObjectClass = class of TSabObject;
  TSabObject = class(TObject)
  strict private
    FMark: Boolean;
    FId: TSabId;
    FModified: TDateTime;
    FModifiedUTC: TDateTime;

    FAttachedObject: TObject;
    FOwnedObject: Boolean;

    function GetSize: Integer; protected
    function GetJSONString: string;
    function GetJSONObject: TJSONObject;
  strict protected
    FName: string;

    FSecureFieldsUsed: Boolean;

    class function ParseObjects(const Strm: TCustomMemoryStream; const AObjectName: string = ''; const AClass: TSabObjectClass = nil; const List: TSabList = nil; const ClearList: Boolean = True; const HasPreamble: Boolean = False): TSABObject;

    function GetSabObjType: TSabObjectType; virtual; abstract;

  public
    constructor Create(const JSON: TJSONObject); overload; virtual;
    constructor Create(const AName: string); overload;
    destructor Destroy; override;

    class procedure AddObjects(const List: TSabList; const Strm: TCustomMemoryStream; const AObjectName: string; const ClearList: Boolean = True; const HasPreamble: Boolean = False);
    class function CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string; const HasPreamble: Boolean = False; const AClass: TSabObjectClass = nil): TSabObject; overload; virtual;
    class function CreateObject(const JSON: TJSONObject; const FreeJSON: Boolean = False): TSabObject; overload;
    class function JSToStr(const S: string): string;
    class function TextToJS(const S: string): string;

    function AssignNonEmptyFieldsFrom(const Obj: TSabObject): Boolean; virtual;
    procedure GetSecureFieldsFrom(Obj: TSabObject); virtual;
    procedure SetId(Id: TSabId);

    procedure SaveToList(const List: TStringList); virtual;
    function SaveTo(const AUpdate: Boolean = True; const AFileContent: Boolean = False): TJSONObject; virtual;

    procedure AttachObject(const Obj: TObject; AOwnedObject: Boolean = True);
    procedure DetachObject;

    procedure WipeStrings; virtual;

    property JSONStr: string read GetJSONString;
    property JSONObj: TJSONObject read GetJSONObject;

    property ObjType: TSabObjectType read GetSabObjType;
    property Id: TSabId read FId;
    property Modified: TDateTime read FModified;
    property ModifiedUTC: TDateTime read FModifiedUTC;
    property Name: string read FName write FName;

    property Mark: Boolean read FMark write FMark;
    property Size: Integer read GetSize;

    property AttachedObject: TObject read FAttachedObject;

    property SecureFieldsUsed: Boolean read FSecureFieldsUsed;
  end;

  TSabStation = class(TSabObject)
  strict private
    // FName == stationName
    FConfigName: string;  // DEFAULT, etc
    FPlatform: string;
    FUserName: string;
    FDefaultUser: string;
    FDefaultDomain: string;
    FStationIP: string;
    FTimeZone: string;
    FRegistered: TDateTime;
    FLastAccess: TDateTime;

    FDefUser: string;
    FCurrentPassword: string;
    FPswdDate: TDateTime;
    function GetDefUser: string;
  strict protected
    function GetSabObjType: TSabObjectType; override;
  public
    constructor Create(const JSON: TJSONObject); override;

    procedure SaveToList(const List: TStringList); override;
    procedure SetPassword(const APassword: string; ADate: TDateTime);

    procedure WipeStrings; override;

    property StationName: string read FName;
    property ConfigName: string read FConfigName;
    property UserName: string read FUserName;
    property OsPlatform: string read FPlatform;
    property StationIP: string read FStationIP;
    property Registered: TDateTime read FRegistered write FRegistered;
    property LastAccess: TDateTime read FLastAccess write FLastAccess;

    property CurrentPassword: string read FCurrentPassword;
    property PasswordDate: TDateTime read FPswdDate;
    property DefUser: string read GetDefUser write FDefUser;
  end;

  TSabConfiguration = class(TSabObject)
  strict private
    FSeedBase: string;
    FPattern: string;
    FPlatforms: TSabPlatforms;
    FPasswordTarget: string;
    FAdminUserName: string;
    FAdminGroupName: string;
    FStationNameMethod: string;
    FUpdateDefaultPassword: Boolean;
    FOnlineConfig: Boolean;
    FLinuxPasswordTarget: string;
    FLinuxAdminUserName: string;
    FLinuxAdminGroupName: string;
    FMacPasswordTarget: string;
    FMacAdminUserName: string;
    FMacAdminGroupName: string;
    FLdapPasswordTarget: string;
    FLdapAdminUserName: string;
    FLdapAdminGroupName: string;
    FPatternChars1: string;
    FPatternChars2: string;
    FPatternChars3: string;
    FPatternChars: string;
    FSimplePattern: string;
  strict protected
    function GetSabObjType: TSabObjectType; override;
  public
    constructor Create(const JSON: TJSONObject); override;

    procedure SaveToList(const List: TStringList); override;

    property SeedBase: string read FSeedBase;
    property Pattern: string read FPattern;      
    property Platforms: TSabPlatforms read FPlatforms;
    property PasswordTarget: string read FPasswordTarget;
    property AdminUserName: string read FAdminUserName;
    property AdminGroupName: string read FAdminGroupName;
    property StationNameMethod: string read FStationNameMethod;
    property UpdateDefaultPassword: Boolean read FUpdateDefaultPassword;
    property OnlineConfig: Boolean read FOnlineConfig;
    property LinuxPasswordTarget: string read FLinuxPasswordTarget;
    property LinuxAdminUserName: string read FLinuxAdminUserName;
    property LinuxAdminGroupName: string read FLinuxAdminGroupName;
    property MacPasswordTarget: string read FMacPasswordTarget;
    property MacAdminUserName: string read FMacAdminUserName;
    property MacAdminGroupName: string read FMacAdminGroupName;
    property LdapPasswordTarget: string read FLdapPasswordTarget;
    property LdapAdminUserName: string read FLdapAdminUserName;
    property LdapAdminGroupName: string read FLdapAdminGroupName;
    property PatternChars1: string read FPatternChars1;
    property PatternChars2: string read FPatternChars2;
    property PatternChars3: string read FPatternChars3;
    property PatternChars: string read FPatternChars;
    property SimplePattern: string read FSimplePattern;
  end;

  TSabSafeBox = class(TSabObject)
  strict private
    FPasswordPattern: string;
    FDescription: string;
    FPrivateBox: Boolean;
    FParentId: TSabId;
    FOwnerName: string;
    FExternalFilesPath: string;
    FBoxACL: string;

    FWatchedChanges: Boolean;
    FWatchedAccesses: Boolean;

    FWatchedAccessesModified: Boolean;
    FWatchedChangesModified: Boolean;

    function GetHasPermissionToCreate: Boolean;
    function GetHasPermissionToDelete: Boolean;
    function GetHasPermissionToModify: Boolean;

    procedure SetWatchedAccessses(const Value: Boolean);
    procedure SetWatchedChanges(const Value: Boolean);
    function GetWatchedModified: Boolean;
  strict protected
    function GetSabObjType: TSabObjectType; override;
  public
    constructor Create(const JSON: TJSONObject); overload; override;
    constructor Create(const AName, APasswordPattern: string; APrivateBox: Boolean; AParentId: TSabId); overload;

    function SaveTo(const AUpdate: Boolean = True; const AFileContent: Boolean = False): TJSONObject; override;
    procedure SaveToList(const List: TStringList); override;

    procedure ClearWatched(const AWatchTypes: TWatchedTypes = [wtChanges, wtAccesses]);

    procedure WipeStrings; override;

    property Description: string read FDescription write FDescription;
    property PasswordPattern: string read FPasswordPattern write FPasswordPattern;
    property PrivateBox: Boolean read FPrivateBox write FPrivateBox;
    property ParentId: TSabId read FParentId write FParentId;
    property BoxACL: string read FBoxACL;
    property ExternalFilesPath: string read FExternalFilesPath write FExternalFilesPath;

    property WatchedChanges: Boolean read FWatchedChanges write SetWatchedChanges;
    property WatchedAccesses: Boolean read FWatchedAccesses write SetWatchedAccessses;

    property WatchedAccessesModified: Boolean read FWatchedAccessesModified;
    property WatchedChangesModified: Boolean read FWatchedChangesModified;
    property WatchedModified: Boolean read GetWatchedModified;

    property OwnerName: string read FOwnerName;

    property HasPermissionToCreate: Boolean read GetHasPermissionToCreate;
    property HasPermissionToDelete: Boolean read GetHasPermissionToDelete;
    property HasPermissionToModify: Boolean read GetHasPermissionToModify;
  end;

  TSabSafeBoxGroup = class(TSabSafeBox)
  strict protected
    function GetSabObjType: TSabObjectType; override;
  end;

  TSabRecord = class(TSabObject)
  strict private
    FDescription: string;
    FTags: string;
    FLauncher: TSabLauncher;
    FNote: string;
    FSite: string;

    FSecNote: string;          // Secure Attribute
  public
    constructor Create(const JSON: TJSONObject); overload; override;
    constructor Create(const AName, ADescription, ATags, ANote: string); overload;

    function AssignNonEmptyFieldsFrom(const Obj: TSabObject): Boolean; override;

    procedure GetSecureFieldsFrom(Obj: TSabObject); override;
    function SaveTo(const AUpdate: Boolean = True; const AFileContent: Boolean = False): TJSONObject; override;
    procedure SaveToList(const List: TStringList); override;

    procedure WipeStrings; override;

    property Description: string read FDescription write FDescription;
    property Tags: string read FTags write FTags;
    property Note: string read FNote write FNote;
    property Site: string read FSite write FSite;
    property Launcher: TSabLauncher read FLauncher write FLauncher;

    property SecNote: string read FSecNote write FSecNote;
  end;

  TSabAccountCommon = class(TSabRecord)
  strict private
    FLoginSite: string;        // attributes
    FPasswordPattern: string;  // attributes
    FPassword: string;         // Secure Attribute
    FPasswordRe: string;       // Secure Attribute

    procedure SetPassword(const Value: string); protected
  public
    constructor Create(const JSON: TJSONObject); override;
    constructor Create(const AName, ADescription, ATags, ANote, APasswordPattern, APassword, ASite: string); overload;

    function AssignNonEmptyFieldsFrom(const Obj: TSabObject): Boolean; override;

    procedure GetSecureFieldsFrom(Obj: TSabObject); override;
    function SaveTo(const AUpdate: Boolean = True; const AFileContent: Boolean = False): TJSONObject; override;
    function SavePasswords: TJSONObject;
    procedure SaveToList(const List: TStringList); override;

    procedure WipeStrings; override;

    property LoginSite: string read FLoginSite write FLoginSite;
    property PasswordPattern: string read FPasswordPattern write FPasswordPattern;
    property Password: string read FPassword write SetPassword;
    property PasswordRe: string read FPasswordRe;
  end;

  TSabAccount = class(TSabAccountCommon)
  strict private
    FLoginName: string;      // attributes
    FServerAddress: string;  // attributes
  strict protected
    function GetSabObjType: TSabObjectType; override;
  public
    constructor Create(const JSON: TJSONObject); override;
    constructor Create(const AName, ADescription, ATags, ANote, APasswordPattern, APassword, ALoginName, AServerAddress, ASite: string); overload;

    class function CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string; const HasPreamble: Boolean = False; const AClass: TSabObjectClass = nil): TSabObject; overload; override;

    function AssignNonEmptyFieldsFrom(const Obj: TSabObject): Boolean; override;

    procedure GetSecureFieldsFrom(Obj: TSabObject); override;
    function SaveTo(const AUpdate: Boolean = True; const AFileContent: Boolean = False): TJSONObject; override;
    procedure SaveToList(const List: TStringList); override;

    property LoginName: string read FLoginName write FLoginName;
    property ServerAddress: string read FServerAddress write FServerAddress;
  end;

  TSabSecAccount = class(TSabAccountCommon)
  strict private
    FSecLoginName: string;      // Secret Attributes
    FSecServerAddress: string;  // Secretd Attributes
  strict protected
    function GetSabObjType: TSabObjectType; override;
  public
    constructor Create(const JSON: TJSONObject); override;
    constructor Create(const AName, ADescription, ATags, ANote, APasswordPattern, APassword, ASecLoginName, ASecServerAddress, ASite, ASecNote: string); overload;

    class function CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string; const HasPreamble: Boolean = False; const AClass: TSabObjectClass = nil): TSabObject; overload; override;

    function AssignNonEmptyFieldsFrom(const Obj: TSabObject): Boolean; override;

    procedure GetSecureFieldsFrom(Obj: TSabObject); override;
    function SaveTo(const AUpdate: Boolean = True; const AFileContent: Boolean = False): TJSONObject; override;
    procedure SaveToList(const List: TStringList); override;

    procedure WipeStrings; override;

    property SecLoginName: string read FSecLoginName write FSecLoginName;
    property SecServerAddress: string read FSecServerAddress write FSecServerAddress;
  end;

  TSabFile = class(TSabRecord)
  strict protected
    FFileName: string;           // Secure Attribute
    FFileSizeApprox: Int64;
    FStream: TBytesStream;
    FOwnStream: Boolean;

    function GetFilename: string; virtual;
    procedure SetFilename(const Value: string); virtual;

    function GetSabObjType: TSabObjectType; override;
    function GetFileSizeApprox: Int64;
  public
    constructor Create(const JSON: TJSONObject); override;
    constructor Create(const AName, ADescription, ATags, ANote, ASecNote, AFilename: string; const AStream: TBytesStream = nil; const AOwnStream: Boolean = False); overload;
    constructor Create(const AName, ADescription, ATags, ANote, ASecNote, AFilename: string; const AData: TBytes); overload;
    constructor Create(const AName, ADescription, ATags, ANote, ASecNote, AFilename: string; const AData: PByte; ADataLen: UInt32); overload;
    constructor Create(const FilePath, AName, ADescription, ATags, ANote, ASecNote, AFilename: string); overload;
    destructor Destroy; override;

    class function CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string; const HasPreamble: Boolean = False; const AClass: TSabObjectClass = nil): TSabObject; overload; override;

    procedure GetSecureFieldsFrom(Obj: TSabObject); override;

    function BuildPath(const Dir, ALocalFile: string): string;

    procedure LoadFile(const AStream: TBytesStream; const AFilename: string); overload;
    procedure LoadFile(const AFilename: string); overload;
    function SaveTo(const AUpdate: Boolean = True; const AFileContent: Boolean = False): TJSONObject; override;
    procedure SaveToList(const List: TStringList); override;
    procedure SaveToFile(const Dir: string; const ALocalFile: string = ''; const AEncrypted: Boolean = False);

    procedure SetStream(const AStream: TBytesStream; AOwnStream: Boolean = False);

    property Filename: string read GetFilename write SetFilename;
    property FileSizeApprox: Int64 read GetFileSizeApprox write FFileSizeApprox;
    property Stream: TBytesStream read FStream;
    property OwnStream: Boolean read FOwnStream;
  end;

  TSabCertificate = class(TSabFile)
  strict private
    FAlias: string;           // Secure Attribute
    FFilePassword: string;    // Secure Attribute

  strict protected
    function GetSabObjType: TSabObjectType; override;
  public
    constructor Create(const JSON: TJSONObject); override;
    constructor Create(const AName, ADescription, ATags, ANote, ASecNote, AFilename, AAlias, AFilePassword: string; const AStream: TBytesStream= nil; const AOwnStream: Boolean = False); overload;
    constructor Create(const AName, ADescription, ATags, ANote, ASecNote, AFilename, AAlias, AFilePassword: string; const AData: TBytes); overload;
    constructor Create(const AName, ADescription, ATags, ANote, ASecNote, AFilename, AAlias, AFilePassword: string; const AData: PByte; ADataLen: UInt32); overload;
    constructor Create(const FilePath, AName, ADescription, ATags, ANote, ASecNote, AFilename, AAlias, AFilePassword: string); overload;

    class function CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string; const HasPreamble: Boolean = False; const AClass: TSabObjectClass = nil): TSabObject; overload; override;

    procedure GetSecureFieldsFrom(Obj: TSabObject); override;
    function SaveTo(const AUpdate: Boolean = True; const AFileContent: Boolean = False): TJSONObject; override;
    procedure SaveToList(const List: TStringList); override;

    procedure WipeStrings; override;

    property Alias: string read FAlias write FAlias;
    property CertificateFile: string read FFilename write FFilename;
    property FilePassword: string read FFilePassword write FFilePassword;
  end;

  TSabExternalFile = class;

  TExternalFileHeader = packed record
    Sign: UInt32;
//    RecordId: UInt32;

    procedure Init;
//    procedure Init(const RecId: UInt32);
    function Valid: Boolean;
  end;

  TSabExternalFileAttributes = class(TObject)
  public
    efOrigFileName: string;
    efFilePath: string;
    efKey: string;
    efIV: string;
    efAAD: string;

    efOldFilePath: string;

    constructor Create(const Rec: TSabRecord = nil);

    procedure Assign(const Rec: TSabRecord);
    procedure Clear;
  end;

  TSabExternalFile = class(TSabFile)
  strict private
    FefOrigFileName: string;
    FefFilePath: string;
    FefKey: string;
    FefIV: string;
    FefAAD: string;
    FFileSize: Int64;

{$IFDEF EXTERNAL_FILE_HAS_PASSWORD}
    FFilePassword: string;
{$ENDIF}
  strict private
    function GetAttributes: TSabExternalFileAttributes; protected
    procedure SetEfFilePath(const Value: string);

    function GetSabObjType: TSabObjectType; override;
    function GetFilename: string; override;
    procedure SetFilename(const Value: string); override;
  public
    constructor Create(const JSON: TJSONObject); override;
    constructor Create(const AName, ADescription, ATags, ANote, AefOrigFileName, AefFilePath, AefKey, AefIV, AefAAD: string; const
      AFileSize: Int64); overload;

    class function CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string; const HasPreamble: Boolean = False; const AClass: TSabObjectClass = nil): TSabObject; overload; override;

    procedure Assign(const Attrs: TSabExternalFileAttributes);

    procedure GetSecureFieldsFrom(Obj: TSabObject); override;

    function SaveTo(const AUpdate: Boolean = True; const AFileContent: Boolean = False): TJSONObject; override;
    procedure SaveToList(const List: TStringList); override;

    property efOrigFilename: string read FefOrigFileName write FefOrigFileName;
    property efFilePath: string read FefFilePath write SetEfFilePath;
    property efKey: string read FefKey write FefKey;
    property efIV: string read FefIV write FefIV;
    property efAAD: string read FefAAD write FefAAD;
    property FileSize: Int64 read FFileSize;

    property Attributes: TSabExternalFileAttributes read GetAttributes;

{$IFDEF EXTERNAL_FILE_HAS_PASSWORD}
    property FilePassword: string read FFilePassword write FFilePassword;
{$ENDIF}
  end;

  TSabCreditCard = class(TSabRecord)
  strict private
    FNumber: string;       // Secure Attributes
    FExpiration: string;   // Secure Attributes
    FCvv: string;          // Secure Attributes
    FPin: string;          // Secure Attributes

    function GetExpirationDate(Local: Boolean): TDateTime;
    procedure SetExpirationDate(Local: Boolean; const Value: TDateTime); protected
    function GetSabObjType: TSabObjectType; override;
  public
    constructor Create(const JSON: TJSONObject); override;
    constructor Create(const AName, ADescription, ATags, ANote, ASecNote, ANumber, AExpiration, ACvv, APin: string); overload;

    class function CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string; const HasPreamble: Boolean = False; const AClass: TSabObjectClass = nil): TSabObject; overload; override;

    procedure GetSecureFieldsFrom(Obj: TSabObject); override;
    function SaveTo(const AUpdate: Boolean = True; const AFileContent: Boolean = False): TJSONObject; override;
    procedure SaveToList(const List: TStringList); override;

    procedure WipeStrings; override;

    property Number: string read FNumber write FNumber;
    property Expiration: string read FExpiration write FExpiration;
    property Cvv: string read FCvv write FCvv;
    property Pin: string read FPin write FPin;

    property ExpirationDate[Local: Boolean]: TDateTime read GetExpirationDate write SetExpirationDate;
  end;

{$IFDEF FILE_STREAM_EX}
function GetRemoteMode(const AFileName: string): DWORD;
{$ENDIF}

implementation

uses
  System.Types, System.DateUtils,
  System.RTLConsts,
{$IFDEF TIME_ZONES_CONSIDERED}
  TZDB,  // see github.com/pavkam/tzdb
{$ENDIF}
  TDP.CommonCode,
  SAB.Utils, System.StrUtils;

function StringListCompareStrings(List: TSabList; Index1, Index2: Integer): Integer;
begin
  Result := List.CompareStrings(List[Index1].Name, List[Index2].Name);
end;

{$IFDEF FILE_STREAM_EX}
function GetRemoteMode(const AFileName: string): DWORD;
var
  DrvType: Cardinal;
begin
  Result := 0;
  DrvType := WinApi.Windows.GetDriveType(PChar(System.SysUtils.ExtractFileDrive(AFileName)));
  if DrvType = DRIVE_REMOTE then
    Result := SAB.Classes.fmOpenSync + SAB.Classes.fmOpenDirect;
end;

const
  AccessModes: array[0..2] of DWORD = (
    GENERIC_READ,
    GENERIC_WRITE,
    GENERIC_READ or GENERIC_WRITE);

  ShareModes: array[0..4] of DWORD = (
    0,
    0,
    FILE_SHARE_READ,
    FILE_SHARE_WRITE,
    FILE_SHARE_READ or FILE_SHARE_WRITE);

  OpenFlags: array[0..3] of DWORD  = (
    0,
    FILE_FLAG_WRITE_THROUGH,                             // fmOpenSync
    FILE_FLAG_NO_BUFFERING,                              // fmOpenDirect
    FILE_FLAG_WRITE_THROUGH or FILE_FLAG_NO_BUFFERING);  // fmOpenSync + fmOpenDirect

  EncryptAttrs: array[0..1] of DWORD = (
    0,
    FILE_ATTRIBUTE_ENCRYPTED
  );

function mbFileCreate(const FileName: String; Mode: DWORD): System.THandle;
begin
//  Result:= CreateFileW(PChar(UTF16LongName(FileName)), GENERIC_READ or GENERIC_WRITE,
  Result := CreateFile(PChar(FileName),
                       GENERIC_READ or GENERIC_WRITE,
                       ShareModes[(Mode and $F0) shr 4], nil, CREATE_ALWAYS,
                       FILE_ATTRIBUTE_NORMAL + OpenFlags[(Mode shr 16) and 3] + EncryptAttrs[(Mode shr 20) and 1], 0);
end;

function mbFileOpen(const FileName: String; Mode: DWORD): System.THandle;
const
  ft: TFileTime = ( dwLowDateTime: $FFFFFFFF; dwHighDateTime: $FFFFFFFF; );
begin
//  Result:= CreateFileW(PWideChar(UTF16LongName(FileName)),
  Result := CreateFile(PChar(FileName),
                       AccessModes[Mode and 3] or ((Mode and fmOpenNoATime) shr 10),
                       ShareModes[(Mode and $F0) shr 4], nil, OPEN_EXISTING,
                       FILE_ATTRIBUTE_NORMAL + OpenFlags[(Mode shr 16) and 3], 0);
  if (Mode and fmOpenNoATime <> 0) then
  begin
    if (Result <> INVALID_HANDLE_VALUE) then
      SetFileTime(Result, nil, @ft, @ft)
    else if GetLastError = ERROR_ACCESS_DENIED then
      Result := mbFileOpen(FileName, Mode and not fmOpenNoATime);
  end;
end;

{ TFileStreamEx }

constructor TFileStreamEx.Create(const AFileName: string; Mode: DWORD);
var
  AHandle: System.THandle;
begin
  if (Mode and fmCreate) = fmCreate then
  begin
    AHandle := mbFileCreate(AFileName, Mode);
    if AHandle = INVALID_HANDLE_VALUE then
      raise EFCreateError.CreateFmt(SFCreateError, [AFileName])
    else
      inherited Create(AHandle);
  end else
  begin
    AHandle := mbFileOpen(AFileName, Mode);
    if AHandle = INVALID_HANDLE_VALUE then
      raise EFOpenError.CreateFmt(SFOpenError, [AFilename])
    else
      inherited Create(AHandle);
  end;
  FFileName := AFileName;
end;

destructor TFileStreamEx.Destroy;
begin
  if Handle <> INVALID_HANDLE_VALUE then
    CloseHandle(THandle(Handle));
  inherited;
end;
{$ENDIF}

{ TSearchResult }

constructor TSearchResult.Create(const AId: TSabId; const AName, ADescription,
  AType: string; OwnerId: Integer; ATemplate, AParent: string);
var
  T: TSabObjectType;
begin
  inherited Create;

  FId := AId;
  FName := AName;
  FDescription := ADescription;
  FType := AType;
  FOwnerId := OwnerId;
  FTemplate := ATemplate;
  FParent := AParent;

  FSabObjType := sabConfiguration;
  for T := Low(SabTypeValues) to High(SabTypeValues) do
    if AnsiSameText(AType, SabTypeValues[T]) then
    begin
      FSabObjType := T;
      Break;
    end;
end;

class function TSearchResult.CreateObject(const JSON: TJSONObject): TSearchResult;
var
  JPair: TJSONPair;
  STemplate, SName, SType, SParent: string;
  SId: TSabId;
  SOwnerId: Integer;
begin
  STemplate := '';
  SName := '';
  SType := '';
  SParent := '';
  SId := 1;
  SOwnerId := -1;

  JPair := JSON.Get(SSabObjId);
  if Assigned(JPair) then
{$IFDEF LONG_ID}
    SId := StrToInt64Def(JPair.JsonValue.Value, 0);
{$ELSE}
    SId := StrToIntDef(JPair.JsonValue.Value, 0);
{$ENDIF}

  JPair := JSON.Get(SSabObjName);
  if Assigned(JPair) then
    SName := TSabObject.JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabSafeBoxTypeFld);
  if Assigned(JPair) then
    SType := TSabObject.JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSearchOwnerId);
  if Assigned(JPair) then
    SOwnerId := StrToIntDef(JPair.JsonValue.Value, 0);

  JPair := JSON.Get(SSabRecordTemplateFld);
  if Assigned(JPair) then
    STemplate := TSabObject.JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSearchParent);
  if Assigned(JPair) then
    SParent := TSabObject.JSToStr(JPair.JsonValue.Value);

  Result := TSearchResult.Create(SId, SName, '', SType, SOwnerId, STemplate, SParent);
end;

{ TSearchList }

function TSearchResultList.GetResultItem(Index: Integer): TSearchResult;
begin
  Result := TSearchResult(inherited Items[Index]);
end;

{ TDeletedItem }

constructor TDeletedItem.Create(const AId: TSabId; const AName, AOwner,
  ADomain, AType, ADescription, ALogin, AFromSite, ATemplate, AParent: string;
  AOwnerId: Integer; ADate: TDateTime; ADeletedParent: Boolean;
  const ADeletedParentType: string; ADeletedParentId: TSabId;
  const ADeletedParentName: string; AIcon: HICON);
var
  T: TSabObjectType;
begin
  inherited Create;

  FId := AId;
  FName := AName;
  FOwner := AOwner;
  FDomain := ADomain;
  FType := AType;
  FDescription := ADescription;
  FLogin := ALogin;
  FFromSite := AFromSite;
  FTemplate := ATemplate;
  FParent := AParent;
  FOwnerId := AOwnerId;
  FDate := ADate;
  FDeletedParent := ADeletedParent;
  FDeletedParentType := ADeletedParentType;
  FDeletedParentId := ADeletedParentId;
  FDeletedParentName := ADeletedParentName;
  FIcon := AIcon;

  FSabObjType := sabConfiguration;
  for T := Low(SabTypeValues) to High(SabTypeValues) do
    if System.SysUtils.SameText(FType, SabTypeValues[T]) then
    begin
      FSabObjType := T;
      Break;
    end;

  if FSabObjType = sabAccount then
  begin
    for T := Low(SabTypeValues) to High(SabTypeValues) do
      if System.SysUtils.SameText(FTemplate, SabTypeValues[T]) then
      begin
        FSabObjType := T;
        Break;
      end;
  end;
end;

class function TDeletedItem.CreateObject(const JSON: TJSONObject): TDeletedItem;
var
  JPair: TJSONPair;
  STemplate, SName, SDomain, SType, SParent: string;
  SDescription, SLogin, SOwner, SFromSite, SDeletedParentType, SDeletedParentName: string;
  SId: TSabId;
  SOwnerId: Integer;
  SIcon: HIcon;
  SDeletedParentId: TSabId;
  SWhiteBox, SDeletedParent: Boolean;
  SDate: TDateTime;
  UMSec: Int64;
begin
  SId := -1;
  SName := '';
  SOwner := '';
  SOwnerId := -1;
  SDomain := '';
  SType := '';
  SIcon := 0;
  SDescription := '';
  SWhiteBox := False;
  SLogin := '';
  SFromSite := '';
  STemplate := '';
  SParent := '';
  SDate := 0;
  SDeletedParent := False;
  SDeletedParentType := '';
  SDeletedParentId := -1;
  SDeletedParentName := '';

  JPair := JSON.Get(SSabObjId);
  if Assigned(JPair) then
{$IFDEF LONG_ID}
    SId := StrToInt64Def(JPair.JsonValue.Value, 0);
{$ELSE}
    SId := StrToIntDef(JPair.JsonValue.Value, 0);
{$ENDIF}

  JPair := JSON.Get(SSabObjName);
  if Assigned(JPair) then
    SName := TSabObject.JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabDeletedOwner);
  if Assigned(JPair) then
    SOwner := TSabObject.JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SDomain);
  if Assigned(JPair) then
    SDomain := TSabObject.JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabOwnerId);
  if Assigned(JPair) then
    SOwnerId := StrToIntDef(JPair.JsonValue.Value, 0);

  JPair := JSON.Get(SSabSafeBoxTypeFld);
  if Assigned(JPair) then
    SType := TSabObject.JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabIcon);
  if Assigned(JPair) then
    SIcon := StrToIntDef(JPair.JsonValue.Value, 0);

  JPair := JSON.Get(SSabRecordDescription);
  if Assigned(JPair) then
    SDescription := TSabObject.JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabWhiteBox);
  if Assigned(JPair) then
    if not TryStrToBool(JPair.JsonValue.Value, SWhiteBox) then
      SWhiteBox := False;

  JPair := JSON.Get(SSabLogin);
  if Assigned(JPair) then
    SLogin := TSabObject.JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabFromSite);
  if Assigned(JPair) then
    SFromSite := TSabObject.JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabRecordTemplateFld);
  if Assigned(JPair) then
    STemplate := TSabObject.JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSearchParent);
  if Assigned(JPair) then
    SParent := TSabObject.JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabDate);
  if Assigned(JPair) then
  begin
    UMSec := StrToInt64Def(JPair.JsonValue.Value, 0);
    SDate := System.DateUtils.UnixToDateTime(Round(UMSec/1000));
  end;

  JPair := JSON.Get(SSabDeletedParent);
  if Assigned(JPair) then
    if not TryStrToBool(JPair.JsonValue.Value, SDeletedParent) then
      SDeletedParent:= False;

  JPair := JSON.Get(SDeletedParentType);
  if Assigned(JPair) then
    SDeletedParentType := TSabObject.JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabDeletedParentId);
  if Assigned(JPair) then
{$IFDEF LONG_ID}
    SDeletedParentId := StrToInt64Def(JPair.JsonValue.Value, 0);
{$ELSE}
    SDeletedParentId := StrToIntDef(JPair.JsonValue.Value, 0);
{$ENDIF}

  JPair := JSON.Get(SSabDeletedParentName);
  if Assigned(JPair) then
    SDeletedParentName := TSabObject.JSToStr(JPair.JsonValue.Value);

  Result := TDeletedItem.Create(SId, SName, SOwner, SDomain, SType, SDescription, SLogin, SFromSite, STemplate, SParent,
    SOwnerId, SDate, SDeletedParent, SDeletedParentType, SDeletedParentId, SDeletedParentName, SIcon);
end;

{ TDeletedItemList }

function TDeletedItemList.FindBy(const AName: string; out AIndex: Integer): Boolean;
var
  I: Integer;
begin
  Result := True;
  for I := 0 to Count - 1 do
    if System.StrUtils.StartsText(Items[I].Name, AName) then
    begin
      AIndex := I;
      Exit;
    end;
  Result := False;
end;

function TDeletedItemList.GetResultItem(Index: Integer): TDeletedItem;
begin
  Result := TDeletedItem(inherited Items[Index]);
end;

{ TSabList }

constructor TSabList.Create;
begin
  inherited Create(True);

  FDuplicates := dupError;
end;

procedure TSabList.Clear;
begin
  inherited Clear;
  FSorted := False;
end;

function TSabList.CompareStrings(const S1, S2: string): Integer;
begin
  Result := AnsiCompareText(S1, S2);
end;

function TSabList.Find(const S: string; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;

  if not FSorted then
    Exit;

  L := 0;
  H := Count - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := CompareStrings(Self[I].Name, S);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        if Duplicates <> dupAccept then L := I;
      end;
    end;
  end;
  Index := L;
end;

function TSabList.FindBy(const Id: TSabId): TSabObject;
begin
  for Result in Self do
  begin
    if Result.Id = Id then
      Exit;
  end;
  Result := nil;
end;

function TSabList.FindSafeBox(const Name: string; out Id: TSabId): Boolean;
var
  Obj: TSabObject;
begin
  Result := False;
  for Obj in Self do
  begin
    if Obj.ObjType in [sabSafeBoxGroup, sabSafeBox] then
    begin
      if ContainsText(Name, TSabSafeBox(Obj).Name) then
      begin
        Id := Obj.Id;
        Exit(True);
      end;
    end;
  end;
end;

function TSabList.GetNumberOfMarked: Integer;
var
  Obj: TSabObject;
begin
  Result := 0;
  for Obj inSelf do
    if Obj.Mark then
      Result := Result + 1;
end;

function TSabList.GetSize: Integer;
var
  Obj: TSabObject;
begin
  Result := 0;
  for Obj inSelf do
    Result := Result + Obj.Size;
end;

function TSabList.FindSafeBox(Id: TSabId; out Name: string): Boolean;
var
  Obj: TSabObject;
begin
  Result := False;
  Name := '';
  for Obj in Self do
  begin
    if Obj.ObjType in [sabSafeBoxGroup, sabSafeBox] then
    begin
      if TSabSafeBox(Obj).Id = Id then
      begin
        Name := Obj.Name;
        Exit(True);
      end;
    end;
  end;
end;

procedure TSabList.Populate(const List: TStringList; const WhatObjects: TSabObjectSet; const AllObjects: Boolean);
var
  Obj: TSabObject;
  Box: TSabSafeBox;

  S, Name: string;
  I: Integer;
  Id: TSabId;
  ObjType: TSabObjectType;
begin
  List.Clear;
  List.CaseSensitive := False;

  for Obj in Self do
  begin
    Name := Obj.Name;
    Id := Obj.Id;
    ObjType := Obj.ObjType;

    if (AllSabBoxes * WhatObjects <> []) and (ObjType in AllSabBoxes) then
    begin
      Box := TSabSafeBox(Obj);

      if FindSafeBox(Box.ParentId, S) then
        Name := Name + ' / ' + S + '';

      if Box.OwnerName <> '' then
        Name := Name + ' (' + Box.OwnerName + ')';

      if Box.HasPermissionToCreate or AllObjects then
      begin
        case Box.ObjType of
          sabSafeBoxGroup:
            if AllSabBoxes * WhatObjects <> [sabSafeBoxGroup] then // not OnlyGroups
              Name := '';
          sabSafeBox:
            if AllSabBoxes * WhatObjects = [sabSafeBoxGroup] then // OnlyGroups
              Name := '';
        else
          Name := '';
        end
      end else
        Name := '';

      if Name <> '' then
        if List.Find(Name, I) then
          if Box.PrivateBox then
            Name := Name + ' - private'
          else
            Name := Name + ' - shared';
    end;

    if (Name <> '') and (ObjType in WhatObjects) then
      if List.OwnsObjects then
        List.Add(Name)
      else
        List.AddObject(Name, TObject(Id));
  end;
end;

procedure TSabList.QuickSort(L, R: Integer; SCompare: TSabListSortCompare);
var
  I, J, P: Integer;
begin
  repeat
    I := L;
    J := R;
    P := (L + R) shr 1;
    repeat
      while SCompare(Self, I, P) < 0 do Inc(I);
      while SCompare(Self, J, P) > 0 do Dec(J);
      if I <= J then
      begin
        if I <> J then
          Exchange(I, J);
        if P = I then
          P := J
        else if P = J then
          P := I;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J, SCompare);
    L := I;
  until I >= R;
end;

procedure TSabList.SortByName;
begin
  if not Sorted and (Count > 1) then
  begin
    QuickSort(0, Count - 1, StringListCompareStrings);
    FSorted := True;
  end;
end;

{ TSabObject }

constructor TSabObject.Create(const JSON: TJSONObject);
var
  JPair: TJSONPair;
  UMSec: Int64;
begin
  inherited Create;

  JPair := JSON.Get(SSabObjId);
  if Assigned(JPair) then
{$IFDEF LONG_ID}
    FId := StrToInt64Def(JPair.JsonValue.Value, 0);
{$ELSE}
    FId := StrToIntDef(JPair.JsonValue.Value, 0);
{$ENDIF}

  JPair := JSON.Get(SSabObjName);
  if Assigned(JPair) then
    FName := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabObjModified);
  if Assigned(JPair) then
  begin
    UMSec := StrToInt64Def(JPair.JsonValue.Value, 0);
    FModified := System.DateUtils.UnixToDateTime(Round(UMSec/1000), False);
    FModifiedUTC := System.DateUtils.UnixToDateTime(Round(UMSec/1000));
  end;
end;

constructor TSabObject.Create(const AName: string);
begin
  inherited Create;
  FName := AName;
end;

class procedure TSabObject.AddObjects(const List: TSabList; const Strm: TCustomMemoryStream; const AObjectName: string; const ClearList, HasPreamble: Boolean);
begin
  ParseObjects(Strm, AObjectName, nil, List, ClearList, HasPreamble);
end;

class function TSabObject.CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string; const HasPreamble: Boolean; const AClass: TSabObjectClass): TSabObject;
begin
  Result := ParseObjects(Strm, AObjectName, AClass, nil, False, HasPreamble);
end;

class function TSabObject.CreateObject(const JSON: TJSONObject; const FreeJSON: Boolean): TSabObject;
var
  JPair: TJSONPair;
  Value: string;
  SabClass: TSabObjectClass;
begin
  JPair := JSON.Get(SSabRecordTemplateFld);
  if Assigned(JPair) then
  begin
    Value := JSToStr(JPair.JsonValue.Value);
  end else
    Value := '';

  if Value = '' then
  begin
    JPair := JSON.Get(SSabSafeBoxTypeFld);
    if Assigned(JPair) then
    begin
      Value := JSToStr(JPair.JsonValue.Value);
    end else
      raise SabException.Create('SAB object template/type field not found');
  end;

  if Value = '' then
    raise SabException.Create('SAB object type/template is empty');

  SabClass := GetSabObjectClass(Value);
  Result := SabClass.Create(JSON);
  if FreeJSON then
    JSON.Free;
end;

destructor TSabObject.Destroy;
begin
  DetachObject;
  WipeStrings;
  inherited;
end;

function TSabObject.AssignNonEmptyFieldsFrom(const Obj: TSabObject): Boolean;
begin
  Result := False;
  if ObjType <> Obj.ObjType then
    Exit;

//  NEVER modify Id, NEVER modify Name !!!

  FMark := Obj.Mark;
  FModified := Obj.Modified;
  FModifiedUTC := Obj.ModifiedUTC;

  Result := True;
end;

procedure TSabObject.AttachObject(const Obj: TObject; AOwnedObject: Boolean);
begin
  DetachObject;
  FAttachedObject := Obj;
  FOwnedObject := AOwnedObject;
end;

class function TSabObject.JSToStr(const S: string): string;
const
  NULLString: string = 'null';
begin
  if S = NULLString then
    Result := ''
  else
    Result := StringReplace(S, LineFeed, SLineBreak, [rfReplaceAll, rfIgnoreCase]);
end;

procedure TSabObject.DetachObject;
begin
  if Assigned(FAttachedObject) and FOwnedObject then
    FreeAndNil(FAttachedObject);
end;

function TSabObject.GetJSONObject: TJSONObject;
begin
  Result := SaveTo(False);
end;

function TSabObject.GetJSONString: string;
var
  JSON: TJSONObject;
begin
  JSON := Self.JSONObj;
  try
    Result := JSON.ToString;
  finally
    JSON.Free;
  end;
end;

procedure TSabObject.GetSecureFieldsFrom(Obj: TSabObject);
begin
  Assert(Assigned(Obj), 'Obj = nil');

  FSecureFieldsUsed := True;
  if Obj.Id > 0 then
    FId := Obj.Id;
  if Obj.Name <> '' then
    FName := Obj.Name;
end;

function TSabObject.GetSize: Integer;
begin
  Result := Length(SaveTo.ToString);
end;

class function TSabObject.ParseObjects(const Strm: TCustomMemoryStream; const AObjectName: string; const AClass: TSabObjectClass; const
  List: TSabList; const ClearList, HasPreamble: Boolean): TSABObject;
var
  Stream: TBytesStream;
  JMain: TJSONObject;
  JPair: TJSONPair;

  procedure DoClearList(const AList: TSabList; AClear: Boolean);
  begin
    if Assigned(AList) and AClear then
      AList.Clear;
  end;

  function GetSabObjectList(const JArray: TJSONArray; const AList: TSabList): TSabObject;  // get safeboxes, records
  var
    I: Integer;
    JValue: TJSONValue;
    Obj: TSabObject;
  begin
    Result := nil;
    for I := 0 to JArray.Count - 1 do
    begin
      JValue := JArray.Items[I];
      if Assigned(JValue) and (JValue is TJSONObject) then
      begin
        Obj := TSabObject.CreateObject(TJSONObject(JValue));

        if Assigned(List) then
          List.Add(Obj)
        else
          Exit(Obj);
      end;
    end;
  end;

  function GetSpecialSabObjectList(const ASpecialClass: TSabObjectClass; const JArray: TJSONArray; const AList: TSabList): TSabObject;  // get stations, configs
  var
    I: Integer;
    JValue: TJSONValue;
    Obj: TSabObject;
  begin
    Result := nil;
    for I := 0 to JArray.Count - 1 do
    begin
      JValue := JArray.Items[I];
      if Assigned(JValue) and (JValue is TJSONObject) then
      begin
        Obj := ASpecialClass.Create(TJSONObject(JValue));

        if Assigned(List) then
          List.Add(Obj)
        else
          Exit(Obj);
      end;
    end;
  end;

  function GetSingleSabObject(const JObject: TJSONObject; const AList: TSabList): TSabObject;  // get single safebox, record
  var
    Obj: TSabObject;
  begin
    Result := nil;

    Obj := TSabObject.CreateObject(JObject);

    if Assigned(List) then
      List.Add(Obj)
    else
      Result := Obj;
  end;

  function GetPassword(const AClass: TSabObjectClass; JPair: TJSONPair): TSabObject;
  var
    Val: string;
  begin
    if Assigned(AClass) then
    begin
      Val := JSToStr(JPair.JsonValue.Value);
      if AClass = TSabAccount then
        Result := TSABAccount.Create('', '', '', '', '', Val, '', '', '')
      else if AClass = TSabSecAccount then
        Result := TSABSecAccount.Create('', '', '', '', '', Val, '', '', '', '')
      else if AClass = TSabFile then
        Result := TSABFile.Create('', '', '', '', '', Val)
      else if AClass = TSabCreditCard then
        Result := TSABCreditCard.Create('', '', '', '', '', Val, '', '', '')
      else if AClass = TSabCertificate then
        Result := TSABCertificate.Create('', '', '', '', '', Val, '')
      else
        Result := nil;
    end else
      Result := nil;
  end;

begin
  if not HasPreamble then
    Stream := TBytesStream.Create(TEncoding.UTF8.GetPreamble)
  else
    Stream := TBytesStream.Create;

  JMain := TJSONObject.Create;
  try
    Strm.Position := 0;
    Stream.Seek(0, soEnd);
    Stream.CopyFrom(Strm, Strm.Size);

    JMain.Parse(Stream.Bytes, 0);

    if AObjectName <> '' then
    begin
      JPair := JMain.Get(AObjectName);
      if Assigned(JPair) then
      begin
        if SameText(AObjectName, SSabSafeBoxes_Sign) and (JPair.JsonValue is TJSONArray) then   // safeboxes
        begin
          DoClearList(List, ClearList);
          Result := GetSabObjectList(TJSONArray(JPair.JsonValue), List);
          Exit;
        end else
        if SameText(AObjectName, SSabSafeBox_Sign) and (JPair.JsonValue is TJSONObject) then    // single safebox
        begin
          DoClearList(List, ClearList);
          Result := GetSingleSabObject(TJSONObject(JPair.JsonValue), List);
          Exit;
        end else
        if SameText(AObjectName, SSabRecords_Sign) and (JPair.JsonValue is TJSONArray) then     // records
        begin
          DoClearList(List, ClearList);
          Result := GetSabObjectList(TJSONArray(JPair.JsonValue), List);
          Exit;
        end else
        if SameText(AObjectName, SSabRecord_Sign) and (JPair.JsonValue is TJSONArray) then       // single record
        begin
          DoClearList(List, ClearList);
          Result := GetSingleSabObject(TJSONObject(JPair.JsonValue), List);
          Exit;
        end else
        if SameText(AObjectName, SSabStations_Sign) and (JPair.JsonValue is TJSONArray) then     // station
        begin
          DoClearList(List, ClearList);
          Result := GetSpecialSabObjectList(TSabStation, TJSONArray(JPair.JsonValue), List);
          Exit;
        end else
        if SameText(AObjectName, SSabConfigs_Sign) and (JPair.JsonValue is TJSONArray) then     // config
        begin
          DoClearList(List, ClearList);
          Result := GetSpecialSabObjectList(TSabConfiguration, TJSONArray(JPair.JsonValue), List);
          Exit;
        end else
        if SameText(AObjectName, SSabAccountPassword) and ((JPair.JsonValue is TJSONString) or (JPair.JsonValue is TJSONNull)) then     // password field
        begin
          Result := GetPassword(AClass, JPair);
          Exit;
        end;
      end;
    end;

    JPair := JMain.Get(SSabSafeBoxes_Sign);
    if Assigned(JPair) and (JPair.JsonValue is TJSONArray) then
    begin
      DoClearList(List, ClearList);
      Result := GetSabObjectList(TJSONArray(JPair.JsonValue), List);
      Exit;
    end else
    begin
      JPair := JMain.Get(SSabSafeBox_Sign);
      if Assigned(JPair) and (JPair.JsonValue is TJSONObject) then
      begin
        DoClearList(List, ClearList);
        Result := GetSingleSabObject(TJSONObject(JPair.JsonValue), List);
        Exit;
      end else
      begin
        JPair := JMain.Get(SSabRecords_Sign);
        if Assigned(JPair) and (JPair.JsonValue is TJSONArray) then
        begin
          DoClearList(List, ClearList);
          Result := GetSabObjectList(TJSONArray(JPair.JsonValue), List);
          Exit;
        end else
        begin
          JPair := JMain.Get(SSabRecord_Sign);  // record
          if not Assigned(JPair) then
            JPair := JMain.Get(SSabDefaults_Sign);  // defaults GET RECORD FORM

          if Assigned(JPair) and (JPair.JsonValue is TJSONObject) then
          begin
            DoClearList(List, ClearList);
            Result := GetSingleSabObject(TJSONObject(JPair.JsonValue), List);
            Exit;
          end else
          begin
            JPair := JMain.Get(SSabStations_Sign);
            if Assigned(JPair) and (JPair.JsonValue is TJSONArray) then
            begin
              DoClearList(List, ClearList);
              Result := GetSpecialSabObjectList(TSabStation, TJSONArray(JPair.JsonValue), List);
              Exit;
            end else
            begin
              JPair := JMain.Get(SSabConfigs_Sign);
              if Assigned(JPair) and (JPair.JsonValue is TJSONArray) then
              begin
                DoClearList(List, ClearList);
                Result := GetSpecialSabObjectList(TSabConfiguration, TJSONArray(JPair.JsonValue), List);
                Exit;
              end else
              begin
                JPair := JMain.Get(SSabAccountPassword);
                if Assigned(JPair) and ((JPair.JsonValue is TJSONString) or (JPair.JsonValue is TJSONNull)) then
                begin
                  Result := GetPassword(AClass, JPair);
                end else
                  raise SabException.Create('SAB object not found');
              end;
            end;
          end;
        end;
      end;
    end;
  finally
    JMain.Free;
    Stream.Free;
  end;
end;

function TSabObject.SaveTo(const AUpdate, AFileContent: Boolean): TJSONObject;
begin
  Result := TJSONObject.Create;

  Result.AddPair(TJSONPair.Create(SSabObjId, TJSONNumber.Create(FId)));

  if not (ObjType in [sabSafeBox, sabSafeBoxGroup]) then
    Result.AddPair(TJSONPair.Create(SSabRecordTemplateFld, TJSONString.Create(SabTypeValues[ObjType])));

  Result.AddPair(TJSONPair.Create(SSabObjName, TJSONString.Create(FName)));
end;

procedure TSabObject.SaveToList(const List: TStringList);
begin
  Assert(Assigned(List), 'List = nil');

  List.Clear;
  List.Add(SSabObjId + '= ' + IntToStr(FId));
  List.Add(SSabObjName + ' = ' + FName);
end;

procedure TSabObject.SetId(Id: TSabId);
begin
  FId := Id;
end;

class function TSabObject.TextToJS(const S: string): string;
begin
  Result := StringReplace(S, SLineBreak, LineFeed, [rfReplaceAll, rfIgnoreCase]);
end;

procedure TSabObject.WipeStrings;
begin
end;

{ TSabStation }

constructor TSabStation.Create(const JSON: TJSONObject);
var
  JPair: TJSONPair;
  DT: Int64;
{$IFDEF TIME_ZONES_CONSIDERED}
  TZ: TBundledTimeZone;
{$ENDIF}
begin
  inherited;

  JPair := JSON.Get(SSabStationName);
  if Assigned(JPair) then
    FName := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabStationConfigName);
  if Assigned(JPair) then
    FConfigName := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabStationPlatform);
  if Assigned(JPair) then
    FPlatform := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabStationUserName);
  if Assigned(JPair) then
    FUserName := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabStationDefUser);
  if Assigned(JPair) then
    FDefaultUser := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabStationDefDomain);
  if Assigned(JPair) then
    FDefaultDomain := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabStationIp);
  if Assigned(JPair) then
    FStationIP := JSToStr(JPair.JsonValue.Value);

{$IFDEF DEBUG}
  if FStationIP = '' then
    FStationIP := '128.0.0.1';
{$ENDIF}

  JPair := JSON.Get(SSabStationTimeZone);
  if Assigned(JPair) then
    FTimeZone := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabStationRegistered);
  if Assigned(JPair) then
  begin
    DT := StrToInt64(JPair.JsonValue.Value);
    FRegistered := UnixTimeToDateTime(DT);
{$IFDEF TIME_ZONES_CONSIDERED}
    if FTimeZone <> '' then
    begin
      TZ := TBundledTimeZone.GetTimeZone(FTimeZone);
      FRegistered := TZ.ToLocalTime(FRegistered);
    end;
{$ENDIF}
  end;

  JPair := JSON.Get(SSabStationLastAccess);
  if Assigned(JPair) then
  begin
    DT := StrToInt64(JPair.JsonValue.Value);
    FLastAccess := UnixTimeToDateTime(DT);
{$IFDEF TIME_ZONES_CONSIDERED}
    if FTimeZone <> '' then
    begin
      TZ := TBundledTimeZone.GetTimeZone(FTimeZone);
      FLastAccess := TZ.ToLocalTime(FLastAccess);
    end;
{$ENDIF}
  end;
end;

function TSabStation.GetDefUser: string;
begin
  if FDefUser <> '' then
    Result := FDefUser
  else
    Result := FUserName;
end;

function TSabStation.GetSabObjType: TSabObjectType;
begin
  Result := sabStation;
end;

procedure TSabStation.SaveToList(const List: TStringList);
begin
  inherited;

  raise Exception.Create('N/A');
//    FConfigName: string;  // DEFAULT, etc
//    FPlatform: string;
//    FUserName: string;
//    FDefaultUser: string;
//    FDefaultDomain: string;
//    FStationIP: string;
//    FTimeZone: string;
//    FRegistered: TDateTime;
//    FLastAccess: TDateTime;

end;

procedure TSabStation.SetPassword(const APassword: string; ADate: TDateTime);
begin
  FCurrentPassword := APassword;
  FPswdDate := ADate;
end;

procedure TSabStation.WipeStrings;
begin
  inherited;

  TDP.CommonCode.WipeString(FCurrentPassword);
end;

{ TSabConfiguration }

constructor TSabConfiguration.Create(const JSON: TJSONObject);
var
  JPair: TJSONPair;
begin
  inherited;

  JPair := JSON.Get(SSabConfigSeedBase);
  if Assigned(JPair) then
    FSeedBase := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabConfigPattern);
  if Assigned(JPair) then
    FPattern := JSToStr(JPair.JsonValue.Value);
end;

function TSabConfiguration.GetSabObjType: TSabObjectType;
begin
  Result := sabConfiguration;
end;

procedure TSabConfiguration.SaveToList(const List: TStringList);
begin
  inherited;

  List.Add(SSabConfigSeedBase + ' = ' + FSeedBase);
  List.Add(SSabConfigPattern + ' = ' + FPattern);
end;

{ TSabSafeBox }

constructor TSabSafeBox.Create(const JSON: TJSONObject);
var
  JPair: TJSONPair;
begin
  inherited;

  JPair := JSON.Get(SSabBoxDescription);
  if Assigned(JPair) then
    FDescription := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabBoxPswdPattern);
  if Assigned(JPair) then
    FPasswordPattern := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabBoxPrivate);
  if Assigned(JPair) then
    if not TryStrToBool(JPair.JsonValue.Value, FPrivateBox) then
      FPrivateBox := False;

  JPair := JSON.Get(SSabBoxParentId);
  if Assigned(JPair) then
{$IFDEF LONG_ID}
    FParentId := StrToInt64Def(JPair.JsonValue.Value, 0);
{$ELSE}
    FParentId := StrToIntDef(JPair.JsonValue.Value, 0);
{$ENDIF}

  JPair := JSON.Get(SSabBoxOwnerName);
  if Assigned(JPair) then
    FOwnerName := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabBoxACL);
  if Assigned(JPair) then
    FBoxACL := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabBoxWatched);
  if Assigned(JPair) then
    if not StrToBoolDef(JPair.JsonValue.Value, FWatchedChanges) then
      FWatchedChanges := False;

  JPair := JSON.Get(SSabBoxWatchedAccess);
  if Assigned(JPair) then
    if not TryStrToBool(JPair.JsonValue.Value, FWatchedAccesses) then
      FWatchedAccesses := False;

  JPair := JSON.Get(SSabBoxExternalFile);
  if Assigned(JPair) then
    FExternalFilesPath := JSToStr(JPair.JsonValue.Value);
end;

procedure TSabSafeBox.ClearWatched(const AWatchTypes: TWatchedTypes);
begin
  if wtChanges in AWatchTypes then
    FWatchedChangesModified := False;

  if wtAccesses in AWatchTypes then
    FWatchedAccessesModified := False;
end;

constructor TSabSafeBox.Create(const AName, APasswordPattern: string; APrivateBox: Boolean; AParentId: TSabId);
begin
  inherited Create(AName);
  FPasswordPattern := APasswordPattern;
  FPrivateBox := APrivateBox;
end;

function TSabSafeBox.GetHasPermissionToCreate: Boolean;
begin
  Result := System.Pos('C', FBoxACL) > 0;
end;

function TSabSafeBox.GetHasPermissionToDelete: Boolean;
begin
  Result := System.Pos('D', FBoxACL) > 0;
end;

function TSabSafeBox.GetHasPermissionToModify: Boolean;
begin
  Result := System.Pos('M', FBoxACL) > 0;
end;

function TSabSafeBox.GetSabObjType: TSabObjectType;
begin
  Result := sabSafeBox;
end;

function TSabSafeBox.GetWatchedModified: Boolean;
begin
  Result := FWatchedAccessesModified or FWatchedChangesModified;
end;

function TSabSafeBox.SaveTo(const AUpdate, AFileContent: Boolean): TJSONObject;
begin
  Result := inherited SaveTo(AUpdate, AFileContent);

  Result.AddPair(TJSONPair.Create(SSabBoxDescription, TJSONString.Create(FDescription)));
  Result.AddPair(TJSONPair.Create(SSabBoxPswdPattern, TJSONString.Create(FPasswordPattern)));
  if FPrivateBox then
    Result.AddPair(TJSONPair.Create(SSabBoxPrivate, TJSONTrue.Create))
  else
    Result.AddPair(TJSONPair.Create(SSabBoxPrivate, TJSONFalse.Create));
  Result.AddPair(TJSONPair.Create(SSabBoxParentId, TJSONNumber.Create(FParentId)));
  Result.AddPair(TJSONPair.Create(SSabSafeBoxTypeFld, TJSONString.Create(SabTypeValues[ObjType])));

  Result.AddPair(TJSONPair.Create(SSabBoxExternalFile, TJSONString.Create(FExternalFilesPath)));
end;

procedure TSabSafeBox.SaveToList(const List: TStringList);
begin
  inherited;

  List.Add(SSabBoxDescription + ' = '+ FDescription);
  List.Add(SSabBoxPswdPattern + ' = '+ FPasswordPattern);
  List.Add(SSabBoxPrivate + ' = '+  BoolToStr(FPrivateBox, True));
  List.Add(SSabBoxParentId + ' = '+   IntToStr(FParentId));
  List.Add(SSabBoxExternalFile + ' = ' + FExternalFilesPath);
end;

procedure TSabSafeBox.SetWatchedAccessses(const Value: Boolean);
begin
  FWatchedAccesses := Value;
  FWatchedAccessesModified := True;
end;

procedure TSabSafeBox.SetWatchedChanges(const Value: Boolean);
begin
  FWatchedChanges := Value;
  FWatchedChangesModified := True;
end;

procedure TSabSafeBox.WipeStrings;
begin
  inherited;

  TDP.CommonCode.WipeString(FPasswordPattern);
end;

{ TSabSafeBoxGroup }

function TSabSafeBoxGroup.GetSabObjType: TSabObjectType;
begin
  Result := sabSafeBoxGroup;
end;

{ TSabRecord }

constructor TSabRecord.Create(const JSON: TJSONObject);
var
  I: Integer;
  JPair: TJSONPair;
  JArray: TJSONArray;
  JValue: TJSONValue;
  JObj: TJSONObject;
  S: string;
begin
  inherited;

  JPair := JSON.Get(SSabRecordDescription);
  if Assigned(JPair) then
    FDescription := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabRecordTags);
  if Assigned(JPair) then
  begin
    if JPair.JsonValue is TJSONArray then
    begin
      FTags := '';
      JArray := TJSONArray(JPair.JsonValue);
      for I := 0 to JArray.Count - 1 do
      begin
        JValue := JArray.Items[I];
        Assert(JValue is TJSONString, 'JSON tag mismatch');
        S := TJSONString(JValue).Value;
        if FTags = '' then
          FTags := System.SysUtils.AnsiQuotedStr(S, '"')
        else
          FTags := FTags + ', ' + System.SysUtils.AnsiQuotedStr(S, '"');  // 22/06/21, MUST comma separated array
      end;
    end else
    begin
      Assert(JPair.JsonValue is TJSONString, 'JSON tag mismatch');
      FTags := TJSONString(JPair.JsonValue).Value;
    end;
  end;

  JPair := JSON.Get(SSabRecordLauncher);
  if Assigned(JPair) then
  begin
    S := JSToStr(JPair.JsonValue.Value);
    FLauncher := StrToLauncher(S);
  end;

  JPair := JSON.Get(SSabRecordNote);
  if Assigned(JPair) then
    FNote := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabRecordSite);
  if Assigned(JPair) then
    FSite := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SObjAttributes);
  if Assigned(JPair) and (JPair.JsonValue is TJSONObject) then
  begin
    JObj := JPair.JsonValue as TJSONObject;
    JPair := JObj.Get(SSabRecordSecNote);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FSecNote := JSToStr(JPair.JsonValue.Value);
    end;
  end;
end;

function TSabRecord.AssignNonEmptyFieldsFrom(const Obj: TSabObject): Boolean;
begin
  Result := inherited AssignNonEmptyFieldsFrom(Obj);
  if not Result then
    Exit;

  if TSabRecord(Obj).Description <> '' then
    FDescription := TSabRecord(Obj).Description;

  if TSabRecord(Obj).Tags <> '' then
    FTags := TSabRecord(Obj).Tags;

  if TSabRecord(Obj).Note <> '' then
    FNote := TSabRecord(Obj).Note;

  if TSabRecord(Obj).SecNote <> '' then
    FSecNote := TSabRecord(Obj).SecNote;
end;

constructor TSabRecord.Create(const AName, ADescription, ATags, ANote: string);
begin
  inherited Create(AName);

  FDescription := ADescription;
  FTags := ATags;
  FNote := ANote;
end;

procedure TSabRecord.GetSecureFieldsFrom(Obj: TSabObject);
var
  Rec: TSabRecord;
begin
  inherited;
  if Obj is TSabRecord then
  begin
    Rec := TSabRecord(Obj);

    if Rec.Description <> '' then
      FDescription := Rec.Description;

    if Rec.Tags <> '' then
      FTags := Rec.Tags;

    if Rec.Note <> '' then
      FNote := Rec.Note;

    if Rec.SecNote <> '' then
      FSecNote := Rec.SecNote;
  end;
end;

function TSabRecord.SaveTo(const AUpdate, AFileContent: Boolean): TJSONObject;
begin
  Result := inherited SaveTo(AUpdate, AFileContent);

  Result.AddPair(TJSONPair.Create(SSabRecordDescription, TJSONString.Create(FDescription)));
  Result.AddPair(TJSONPair.Create(SSabRecordTags, TJSONString.Create(FTags)));
  Result.AddPair(TJSONPair.Create(SSabRecordNote, TJSONString.Create(TextToJS(FNote))));
  Result.AddPair(TJSONPair.Create(SSabRecordSite, TJSONString.Create(TextToJS(FSite))));

  if FLauncher <> sabLauncherNone then
  begin
    case ObjType of
      sabAccount, sabAccountSec:
        Result.AddPair(TJSONPair.Create(SSabRecordLauncher, TJSONString.Create(SabLauncherSign[FLauncher])));
    end;
  end;

  case ObjType of
    sabAccountSec, sabFile, sabCreditCard, sabCertificate:
      if SecureFieldsUsed then
        Result.AddPair(TJSONPair.Create(SSabRecordSecNote, TJSONString.Create(TextToJS(FSecNote))));
  end;
end;

procedure TSabRecord.SaveToList(const List: TStringList);
begin
  inherited;

  List.Add(SSabRecordDescription + ' = ' + FDescription);
  List.Add(SSabRecordTags + ' = ' + FTags);
  List.Add(SSabRecordNote + ' = ' + FNote);
  List.Add(SSabRecordSite + ' = ' + FSite);

  if FLauncher <> sabLauncherNone then
  begin
    case ObjType of
      sabAccount, sabAccountSec:
        List.Add(SSabRecordLauncher + ' = ' + SabLauncherSign[FLauncher]);
    end;
  end;

  case ObjType of
    sabAccountSec, sabFile, sabCreditCard:
      if SecureFieldsUsed then
        List.Add(SSabRecordSecNote + ' = ' + FSecNote);
  end;
end;

procedure TSabRecord.WipeStrings;
begin
  inherited;

  FSecureFieldsUsed := False;
  TDP.CommonCode.WipeString(FSecNote);
end;

{ TSabAccountCommon }

constructor TSabAccountCommon.Create(const JSON: TJSONObject);
var
  JPair: TJSONPair;
  JObj: TJSONObject;
begin
  inherited;

  JPair := JSON.Get(SSabAccountLoginSite);
  if Assigned(JPair) then
    FLoginSite := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabAccountPassword);
  if Assigned(JPair) then
  begin
    FSecureFieldsUsed := True;
    FPassword := JSToStr(JPair.JsonValue.Value);
  end;

  JPair := JSON.Get(SSabAccountPasswordRe);
  if Assigned(JPair) then
  begin
    FSecureFieldsUsed := True;
    FPasswordRe := JSToStr(JPair.JsonValue.Value);
  end;

  JPair := JSON.Get(SObjAttributes);
  if Assigned(JPair) and (JPair.JsonValue is TJSONObject) then
  begin
    JObj := JPair.JsonValue as TJSONObject;
    JPair := JObj.Get(SSabBoxPswdPattern);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FPasswordPattern := JSToStr(JPair.JsonValue.Value);
    end;
  end;

  if Trim(FPasswordPattern) = '' then
  begin
    JPair := JSON.Get(SSabBoxPswdPattern);
    if Assigned(JPair) then
      FPasswordPattern := JSToStr(JPair.JsonValue.Value);
  end;
end;

function TSabAccountCommon.AssignNonEmptyFieldsFrom(const Obj: TSabObject): Boolean;
begin
  Result := inherited AssignNonEmptyFieldsFrom(Obj);
  if not Result then
    Exit;

  if TSabAccountCommon(Obj).PasswordPattern <> '' then
    FPasswordPattern := TSabAccountCommon(Obj).PasswordPattern;

  if TSabAccountCommon(Obj).Password <> '' then
    FPassword := TSabAccountCommon(Obj).Password;

  if TSabAccountCommon(Obj).PasswordRe <> '' then
    FPasswordRe := TSabAccountCommon(Obj).PasswordRe;
end;

constructor TSabAccountCommon.Create(const AName, ADescription, ATags, ANote, APasswordPattern, APassword, ASite: string);
begin
  inherited Create(AName, ADescription, ATags, ANote);
  FPasswordPattern := APasswordPattern;
  FPassword := APassword;
  FPasswordRe := APassword;
  FLoginSite := ASite;
  FSecureFieldsUsed := True;
end;

procedure TSabAccountCommon.GetSecureFieldsFrom(Obj: TSabObject);
var
  Acc: TSabAccountCommon;
begin
  inherited;
  if Obj is TSabAccountCommon then
  begin
    Acc := TSabAccountCommon(Obj);

    if Acc.PasswordPattern <> '' then
      FPasswordPattern := Acc.PasswordPattern;

    if Acc.Password <> '' then
      FPassword := Acc.Password;

    if Acc.PasswordRe <> '' then
      FPasswordRe := Acc.PasswordRe;
  end;
end;

function TSabAccountCommon.SavePasswords: TJSONObject;
begin
  Result := TJSONObject.Create;
  Result.AddPair(TJSONPair.Create(SSabAccountPassword, TJSONString.Create(FPassword)));
  Result.AddPair(TJSONPair.Create(SSabAccountPasswordRe, TJSONString.Create(FPasswordRe)));
end;

function TSabAccountCommon.SaveTo(const AUpdate, AFileContent: Boolean): TJSONObject;
begin
  Result := inherited SaveTo(AUpdate, AFileContent);

  Result.AddPair(TJSONPair.Create(SSabAccountLoginSite, TJSONString.Create(FLoginSite)));

  if SecureFieldsUsed and not AUpdate then
  begin
    Result.AddPair(TJSONPair.Create(SSabAccountPassword, TJSONString.Create(FPassword)));
    Result.AddPair(TJSONPair.Create(SSabAccountPasswordRe, TJSONString.Create(FPasswordRe)));
  end;
end;

procedure TSabAccountCommon.SaveToList(const List: TStringList);
begin
  inherited;

  List.Add(SSabBoxPswdPattern + ' = ' + FPasswordPattern);
  if SecureFieldsUsed then
  begin
    List.Add(SSabAccountPassword + ' = ' + FPassword);
    List.Add(SSabAccountPasswordRe + ' = ' + FPasswordRe);
  end;
end;

procedure TSabAccountCommon.SetPassword(const Value: string);
begin
  FPassword := Value;
  FPasswordRe := Value;
end;

procedure TSabAccountCommon.WipeStrings;
begin
  inherited;

  TDP.CommonCode.WipeString(FPassword);
  TDP.CommonCode.WipeString(FPasswordRe);
end;

{ TSabAccount }

constructor TSabAccount.Create(const JSON: TJSONObject);
var
  JPair: TJSONPair;
begin
  inherited;

  JPair := JSON.Get(SSabAccountLoginName);
  if Assigned(JPair) then
    FLoginName := JSToStr(JPair.JsonValue.Value);

  JPair := JSON.Get(SSabAccountServerAddress);
  if Assigned(JPair) then
    FServerAddress := JSToStr(JPair.JsonValue.Value);
end;

constructor TSabAccount.Create(const AName, ADescription, ATags, ANote, APasswordPattern, APassword, ALoginName, AServerAddress, ASite: string);
begin
  inherited Create(AName, ADescription, ATags, ANote, APasswordPattern, APassword, ASite);
  FLoginName := ALoginName;
  FServerAddress := AServerAddress;
end;

class function TSabAccount.CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string; const HasPreamble: Boolean; const AClass: TSabObjectClass): TSabObject;
begin
  Result := inherited CreateObject(Strm, AObjectName, HasPreamble, TSabAccount);
end;

function TSabAccount.AssignNonEmptyFieldsFrom(const Obj: TSabObject): Boolean;
begin
  Result := inherited AssignNonEmptyFieldsFrom(Obj);
  if not Result then
    Exit;

  if TSabAccount(Obj).LoginName <> '' then
    FLoginName := TSabAccount(Obj).LoginName;

  if TSabAccount(Obj).ServerAddress <> '' then
    FServerAddress := TSabAccount(Obj).ServerAddress;
end;

function TSabAccount.GetSabObjType: TSabObjectType;
begin
  Result := sabAccount;
end;

procedure TSabAccount.GetSecureFieldsFrom(Obj: TSabObject);
var
  Acc: TSabAccount;
begin
  inherited;
  if Obj is TSabAccount then
  begin
    Acc := TSabAccount(Obj);

    if Acc.LoginName <> '' then
      FLoginName := Acc.LoginName;

    if Acc.ServerAddress <> '' then
      FServerAddress := Acc.ServerAddress;
  end;
end;

function TSabAccount.SaveTo(const AUpdate, AFileContent: Boolean): TJSONObject;
var
  JObj: TJSONObject;
begin
  Result := inherited SaveTo(AUpdate, AFileContent);

  Result.AddPair(TJSONPair.Create(SSabAccountLoginName, TJSONString.Create(FLoginName)));
  Result.AddPair(TJSONPair.Create(SSabAccountServerAddress, TJSONString.Create(FServerAddress)));

  JObj := TJSONObject.Create;
  JObj.AddPair(TJSONPair.Create(SSabBoxPswdPattern, TJSONString.Create(PasswordPattern)));
  Result.AddPair(TJSONPair.Create(SObjAttributes, JObj));
end;

procedure TSabAccount.SaveToList(const List: TStringList);
begin
  inherited;

  List.Add(SSabAccountLoginName + ' = ' + FLoginName);
  List.Add(SSabAccountServerAddress + ' = ' + FServerAddress);
end;

{ TSabSecAccount }

constructor TSabSecAccount.Create(const JSON: TJSONObject);
var
  JPair: TJSONPair;
  JObj: TJSONObject;
begin
  inherited;

  JPair := JSON.Get(SObjAttributes);
  if Assigned(JPair) and (JPair.JsonValue is TJSONObject) then
  begin
    JObj := JPair.JsonValue as TJSONObject;

    JPair := JObj.Get(SSabAccountSecLoginName);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FSecLoginName := JSToStr(JPair.JsonValue.Value);
    end;

    JPair := JObj.Get(SSabAccountSecServerAddr);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FSecServerAddress := JSToStr(JPair.JsonValue.Value);
    end;
  end;
end;

constructor TSabSecAccount.Create(const AName, ADescription, ATags, ANote, APasswordPattern, APassword, ASecLoginName, ASecServerAddress, ASite, ASecNote: string);
begin
  inherited Create(AName, ADescription, ATags, ANote, APasswordPattern, APassword, ASite);
  FSecLoginName := ASecLoginName;
  FSecServerAddress := ASecServerAddress;
  SecNote := ASecNote;
  FSecureFieldsUsed := True;
end;

class function TSabSecAccount.CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string;
  const HasPreamble: Boolean; const AClass: TSabObjectClass): TSabObject;
begin
  Result := inherited CreateObject(Strm, AObjectName, HasPreamble, TSabSecAccount);
end;

function TSabSecAccount.GetSabObjType: TSabObjectType;
begin
  Result := sabAccountSec;
end;

function TSabSecAccount.AssignNonEmptyFieldsFrom(const Obj: TSabObject): Boolean;
begin
  Result := inherited AssignNonEmptyFieldsFrom(Obj);
  if not Result then
    Exit;

  if TSabSecAccount(Obj).SecLoginName <> '' then
    FSecLoginName := TSabSecAccount(Obj).SecLoginName;

  if TSabSecAccount(Obj).SecServerAddress <> '' then
    FSecServerAddress := TSabSecAccount(Obj).SecServerAddress;
end;

procedure TSabSecAccount.GetSecureFieldsFrom(Obj: TSabObject);
var
  Sec: TSabSecAccount;
begin
  inherited;
  if Obj is TSabSecAccount then
  begin
    Sec := TSabSecAccount(Obj);

    if Sec.SecLoginName <> '' then
      FSecLoginName := Sec.SecLoginName;

    if Sec.SecServerAddress <> '' then
      FSecServerAddress := Sec.SecServerAddress;
  end;
end;

function TSabSecAccount.SaveTo(const AUpdate, AFileContent: Boolean): TJSONObject;
var
  JObj: TJSONObject;
begin
  Result := inherited SaveTo(AUpdate, AFileContent);

  if SecureFieldsUsed then
  begin
    JObj := TJSONObject.Create;
    JObj.AddPair(TJSONPair.Create(SSabBoxPswdPattern, TJSONString.Create(PasswordPattern)));
    JObj.AddPair(TJSONPair.Create(SSabAccountSecServerAddr, TJSONString.Create(FSecServerAddress)));
    JObj.AddPair(TJSONPair.Create(SSabAccountSecLoginName, TJSONString.Create(FSecLoginName)));
    JObj.AddPair(TJSONPair.Create(SSabRecordSecNote, TJSONString.Create(SecNote)));

    Result.AddPair(TJSONPair.Create(SObjAttributes, JObj));
  end;
end;

procedure TSabSecAccount.SaveToList(const List: TStringList);
begin
  inherited;

  if SecureFieldsUsed then
  begin
    List.Add(SSabAccountSecLoginName + ' = ' + FSecLoginName);
    List.Add(SSabAccountSecServerAddr + ' = ' + FSecServerAddress);
  end;
end;

procedure TSabSecAccount.WipeStrings;
begin
  inherited;

  TDP.CommonCode.WipeString(FSecLoginName);
  TDP.CommonCode.WipeString(FSecServerAddress);
end;

{ TSabFile }

constructor TSabFile.Create(const JSON: TJSONObject);
var
  JPair: TJSONPair;
  Enc: TBase64Encoding;
  DecodedFile: TStringStream;
  JObj: TJSONObject;
begin
  inherited;

  if ObjType in [sabFile, sabCertificate] then
  begin
    JPair := JSON.Get(SObjAttributes);
    if Assigned(JPair) and (JPair.JsonValue is TJSONObject) then
    begin
      JObj := JPair.JsonValue as TJSONObject;

      JPair := JObj.Get(SSabFileName);
      if Assigned(JPair) then
      begin
        FSecureFieldsUsed := True;
        FFilename := JSToStr(JPair.JsonValue.Value);
      end;
    end;

    JPair := JSON.Get(SSabFileSizeApprox);
    if Assigned(JPair) then
      FFileSizeApprox := StrToInt64(JPair.JsonValue.Value);

    // parsing JSON from get-records; if fileInfo=True
    if FFilename = '' then
    begin
      JPair := JSON.Get(SSabRecordsFileName);
      if Assigned(JPair) then
        FFilename := JSToStr(JPair.JsonValue.Value);
    end;

    FOwnStream := True;
    FStream := TBytesStream.Create;

    JPair := JSON.Get(SSabFileContent);
    if Assigned(JPair) then
    begin
      Enc := TBase64Encoding.Create(0);
      DecodedFile := TStringStream.Create(JPair.JSONValue.Value);
      try
        Enc.Decode(DecodedFile, FStream);
      finally
        Enc.Free;
        DecodedFile.Free;
      end;
    end;
  end;
end;

constructor TSabFile.Create(const AName, ADescription, ATags, ANote, ASecNote, AFilename: string; const AStream: TBytesStream; const AOwnStream: Boolean);
begin
  inherited Create(AName, ADescription, ATags, ANote);
  FFilename := AFilename;
  FOwnStream := AOwnStream;
  SecNote := ASecNote;
  FSecureFieldsUsed := True;

  if AOwnStream then
  begin
    FStream := TBytesStream.Create;
    FStream.CopyFrom(AStream, -1);
  end else
    FStream := AStream;
end;

constructor TSabFile.Create(const AName, ADescription, ATags, ANote, ASecNote, AFilename: string; const AData: TBytes);
begin
  inherited Create(AName, ADescription, ATags, ANote);
  FFilename := AFilename;
  FStream := TBytesStream.Create(AData);
  FOwnStream := True;
  SecNote := ASecNote;
  FSecureFieldsUsed := True;
end;

// Created TSabFile object, load data from Filename file
constructor TSabFile.Create(const FilePath, AName, ADescription, ATags, ANote, ASecNote, AFilename: string);
begin
  inherited Create(AName, ADescription, ATags, ANote);
  SecNote := ASecNote;
  FSecureFieldsUsed := True;
  LoadFile(FilePath);
  if AFilename <> '' then
    FFilename := AFilename;
end;

constructor TSabFile.Create(const AName, ADescription, ATags, ANote, ASecNote, AFilename: string; const AData: PByte; ADataLen: UInt32);
var
  Data: TBytes;
begin
  SetLength(Data, ADataLen);
  if ADataLen > 0 then
    Move(AData[0], Data[0], Length(Data));
  try
    Create(AName, ADescription, ATags, ANote, ASecNote, AFilename, Data);
  finally
    SetLength(Data, 0);
  end;
end;

class function TSabFile.CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string; const HasPreamble: Boolean; const AClass: TSabObjectClass): TSabObject;
begin
  Result := inherited CreateObject(Strm, AObjectName, HasPreamble, TSabFile);
end;

destructor TSabFile.Destroy;
begin
  if FOwnStream then
    FStream.Free;
  inherited;
end;

function TSabFile.BuildPath(const Dir, ALocalFile: string): string;
begin
  if ALocalFile <> '' then
    Result := ALocalFile
  else if Dir = '' then
    Result := Filename
  else
    Result := IncludeTrailingPathDelimiter(Dir) + Filename;
end;

function TSabFile.GetFilename: string;
begin
  Result := FFileName;
end;

function TSabFile.GetFileSizeApprox: Int64;
begin
  if (FFileSizeApprox <= 0) and Assigned(FStream) then
    FFileSizeApprox := FStream.Size;

  Result := FFileSizeApprox;
end;

function TSabFile.GetSabObjType: TSabObjectType;
begin
  Result := sabFile;
end;

procedure TSabFile.GetSecureFieldsFrom(Obj: TSabObject);
var
  Sec: TSabFile;
begin
  inherited;

  if Obj is TSabFile then
  begin
    Sec := TSabFile(Obj);

    if Sec.Filename <> '' then
      FFilename := Sec.Filename;
  end;
end;

procedure TSabFile.LoadFile(const AFilename: string);
var
  FileStream: TBytesStream;
begin
  FileStream := TBytesStream.Create;
  try
    FileStream.LoadFromFile(AFilename);
    LoadFile(FileStream, AFilename);
  finally
    FileStream.Free;
  end;
end;

procedure TSabFile.LoadFile(const AStream: TBytesStream; const AFilename: string);
begin
  SetStream(AStream, True);
  FFilename := ExtractFileName(AFilename);
end;

function TSabFile.SaveTo(const AUpdate, AFileContent: Boolean): TJSONObject;
var
  Enc: TBase64Encoding;
  EncodedFile: TStringStream;
  JObj: TJSONObject;
begin
  Result := inherited SaveTo(AUpdate, AFileContent);

  if ObjType in [sabFile, sabCertificate] then
  begin
    if SecureFieldsUsed then
    begin
      if not (Self is TSabCertificate) then
      begin
        JObj := TJSONObject.Create;
        JObj.AddPair(TJSONPair.Create(SSabFileName, TJSONString.Create(FFilename)));
        JObj.AddPair(TJSONPair.Create(SSabRecordSecNote, TJSONString.Create(TextToJS(SecNote))));

        Result.AddPair(TJSONPair.Create(SObjAttributes, JObj));
      end;
    end;

    if Assigned(FStream) and AFileContent then
    begin
      Enc := TBase64Encoding.Create(0);
      EncodedFile := TStringStream.Create;
      try
        FStream.Position := 0;
        Enc.Encode(FStream, EncodedFile);
        Result.AddPair(TJSONPair.Create(SSabFileContent, TJSONString.Create(EncodedFile.DataString)));
      finally
        Enc.Free;
        EncodedFile.Free;
      end;
    end;
  end;
end;

procedure TSabFile.SaveToFile(const Dir: string; const ALocalFile: string; const AEncrypted: Boolean);
// if AEncrypted = True then must save to C:\users\user\appdata\file
var
  FileNamePrefix: string;
{$IFDEF FILE_STREAM_EX}
  FilStreamEx: TFileStreamEx;
{$ENDIF}
begin
  if ALocalFile <> '' then
    FileNamePrefix := ALocalFile
  else if Dir = '' then
    FileNamePrefix := FFilename
  else
    FileNamePrefix := IncludeTrailingPathDelimiter(Dir) + FFilename;

{$IFDEF FILE_STREAM_EX}
  if AEncrypted then
  begin
    FilStreamEx := TFileStreamEx.Create(FileNamePrefix, fmCreate + fmEncrypted + GetRemoteMode(FileNamePrefix));
    try
      FStream.SaveToStream(FilStreamEx);
    finally
      FilStreamEx.Free;
    end;
  end else
  begin
    FilStreamEx := TFileStreamEx.Create(FileNamePrefix, fmCreate + GetRemoteMode(FileNamePrefix));
    try
      FStream.SaveToStream(FilStreamEx)
    finally
      FilStreamEx.Free;
    end;
  end;
{$ELSE}
  FStream.SaveToFile(FileNamePrefix);
{$ENDIF}
end;

procedure TSabFile.SaveToList(const List: TStringList);
begin
  inherited;

  if ObjType in [sabFile, sabCertificate] then
  begin
    if SecureFieldsUsed then
      if not (Self is TSabCertificate) then
        List.Add(SSabFileName + ' = ' + FFilename);
  end;
end;

procedure TSabFile.SetFilename(const Value: string);
begin
  FFileName := Value;
end;

procedure TSabFile.SetStream(const AStream: TBytesStream; AOwnStream: Boolean);
begin
  if Assigned(FStream) then
  begin
    if FOwnStream then
      FreeAndNil(FStream);
    FStream := nil;
  end;
  FOwnStream := AOwnStream;
  if FOwnStream and Assigned(AStream) then
  begin
    AStream.Position := 0;
    FStream := TBytesStream.Create;
    FStream.CopyFrom(AStream, -1);
  end else
    FStream := AStream;
end;

{ TSabCertificate }

constructor TSabCertificate.Create(const JSON: TJSONObject);
var
  JPair: TJSONPair;
  JObj: TJSONObject;
begin
  inherited;

  JPair := JSON.Get(SObjAttributes);
  if Assigned(JPair) and (JPair.JsonValue is TJSONObject) then
  begin
    JObj := JPair.JsonValue as TJSONObject;

    JPair := JObj.Get(SSabCertifAlias);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FAlias := JSToStr(JPair.JsonValue.Value);
    end;

    JPair := JObj.Get(SSabCertifFilePswd);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FFilePassword := JSToStr(JPair.JsonValue.Value);
    end;
  end;
end;

constructor TSabCertificate.Create(const AName, ADescription, ATags, ANote, ASecNote, AFilename, AAlias, AFilePassword: string; const
  AStream: TBytesStream; const AOwnStream: Boolean);
begin
  inherited Create(AName, ADescription, ATags, ANote, ASecNote, AFilename, AStream, AOwnStream);
  FAlias := AAlias;
  FFilePassword := AFilePassword;
  FSecureFieldsUsed := True;
end;

constructor TSabCertificate.Create(const AName, ADescription, ATags, ANote, ASecNote, AFilename, AAlias, AFilePassword: string; const AData: TBytes);
begin
  inherited Create(AName, ADescription, ATags, ANote, ASecNote, AFilename, AData);
  FAlias := AAlias;
  FFilePassword := AFilePassword;
  FSecureFieldsUsed := True;
end;

constructor TSabCertificate.Create(const FilePath, AName, ADescription, ATags, ANote, ASecNote, AFilename, AAlias, AFilePassword: string);
begin
  inherited Create(FilePath, AName, ADescription, ATags, ANote, ASecNote, AFilename);
  FAlias := AAlias;
  FFilePassword := AFilePassword;
  FSecureFieldsUsed := True;
end;

constructor TSabCertificate.Create(const AName, ADescription, ATags, ANote, ASecNote, AFilename, AAlias, AFilePassword: string; const
  AData: PByte; ADataLen: UInt32);
begin
  inherited Create(AName, ADescription, ATags, ANote, ASecNote, AFilename, AData, ADataLen);
  FAlias := AAlias;
  FFilePassword := AFilePassword;
  FSecureFieldsUsed := True;
end;

class function TSabCertificate.CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string;
  const HasPreamble: Boolean; const AClass: TSabObjectClass): TSabObject;
begin
  Result := inherited CreateObject(Strm, AObjectName, HasPreamble, TSabCertificate);
end;

function TSabCertificate.GetSabObjType: TSabObjectType;
begin
  Result := sabCertificate;
end;

procedure TSabCertificate.GetSecureFieldsFrom(Obj: TSabObject);
var
  Rec: TSabCertificate;
begin
  inherited;
  if Obj is TSabCertificate then
  begin
    Rec := TSabCertificate(Obj);

    if Rec.Alias <> '' then
      FAlias := Rec.Alias;

    if Rec.FilePassword <> '' then
      FFilePassword := Rec.FilePassword;
  end;
end;

function TSabCertificate.SaveTo(const AUpdate, AFileContent: Boolean): TJSONObject;
var
  JObj: TJSONObject;
begin
  Result := inherited SaveTo(AUpdate, AFileContent);

  if SecureFieldsUsed then
  begin
    JObj := TJSONObject.Create;
    JObj.AddPair(TJSONPair.Create(SSabCertifAlias, TJSONString.Create(FAlias)));
    JObj.AddPair(TJSONPair.Create(SSabFileName, TJSONString.Create(FFilename)));
    JObj.AddPair(TJSONPair.Create(SSabCertifFilePswd, TJSONString.Create(FFilePassword)));
    JObj.AddPair(TJSONPair.Create(SSabRecordSecNote, TJSONString.Create(TextToJS(SecNote))));

    Result.AddPair(TJSONPair.Create(SObjAttributes, JObj));
  end;
end;

procedure TSabCertificate.SaveToList(const List: TStringList);
begin
  inherited;

  if SecureFieldsUsed then
  begin
    List.Add(SSabCertifAlias + ' = ' + FAlias);
    List.Add(SSabCertifFile + ' = ' + FFilename);
    List.Add(SSabCertifFilePswd + ' = ' + FFilePassword);
  end;
end;

procedure TSabCertificate.WipeStrings;
begin
  inherited;

  TDP.CommonCode.WipeString(FFilePassword);
  TDP.CommonCode.WipeString(FAlias);
  TDP.CommonCode.WipeString(FFilename);
end;

{ TExternalFileHeader }

procedure TExternalFileHeader.Init;
begin
  Sign := EXTFILE_SIGNATURE;
end;

function TExternalFileHeader.Valid: Boolean;
begin
  Result := Sign = EXTFILE_SIGNATURE;
end;

{ TSabExternalAttributes }

constructor TSabExternalFileAttributes.Create(const Rec: TSabRecord);
begin
  inherited Create;
  if Assigned(Rec) then
    Assign(Rec);
end;

procedure TSabExternalFileAttributes.Assign(const Rec: TSabRecord);
var
  Ext: TSabExternalFile;
begin
  if Rec.ObjType = sabExternalFile then
  begin
    Ext := Rec as TSabExternalFile;
    efOrigFileName := Ext.efOrigFilename;
    efFilePath := Ext.efFilePath;
    efKey := Ext.efKey;
    efIV := Ext.efIV;
    efAAD := Ext.efAAD;
    efOldFilePath := '';
  end;
end;

procedure TSabExternalFileAttributes.Clear;
begin
  efOrigFileName := '';
  efFilePath := '';
  efKey := '';
  efIV := '';
  efAAD := '';
  efOldFilePath := '';
end;

{ TSabExternalFile }

constructor TSabExternalFile.Create(const JSON: TJSONObject);
var
  JPair: TJSONPair;
  JObj: TJSONObject;
begin
  inherited;

  JPair := JSON.Get(SObjAttributes);
  if Assigned(JPair) and (JPair.JsonValue is TJSONObject) then
  begin
    JObj := JPair.JsonValue as TJSONObject;

    JPair := JObj.Get(SSabExternalFileName);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FefOrigFileName := JSToStr(JPair.JsonValue.Value);
    end;

    JPair := JObj.Get(SSabExternalFilePath);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FefFilePath := JSToStr(JPair.JsonValue.Value);
    end;

    JPair := JObj.Get(SSabExternalFileKey);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FefKey := JSToStr(JPair.JsonValue.Value);
    end;

    JPair := JObj.Get(SSabExternalFileIV);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FefIV  := JSToStr(JPair.JsonValue.Value);
    end;

    JPair := JObj.Get(SSabExternalFileAAD);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FefAAD  := JSToStr(JPair.JsonValue.Value);
    end;

    JPair := JObj.Get(SSabExternalFileSize);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FFileSize  := StrToIntDef(JPair.JsonValue.Value, 0);
    end;
  end;
end;

constructor TSabExternalFile.Create(const AName, ADescription, ATags, ANote,
  AefOrigFileName, AefFilePath, AefKey, AefIV, AefAAD: string; const AFileSize: Int64);
begin
  inherited Create(AName, ADescription, ATags, ANote);

  FSecureFieldsUsed := True;

  FefOrigFileName := AefOrigFileName;
  efFilePath := AefFilePath;
  FefKey := AefKey;
  FefIV := AefIV;
  FefAAD := AefAAD;

  FFileSize := AFileSize;
end;

class function TSabExternalFile.CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string; const HasPreamble: Boolean;
  const AClass: TSabObjectClass): TSabObject;
begin
  Result := TSabRecord.CreateObject(Strm, AObjectName, HasPreamble, TSabExternalFile);
end;

procedure TSabExternalFile.Assign(const Attrs: TSabExternalFileAttributes);
begin
  FefOrigFileName := Attrs.efOrigFileName;
  FefFilePath := Attrs.efFilePath;
  FefKey := Attrs.efKey;;
  FefIV := Attrs.efIV;
  FefAAD := Attrs.efAAD;
end;

function TSabExternalFile.GetAttributes: TSabExternalFileAttributes;
begin
  Result := TSabExternalFileAttributes.Create(Self);
end;

function TSabExternalFile.GetFilename: string;
begin
  Result := FefOrigFileName;
end;

function TSabExternalFile.GetSabObjType: TSabObjectType;
begin
  Result := sabExternalFile;
end;

procedure TSabExternalFile.GetSecureFieldsFrom(Obj: TSabObject);
var
  Sec: TSabExternalFile;
begin
  inherited;

  if Obj is TSabExternalFile then
  begin
    Sec := TSabExternalFile(Obj);

    if Sec.efOrigFilename <> '' then
      FefOrigFileName := Sec.efOrigFilename;

    if Sec.efFilePath <> '' then
      FefFilePath := Sec.efFilePath;

    if Sec.efKey <> '' then
      FefKey := Sec.efKey;

    if Sec.efIV <> '' then
      FefIV := Sec.efIV;

    if Sec.efAAD <> '' then
      FefAAD := Sec.efAAD;
  end;
end;

function TSabExternalFile.SaveTo(const AUpdate, AFileContent: Boolean): TJSONObject;
var
  JObj: TJSONObject;
begin
  Result := inherited;


  if SecureFieldsUsed then
  begin
    JObj := TJSONObject.Create;

    JObj.AddPair(TJSONPair.Create(SSabExternalFileName, TJSONString.Create(FefOrigFileName)));
    JObj.AddPair(TJSONPair.Create(SSabExternalFilePath, TJSONString.Create(FefFilePath)));
    JObj.AddPair(TJSONPair.Create(SSabExternalFileKey, TJSONString.Create(FefKey)));
    JObj.AddPair(TJSONPair.Create(SSabExternalFileIV, TJSONString.Create(FefIV)));
    JObj.AddPair(TJSONPair.Create(SSabExternalFileAAD, TJSONString.Create(FefAAD)));
    JObj.AddPair(TJSONPair.Create(SSabExternalFileSize, TJSONNumber.Create(FFileSize)));

    Result.AddPair(TJSONPair.Create(SObjAttributes, JObj));
  end;
end;

procedure TSabExternalFile.SaveToList(const List: TStringList);
begin
  inherited;

  if SecureFieldsUsed then
  begin
    List.Add(SSabExternalFileName + ' = ' + FefOrigFileName);
    List.Add(SSabExternalFilePath + ' = ' + FefFilePath);
    List.Add(SSabExternalFileKey + ' = ' + FefKey);
    List.Add(SSabExternalFileIV + ' = ' + FefIV);
    List.Add(SSabExternalFileAAD + ' = ' + FefAAD);
    List.Add(SSabExternalFileSize + ' = ' + IntToStr(FFileSize));
  end;
end;

procedure TSabExternalFile.SetEfFilePath(const Value: string);
begin
  FefFilePath := StringReplace(Value, '/', '\', [rfReplaceAll]);
end;

procedure TSabExternalFile.SetFilename(const Value: string);
begin
  FefOrigFileName := Value;
end;

{ TSabCreditCard }

constructor TSabCreditCard.Create(const JSON: TJSONObject);
var
  JPair: TJSONPair;
  JObj: TJSONObject;
begin
  inherited;

  JPair := JSON.Get(SObjAttributes);
  if Assigned(JPair) and (JPair.JsonValue is TJSONObject) then
  begin
    JObj := JPair.JsonValue as TJSONObject;

    JPair := JObj.Get(SSabCCardNumber);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FNumber:= JSToStr(JPair.JsonValue.Value);
    end;

    JPair := JObj.Get(SSabCCardExpiration);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FExpiration := JSToStr(JPair.JsonValue.Value);
    end;

    JPair := JObj.Get(SSabCCardCvv);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FCvv := JSToStr(JPair.JsonValue.Value);
    end;

    JPair := JObj.Get(SSabCCardPin);
    if Assigned(JPair) then
    begin
      FSecureFieldsUsed := True;
      FPin := JSToStr(JPair.JsonValue.Value);
    end;
  end;
end;

constructor TSabCreditCard.Create(const AName, ADescription, ATags, ANote, ASecNote, ANumber, AExpiration, ACvv, APin: string);
begin
  inherited Create(AName, ADescription, ATags, ANote);
  FNumber := ANumber;
  FExpiration := AExpiration;
  FCvv := ACvv;
  FPin := APin;
  FSecureFieldsUsed := True;

  SecNote := ASecNote;
end;

class function TSabCreditCard.CreateObject(const Strm: TCustomMemoryStream; const AObjectName: string;
  const HasPreamble: Boolean; const AClass: TSabObjectClass): TSabObject;
begin
  Result := inherited CreateObject(Strm, AObjectName, HasPreamble, TSabCreditCard);
end;

function TSabCreditCard.GetExpirationDate(Local: Boolean): TDateTime;
begin
  try
    Result := ISO8601ToDate(FExpiration, not Local)
  except
    // if FExpiration has not "right" format return 0
    Result := 0;
  end;
end;

function TSabCreditCard.GetSabObjType: TSabObjectType;
begin
  Result := sabCreditCard;
end;

procedure TSabCreditCard.GetSecureFieldsFrom(Obj: TSabObject);
var
  Rec: TSabCreditCard;
begin
  inherited;
  if Obj is TSabCreditCard then
  begin
    Rec := TSabCreditCard(Obj);

    if Rec.Number <> '' then
      FNumber := Rec.Number;

    if Rec.Expiration <> '' then
      FExpiration := Rec.Expiration;

    if Rec.Cvv <> '' then
      FCvv := Rec.Cvv;

    if Rec.Pin <> '' then
      FPin := Rec.Pin;

    if Rec.SecNote <> '' then
      SecNote := Rec.SecNote;
  end;
end;

function TSabCreditCard.SaveTo(const AUpdate, AFileContent: Boolean): TJSONObject;
var
  JObj: TJSONObject;
begin
  Result := inherited SaveTo(AUpdate, AFileContent);

  if SecureFieldsUsed then
  begin
    JObj := TJSONObject.Create;
    JObj.AddPair(TJSONPair.Create(SSabCCardNumber, TJSONString.Create(FNumber)));
    JObj.AddPair(TJSONPair.Create(SSabCCardExpiration, TJSONString.Create(FExpiration)));
    JObj.AddPair(TJSONPair.Create(SSabCCardCvv, TJSONString.Create(FCvv)));
    JObj.AddPair(TJSONPair.Create(SSabCCardPin, TJSONString.Create(FPin)));
    JObj.AddPair(TJSONPair.Create(SSabRecordSecNote, TJSONString.Create(TextToJS(SecNote))));

    Result.AddPair(TJSONPair.Create(SObjAttributes, JObj));
  end;
end;

procedure TSabCreditCard.SaveToList(const List: TStringList);
begin
  inherited;

  if SecureFieldsUsed then
  begin
    List.Add(SSabCCardNumber + ' = ' + FNumber);
    List.Add(SSabCCardExpiration + ' = ' + FExpiration);
    List.Add(SSabCCardCvv + ' = ' + FCvv);
    List.Add(SSabCCardPin + ' = ' + FPin);
  end;
end;

procedure TSabCreditCard.SetExpirationDate(Local: Boolean; const Value: TDateTime);
begin
  try
    FExpiration := DateToISO8601(Value, not Local);
  except
    FExpiration := '';
  end;
end;

procedure TSabCreditCard.WipeStrings;
begin
  inherited;

  TDP.CommonCode.WipeString(FNumber);
  TDP.CommonCode.WipeString(FExpiration);
  TDP.CommonCode.WipeString(FCvv);
  TDP.CommonCode.WipeString(FPin);
end;

end.
