
{*******************************************************}
{                                                       }
{       SAB Object Framework                            }
{       SAB Client                                      }
{                                                       }
{       Copyright (c) 2017-23 TDP                       }
{                                                       }
{****************************************************tk*}

unit SAB.Clients;

{ DEFINE API_1_1_3}
{ DEFINE PREV_SAB_4_17_1} // new SAB 4.17.1 can fileInfo=True without calling get-record

{ DEFINE GET_RECORD_FORM}

{$IFNDEF USE_SSL_0_9}
  {$DEFINE SSL_1_1_1}
{$ENDIF}

interface

uses
  System.SysUtils, System.Classes, WinApi.Windows,
  TDP.IntList,
  blcksock, SAB.RESTsend,
  SAB.Types, SAB.Classes, SAB.Uri, SAB.OpenSSL;

const
  PasswordQualityByEntropy: array[0..5] of Double = (
     0.00, // prilis kratke/too short
     3.00, // velmi slabe/very week
    20.00, // slabe/weak
    35.00, // dobre/good
    40.00, // silne/strong
    50.00  // velmi silne/very strong
    );

type
  TSabClient = class;

  TWhatWipe = (wipPassword, wipAccessCode);
  TWhatWipeSet = set of TWhatWipe;

  TCheckPasswordResult = record
    Len: Integer;
    TimeToCrackOffline: string;
    TimeToCrackOnline: string;
    SecondsToCrackOffline: Int64;
    SecondsToCrackOnline: Int64;
    Entropy: Extended;
    LowerCaseCount: Integer;
    UpperCaseCount: Integer;
    DigitCount: Integer;
    SpecialCharCount: Integer;
    LowerCaseBad, UppercaseBad, DigitsBad, SpecialBad: Boolean;
    EntropyBad: Boolean;

    procedure Init;
    function OK: Boolean;
  end;

  TRequirements = record
    MinLength,
    MinUpper,
    MinLower,
    MinDigits,
    MinSpecials: Integer;
    MinEntropy: Double;
  end;

  TSabVersion = record
    Version, Revision, Build: Integer;
    Encoded: Int64;
  end;

  TLoginGetCertificate = function(const AStream: TMemoryStream): Boolean of object;
  TLoginUserCallback = function(const ABaseURL, ADomain, AUserName: string; out APassword: string): Boolean of object;
  TClientCheckSecondFactor = function(const AClient: TSabClient): Boolean of object;
  TGetAccessCodeCallback = function(out AAcessCode: string; Invalid: Boolean = False): Boolean of object;

  TSabClient = class(TSabUri)
  private
    FHTTP: TRESTsend;

    FBaseURL: string;
    FDomain: string;
    FUserName: string;
    FPassword: string;
    FAccessCode: string;

    FErrorKey: string;
    FRecordId: TSabId;
    FErrorMsg: string;
    FErrorSubj: string;
    FErrorCode: Integer;

    FConnected: Boolean;
    FConnectedByCertificate: Boolean;

    FSabProduct: string;
    FSabApiVersion: string;
    FSabGmtTime: Int64;
    FSabServerTime: string;
    FSabVersion: string;

    FResponseSize: Integer;
    FConnectionTimeout: Integer;
    FAccessCodeValidityTimeoutSec: Integer;
    FSecondFactorRequired: Boolean;
    FHasAccessCode: Boolean;
    FSabUser: Boolean;
    FSabAdmin: Boolean;
    FCfgAdmin: Boolean;
    FSafeBoxUser: Boolean;
    FSafeBoxAdmin: Boolean;
    FAuditor: Boolean;

    FMustConfigureSecondFactor: Boolean;
    FHasFeatureCbtClient: Boolean;
    FHasFeatureFileManager: Boolean;

    FUid: string;

    FLanguage: TSabLanguage;
    FEnterpriseEdition: Boolean;

    FLoginPasswordRequirements: TRequirements;
    FAccessCodeRequirements: TRequirements;
    FVerifyCertificate: Boolean;
    FCertificateFile: string;
    FCertCAFile: string;
    FCertificate: RawByteString;
    FCertCA: RawByteString;
    FCertCABytes: TBytes;
    FImprint: RawByteString;

    FLoginCallbackRuns: Boolean;
    FAcCallbackRuns: Boolean;

    FOnGetCertificate: TLoginGetCertificate;
    FOnUserLogin: TLoginUserCallback;
    FOnGetAccessCode: TGetAccessCodeCallback;
    FOnCheckSecondFactor: TClientCheckSecondFactor;

    FPFX: RawByteString;

    procedure CheckNil(const Obj: TObject);
    procedure CheckId(const Obj: TSabObject);
    procedure CheckAccount(const Obj: TSabObject);
    procedure CheckHTTP;

    function CheckRestFunc(AFuncResult: Boolean; const AResponse: TBytesStream): Boolean;
{$IFDEF INET_INDY}
    function CheckLoggedLocation(const ALocation: string): Boolean;
{$ENDIF}
    function DetectError(const AErrorMessage: string): Boolean; overload;
    function DetectError(const AErrorStream: TBytesStream): Boolean; overload;
    procedure SetError(ACode: Integer; const AMsg: string; const AErrorKey: string = '');
    procedure SetConnected(const Value: Boolean);
    function GetErrorMsg: string;
    procedure SetAccessCode(const Value: string);

    procedure LoadErrorMessages(const FreeTable: Boolean = False);
    procedure SetBaseURLFromInvitation(const AProt, AHost, APort, APath: string);
    procedure SetBaseURL(const Value: string);
    procedure SetConnectionTimeout(const Value: Integer);
    procedure SetVerifyCertificate(const Value: Boolean);
    procedure SetCertificateFile(const Value: string);
    procedure SetCerticate(const Value: RawByteString);
    procedure SetCertCAFile(const Value: string);
    procedure SetCertCA(const Value: RawByteString);
    procedure SetCertCABytes(const Value: TBytes);
    procedure SetPFX(const Value: RawByteString);
    procedure SetImprint(const Value: RawByteString);
    function GetSabVersionEncoded: TSabVersion;
    function GetCertifPubKey: string;

    function HandleLoginFailedError: Boolean;
    function GetAuthenticated: Boolean;

    function RelatedOperationWithDeletedRecords(Remove: Boolean; const List: TSabIdList): Boolean;
  public
    constructor Create; overload;
    constructor Create(const ABaseURL, AUserName, APassword, AAccessCode: string; const ADomain: string = 'SYSTEM'; ALanguage: TSabLanguage = sablngEnglish); overload;
    constructor Create(const ABaseURL: string; ALanguage: TSabLanguage = sablngEnglish); overload;
    destructor Destroy; override;

    class function CheckSSL(const APath: string; const AForceInit: Boolean = False): Boolean;
    class function CheckSSLLib(AWhat: Integer; out LibPath: string): Integer;

    function LoginStatus(out Authenticated: Boolean): Boolean;
    function Status(const DoLoadErrorMessages: Boolean = True): Boolean;
    function Connect(const ACertificate: TStream = nil; const ForceDisconnect: Boolean = True): Boolean;
    procedure Disconnect(const AWipeStrings: Boolean = True);

    procedure ClearError(IncludeErrorKey: Boolean = True);
{$IFDEF CPUX86}
    procedure CancelTraffic;
{$ENDIF}
    procedure RestoreTraffic;

    procedure WipeStrings(const WhatWipe: TWhatWipeSet = [wipPassword, wipAccessCode]);

    function GetAbout(const ADocument: TStringStream): Boolean;
    function GetStations(const AList: TSabList; const AConfigName: string = 'DEFAULT'): Boolean;
    function GetConfigurations(const AList: TSabList): Boolean;
    function GetAgentPassword(const AName, AConfiguration, APlatform, AUser: string; ADate: TDateTime; out APassword: string): Boolean;
    function GetStationPassword(const AName, AConfiguration, APlatform, AUser: string; ADate: TDateTime; out APassword: string): Boolean;
    function GetTicketPassword(const ATicketId, AStationName, APlatform, AUserName: string; out APassword: string): Boolean;
    function GetFileSizeLimit: Int64;
    function GetExternalFileParams(const ABoxId: TSabId; out AKey, AIv, AAad, AFilePath: string): Boolean;

    function Search(const APattern: string; const List: TSearchResultList): Boolean;
    function SearchByNote(const APattern: string; const List: TSabIdList): Boolean;

    function GetRecord(RecId: TSabId; InclSecretFields: Boolean = False; Download: Boolean = False; const AFileMonitor: THookMonitor = nil): TSabRecord;
    function ReadSecretFields(const AObj: TSABObject; const ADownload: Boolean = False; const AFileMonitor: THookMonitor = nil): Boolean;

    function GetRecords(BoxId: TSabId; const List: TSabList; SecretFields: Boolean = False; Download: Boolean = False; GetFileSizes: Boolean = False): Boolean;
    function GetSafeBox(BoxId: TSabId): TSabSafeBox;
    function GetSafeBoxes(BoxId: TSabId; const List: TSabList; ExpandAll: Boolean = True; IncludeGroupBoxes: Boolean = False): Boolean;

    function GetDeletedRecords(const List: TDeletedItemList): Boolean;
    function RemoveDeletedRecord(const Id: TSabId): Boolean;
    function RemoveDeletedRecords(const List: TSabIdList): Boolean;
    function RestoreDeletedRecords(const List: TSabIdList): Boolean; // undelete
    function RestoreDeletedRecord(const Id: TSabId): Boolean;

    function Add(BoxId: TSabId; const SabObj: TSabObject; out Id: TSabId; const ForceAdd: Boolean = True; const FreeObj: Boolean = False; const AFileMonitor: THookMonitor = nil): Boolean;
    function Update(const Obj: TSabObject; const AUploadFileOrChangePasswords: Boolean = False): Boolean;
    function Delete(const Obj: TSabObject): Boolean;

    function UpdatePassword(const Acc: TSabAccountCommon): Boolean;
    function UploadFile(const SabObj: TSabFile; const Stream: TBytesStream): Boolean;
    function DownloadFile(const SabObj: TSabFile; const Stream: TBytesStream; const AFileMonitor: THookMonitor = nil): Boolean;

    function SetWatched(const ABox: TSabSafeBox; const WithRecords: Boolean): Boolean;
    function GetUserConfiguration(const AUserCodeInURL: string = ''; ACheckBase: Boolean = True): Boolean;

    function GeneratePasswordOrAccCode(AType: TGenerationType; const AUserCodeInURL: string; out ANew: string; ACheckBase: Boolean = True): Boolean;
    function CheckPasswordCorrectness(AType: TGenerationType; const AUserCodeInURL: string; const APassword: string; out AResult: TCheckPasswordResult; ACheckBase: Boolean = True): Boolean;
    function SetPasswordToInvitedUser(const AUserCodeInURL, ANewPassword: string; ACheckBase: Boolean = True): Boolean;

    function SetNewAccessCode(const ANewAccessCode: string): Boolean;

//    Registration of Second Factor: Step One
    function RegisterSecondFactor(out ASecretKeyUri: string; const ASecurityKey: string; const AIssuer: string = ''; const AUsername: string = '';  const ADomain: string = ''): Boolean; // Step One
//    Registration of Second Factor: Step Two
    function GetSecondFactorQR(const ASecretKeyUri: string; const APngImage: TStream): Boolean;
//    if ATOTPKey <> '' then method means Registration of Second Factor: Step Three
    function SecondFactor(const ASecurityCode: string; const ATOTPKey: string = ''): Boolean;

    property Connected: Boolean read FConnected write SetConnected;
    property ConnectedByCertificate: Boolean read FConnectedByCertificate;
    property Authenticated: Boolean read GetAuthenticated;
    property HTTP: TRESTsend read FHTTP;

    property BaseURL: string read FBaseURL write SetBaseURL;
    property Domain: string read FDomain write FDomain;
    property UserName: string read FUsername write FUserName;
    property Password: string read FPassword write FPassword;
    property AccessCode: string read FAccessCode write SetAccessCode;

    property ErrorCode: Integer read FErrorCode;
    property ErrorMsg: string read GetErrorMsg;
    property ErrorSubj: string read FErrorSubj;

    property ErrorKey: string read FErrorKey;
    property RecordId: TSabId read FRecordId;

    property SabProduct: string read FSabProduct;
    property SabApiVersion: string read FSabApiVersion;
    property SabGmtTime: Int64 read FSabGmtTime;
    property SabServerTime: string read FSabServerTime;
    property SabVersion: string read FSabVersion;
    property SabVersionEncoded: TSabVersion read GetSabVersionEncoded;

    property VerifyCertificate: Boolean read FVerifyCertificate write SetVerifyCertificate;
    property CertificateFile: string read FCertificateFile write SetCertificateFile;
    property Certificate: RawByteString read FCertificate write SetCerticate;
    property CertCAFile: string read FCertCAFile write SetCertCAFile;
    property CertCA: RawByteString read FCertCA write SetCertCA;
    property CertCABytes: TBytes read FCertCABytes write SetCertCABytes;
    property PFX: RawByteString read FPFX write SetPFX;
    property Imprint: RawByteString read FImprint write SetImprint;

    property ResponseSize: Integer read FResponseSize;
    property ConnectionTimeout: Integer read FConnectionTimeout write SetConnectionTimeout;
    property SecondFactorRequired: Boolean read FSecondFactorRequired;
    property HasAccessCode: Boolean read FHasAccessCode;

    property SabUser: Boolean read FSabUser;
    property SabAdmin: Boolean read FSabAdmin;
    property CfgAdmin: Boolean read FCfgAdmin;
    property SafeBoxUser: Boolean read FSafeBoxUser;
    property SafeBoxAdmin: Boolean read FSafeBoxAdmin;
    property Auditor: Boolean read FAuditor;

    property MustConfigureSecondFactor: Boolean read FMustConfigureSecondFactor;
    property HasFeatureCbtClient: Boolean read FHasFeatureCbtClient;
    property HasFeatureFileManager: Boolean read FHasFeatureFileManager;

    property FileSizeLimit: Int64 read GetFileSizeLimit;
    property Language: TSabLanguage read FLanguage;
    property EnterpriseEdition: Boolean read FEnterpriseEdition;
    property AccessCodeValidityTimoutSec: Integer read FAccessCodeValidityTimeoutSec;

    property Uid: string read FUid;

    property LoginPasswordRequirements: TRequirements read FLoginPasswordRequirements;
    property AccessCodeRequirements: TRequirements read FAccessCodeRequirements;

    property CertifPubKey: string read GetCertifPubKey;

    property OnGetCertificate: TLoginGetCertificate read FOnGetCertificate write FOnGetCertificate;
    property OnUserLogin: TLoginUserCallback read FOnUserLogin write FOnUserLogin;
    property OnCheckSecondFactor: TClientCheckSecondFactor read FOnCheckSecondFactor write FOnCheckSecondFactor;
    property OnGetAccessCode: TGetAccessCodeCallback read FOnGetAccessCode write FOnGetAccessCode;
  end;

function EncodeSabVersion(const Version, Revision, Build: Integer): Int64;

function IsRestResultOk(const AErrorCode: Integer): Boolean;

implementation

uses
{$IFDEF SSL_1_1_1}
  ssl_openssl11, ssl_openssl11_lib,
{$ELSE}
  ssl_openssl, ssl_openssl_lib,
{$ENDIF}
  synautil, httpsend,
  System.Types, System.Contnrs, System.StrUtils, System.JSON, System.NetEncoding, System.Generics.Collections,
  TDP.CommonCode,
  OpenSSLHeadersExt,
  SAB.Consts, SAB.Errors, SAB.Utils;


{ IsRestResultOk }

function IsRestResultOk(const AErrorCode: Integer): Boolean;
begin
  Result := (AErrorCode = 0) or (AErrorCode = HTTP_OK);
end;

{ EncodeSabVersion }

function EncodeSabVersion(const Version, Revision, Build: Integer): Int64;
begin
  Result := Version * 1000 * 1000 + Revision * 1000 + Build;
end;

{ TCheckPasswordResult }

procedure TCheckPasswordResult.Init;
begin
  TimeToCrackOffline := '';
  TimeToCrackOnline := '';
  FillChar(Self, SizeOf(Self), 0);
end;

function TCheckPasswordResult.OK: Boolean;
begin
  Result := not (LowerCaseBad or UppercaseBad or DigitsBad or SpecialBad or EntropyBad);
end;

{ TSabClient }

constructor TSabClient.Create;
begin
  Create('');
end;

constructor TSabClient.Create(const ABaseURL, AUserName, APassword, AAccessCode, ADomain: string; ALanguage: TSabLanguage);
begin
  inherited Create;

  FBaseURL := PrepareURL(ABaseURL);
  FDomain := ADomain;
  FUserName := AUserName;
  FPassword := APassword;
  FAccessCode := AAccessCode;

  FLanguage := ALanguage;
  FAccessCodeValidityTimeoutSec := -1;
  FVerifyCertificate := False;
end;

constructor TSabClient.Create(const ABaseURL: string; ALanguage: TSabLanguage);
begin
  Create(ABaseURL, '', '', '');
  FLanguage := ALanguage;
end;

destructor TSabClient.Destroy;
begin
  WipeStrings;
  Connected := False;
  FHttp.Free;

  if IsSSLLoaded then
    DestroySSLInterface;

  inherited;
end;

function TSabClient.Add(BoxId: TSabId; const SabObj: TSabObject; out Id: TSabId; const ForceAdd, FreeObj: Boolean; const AFileMonitor: THookMonitor): Boolean;
var
  URI: string;
  Src, Stream: TBytesStream;
  JObj: TJSONObject;
  Response, UTF8Bytes: TBytesStream;
  JMain: TJSONObject;
  Val: TJSONNumber;
  Len: Int64;
begin
  Result := False;

  CheckHTTP;

  CheckNil(SabObj);
  Id := SabObj.Id;
  if not ForceAdd and (SabObj.Id > 0) then
  begin
    Result := Update(SabObj);
    Exit;
  end;

  SabObj.SetId(0);

  ClearError;
  case SabObj.ObjType of
    sabSafeBox,
    sabSafeBoxGroup:
      URI := BaseURL + URI_AddBox(BoxId);
    sabExternalFile,
    sabAccount,
    sabAccountSec,
    sabCertificate,
    sabCreditCard,
    sabStation,
    sabConfiguration:
      URI := BaseURL + URI_AddRecord(BoxId);
    sabFile:
      URI := BaseURL + URI_AddRecord(BoxId);
  else
    URI := '';
  end;

  Stream := TBytesStream.Create;
  Src := TBytesStream.Create;
  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
  JMain := TJSONObject.Create;
  JObj := SabObj.SaveTo(False);
  try
    Src.Size := JObj.EstimatedByteSize;

    HTTP.AccessCode := AccessCode;

    Len := JObj.ToBytes(Src.Bytes, 0);
    Src.Size := Len;
    Src.Position := 0;
    try
      while not CheckRestFunc(HTTP.Post(URI, Src, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      // obtained Id
      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);
      JMain.Parse(UTF8Bytes.Bytes, 0);

      Val := JMain.Values[SSabObjId] as TJSONNumber;
      if Assigned(Val) then
      begin
{$IFDEF LONG_ID}
        SabObj.SetId(Val.AsInt64);
{$ELSE}
        SabObj.SetId(Val.AsInt);
{$ENDIF}
        Id := SabObj.Id;
      end;

      if (SabObj.ObjType in [sabFile, sabCertificate]) and Assigned(TSabFile(SabObj).Stream) and
        (TSabFile(SabObj).Stream.Size > 0) then
      begin
        if Assigned(AFileMonitor) then
          HTTP.Sock.OnMonitor := AFileMonitor;
        Result := UploadFile(TSabFile(SabObj), TSabFile(SabObj).Stream);
      end else
        Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    HTTP.Sock.OnMonitor := nil;
    if FreeObj then
      SabObj.Free;
    Src.Free;
    Stream.Free;
    Response.Free;
    UTF8Bytes.Free;
    JObj.Free;
    JMain.Free;
  end;
end;

{$IFDEF CPUX86}
procedure TSabClient.CancelTraffic;
begin
  if Assigned(HTTP) then
    HTTP.CancelTraffic;
end;
{$ENDIF}

procedure TSabClient.CheckAccount(const Obj: TSabObject);
begin
  if not Assigned(Obj) then
    raise Exception.Create('Object is nil');

  if not (Obj.ObjType in [sabAccount, sabAccountSec]) then
    raise Exception.Create('Object must be Account or Secret Account');
end;

procedure TSabClient.CheckHTTP;
begin
  if not Assigned(HTTP) then
  begin
    FHTTP := TRESTsend.Create(Self, AccessCode);
    if FConnectionTimeout <> 0 then
      FHTTP.Sock.ConnectionTimeout := FConnectionTimeout;
  end;
  // SetError(HTTP_UNAUTHORIZED, SLoggedOutMessage, SLoginFailed);
end;

procedure TSabClient.CheckId(const Obj: TSabObject);
begin
  if not Assigned(Obj) then
    raise Exception.Create('Object is nil');

  if Obj.Id <= 0 then
    raise Exception.Create('Object.Id = 0');
end;

{$IFDEF INET_INDY}
function TSabClient.CheckLoggedLocation(const ALocation: string): Boolean;
begin
  Result := {ContainsText(ALocation, FPureURL) and}
    not (ContainsText(ALocation, 'failed') and
      ContainsText(ALocation, 'static/login.page'));
end;
{$ENDIF}

procedure TSabClient.CheckNil(const Obj: TObject);
begin
  if not Assigned(Obj) then
    raise Exception.Create('Object is nil');
end;

function TSabClient.CheckPasswordCorrectness(AType: TGenerationType; const AUserCodeInURL, APassword: string; out AResult: TCheckPasswordResult; ACheckBase: Boolean): Boolean;
var
  URI: string;
  Response, UTF8Bytes: TBytesStream;
  List: TStringList;
  Params: TStringDynArray;
  Msg: string;
  E: Integer;
  Key: string;
  JObj: TJSONObject;
  Val: TJSONString;
  Flag: TJSONBool;
  Num: TJSONNumber;
  URL, Prot, User, Pass, Host, Port, Path, Param, Item: string;
  Tkn: TStringTokenizer;
  I: Integer;
begin
  Result := False;
  AResult.Init;

  if Connected and (AUserCodeInURL <> '') then
    Exit;

  CheckHTTP;

  List := TStringList.Create;
  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
  JObj := TJSONObject.Create;
  Tkn := TStringTokenizer.Create('&');
  try
    List.Sorted := False;

    LoadErrorMessages;

    URL := TSabClient.ParseURL(AUserCodeInURL, Prot, User, Pass, Host, Port, Path, Param, False);

    if ACheckBase and (BaseURL <> '') and (AUserCodeInURL <> '') then
    begin
      if not ContainsText(BaseUrl, Host) then
      begin
        FErrorCode := -1;
        Exit;
      end;
    end;

    if BaseURL = '' then
      SetBaseURLFromInvitation(Prot, Host, Port, Path);

    try
      Tkn.Str := Param;
      for I := 0 to Tkn.Count - 1 do
      begin
        Item := Tkn[I];
        if Pos('code=', Item) > 0 then
        begin
          Param := Copy(Item, Pos('code=', Item) + 5, Length(Item));
          Break;
        end;
      end;
      URI := BaseURL + URI_CheckPswdValidity;

      if Param <> '' then
        List.Add('code=' + Param);
      List.Add('password=' + APassword);
      if AType = gtPassword then
        List.Add('type=login')
      else
        List.Add('type=accessCode');

      while not CheckRestFunc(HTTP.Options(URI, List, Response, True), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);

      JObj.Parse(UTF8Bytes.Bytes, 0);

      Val := JObj.Values[SPswdChkTimeToCrackOffline] as TJSONString;
      if Assigned(Val) then
      begin
        Item := Val.Value;
        if ParseErrorMessage(Item, Key, Params) then
          if GetSabError(Key, Params, Msg, E) then
            AResult.TimeToCrackOffline := Msg;
      end;

      Val := JObj.Values[SPswdChkTimeToCrackOnline] as TJSONString;
      if Assigned(Val) then
      begin
        Item := Val.Value;
        if ParseErrorMessage(Item, Key, Params) then
          if GetSabError(Key, Params, Msg, E) then
            AResult.TimeToCrackOnline := Msg;
      end;

      Num := JObj.Values[SPswdChkSecondsToCrackOff] as TJSONNumber;
      if Assigned(Num) then
      begin
        try
          AResult.SecondsToCrackOffline := Num.AsInt64;
        except
          AResult.SecondsToCrackOffline := MAX_INT64;
        end;
      end;

      Num := JObj.Values[SPswdChkSecondsToCrackOn] as TJSONNumber;
      if Assigned(Num) then
      begin
        try
          AResult.SecondsToCrackOnline := Num.AsInt64;
        except
          AResult.SecondsToCrackOnline := MAX_INT64;
        end;
      end;

      Num := JObj.Values[SPswdChkEntropy] as TJSONNumber;
      if Assigned(Num) then
      begin
        try
          AResult.Entropy := Num.AsDouble;
        except
          AResult.Entropy := 999999999999999999999.0;
        end;
      end;

      Num := JObj.Values[SPswdChkLowercase] as TJSONNumber;
      if Assigned(Num) then
        AResult.LowerCaseCount := Num.AsInt;

      Num := JObj.Values[SPswdChkUppercase] as TJSONNumber;
      if Assigned(Num) then
        AResult.UpperCaseCount := Num.AsInt;

      Num := JObj.Values[SPswdChkDigits] as TJSONNumber;
      if Assigned(Num) then
        AResult.DigitCount := Num.AsInt;

      Num := JObj.Values[SPswdChkSpecial] as TJSONNumber;
      if Assigned(Num) then
        AResult.SpecialCharCount := Num.AsInt;

      Num := JObj.Values[SPswdChkLength] as TJSONNumber;
      if Assigned(Num) then
        AResult.Len := Num.AsInt;

      Flag := JObj.Values[SPswdChkLowerCaseBad] as TJSONBool;
      if Assigned(Flag) then
        AResult.LowerCaseBad := Flag.AsBoolean;

      Flag := JObj.Values[SPswdChkUpperCaseBad] as TJSONBool;
      if Assigned(Flag) then
        AResult.UppercaseBad := Flag.AsBoolean;

      Flag := JObj.Values[SPswdChkDigitsBad] as TJSONBool;
      if Assigned(Flag) then
        AResult.DigitsBad := Flag.AsBoolean;

      Flag := JObj.Values[SPswdChkSpecialbad] as TJSONBool;
      if Assigned(Flag) then
        AResult.SpecialBad := Flag.AsBoolean;

      Flag := JObj.Values[SPswdChkEntropyBad] as TJSONBool;
      if Assigned(Flag) then
        AResult.EntropyBad := Flag.AsBoolean;

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    List.Free;
    Response.Free;
    UTF8Bytes.Free;
    JObj.Free;
    Tkn.Free;
  end;
end;

class function TSabClient.CheckSSL(const APath: string; const AForceInit: Boolean): Boolean;
const
  DLLSSLNameStr =  'ssleay32.dll';

{$IFDEF SSL_1_1_1}
{$ELSE}
  DLLSSLName2Str = 'libssl32.dll';
{$ENDIF}
  DLLUtilNameStr = 'libeay32.dll';
var
  Path: string;
begin
  if IsSSLLoaded and AForceInit then
    DestroySSLInterface;

  if not IsSSLLoaded then
  begin
    if APath = '' then
      Path := IncludeTrailingPathDelimiter(ExtractFilePath(GetModuleName(0)))
    else
      Path := IncludeTrailingPathDelimiter(APath);

    DLLUtilName := Path + DLLUtilNameStr;
    DLLSSLName := Path + DLLSSLNameStr;
{$IFDEF SSL_1_1_1}
    Result := ssl_openssl11_lib.InitSSLInterface;
{$ELSE}
    Result := ssl_openssl_lib.InitSSLInterface;
    DLLSSLName2 := Path + DLLSSLName2Str;
{$ENDIF}

    if Result then
    begin
      OpenSSLHeadersExt.Load {$IFDEF SSL_1_1_1}(True){$ENDIF};

      SSLImplementation := TSabOpenSSL;
    end;
  end else
    Result := True;
end;

class function TSabClient.CheckSSLLib(AWhat: Integer; out LibPath: string): Integer;
var
  Handle: THandle;
begin
  Result := 0;
  SetLastError(ERROR_SUCCESS);
  case AWhat of
    1:
      begin
        Handle := LoadLibrary(PChar(DLLUtilName));
        LibPath := DLLUtilName;
      end;
    2:
      begin
        Handle := LoadLibrary(PChar(DLLSSLName));
        LibPath := DLLSSLName;
      end
  else
{$IFDEF SSL_1_1_1}
    Handle := LoadLibrary(PChar(DLLSSLName));
    LibPath := DLLSSLName;
{$ELSE}
    Handle := LoadLibrary(PChar(DLLSSLName2));
    LibPath := DLLSSLName2;
{$ENDIF}
  end;

  if Handle = 0 then
  begin
    Result := GetLastError;
  end else
    FreeLibrary(Handle);
end;

function TSabClient.CheckRestFunc(AFuncResult: Boolean; const AResponse: TBytesStream): Boolean;
var
  ResultString: string;
begin
  Result := False;
  if AFuncResult then
  begin
    if HTTP.ResultCode <> 200 then
    begin
      if not DetectError(AResponse) then
      begin
        if GetHttpError(HTTP.ResultCode, ResultString, FLanguage) then
          SetError(HTTP.ResultCode, ResultString)
        else
          SetError(HTTP.ResultCode, HTTP.ResultString);
      end else
        if (ErrorKey = SInvalidAccessCodeKey) or (ErrorKey = SAccessCodeBannedKey) then
          WipeStrings([wipAccessCode]);
    end else
      Result := True;
  end else
    SetError(HTTP.Sock.LastError, HTTP.Sock.LastErrorDesc);
end;

procedure TSabClient.ClearError(IncludeErrorKey: Boolean);
var
  SaveErrorKey: string;
begin
  FResponseSize := 0;

  SaveErrorKey := FErrorKey;
  try
    SetError(0, '');
  finally
    if not IncludeErrorKey then
      FErrorKey := SaveErrorKey;
  end;

  RestoreTraffic;
end;

function TSabClient.Connect(const ACertificate: TStream; const ForceDisconnect: Boolean): Boolean;
var
  URI, Login: string;
  Response, UTF8Bytes: TBytesStream;
  EncodedCert: TStringStream;
  Stream: TStringStream;
  List, ErrList: TStringList;
  Enc: TBase64Encoding;
  JObj: TJSONObject;
  Flag: TJSONBool;
  Val: TJSONNumber;
begin
  Result := False;
  if FConnected then
    Exit(True);

  CheckHTTP;

  Stream := TStringStream.Create('', TEncoding.UTF8);
  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create;
  List := TStringList.Create;
  ErrList := TStringList.Create;
  EncodedCert := TStringStream.Create;
  JObj := TJSONObject.Create;
  Enc := TBase64Encoding.Create(0);
  try
    ClearError;
    try
      List.Sorted := False;
      ErrList.Sorted := False;

      LoadErrorMessages;

      if Assigned(ACertificate) then
      begin
        Enc.Encode(ACertificate, EncodedCert);
        List.Add('authCert=' + EncodedCert.DataString);

        URI := FBaseURL + URI_CertLogin;
      end else
      begin
        if FDomain <> '' then
        begin
          Login := FDomain + '\' + FUsername;
          List.Add('username=' + FUsername);
          List.Add('domain=' + FDomain);
        end else
          Login := FUsername;

        List.Add('login=' + Login);
        List.Add('password=' + FPassword);

        URI := FBaseURL + URI_Login;
      end;

      Result := CheckRestFunc(HTTP.Post(URI, List, Response), Response);

      if Result then
      begin
        FResponseSize := Response.Size;
        Response.Position := 0;
        UTF8Bytes.CopyFrom(Response, Response.Size);
        JObj.Parse(UTF8Bytes.Bytes, 0);

        Flag := JObj.Values[SSabStatus2ndFactor] as TJSONBool;
        if Assigned(Flag) then
          FSecondFactorRequired := Flag.AsBoolean
        else
          FSecondFactorRequired := False;

        Flag := JObj.Values[SMustConfigureSecondFactor] as TJSONBool;
        if Assigned(Flag) then
          FMustConfigureSecondFactor := Flag.AsBoolean
        else
          FMustConfigureSecondFactor := False;

        Flag := JObj.Values[SHasAccessCode] as TJSONBool;
        if Assigned(Flag) then
          FHasAccessCode := Flag.AsBoolean
        else
          FHasAccessCode := False;

        Val := JObj.Values[SAccessCodeTimeout] as TJSONNumber;
        if Assigned(Val) then
          FAccessCodeValidityTimeoutSec := Val.AsInt;
      end;

      if Result then
        TDP.CommonCode.WipeString(FPassword);

      FConnected := Result;
      if Result and Assigned(ACertificate) then
        FConnectedByCertificate := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    if not Result and ForceDisconnect then
    begin
      if Connected then
        Disconnect;
    end;
    Stream.Free;
    Response.Free;
    UTF8Bytes.Free;
    List.Free;
    ErrList.Free;
    EncodedCert.Free;
    Enc.Free;
    JObj.Free;
  end;
end;

function TSabClient.Delete(const Obj: TSabObject): Boolean;
var
  URI: string;
  Response: TBytesStream;
begin
  Result := False;

  CheckHTTP;

  CheckId(Obj);

  ClearError;
  Response := TBytesStream.Create;
  try
    try
      if Obj.ObjType in AllSabBoxes then
        URI := BaseURL + URI_ReadDeleteSafeBox(Obj.Id)
      else
        URI := BaseURL + URI_DeleteRecord(Obj.Id);

      while not CheckRestFunc(HTTP.Delete(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
  end;
end;

function TSabClient.DetectError(const AErrorStream: TBytesStream): Boolean;
var
  Str: TStringStream;
begin
  Str := TStringStream.Create(TEncoding.UTF8.GetPreamble + AErrorStream.Bytes);
  try
    Result := DetectError(Str.DataString);
  finally
    Str.Free;
  end;
end;

function TSabClient.DetectError(const AErrorMessage: string): Boolean;
var
  E: Integer;
  S: string;
  Bytes: TBytes;
  JMain: TJSONObject;
  JPair: TJSONPair;
  JArray: TJSONArray;
  JObj: TJSONObject;
  JValue: TJSONValue;
  Severity, Msg, Msg2: string;
  Params: TStringDynArray;
begin
  Result := False;
  FErrorKey := '';

  if StartsText('err,', AErrorMessage) then
  begin
    FErrorCode := HTTP.ResultCode;
    FErrorMsg := Copy(AErrorMessage, 5, Length(AErrorMessage));
    FErrorSubj := '';
    Result := True;
  end else
  begin
    FErrorCode := HTTP.ResultCode;

    JMain := TJSONObject.Create;
    try
      Bytes := StringToBytes(AErrorMessage);
      JMain.Parse(Bytes, 0);

      JPair := JMain.Get('msg');
      if Assigned(JPair) and (JPair.JsonValue is TJSONArray) then
      begin
        JArray := TJSONArray(JPair.JsonValue);

        JValue := JArray.Items[0];
        if Assigned(JValue) and (JValue is TJSONObject) then
        begin
          JObj := TJSONObject(JValue);

          JPair := JObj.Get('severity');
          Severity := TJSONString(JPair.JsonValue).Value;
          JPair := JObj.Get('message');
          Msg := TJSONString(JPair.JsonValue).Value;
          JPair := JObj.Get('translationKey');
          Msg2 := TJSONString(JPair.JsonValue).Value;
          JPair := JObj.Get('recordId');
          if Assigned(JPair) then
          begin
            if not (JPair.JsonValue is TJSONNull) then
{$IFDEF LONG_ID}
              FRecordId := TJSONNumber(JPair.JsonValue).AsInt64
{$ELSE}
              FRecordId := TJSONNumber(JPair.JsonValue).AsInt
{$ENDIF}
            else
              FRecordId := 0;
          end else
            FRecordId := 0;

          if SameText(Severity, 'ERROR') then
          begin
            if ParseErrorMessage(Msg, FErrorKey, Params) then
            begin
              if not GetSabError(FErrorKey, Params, FErrorMsg, E) then
              begin
                if GetHttpError(FErrorCode, S, FLanguage) then
                  FErrorMsg := System.SysUtils.Format('%s (%s)', [S, FErrorMsg])
                else
                  FErrorMsg := System.SysUtils.Format('%d %s', [FErrorCode, FErrorMsg]);
              end;
            end else
            if ParseErrorMessage(Msg2, FErrorKey, Params) then
            begin
              if not GetSabError(FErrorKey, Params, FErrorMsg, E) then
              begin
                if GetHttpError(FErrorCode, S, FLanguage) then
                  FErrorMsg := System.SysUtils.Format('%s (%s)', [S, FErrorMsg])
                else
                  FErrorMsg := System.SysUtils.Format('%d %s', [FErrorCode, FErrorMsg]);
              end;
            end else
              if GetHttpError(FErrorCode, S, FLanguage) then
                FErrorMsg := System.SysUtils.Format('%s (%s)', [S, Msg])
              else
                FErrorMsg := System.SysUtils.Format('%d %s', [FErrorCode, Msg]);
          end else
            FErrorMsg := System.SysUtils.Format('%d %s', [FErrorCode, AErrorMessage]);
        end else
          FErrorMsg := System.SysUtils.Format('%d %s', [FErrorCode, AErrorMessage]);

        Result := True;
      end;
    finally
      JMain.Free;
      SetLength(Bytes, 0);
    end;
  end;
end;

procedure TSabClient.Disconnect(const AWipeStrings: Boolean);
begin
  FConnectedByCertificate := False;
  ClearError;
  if Assigned(FHTTP) then
  begin
    FreeAndNil(FHTTP);
    FConnected := False;
    if AWipeStrings then
      WipeStrings;
  end;
end;

function TSabClient.DownloadFile(const SabObj: TSabFile; const Stream: TBytesStream; const AFileMonitor: THookMonitor): Boolean;
var
  URI: string;
  Response: TBytesStream;
  List: TStringList;
begin
  Result := False;

  CheckHTTP;

  CheckId(SabObj);

  if not (Assigned(Stream) or Assigned(SabObj.Stream)) then
    raise Exception.Create('Stream not assigned');

  Response := TBytesStream.Create;
  List := TStringList.Create;
  try
    ClearError;
    try
      URI := BaseURL + URI_DownloadFileFromRecord(SabObj.Id);
      List.Add('SAB-Access-Code=' + AccessCode);

      if Assigned(AFileMonitor) then
        HTTP.Sock.OnMonitor := AFileMonitor;

      while not CheckRestFunc(HTTP.GetFile(URI, List, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Stream.CopyFrom(Response, -1);
      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    HTTP.Sock.OnMonitor := nil;
    List.Free;
    Response.Free;
  end;
end;

function TSabClient.GeneratePasswordOrAccCode(AType: TGenerationType; const AUserCodeInURL: string; out ANew: string; ACheckBase: Boolean): Boolean;
var
  URI: string;
  Response, UTF8Bytes: TBytesStream;
  JMain: TJSONObject;
  Val: TJSONString;
  URL, Prot, User, Pass, Host, Port, Path, Param, Item: string;
  Tkn: TStringTokenizer;
  I: Integer;
begin
  Result := False;
  if Connected and (AUserCodeInURL <> '') then
    Exit;

  CheckHTTP;

  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
  JMain := TJSONObject.Create;
  Tkn := TStringTokenizer.Create('&');
  try
    LoadErrorMessages;

    URL := Self.ParseURL(AUserCodeInURL, Prot, User, Pass, Host, Port, Path, Param, False);

    if ACheckBase and (BaseURL <> '') and (AUserCodeInURL <> '') then
    begin
      if not ContainsText(BaseUrl, Host) then
      begin
        FErrorCode := -1;
        Exit;
      end;
    end;

    if BaseURL = '' then
      SetBaseURLFromInvitation(Prot, Host, Port, Path);

    try
      Tkn.Str := Param;
      for I := 0 to Tkn.Count - 1 do
      begin
        Item := Tkn[I];
        if Pos('code=', Item) > 0 then
        begin
          Param := Copy(Item, Pos('code=', Item) + 5, Length(Item));
          Break;
        end;
      end;
      URI := BaseURL + URI_GeneratePswdOrAC(AType, Param);

      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);
      JMain.Parse(UTF8Bytes.Bytes, 0);

      Val := JMain.Values[SSabUserPassword] as TJSONString;
      if Assigned(Val) then
        ANew := Val.Value;

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
    UTF8Bytes.Free;
    JMain.Free;
    Tkn.Free;
  end;
end;

function TSabClient.GetAbout(const ADocument: TStringStream): Boolean;
var
  HTTP: TRESTsend;
  URI: string;
begin
  Result := False;
  if not Assigned(ADocument) then
    Exit;

  HTTP := TRESTsend.Create(Self);
  try
    URI := BaseURL + URI_About;
    if not HTTP.GetPage(URI) then
      Exit;

    ADocument.CopyFrom(HTTP.Document, HTTP.Document.Size);

    Result := True;
  finally
    HTTP.Free;
  end;
end;

function TSabClient.GetAgentPassword(const AName, AConfiguration, APlatform, AUser: string; ADate: TDateTime; out APassword: string): Boolean;
var
  URI: string;
  Response: TBytesStream;
  UTF8Bytes: TBytesStream;
  JMain: TJSONObject;
  JPair: TJSONPair;
begin
  Result := False;

  CheckHTTP;

  APassword := '';

  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
  JMain := TJSONObject.Create;
  try
    ClearError;
    try
      URI := BaseURL + URI_GetAgentPassword(AName, AConfiguration, APlatform, AUser, ADate);
      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);

      JMain.Parse(UTF8Bytes.Bytes, 0);

      JPair := JMain.Get(SSabAccountPassword);
      if Assigned(JPair) then
        APassword := TSabObject.JSToStr(JPair.JsonValue.Value);

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
    UTF8Bytes.Free;
  end;
end;

function TSabClient.GetAuthenticated: Boolean;
var
  IsAuthenticated: Boolean;
begin
  if Connected then
  begin
    Result := LoginStatus(IsAuthenticated);
    Result := Result and IsAuthenticated;
  end else
    Result := False;
end;

function TSabClient.GetCertifPubKey: string;
var
  Cert: PX509;
  Name: PX509_NAME;
begin
  Result := '';
  if (BaseURL <> '') and Assigned(HTTP) and Assigned(HTTP.Sock) and Assigned(HTTP.Sock.SSL) and (HTTP.Sock.SSL is TSSLOpenSSL) and Assigned(TSSLOPenSSL(HTTP.Sock.SSL).Ssl)  then
  begin
    Cert := SslGetPeerCertificate(TSSLOPenSSL(HTTP.Sock.SSL).Ssl);
    if Assigned(Cert) then
    begin
      Name := X509GetSubjectName(Cert);
      if Assigned(PX509_NAME_ENTRY(Name).Value) then
        Result := string(PX509_NAME_ENTRY(Name).Value);
    end;
  end;
end;

function TSabClient.GetConfigurations(const AList: TSabList): Boolean;
var
  URI: string;
  Response: TBytesStream;
begin
  Result := False;

  CheckHTTP;

  AList.Clear;

  Response := TBytesStream.Create;
  try
    ClearError;
    try
      URI := BaseURL + URI_GetConfigurations;
      while not CheckRestFunc(HTTP.Options(URI, nil, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      TSabObject.AddObjects(AList, Response, SSabConfigs_Sign);

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
  end;
end;

function TSabClient.GetDeletedRecords(const List: TDeletedItemList): Boolean;
var
  URI: string;
  Response, UTF8Bytes: TBytesStream;
  JObj: TJSONObject;
  JPair: TJSONPair;

  procedure GetResultList(const JArray: TJSONArray; const AList: TDeletedItemList);
  var
    I: Integer;
    JValue: TJSONValue;
    Res: TDeletedItem;
  begin
    for I := 0 to JArray.Count - 1 do
    begin
      JValue := JArray.Items[I];
      if Assigned(JValue) and (JValue is TJSONObject) then
      begin
        Res := TDeletedItem.CreateObject(TJSONObject(JValue));
        AList.Add(Res)
      end;
    end;
  end;

begin
  Result := False;

  CheckHTTP;

  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
  JObj := TJSONObject.Create;
  try
    ClearError;
    try
      List.Clear;

      URI := BaseURL + URI_ReadRemoveRestoreDeletedRecords;
      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);

      JObj.Parse(UTF8Bytes.Bytes, 0);
      JPair := JObj.Get(SSabItems_Sign);
      if Assigned(JPair) and (JPair.JsonValue is TJSONArray) then
        GetResultList(TJSONArray(JPair.JsonValue), List);

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
    UTF8Bytes.Free;
    JObj.Free;
  end;
end;

function TSabClient.GetErrorMsg: string;
begin
  Result := FErrorMsg;
  if (Result <> '') and (FErrorSubj <> '') then
    Result := Result + ' ('+ FErrorSubj + ')';
end;

function TSabClient.GetExternalFileParams(const ABoxId: TSabId; out AKey, AIv, AAad, AFilePath: string): Boolean;
var
  URI: string;
  Response: TBytesStream;
  UTF8Bytes: TBytesStream;
  JMain: TJSONObject;
  JPair: TJSONPair;
begin
  Result := False;

  CheckHTTP;

  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
  JMain := TJSONObject.Create;
  try
    ClearError;
    try
      URI := BaseURL + self.URI_GetExternalFileParams(ABoxId);
      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);

      JMain.Parse(UTF8Bytes.Bytes, 0);

      JPair := JMain.Get(SSabExtFileParamPath);
      if Assigned(JPair) then
      begin
        AFilePath := StringReplace(TSabObject.JSToStr(JPair.JsonValue.Value), '/', '\', [rfReplaceAll]);
      end else
        Exit;

      JPair := JMain.Get(SSabExtFileParamKey);
      if Assigned(JPair) then
        AKey := TSabObject.JSToStr(JPair.JsonValue.Value)
      else
        Exit;

      JPair := JMain.Get(SSabExtFileParamIV);
      if Assigned(JPair) then
        AIv := TSabObject.JSToStr(JPair.JsonValue.Value)
      else
        Exit;

      JPair := JMain.Get(SSabExtFileParamAAD);
      if Assigned(JPair) then
        AAad  := TSabObject.JSToStr(JPair.JsonValue.Value)
      else
        Exit;

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
    UTF8Bytes.Free;
  end;
end;

function TSabClient.GetFileSizeLimit: Int64;
var
  URI: string;
  Response, UTF8Bytes: TBytesStream;
  JObj: TJSONObject;
  Num: TJSONNumber;
begin
  Result := -1;

  CheckHTTP;

  if FBaseURL <> '' then
  begin
    Response := TBytesStream.Create;
    UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
    JObj := TJSONObject.Create;
    try
      URI := BaseURL + URI_GetFileSizeLimit;
      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);

      JObj.Parse(UTF8Bytes.Bytes, 0);

      Num := JObj.Values[SSabFileSizeLimit] as TJSONNumber;
      if Assigned(Num) then
      begin
        Result := Num.AsInt64;
      end;
    finally
      Response.Free;
      Utf8Bytes.Free;
      JObj.Free;
    end;
  end;
end;

// if SecretFields==True, load fields and passwords together
function TSabClient.GetRecord(RecId: TSabId; InclSecretFields, Download: Boolean; const AFileMonitor: THookMonitor): TSabRecord;
var
  URI: string;
  Res: TSabObject;
  Response: TBytesStream;
begin
  Result := nil;

  CheckHTTP;

  Res := nil;
  Response := TBytesStream.Create;
  try
    ClearError;
    try
      URI := BaseURL + URI_ReadRecord(RecId);
      while not CheckRestFunc(HTTP.Get(URI, Response, False), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Res := TSabObject.CreateObject(Response, SSabRecord_Sign) as TSabRecord;
      if Assigned(Res) and (InclSecretFields or Download) then
      begin
        if not ReadSecretFields(Res, Download) then
          Exit;
      end;
      Result := Res as TSabRecord;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    if not Assigned(Result) then
    begin
      Res.Free;
    end;
    Response.Free;
  end;
end;

function TSabClient.GetRecords(BoxId: TSabId; const List: TSabList; SecretFields, Download, GetFileSizes: Boolean): Boolean;
var
  URI: string;
  Response: TBytesStream;
{$IFDEF PREV_SAB_4_17_1}
  SABObj: TSABObject;
  I: Integer;
{$ENDIF}
begin
  Result := False;

  CheckHTTP;

  Response := TBytesStream.Create;
  try
    ClearError;
    try
      List.Clear;

      URI := BaseURL + URI_ReadRecords(BoxId, GetFileSizes);
      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      TSabObject.AddObjects(List, Response, SSabRecords_Sign);

{$IFDEF PREV_SAB_4_17_1}
      if SecretFields or Download then
      begin
        for I := 0 to List.Count - 1 do
        begin
          SABObj := List[I];
          if not ReadSecretFields(SABObj, Download) then
            Exit;
        end;
      end;
{$ENDIF}

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
  end;
end;

function TSabClient.GetSabVersionEncoded: TSabVersion;
var
  S: string;
  Tkn: TStringTokenizer;
  Build: Integer;
  I: Integer;
begin
  Result.Version := 1;
  Result.Revision := 0;
  Result.Build := 0;

  if SabVersion <> '' then
  begin
    Tkn := TStringTokenizer.Create('.');
    try
      Tkn.Str := SabVersion;
      if Tkn.count >= 3 then
      begin
        Result.Version := Tkn.Numbers[0];
        Result.Revision := Tkn.Numbers[1];

        S := Tkn[2];
        if not TryStrToInt(S, Build) then
        begin
          I := 1;
          while (I <= Length(S)) and CharInSet(S[I], ['0'..'9']) do
          begin
            Inc(I);
          end;

          Result.Build := StrToIntDef(Copy(S, 1, I - 1), 0);
        end else
          Result.Build := Build;
      end;
    finally
      Tkn.Free;
    end;
  end;

  Result.Encoded := EncodeSabVersion(Result.Version, Result.Revision, Result.Build);
end;

function TSabClient.GetSafeBox(BoxId: TSabId): TSabSafeBox;
var
  URI: string;
  Response: TBytesStream;
begin
  Result := nil;

  CheckHTTP;

  Response := TBytesStream.Create;
  try
    ClearError;
    try
      URI := BaseURL + URI_ReadSafeBoxWithExtPath(BoxId);
      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Result := TSabObject.CreateObject(Response, SSabSafeBox_Sign) as TSabSafeBox;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
  end;
end;

function TSabClient.GetSafeBoxes(BoxId: TSabId; const List: TSabList; ExpandAll, IncludeGroupBoxes: Boolean): Boolean;
var
  URI: string;
  Response: TBytesStream;
  I: Integer;
begin
  Result := False;

  CheckHTTP;

  Response := TBytesStream.Create;
  try
    ClearError;
    try
      List.Clear;

      URI := BaseURL + URI_ReadSafeBoxes(BoxId);
      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      TSabObject.AddObjects(List, Response, SSabSafeBoxes_Sign);

      I := 0;
      while I < List.Count do
      begin
        if List[I].Id <> BoxId then
        begin
          if List[I].ObjType = sabSafeBoxGroup then
          begin
            if ExpandAll then
            begin
              URI := BaseURL + URI_ReadSafeBoxes(List[I].Id);
              while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
              begin
                if not HandleLoginFailedError then
                  Exit;
              end;

              TSabObject.AddObjects(List, Response, SSabSafeBoxes_Sign, False);
              Inc(I);
            end else
            // if there is no ExpandAll, we don't need groupboxes unless IncludeGroupBoxes is set
            if IncludeGroupBoxes then
              Inc(I)
            else
              List.Delete(I);
          end else
            Inc(I);
        end else
          List.Delete(I);
      end;

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
  end;
end;

function TSabClient.GetSecondFactorQR(const ASecretKeyUri: string; const APngImage: TStream): Boolean;
var
  URI: string;
  Response: TBytesStream;
begin
  Result := False;

  CheckHTTP;

  if not Assigned(APngImage) then
    Exit;

  Response := TBytesStream.Create;
  try
    ClearError;
    try
      URI := BaseURL + URI_GetSecondFactorQrImage(EncodeGetQrTriplet(ASecretKeyUri));
      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      APngImage.Position := 0;
      APngImage.CopyFrom(Response, -1);

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
  end;
end;

function TSabClient.GetStationPassword(const AName, AConfiguration, APlatform, AUser: string; ADate: TDateTime; out APassword: string): Boolean;
var
  URI: string;
  Response: TBytesStream;
  UTF8Bytes: TBytesStream;
  JMain: TJSONObject;
  JPair: TJSONPair;
begin
  Result := False;
  APassword := '';

  CheckHTTP;

  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
  JMain := TJSONObject.Create;
  try
    ClearError;
    try
      URI := BaseURL + URI_GetPassword(AName, AConfiguration, APlatform, AUser, ADate);
      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);

      JMain.Parse(UTF8Bytes.Bytes, 0);

      JPair := JMain.Get(SSabAccountPassword);
      if Assigned(JPair) then
        APassword := TSabObject.JSToStr(JPair.JsonValue.Value);

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
    UTF8Bytes.Free;
  end;
end;

function TSabClient.GetStations(const AList: TSabList; const AConfigName: string): Boolean;
var
  URI: string;
  Response: TBytesStream;
begin
  Result := False;
  AList.Clear;

  CheckHTTP;

  Response := TBytesStream.Create;
  try
    ClearError;
    try
      URI := BaseURL + URI_GetStations(AConfigName);
      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      TSabObject.AddObjects(AList, Response, SSabStations_Sign);

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
  end;
end;

function TSabClient.GetTicketPassword(const ATicketId, AStationName, APlatform, AUserName: string; out APassword: string): Boolean;
// APlatform = (WINDOWS, LINUX, MAC, LDAP)
var
  URI, S: string;
  Response: TBytesStream;
  UTF8Bytes: TBytesStream;
  RS: RawByteString;
  Items: TStringDynArray;
begin
  Result := False;
  APassword := '';

  CheckHTTP;

  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
  try
    ClearError;
    try
      HTTP.TextResponse := True;
      URI := BaseURL + URI_GetTicketPassword(ATicketId, AStationName, APlatform, AUserName);
      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);

      SetLength(RS, UTF8Bytes.Size);
      Move(UTF8Bytes.Memory^, RS[1], UTF8Bytes.Size);

      S := UTF8toString(RS);
      Items := SplitString(S, ',');
      if Length(Items) = 3 then
      begin
        if Items[0] = ANSWER_OK then
        begin
//          AUserName := Items[1];
          APassword := Items[2];
          Result := True;
        end else
          SetError(StrToIntDef(Items[0], HTTP_INTERNAL_SERVER_ERROR), 'Invalid parameters');
      end else
        SetError(StrToIntDef(Items[0], HTTP_INTERNAL_SERVER_ERROR), 'Invalid parameters');
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
    UTF8Bytes.Free;
  end;
end;

function TSabClient.GetUserConfiguration(const AUserCodeInURL: string; ACheckBase: Boolean): Boolean;
var
  URI: string;
  Response, UTF8Bytes: TBytesStream;
  JMain, JObj: TJSONObject;
  Val: TJSONString;
  Num: TJSONNumber;
  Flag: TJSONBool;
  URL, Prot, User, Pass, Host, Port, Path, Param: string;
  Lang: TSabLanguage;
  AppUri: string;
  LoginUri: string;
  UserPasswordGeneratorUri: string;

  function GetRequirements(const Obj: TJSONObject): TRequirements;
  begin
    FillChar(Result, SizeOf(Result), 0);

    Num := Obj.Values[SMinLength] as TJSONNumber;
    if Assigned(Num) then
      Result.MinLength := Num.AsInt;

    Num := Obj.Values[SMinLower] as TJSONNumber;
    if Assigned(Num) then
      Result.MinLower := Num.AsInt;

    Num := Obj.Values[SMinUpper] as TJSONNumber;
    if Assigned(Num) then
      Result.MinUpper := Num.AsInt;

    Num := Obj.Values[SMinDigits] as TJSONNumber;
    if Assigned(Num) then
      Result.MinDigits := Num.AsInt;

    Num := Obj.Values[SMinSpecials] as TJSONNumber;
    if Assigned(Num) then
      Result.MinSpecials := Num.AsInt;

    Num := Obj.Values[SMinEntropy] as TJSONNumber;
    if Assigned(Num) then
      Result.MinEntropy := Num.AsDouble;
  end;

begin
  Result := False;
  if Connected and (AUserCodeInURL <> '') then
    Exit;

  CheckHTTP; // w/o AC

  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
  JMain := TJSONObject.Create;
  try
    LoadErrorMessages;

    if AUserCodeInURL <> '' then
    begin
      URL := Self.ParseURL(AUserCodeInURL, Prot, User, Pass, Host, Port, Path, Param, False);

      if ACheckBase and (BaseURL <> '') then
      begin
        if not ContainsText(BaseUrl, Host) then
        begin
          FErrorCode := -1;
          Exit;
        end;
      end;
    end else
      Param := '';

    if BaseURL = '' then
      SetBaseURLFromInvitation(Prot, Host, Port, Path);

    try
      URI := BaseURL + URI_GetUserConfiguration(Param);

      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);
      JMain.Parse(UTF8Bytes.Bytes, 0);

      Flag := JMain.Values[SEnterpriseEdition] as TJSONBool;
      if Assigned(Flag) then
        FEnterpriseEdition := Flag.AsBoolean;

      Val := JMain.Values[SAppUri] as TJSONString;
      if Assigned(Val) then
        AppUri := Val.Value;

      Val := JMain.Values[SLoginUri] as TJSONString;
      if Assigned(Val) then
        LoginUri := Val.Value;

      Val := JMain.Values[SUsername] as TJSONString;
      if Assigned(Val) then
        FUserName := Val.Value;

      Val := JMain.Values[SDomain] as TJSONString;
      if Assigned(Val) then
        FDomain := Val.Value;

      Flag := JMain.Values[SHasAccessCode] as TJSONBool;
      if Assigned(Flag) then
        FHasAccessCode := Flag.AsBoolean;

      Num := JMain.Values[SAccessCodeTimeout] as TJSONNumber;
      if Assigned(Num) then
        FAccessCodeValidityTimeoutSec := Num.AsInt;

      JObj := JMain.Values[SLoginPasswordRequirements] as TJSONObject;
      if Assigned(JObj) then
        FLoginPasswordRequirements := GetRequirements(JObj);

      JObj := JMain.Values[SAccessCodeRequirements] as TJSONObject;
      if Assigned(JObj) then
        FAccessCodeRequirements := GetRequirements(JObj);

      Val := JMain.Values[SUserPasswordGeneratorPath] as TJSONstring;
      if Assigned(Val) then
        UserPasswordGeneratorUri := Val.Value;

      Flag := JMain.Values[SHasFeatureCbtClient] as TJSONBool;
      if Assigned(Flag) then
        FHasFeatureCbtClient := Flag.AsBoolean;

      Flag := JMain.Values[SHasFeatureFileManager] as TJSONBool;
      if Assigned(Flag) then
        FHasFeatureFileManager := Flag.AsBoolean;

      Val := JMain.Values[SUserLanguage] as TJSONString;
      if Assigned(Val) then
      begin
        Lang := StrToLang(Val.Value);
        if Lang <> FLanguage then
        begin
          FLanguage := Lang;
          LoadErrorMessages(True);
        end;
      end;
      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
    UTF8Bytes.Free;
    JMain.Free;
  end;
end;

{ Returns True, if User is logged in }
{ If error is not login.failed returns False }
{ When SAB session time has expired due to timeout the SAB returns login.failed }
function TSabClient.HandleLoginFailedError: Boolean;
var
  PswdAC: string;
  LoginCertificate: TMemoryStream;
begin
  Result := False;

  if Assigned(FOnUserLogin) and (ErrorCode = HTTP_UNAUTHORIZED) and (ErrorKey = SLoginFailed) then
  begin
    if FLoginCallbackRuns then
      Exit;

    FLoginCallbackRuns := True;
    try
      WipeStrings([wipAccessCode]);

      PswdAC := '';
      while not Result do
      begin
        LoginCertificate := TMemoryStream.Create;
        try
          if Assigned(FOnGetCertificate) and FOnGetCertificate(LoginCertificate) then
          begin
            Disconnect;
            Result := Connect(LoginCertificate);
          end;

          if not Result then
          begin
            if not Assigned(FOnUserLogin) or not FOnUserLogin(BaseURL, Domain, UserName, PswdAC) then
            begin
              ClearError;
              Exit;
            end;

            Disconnect;
            FPassword := PswdAC;
            Result := Connect;
          end;
        finally
          LoginCertificate.Free;
        end;

        if Result then
        begin
          if FSecondFactorRequired then
          begin
            if Assigned(FOnCheckSecondFactor) and FOnCheckSecondFactor(Self) then
            begin
              if ErrorCode <> 0 then
              begin
                if (ErrorKey = SLoginFailed) or (ErrorKey = SSecondFactorInvalid) then
                begin
                  Result := False;
                  // go to new loop
                end else
                begin
                  Result := False;
                  Disconnect;
                  Break;
                end;
              end else
                Break; // login and 2FA are OK: go on with Result == True
            end else
            begin
              Result := False;
              SetError(HTTP_UNAUTHORIZED, SLoggedOutMessage, SLoginFailed);
              Disconnect;
              Break;
            end;
          end;
          WipeString(PswdAC);
        end else
          if ErrorKey <> SLoginFailed then
            Break;

        WipeString(PswdAC);
      end;
      WipeString(PswdAC);
    finally
      FLoginCallbackRuns := False;
    end;
  end else
  if Assigned(FOnGetAccessCode) {and (AccessCode = '') }and ((ErrorKey = SInvalidAccessCodeKey) or (ErrorKey = SMissingAccessCodeKey)) then
  begin
    if FAcCallbackRuns then
      Exit;

    FAcCallbackRuns := True;
    try
      PswdAC := '';
      Result := FOnGetAccessCode(PswdAC, ErrorKey = SInvalidAccessCodeKey);
      if not Result then
        Exit;
      if PswdAC <> '' then
        AccessCode := PswdAC;
      WipeString(PswdAC);
    finally
      FAcCallbackRuns := False;
    end;
  end else
  // 400, missing query parameter (SAB-Access-Code) - very old server bug, fixed, for sure !!!
  if Assigned(FOnGetAccessCode) and ((ErrorCode = HTTP_BAD_REQUEST) and StartsText('missing query parameter (SAB-Access-Code)', ErrorMsg)) then
  begin
    if FAcCallbackRuns then
      Exit;

    FAcCallbackRuns := True;
    try
      PswdAC := '';
      Result := FOnGetAccessCode(PswdAC);
      if not Result then
        Exit;
      if PswdAC <> '' then
        AccessCode := PswdAC;
      WipeString(PswdAC);
    finally
      FAcCallbackRuns := False;
    end;
  end;
end;

procedure TSabClient.LoadErrorMessages(const FreeTable: Boolean);
var
  URI: string;
  Errs: TBytesStream;
begin
  Errs := TBytesStream.Create;
  try
    if FreeTable then
      SAB.Errors.DoneSabErrorTable;

    if not IsSabErrorTableInitialized then
    begin
      URI := FBaseURL + URI_GetErrorMessages(FLanguage);
      if HTTP.Get(URI, Errs) then
        SAB.Errors.InitSabErrorTable(Errs);
    end;
  finally
    Errs.Free;
  end;
end;

function TSabClient.LoginStatus(out Authenticated: Boolean): Boolean;
var
  URI: string;
  Response, UTF8Bytes: TBytesStream;
  JMain: TJSONObject;
  Val: TJSONString;
  Num: TJSONNumber;

  function GetFlag(const Identifier: string; out AFlag: Boolean): Boolean;
  var
    Flag: TJSONBool;
  begin
    Flag := JMain.Values[SAuthenticated] as TJSONBool;
    if Assigned(Flag) then
    begin
      AFlag := Flag.AsBoolean;
      Result := True;
    end else
      Result := False;
  end;

begin
  Result := False;

  UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
  JMain := TJSONObject.Create;
  Response := TBytesStream.Create;
  try
    CheckHTTP; // w/o

    LoadErrorMessages;

    URI := BaseURL + URI_LoginStatus;
    while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
    begin
      if not HandleLoginFailedError then
        Exit;
    end;

    FResponseSize := Response.Size;
    Response.Position := 0;
    UTF8Bytes.CopyFrom(Response, Response.Size);
    JMain.Parse(UTF8Bytes.Bytes, 0);

    if not GetFlag(SAuthenticated, Authenticated) then
      Authenticated := False;

    GetFlag(SSabStatus2ndFactor, FSecondFactorRequired);
    GetFlag(SHasAccessCode, FHasAccessCode);

    Val := JMain.Values[SUid] as TJSONString;
    if Assigned(Val) then
      FUid := Val.Value;

    Num := JMain.Values[SAccessCodeTimeout] as TJSONNumber;
    if Assigned(Num) then
      FAccessCodeValidityTimeoutSec := Num.AsInt;

    GetFlag(SSabUser, FSabUser);
    GetFlag(SSabAdmin, FSabAdmin);
    GetFlag(SCfgAdmin, FCfgAdmin);
    GetFlag(SSafeBoxUser, FSafeBoxUser);
    GetFlag(SSafeBoxAdmin, FSafeBoxAdmin);
    GetFlag(SAuditor, FAuditor);

    Result := True;
  finally
    UTF8Bytes.Free;
    JMain.Free;
    Response.Free;
  end;
end;

function TSabClient.ReadSecretFields(const AObj: TSABObject; const ADownload: Boolean; const AFileMonitor: THookMonitor): Boolean;
var
  URI: string;
  Res2, Res3: TSabObject;
  Response: TBytesStream;
  ErrCode: Integer;

  function SetAccessCode: Boolean;
  var
    PswdAC: string;
  begin
    Result := False;
    if Assigned(FOnGetAccessCode) then
    begin
      PswdAC := '';
      Result := FOnGetAccessCode(PswdAC);
      if not Result then
        Exit;

      if PswdAC <> '' then
        AccessCode := PswdAC;
      WipeString(PswdAC);
    end else
    begin
      GetSabError(SInvalidAccessCodeKey, nil, FErrorMsg, ErrCode);
      FErrorCode := HTTP_INTERNAL_SERVER_ERROR;
      Exit;
    end;
    Result := True;
  end;

begin
  Result := False;

  CheckHTTP;

  Res2 := nil;
  Res3 := nil;
  Response := TBytesStream.Create;
  try
    case AObj.ObjType of
      sabExternalFile,
      sabAccountSec,
      sabCreditCard,
      sabCertificate,
      sabFile:
        begin
          if AccessCode = '' then
          begin
            if not SetAccessCode then
              Exit;
          end;

{$IFDEF GET_RECORD_FORM}
          URI := BaseURL + URI_ReadRecordForm(AObj.Id);   // returns JSON object "defaults"
          while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
          begin
            if not HandleLoginFailedError then
              Exit;
          end;
          Res2 := TSabObject.CreateObject(Response, '');
{$ELSE}
          URI := BaseURL + URI_ReadRecord(AObj.Id); // read record again with AC set
          while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
          begin
            if not HandleLoginFailedError then
              Exit;
          end;
          Res2 := TSabObject.CreateObject(Response, SSabRecord_Sign);
{$ENDIF}
          AObj.GetSecureFieldsFrom(Res2);

          case AObj.ObjType of
            sabAccountSec:
              begin
                URI := BaseURL + URI_ReadRecordPasswords(AObj.Id);
                while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
                begin
                  if not HandleLoginFailedError then
                    Exit;
                end;

                Res3 := TSabSecAccount.CreateObject(Response, '');
                if Assigned(Res3) then
                  AObj.GetSecureFieldsFrom(Res3);
              end;
            sabFile, sabCertificate:
              begin
                if ADownload then
                  if not DownloadFile(TSabFile(AObj), TSabFile(AObj).Stream, AFileMonitor) then
                  begin
                    if TSabFile(AObj).Filename <> '' then
                      Exit;

                    ClearError;
                  end;
              end;
          end;
        end;
      sabAccount:
        begin
          if AccessCode = '' then
          begin
            if not SetAccessCode then
              Exit;
          end;

          URI := BaseURL + URI_ReadRecordPasswords(AObj.Id);
          while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
          begin
            if not HandleLoginFailedError then
              Exit;
          end;

          Res3 := TSabAccount.CreateObject(Response, '');
          if Assigned(Res3) then
            AObj.GetSecureFieldsFrom(Res3);
        end;
    else
      Exit;
    end;
    Result := True;
  finally
    Res2.Free;
    Res3.Free;
    Response.Free;
  end;
end;

function TSabClient.RegisterSecondFactor(out ASecretKeyUri: string; const ASecurityKey, AIssuer, AUsername, ADomain: string): Boolean;
var
  URI: string;
  List: TStringList;
  Response, UTF8Bytes: TBytesStream;
  JObj: TJSONObject;
  Val: TJSONString;
begin
  Result := False;
  ASecretKeyUri := '';

  CheckHTTP;

  ClearError;

  List := TStringList.Create;
  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create;
  JObj := TJSONObject.Create;
  try
    List.Sorted := False;

    URI := BaseURL + URI_RegisterSecondFactor;

    if AIssuer <> '' then
    begin
      List.Add(SSabReg2FAIssuer + '=' + AIssuer);
      if AUsername <> '' then
      begin
        if ADomain <> '' then
          List.Add(SSabReg2FAccount + '=' + AUsername + ' (' + ADomain + ')')
        else
          List.Add(SSabReg2FAccount + '=' + AUsername);
      end else
      begin
        if Self.Domain <> '' then
          List.Add(SSabReg2FAccount + '=' + Self.UserName + ' (' + Self.Domain + ')')
        else
          List.Add(SSabReg2FAccount + '=' + Self.UserName);
      end;
    end;

    List.Add(SSabReg2FAKey + '=' + ASecurityKey);

    try
      if AIssuer = '' then
      begin
        while not CheckRestFunc(HTTP.Post(URI, List, Response), Response) do
        begin
          if not HandleLoginFailedError then
            Exit;
        end;
      end else
      begin
        while not CheckRestFunc(HTTP.Options(URI, List, Response), Response) do
        begin
          if not HandleLoginFailedError then
            Exit;
        end;
      end;

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);

      JObj.Parse(UTF8Bytes.Bytes, 0);

      Val := JObj.Values[SSecretKeyUri] as TJSONString;
      if Assigned(Val) then
        ASecretKeyUri := Val.Value;

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
    UTF8Bytes.Free;
    JObj.Free;
    List.Free;
  end;
end;

function TSabClient.RelatedOperationWithDeletedRecords(Remove: Boolean; const List: TSabIdList): Boolean;
var
  I: Integer;
  URI: string;
  Src: RawByteString;
  JMain: TJSONObject;
  JArray: TJSONArray;
  Response, Source: TBytesStream;
begin
  Result := False;

  CheckHTTP;

  JArray := TJSONArray.Create;
  JMain := TJSONObject.Create;
  Response := TBytesStream.Create;
  Source := TBytesStream.Create;
  try
    for I := 0 to List.Count - 1 do
      JArray.Add(List[I]);
    JMain.AddPair(SSabBoxes_Sign, JArray);

    Src := UTF8Encode(JMain.ToString);
    Source.Write(Src[1], Length(Src));
    Source.Position := 0;

    ClearError;
    try
      URI := BaseURL + URI_ReadRemoveRestoreDeletedRecords;
      if Remove then
      begin
        while not CheckRestFunc(HTTP.Delete(URI, Source, Response), Response) do
        begin
          if not HandleLoginFailedError then
            Exit;
        end;
      end else
      begin
        while not CheckRestFunc(HTTP.Put(URI, Source, Response), Response) do
        begin
          if not HandleLoginFailedError then
            Exit;
        end;
      end;

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
    Source.Free;
    JMain.Free;
  end;
end;

function TSabClient.RemoveDeletedRecord(const Id: TSabId): Boolean;
var
  List: TSabIdList;
begin
  List := TSabIdList.Create;
  try
    List.Add(Id);
    Result := RemoveDeletedRecords(List);
  finally
    List.Free;
  end;
end;

function TSabClient.RemoveDeletedRecords(const List: TSabIdList): Boolean;
begin
  Result := RelatedOperationWithDeletedRecords(True, List);
end;

function TSabClient.RestoreDeletedRecords(const List: TSabIdList): Boolean;
begin
  Result := RelatedOperationWithDeletedRecords(False, List);
end;

function TSabClient.RestoreDeletedRecord(const Id: TSabId): Boolean;
var
  List: TSabIdList;
begin
  List := TSabIdList.Create;
  try
    List.Add(Id);
    Result := RestoreDeletedRecords(List);
  finally
    List.Free;
  end;
end;

procedure TSabClient.RestoreTraffic;
begin
  if Assigned(HTTP) then
    HTTP.RestoreTraffic;
end;

function TSabClient.Search(const APattern: string; const List: TSearchResultList): Boolean;
var
  URI: string;
  Response, UTF8Bytes: TBytesStream;
  JObj: TJSONObject;
  JPair: TJSONPair;

  procedure GetResultList(const JArray: TJSONArray; const AList: TSearchResultList);
  var
    I: Integer;
    JValue: TJSONValue;
    Res: TSearchResult;
  begin
    for I := 0 to JArray.Count - 1 do
    begin
      JValue := JArray.Items[I];
      if Assigned(JValue) and (JValue is TJSONObject) then
      begin
        Res := TSearchResult.CreateObject(TJSONObject(JValue));
        AList.Add(Res)
      end;
    end;
  end;

begin
  Result := False;

  CheckHTTP;

  ClearError;

  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
  JObj := TJSONObject.Create;
  try
    List.Clear;
    if APattern = '' then
      Exit(True);

    URI := BaseURL + URI_Search(APattern);
    try
      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);

      JObj.Parse(UTF8Bytes.Bytes, 0);
      JPair := JObj.Get(SSearchResult);
      if Assigned(JPair) and (JPair.JsonValue is TJSONArray) then
        GetResultList(TJSONArray(JPair.JsonValue), List);

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    JObj.Free;
    Response.Free;
    UTF8Bytes.Free;
  end;
end;

function TSabClient.SearchByNote(const APattern: string; const List: TSabIdList): Boolean;
var
  URI: string;
  Response, UTF8Bytes: TBytesStream;
  JObj: TJSONObject;
  JPair: TJSONPair;

  procedure GetResultList(const JArray: TJSONArray; const AList: TSabIdList);
  var
    I: Integer;
    JValue: TJSONValue;
    Id: TSabId;
    JObj: TJSONObject;
    JPair: TJSONPair;
    ParentId: TSabId;
  begin
    for I := 0 to JArray.Count - 1 do
    begin
      JValue := JArray.Items[I];
      if Assigned(JValue) then
        if JValue is TJSONObject then
        begin
          JObj := TJSONObject(JValue);

          JPair := JObj.Get(SSabObjId);
          if Assigned(JPair) then
          begin
{$IFDEF LONG_ID}
            Id := TJSONNumber(JPair.JsonValue).AsInt64;
{$ELSE}
            Id := TJSONNumber(JPair.JsonValue).AsInt;
{$ENDIF}

            JPair := JObj.Get(SSabBoxParentId);
            if Assigned(JPair) then
{$IFDEF LONG_ID}
              ParentId := TJSONNumber(JPair.JsonValue).AsInt64
{$ELSE}
              ParentId := TJSONNumber(JPair.JsonValue).AsInt
{$ENDIF}
            else
              ParentId := -1;

            if Id > 0 then
              AList.AddDWord(Id, ParentId);
          end;
        end else
        if JValue is TJSONNumber then  // UP_TO_SAB_VERSION_500
        begin
{$IFDEF LONG_ID}
          Id := StrToInt64Def(TJSONNumber(JValue).Value, -1);
{$ELSE}
          Id := StrToIntDef(TJSONNumber(JValue).Value, -1);
{$ENDIF}
          if Id > 0 then
            AList.Add(Id)
        end;
    end;
  end;

begin
  Result := False;

  CheckHTTP;

  ClearError;

  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
  JObj := TJSONObject.Create;
  try
    List.Clear;
    if APattern = '' then
      Exit(True);

    URI := BaseURL + URI_SearchByNote(APattern);
    try
      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);

      JObj.Parse(UTF8Bytes.Bytes, 0);
      JPair := JObj.Get(SSearchByNoteResult);
      if Assigned(JPair) and (JPair.JsonValue is TJSONArray) then
        GetResultList(TJSONArray(JPair.JsonValue), List);

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    JObj.Free;
    Response.Free;
    UTF8Bytes.Free;
  end;
end;

function TSabClient.SecondFactor(const ASecurityCode, ATOTPKey: string): Boolean;
var
  URI: string;
  List: TStringList;
  Response: TBytesStream;
begin
  Result := False;

  CheckHTTP;

  ClearError;

  List := TStringList.Create;
  Response := TBytesStream.Create;
  try
    List.Sorted := False;

    URI := BaseURL + URI_SecondFactor;

    List.Add(SSabSecurityCode + '=' + ASecurityCode);
    if ATOTPKey <> '' then
      List.Add(SSabReg2FAKey + '=' + ATOTPKey);
    try
      while not CheckRestFunc(HTTP.Post(URI, List, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    List.Free;
  end;
end;

procedure TSabClient.SetAccessCode(const Value: string);
begin
  FAccessCode := Value;
  if Assigned(HTTP) then
    HTTP.AccessCode := Value;
end;

procedure TSabClient.SetBaseURL(const Value: string);
begin
  FBaseURL := PrepareURL(Value);
end;

procedure TSabClient.SetBaseURLFromInvitation(const AProt, AHost, APort, APath: string);
var
  I: Integer;
begin
  if (AProt <> '') and (AHost <> '') then
  begin
    FBaseUrl := AProt + '://' + AHost;

    if APort <> '' then
      FBaseURL := FBaseURL + ':'+ APort;

    if APath <> '' then
    begin
      I := Pos('/static/inituser.page', APath);
      if I > 1 then
        FBaseURL := FBaseURL + Copy(APath, 1, I - 1);
    end;

    if (FBaseURL <> '') and (FBaseURL[Length(FBaseURL)] <> '/') then
      FBaseURL := FBaseURL + '/';
  end;
end;

procedure TSabClient.SetCertCA(const Value: RawByteString);
begin
  FCertCA := Value;
  if Value <> '' then
    FVerifyCertificate := True;
{
  if Length(FCertCABytes) > 0 then
    SetLength(FCertCABytes, 0);

  SetLength(FCertCABytes, Length(FCertCA));
  Move(PAnsiChar(FCertCA)^, PAnsiChar(FCertCABytes)^, Length(FCertCABytes));
}
  if Assigned(FHTTP) then
    if Assigned(FHTTP.Sock.SSL) then
      FHTTP.Sock.SSL.CertCA := Value;
end;

procedure TSabClient.SetCertCABytes(const Value: TBytes);
begin
  FCertCABytes := Value;
  if Length(Value) <> 0 then
    FVerifyCertificate := True;
  if Assigned(FHTTP) then
    if Assigned(FHTTP.Sock.SSL) then
      if FHTTP.Sock.SSL is TSabOpenSSL then
        TSabOpenSSL(FHTTP.Sock.SSL).CertCABytes := Value
      else
        FHTTP.Sock.SSL.CertCAFile := TEncoding.ASCII.GetString(Value);
end;

procedure TSabClient.SetCertCAFile(const Value: string);
begin
  FCertCAFile := Value;
  if Value <> '' then
    FVerifyCertificate := True;
  if Assigned(FHTTP) then
    if Assigned(FHTTP.Sock.SSL) then
      FHTTP.Sock.SSL.CertCAFile := Value;
end;

procedure TSabClient.SetCerticate(const Value: RawByteString);
begin
  FCertificate := Value;
  if Value <> '' then
    FVerifyCertificate := True;
  if Assigned(FHTTP) then
    if Assigned(FHTTP.Sock.SSL) then
      FHTTP.Sock.SSL.Certificate := Value;
end;

procedure TSabClient.SetCertificateFile(const Value: string);
begin
  FCertificateFile := Value;
  if Value <> '' then
    FVerifyCertificate := True;
  if Assigned(FHTTP) then
    if Assigned(FHTTP.Sock.SSL) then
      FHTTP.Sock.SSL.CertificateFile := Value;
end;

procedure TSabClient.SetConnected(const Value: Boolean);
begin
  if Value then
  begin
    if not Connected then
      Connect;
  end else
  begin
    if Connected then
      Disconnect;
  end;
end;

procedure TSabClient.SetError(ACode: Integer; const AMsg: string; const AErrorKey: string);
begin
  FErrorMsg := AMsg;
  FErrorSubj := '';
  FErrorCode := ACode;
  FErrorKey := AErrorKey;
end;

procedure TSabClient.SetImprint(const Value: RawByteString);
begin
  FImprint := Value;
  if Assigned(FHTTP) then
    if Assigned(FHTTP.Sock.SSL) then
      if FHTTP.Sock.SSL is TSabOpenSSL then
        TSabOpenSSL(FHTTP.Sock.SSL).Imprint := Value
      else
        raise Exception.Create('SSL is not SabOpenSSL');
end;

procedure TSabClient.SetConnectionTimeout(const Value: Integer);
begin
  FConnectionTimeout := Value;
  if FConnectionTimeout <> 0 then
    if Assigned(FHTTP) then
      FHTTP.Sock.ConnectionTimeout := FConnectionTimeout;
end;

function TSabClient.SetNewAccessCode(const ANewAccessCode: string): Boolean;
var
  URI: string;
  Response: TBytesStream;
  JObj: TJSONObject;
  Src: TBytesStream;
  Len: Integer;
begin
  Result := False;
  if not FConnected then
    Exit;

  Src := TBytesStream.Create;
  Response := TBytesStream.Create;
  JObj := TJSONObject.Create;
  try
    LoadErrorMessages;

    JObj.AddPair(TJSONPair.Create(SSabUserAccessCodeNew, TJSONString.Create(ANewAccessCode)));
    JObj.AddPair(TJSONPair.Create(SSabUserAccessCodeRe, TJSONString.Create(ANewAccessCode)));

    Src.Size := JObj.EstimatedByteSize;
    Len := JObj.ToBytes(Src.Bytes, 0);
    Src.Size := Len;
    try
      URI := BaseURL + URI_SetUserNewAccessCode;

      while not CheckRestFunc(HTTP.Post(URI, Src, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      AccessCode := ANewAccessCode;
      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
    JObj.Free;
  end;
end;

function TSabClient.SetPasswordToInvitedUser(const AUserCodeInURL, ANewPassword: string; ACheckBase: Boolean): Boolean;
var
  URI: string;
  Response, UTF8Bytes: TBytesStream;
  JObj, JMain: TJSONObject;
  Val: TJSONString;
  Src: TBytesStream;
  Len: Integer;
  URL, Prot, User, Pass, Host, Port, Path, Param: string;
begin
  Result := False;

  if FConnected then
    Exit;

  CheckHTTP; // w/o AC

  Src := TBytesStream.Create;
  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create(TEncoding.UTF8.GetPreamble);
  JMain := TJSONObject.Create;
  JObj := TJSONObject.Create;
  try
    LoadErrorMessages;

    URL := Self.ParseURL(AUserCodeInURL, Prot, User, Pass, Host, Port, Path, Param, False);

    if ACheckBase and (BaseURL <> '') and (AUserCodeInURL <> '') then
    begin
      if not ContainsText(BaseUrl, Host) then
      begin
        FErrorCode := -1;
        Exit;
      end;
    end;

    if BaseURL = '' then
      SetBaseURLFromInvitation(Prot, Host, Port, Path);

    JObj.AddPair(TJSONPair.Create(SSabUserPasswordNew, TJSONString.Create(ANewPassword)));
    JObj.AddPair(TJSONPair.Create(SSabUserPasswordRe, TJSONString.Create(ANewPassword)));

    Src.Size := JObj.EstimatedByteSize;
    Len := JObj.ToBytes(Src.Bytes, 0);
    Src.Size := Len;
    try
      URI := BaseURL + URI_SetPasswordToInvitedUser(Param);

      while not CheckRestFunc(HTTP.Post(URI, Src, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FPassword := ANewPassword;
      // response contains msg, domain, username

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);
      JMain.Parse(UTF8Bytes.Bytes, 0);

      Val := JMain.Values[SSabSetPswdDomainRet] as TJSONString;
      if Assigned(Val) then
        FDomain := Val.Value;

      Val := JMain.Values[SSabSetPswdUsernameRet] as TJSONString;
      if Assigned(Val) then
        FUsername := Val.Value;

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
    UTF8Bytes.Free;
    JObj.Free;
    JMain.Free;
  end;
end;

procedure TSabClient.SetPFX(const Value: RawByteString);
begin
  FPFX := Value;
  if Value <> '' then
    FVerifyCertificate := True;
  if Assigned(FHTTP) then
    if Assigned(FHTTP.Sock.SSL) then
      FHTTP.Sock.SSL.PFX := Value;
end;

procedure TSabClient.SetVerifyCertificate(const Value: Boolean);
begin
  FVerifyCertificate := Value;
  if Assigned(FHTTP) then
    if Assigned(FHTTP.Sock.SSL) then
      FHTTP.Sock.SSL.VerifyCert := Value;
end;

function TSabClient.SetWatched(const ABox: TSabSafeBox; const WithRecords: Boolean): Boolean;
var
  URI: string;
  Response: TBytesStream;
begin
  Result := True;

  if not Assigned(ABox) then
    Exit;

  if not ABox.WatchedModified then
    Exit;

  CheckHTTP; // w/o AC

  Response := TBytesStream.Create;
  try
    ClearError;
    try
      if ABox.WatchedChangesModified then
      begin
        URI := BaseURL + URI_SetWatched(ABox.Id, wtChanges, WithRecords, ABox.WatchedChanges);

        while not CheckRestFunc(HTTP.Post(URI, Response), Response) do
        begin
          if not HandleLoginFailedError then
            Exit(False);
        end;

        ABox.ClearWatched([wtChanges]);
      end;

      if ABox.WatchedAccessesModified then
      begin
        URI := BaseURL + URI_SetWatched(ABox.Id, wtAccesses, WithRecords, ABox.WatchedAccesses);

        while not CheckRestFunc(HTTP.Post(URI, Response), Response) do
        begin
          if not HandleLoginFailedError then
            Exit(False);
        end;

        ABox.ClearWatched([wtAccesses]);
      end;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
  end;
end;

function TSabClient.Status(const DoLoadErrorMessages: Boolean): Boolean;
var
  URI: string;
  JMain: TJSONObject;
  Response, UTF8Bytes: TBytesStream;
  Val: TJSONString;
  Num: TJSONNumber;
  Flag: TJSONBool;
begin
  Result := False;

  CheckHTTP; // w/o AC

  Response := TBytesStream.Create;
  UTF8Bytes := TBytesStream.Create;
  JMain := TJSONObject.Create;
  try
    ClearError;
    try
      if DoLoadErrorMessages then
        LoadErrorMessages;

      URI := BaseURL + URI_GetStatus;

      while not CheckRestFunc(HTTP.Get(URI, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Response.Position := 0;
      UTF8Bytes.CopyFrom(Response, Response.Size);
      JMain.Parse(UTF8Bytes.Bytes, 0);

      Val := JMain.Values[SSabStatusProduct] as TJSONString;
      if Assigned(Val) then
        FSabProduct := Val.Value;

      Val := JMain.Values[SSabStatusApiVersion] as TJSONString;
      if Assigned(Val) then
        FSabApiVersion := Val.Value;

      Num := JMain.Values[SSabStatusGmtTime] as TJSONNumber;
      if Assigned(Num) then
        FSabGmtTime := Num.AsInt64;

      Val := JMain.Values[SSabStatusServTime] as TJSONString;
      if Assigned(Val) then
        FSabServerTime := Val.Value;

      Val := JMain.Values[SSabStatusVersion] as TJSONString;
      if Assigned(Val) then
        FSabVersion := Val.Value;

      Flag := JMain.Values[SSabStatus2ndFactor] as TJSONBool;
      if Assigned(Flag) then
        FSecondFactorRequired := Flag.AsBoolean
      else
        FSecondFactorRequired := False;

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    JMain.Free;
    Response.Free;
    UTF8Bytes.Free;
  end;
end;

function TSabClient.Update(const Obj: TSabObject; const AUploadFileOrChangePasswords: Boolean): Boolean;
var
  URI: string;
  Source: TBytesStream;
  Response: TBytesStream;
  JObj: TJSONObject;
  Len: Integer;
begin
  Result := False;

  CheckHTTP;

  CheckId(Obj);

  JObj := Obj.SaveTo;
  Source := TBytesStream.Create;
  Response := TBytesStream.Create;
  try
    ClearError;
    try
      case Obj.ObjType of
        sabSafeBox, sabSafeBoxGroup:
          URI := BaseURL + URI_UpdateSafeBox(Obj.Id);

        sabAccount, sabAccountSec, sabFile, sabCertificate,
        sabCreditCard, sabExternalFile:
          URI := BaseURL + URI_UpdateRecord(Obj.Id);

        sabConfiguration: Exit;
        sabStation: Exit;
      end;
      Source.Size := JObj.EstimatedByteSize;

      Len := JObj.ToBytes(Source.Bytes, 0);
      Source.Size := Len;
      Source.Position := 0;

      while not CheckRestFunc(HTTP.Put(URI, Source, Response, Obj.SecureFieldsUsed), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      if AUploadFileOrChangePasswords then
      begin
        case Obj.ObjType of
          sabAccount, sabAccountSec:
            begin
              if not UpdatePassword(TSabAccountCommon(Obj)) then
                Exit;
            end;
          sabFile, sabCertificate:
            begin
              if Assigned(TSabFile(Obj).Stream) and (TSabFile(Obj).Stream.Size > 0) then
                if not UploadFile(TSabFile(Obj), TSabFile(Obj).Stream) then
                  Exit;
             end;
        end;
      end;

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Source.Free;
    Response.Free;
    JObj.Free;
  end;
end;

function TSabClient.UpdatePassword(const Acc: TSabAccountCommon): Boolean;
var
  URI: string;
  Source: TBytesStream;
  Response: TBytesStream;
  JObj: TJSONObject;
  Len: Integer;
begin
  Result := False;

  CheckHTTP;

  CheckId(Acc);
  CheckAccount(Acc);

  JObj := Acc.SavePasswords;
  Source := TBytesStream.Create;
  Response := TBytesStream.Create;
  try
    ClearError;
    try
      URI := BaseURL + URI_ChangePasswords(Acc.Id);

      Source.Size := JObj.EstimatedByteSize;

      Len := JObj.ToBytes(Source.Bytes, 0);
      Source.Size := Len;
      Source.Position := 0;

      while not CheckRestFunc(HTTP.Post(URI, Source, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Source.Free;
    Response.Free;
    JObj.Free;
  end;
end;

function TSabClient.UploadFile(const SabObj: TSabFile; const Stream: TBytesStream): Boolean;
var
  URI: string;
  Response: TBytesStream;
  FileStream: TBytesStream;
begin
  Result := False;

  CheckHTTP;

  CheckId(SabObj);

  if not (Assigned(Stream) or Assigned(SabObj.Stream)) then
    raise Exception.Create('Stream not assigned');

  FileStream := Stream;
  if Assigned(SabObj.Stream) then
    FileStream := SabObj.Stream;

  Response := TBytesStream.Create;
  try
    ClearError;
    try
      URI := BaseURL + URI_UploadFileToRecord(SabObj.Id);

      while not CheckRestFunc(HTTP.PostFile(URI, SSabFileName, SabObj.Filename, FileStream, Response), Response) do
      begin
        if not HandleLoginFailedError then
          Exit;
      end;

      FResponseSize := Response.Size;
      Result := True;
    except
      on E: Exception do
        SetError(SAB_UNRECOGNIZED_ERROR, E.Message);
    end;
  finally
    Response.Free;
  end;
end;

procedure TSabClient.WipeStrings(const WhatWipe: TWhatWipeSet);
begin
  if wipAccessCode in WhatWipe then
  begin
    TDP.CommonCode.WipeString(FAccessCode);
    FAccessCode := '';
  end;

  if wipPassword in WhatWipe then
  begin
    TDP.CommonCode.WipeString(FPassword);
    FPassword := '';
  end;
end;

end.
