
{*******************************************************}
{                                                       }
{       SAB Object Framework                            }
{       SAB Classes                                     }
{                                                       }
{       Copyright (c) 2017-22 TDP                       }
{                                                       }
{****************************************************tk*}

unit SAB.Filesys;

interface

uses
  System.SysUtils, System.Classes, System.Contnrs,
  System.Generics.Collections, System.JSON, System.SyncObjs,
  SAB.Types, SAB.Consts, SAB.Classes, SAB.Clients;

{$IFDEF ESCAPED_NAMES}
{$ELSE}
const
  EscapedBackSlash = '//';  // '(bslsh)';
  EscapedColon     = ';;';  // '##' ';;'
{$ENDIF}

const
  SabExtension: array[TSabObjectType] of string = (
    '.safeBox',       // sabSafeBox,
    '.safeBoxGroup',  // sabSafeBoxGroup,
    '.account',       // sabAccount,
    '.secretAccount', // sabAccountSec,
    '.file',          // sabFile,
    '.certificate',   // sabCertificate,
    '.creditCard',    // sabCreditCard,
    '.station',       // sabStation,
    '.configuration', // sabConfiguration
    '.extfile'        // sabExternFile
    );

type
  TSabFileSysList = class(TSabList)
{$IFDEF ESCAPED_NAMES}
  private
    class function EscapeString(const S: string; const ACharArray: array of Char): string;
    class function UnescapeString(const S: string; const ACharArray: array of Char): string;
{$ENDIF}
  public
    procedure Populate(const List: TStringList; const WhatObjects: TSabObjectSet = AllSabRecords; const AEscape: Boolean = False; const AllObjects: Boolean = False); reintroduce;

    class function EscapeFsStr(const S: string): string;
  end;

  TThreadedSabClient = class(TSabClient)
  private
    FLock: TCriticalSection;
  public
    constructor Create(const ABaseURL, AUserName, APassword, AAccessKey: string; const ADomain: string = 'SYSTEM'); reintroduce; overload;
    constructor Create(const ABaseURL: string); reintroduce; overload;
    destructor Destroy; override;

    function GetRecord(RecId: TSabId; SecretFields: Boolean = False; Download: Boolean = False): TSabRecord; reintroduce;
    function GetRecords(BoxId: TSabId; const List: TSabList; SecretFields: Boolean = False; Download: Boolean = False; GetFileSizes: Boolean = False): Boolean; reintroduce;
    function GetSafeBox(BoxId: TSabId): TSabSafeBox; reintroduce;
    function GetSafeBoxes(BoxId: TSabId; const List: TSabList; ExpandAll: Boolean = True; IncludeGroupBoxes: Boolean = False): Boolean; reintroduce;

    function Add(BoxId: TSabId; const SabObj: TSabObject; out Id: TSabId; const ForceAdd: Boolean = True; const FreeObj: Boolean = False): Boolean; reintroduce;
    function Update(const Obj: TSabObject; const AUploadFile: Boolean = False): Boolean; reintroduce;
    function Delete(const Obj: TSabObject): Boolean; reintroduce;
  end;

function ObjTypeByFileExt(const Ext: string): TSabObjectType;
function ObjTypeByFilename(const Filename: string): TSabObjectType;

implementation

function ObjTypeByFileExt(const Ext: string): TSabObjectType;
begin
  for Result := Low(SabExtension) to High(SabExtension) do
    if SameStr(SabExtension[Result], LowerCase(Ext)) then
      Exit;
  Result := sabSafeBox;
end;

function ObjTypeByFilename(const Filename: string): TSabObjectType;
begin
  Result := ObjTypeByFileExt(ExtractFileExt(Filename));
end;

{ TSabFileSysList }

class function TSabFileSysList.EscapeFsStr(const S: string): string;
begin
{$IFDEF ESCAPED_NAMES}
  Result := EscapeString(S, ['\', '<', '>', ':', '"', '/', '|', '?', '*']);
{$ELSE}
  Result := StringReplace(S, '\', EscapedBackSlash, [rfReplaceAll]);
  Result := StringReplace(Result, ':', EscapedColon, [rfReplaceAll]);
{$ENDIF}
end;

{$IFDEF ESCAPED_NAMES}
class function TSabFileSysList.EscapeString(const S: string; const ACharArray: array of Char): string;
var
  I: Integer;
begin
  Result := StringReplace(S, '%', '%%', [rfReplaceAll]);
  if Length(ACharArray) > 0 then
  begin
    for I := Low(ACharArray) to High(ACharArray) do
      Result := StringReplace(Result, ACharArray[I], '%' + IntToHex(Byte(Ord(ACharArray[I]))), [rfReplaceAll]);
  end;
end;
{$ENDIF}
procedure TSabFileSysList.Populate(const List: TStringList; const WhatObjects: TSabObjectSet; const AEscape, AllObjects: Boolean);
var
  Obj: TSabObject;
  Box: TSabSafeBox;

  Name: string;
  I: Integer;
  Id: TSabId;
  ObjType: TSabObjectType;
begin
  List.Clear;
  List.CaseSensitive := False;

{ Prepare names for Win/Dos}
  if AEscape then
    for Obj in Self do
    begin
      Obj.Name := EscapeFsStr(Obj.Name);
    end;

  for Obj in Self do
  begin
    Name := Obj.Name;
    Id := Obj.Id;
    ObjType := Obj.ObjType;

    if (AllSabBoxes * WhatObjects <> []) and (ObjType in WhatObjects) then
    begin
      Box := TSabSafeBox(Obj);

      if Box.OwnerName <> '' then
        Name := Name + ' (' + Box.OwnerName + ')';

      if not (Box.HasPermissionToCreate or AllObjects) then
        Name := '';

      if Name <> '' then
      begin
        if List.Find(Name, I) then
          if Box.PrivateBox then
            Name := Name + ' - private'
          else
            Name := Name + ' - shared';
      end;
    end else
    begin
      case Obj.ObjType of
        sabAccount, sabAccountSec, sabCreditCard,
        sabStation, sabConfiguration:
          Name := Name + LowerCase(SabExtension[Obj.ObjType]);
        sabFile, sabCertificate: ;
      end;
    end;

    if (Name <> '') and (ObjType in WhatObjects) then
      if List.OwnsObjects then
        List.Add(Name)
      else
        List.AddObject(Name, TObject(Id));
  end;
end;
{$IFDEF ESCAPED_NAMES}
class function TSabFileSysList.UnescapeString(const S: string; const ACharArray: array of Char): string;
var
  I: Integer;
begin
  Result := S;
  if Length(ACharArray) > 0 then
  begin
    for I := Low(ACharArray) to High(ACharArray) do
      Result := StringReplace(Result, '%' + IntToHex(Byte(Ord(ACharArray[I]))), ACharArray[I], [rfReplaceAll]);
  end;
  Result := StringReplace(S, '%%', '%', [rfReplaceAll]);
end;
{$ENDIF}

{ TThreadedSabClient }

constructor TThreadedSabClient.Create(const ABaseURL: string);
begin
  inherited Create;
  FLock := TCriticalSection.Create;
end;

constructor TThreadedSabClient.Create(const ABaseURL, AUserName, APassword, AAccessKey, ADomain: string);
begin
  inherited Create;
  FLock := TCriticalSection.Create;
end;

destructor TThreadedSabClient.Destroy;
begin
  FLock.Free;
  inherited;
end;

function TThreadedSabClient.Add(BoxId: TSabId; const SabObj: TSabObject; out Id: TSabId; const ForceAdd, FreeObj: Boolean): Boolean;
begin
  FLock.Acquire;
  try
    Result := inherited Add(BoxId, SabObj, Id, ForceAdd, FreeObj);
  finally
    FLock.Release;
  end;
end;

function TThreadedSabClient.Delete(const Obj: TSabObject): Boolean;
begin
  FLock.Acquire;
  try
    Result := inherited Delete(Obj);
  finally
    FLock.Release;
  end;
end;

function TThreadedSabClient.GetRecord(RecId: TSabId; SecretFields, Download: Boolean): TSabRecord;
begin
  FLock.Acquire;
  try
    Result := inherited GetRecord(RecId, SecretFields, Download);
  finally
    FLock.Release;
  end;
end;

function TThreadedSabClient.GetRecords(BoxId: TSabId; const List: TSabList; SecretFields, Download, GetFileSizes: Boolean): Boolean;
begin
  FLock.Acquire;
  try
    Result := inherited GetRecords(BoxId, List, SecretFields, Download, GetFileSizes);
  finally
    FLock.Release;
  end;
end;

function TThreadedSabClient.GetSafeBox(BoxId: TSabId): TSabSafeBox;
begin
  FLock.Acquire;
  try
    Result := inherited GetSafeBox(BoxId);
  finally
    FLock.Release;
  end;
end;

function TThreadedSabClient.GetSafeBoxes(BoxId: TSabId; const List: TSabList; ExpandAll, IncludeGroupBoxes: Boolean): Boolean;
begin
  FLock.Acquire;
  try
    Result := inherited GetSafeBoxes(BoxId, List, ExpandAll, IncludeGroupBoxes);
  finally
    FLock.Release;
  end;
end;

function TThreadedSabClient.Update(const Obj: TSabObject; const AUploadFile: Boolean): Boolean;
begin
  FLock.Acquire;
  try
    Result := inherited Update(Obj, AUploadFile);
  finally
    FLock.Release;
  end;
end;

end.
