
{*******************************************************}
{                                                       }
{       SAB Object Framework                            }
{       SAB OpenSSL Plugin                              }
{                                                       }
{       Copyright (c) 2017-22 TDP                       }
{                                                       }
{****************************************************tk*}

unit SAB.OpenSSL;

{$IFNDEF USE_SSL_0_9}
  {$DEFINE SSL_1_1_1}
{$ENDIF}

interface

uses
  SysUtils, Classes,
  blcksock, synsock, synautil,
{$IFDEF SSL_1_1_1}
  ssl_openssl11_lib, ssl_openssl11,
{$ELSE}
  ssl_openssl_lib, ssl_openssl,
{$ENDIF}
  OpenSSLHeadersExt;

type
  TSabOpenSSL = class(TSSLOpenSSL)
  private
    FCertCABytes: TBytes;
    FImprint: RawByteString;
  protected
    function SetSslKeys: Boolean; override;
    procedure SetVerify; override;

  public
    constructor Create(const Value: TTCPBlockSocket); override;
    destructor Destroy; override;

    property CertCABytes: TBytes read FCertCABytes write FCertCABytes;
    property Imprint: RawByteString read FImprint write FImprint;
  end;

procedure InstallSSLPlugin;

implementation

var
  OpenSSLObject: TSabOpenSSL;

{$HINTS OFF}
function SetVerifyCallback(Ok: Integer; Ctx: PX509_STORE_CTX): Integer; cdecl;
var
  server_cert: PX509;
  fdig: PEVP_MD;
  Depth: Integer;
  cert_sha1_bytes: Integer;
  cert_sha1: AnsiString;
  cert_dn: AnsiString;
  Err: Integer;

  function GetProperties(Ctx: PX509_STORE_CTX; out ACert_DN: AnsiString; out ADepth: Integer): PX509;
  var
    SN: PASN1_INTEGER;
    SNI: Int64;
  begin
    Result := X509_STORE_CTX_get_current_cert(Ctx);
    SetLength(cert_dn, 256);
    X509NameOneline(X509GetSubjectName(Result), cert_dn, 256);
    ACert_DN := PAnsiChar(cert_dn);

    SN := X509_get_serialNumber(Result);
    SNI := ASN1_INTEGER_get(SN);

    ADepth := X509_STORE_CTX_get_error_depth(Ctx);
  end;

begin
  Result := 1;

  if (OK = 1) and Assigned(OpenSSLObject) then
  begin
    if OpenSSLObject.Imprint = '' then
      Exit;

    server_cert := GetProperties(Ctx, cert_dn, Depth);
    if Depth <> 0 then
      Exit;

    fdig := EVP_sha1();

    SetLength(cert_sha1, EVP_MAX_MD_SIZE);
    if X509digest(server_cert, fdig, cert_sha1, cert_sha1_bytes) <> 0 then
    begin
      if cert_sha1_bytes = Length(OpenSSLObject.Imprint) then
      begin
        cert_sha1 := Copy(cert_sha1, 1, Length(OpenSSLObject.Imprint));
        if OpenSSLObject.Imprint = cert_sha1 then
        begin
          Exit;
        end;
      end;
    end;
  end else
  begin
    Err := SslGetVerifyResult(OpenSSLObject.Ssl);
    server_cert := GetProperties(Ctx, cert_dn, Depth);

    Exit;
  end;
  Result := 0;
end;
{$HINTS ON}

{ TSabOpenSSL }

constructor TSabOpenSSL.Create(const Value: TTCPBlockSocket);
begin
  inherited;
end;

destructor TSabOpenSSL.Destroy;
begin
  inherited;
end;

{$HINTS OFF}
function TSabOpenSSL.SetSslKeys: Boolean;
var
  Ret: Integer;
  Bytes: TBytesStream;
  Store: PX509_STORE;
  Cert: PX509;
  CBIO: PBIO;
  Err: Integer;
begin
  Result := False;
  if inherited SetSslKeys then
  begin
    if VerifyCert then
    begin
      if FCertCA <> '' then
      begin
        Store := SSL_CTX_get_cert_store(Fctx);

        CBIO := BIO_new_mem_buf(PAnsiChar(FCertCA), Length(FCertCA));
        Cert := d2i_X509_bio(CBIO, nil);
        if Assigned(Cert) then
        begin
          Ret := X509_STORE_add_cert(Store, Cert);
          if Ret <> 1 then
            Exit;

        end else
          Exit;
      end else
      if Length(FCertCABytes) > 0 then
      begin
        Bytes := TBytesStream.Create(FCertCABytes);
        try
          // get certificate store, its used to verify the SSL connection certificate
          Store := SSL_CTX_get_cert_store(Fctx);

          CBIO := BIO_new_mem_buf(Bytes.Memory, Bytes.Size);
          Cert := d2i_X509_bio(CBIO, nil);

          // if OK, uploaded certificate added to authentication store
          if Assigned(Cert) then
          begin
            Ret := X509_STORE_add_cert(Store, Cert);
            if  Ret <> 1 then
              Exit;
          end else
            Exit;
        finally
          Bytes.Free;
        end;
      end;
    end;
    Result := True;
  end else
    Err := ErrGetError;
end;
{$HINTS ON}

procedure TSabOpenSSL.SetVerify;
begin
  OpenSSLObject := Self;

  SslCtxSetVerify(FCtx, SSL_VERIFY_PEER, @SetVerifyCallback);
end;

{ InstallSSLPlugin }

procedure InstallSSLPlugin;
begin
  if InitSSLInterface then
    SSLImplementation := TSabOpenSSL;
end;

initialization
  OpenSSLObject := nil;

finalization
  OpenSSLObject := nil;

end.

