
{*******************************************************}
{                                                       }
{       SAB Object Framework                            }
{       SAB OpenSSL Plugin                              }
{                                                       }
{       Copyright (c) 2017-23 TDP                       }
{                                                       }
{****************************************************tk*}

unit SAB.OpenSslCrypto;

interface

{$DEFINE OPEN_SSL_LOG}
{$DEFINE BUFFERED_CRYPTO_STREAM_USED}
{$DEFINE FILE_STREAM_EX}

uses
  System.SysUtils, System.Classes, System.RTLConsts,
{$IFDEF FILE_STREAM_EX}
  WinApi.Windows,
{$ENDIF}
  OpenSSLHeadersExt,
  SAB.Types, SAB.Classes;

const
  CRYPT_ERR_BASE = 5000;

  ERR_CIPHER_INIT_FAILED = CRYPT_ERR_BASE + 1;
  ERR_ENCRYPT_INIT_FAILED = CRYPT_ERR_BASE + 2;
  ERR_DECRYPT_INIT_FAILED = CRYPT_ERR_BASE + 3;
  ERR_IVLEN_INIT_FAILED = CRYPT_ERR_BASE + 4;
  ERR_KEY_INIT_FAILED = CRYPT_ERR_BASE + 5;
  ERR_AAD_INIT_FAILED = CRYPT_ERR_BASE + 6;
  ERR_ENCRYPT_UPDATE_FAILED = CRYPT_ERR_BASE + 7;
  ERR_DECRYPT_UPDATE_FAILED = CRYPT_ERR_BASE + 8;
  ERR_ENCRYPT_FINAL_FAILED = CRYPT_ERR_BASE + 9;
  ERR_INVALID_OUT_BUFFER_LEN = CRYPT_ERR_BASE + 10;
  ERR_INVALID_TAG_BUFFER_LEN = CRYPT_ERR_BASE + 11;
  ERR_GET_TAG_FAILED = CRYPT_ERR_BASE + 12;
  ERR_SET_TAG_FAILED = CRYPT_ERR_BASE + 13;
  ERR_TAG_VERIFY_FAILED = CRYPT_ERR_BASE + 14;
  ERR_INVALID_KEY_LEN = CRYPT_ERR_BASE + 15;

  IO_READ_ERROR = 100;
  IO_WRITE_ERROR = 101;
  IO_OPEN_ERROR = 102;
  IO_CREATE_ERROR = 103;
  IO_ERROR = 105;

  ERR_FILE_HEDER_INVALID = CRYPT_ERR_BASE + 16;

type
  TErrorMessagesCrypto = array[0..20] of  record Err: Integer; Msg: string end;

const
  ErrorMessagesCrypto: TErrorMessagesCrypto =
    (
      (Err: IO_READ_ERROR; Msg: 'Read error (%s)';),
      (Err: IO_WRITE_ERROR; Msg: 'Write error/disk full (%s)';),
      (Err: IO_OPEN_ERROR; Msg: 'Open file error (%s)';),
      (Err: IO_CREATE_ERROR; Msg: 'Create file error (%s)';),
      (Err: IO_ERROR; Msg: 'I/O error (%s)';),
      (Err: ERR_CIPHER_INIT_FAILED; Msg: 'Cypher initialization failed';),
      (Err: ERR_ENCRYPT_INIT_FAILED; Msg: 'Encrypting initialization failed (%s)';),
      (Err: ERR_DECRYPT_INIT_FAILED; Msg: 'Decrypting initialization failed (%s)';),
      (Err: ERR_IVLEN_INIT_FAILED; Msg: 'Setting Initialization Vector length failed';),
      (Err: ERR_KEY_INIT_FAILED; Msg: 'Key initialization failed';),
      (Err: ERR_AAD_INIT_FAILED; Msg: 'Additional authenticated data initialization failed';),
      (Err: ERR_ENCRYPT_UPDATE_FAILED; Msg: 'Encrypt udate failed (%s)';),
      (Err: ERR_DECRYPT_UPDATE_FAILED; Msg: 'Decrypt update failed (%s)';),
      (Err: ERR_ENCRYPT_FINAL_FAILED; Msg: 'End of encrypting failed (%s)';),
      (Err: ERR_INVALID_OUT_BUFFER_LEN; Msg: 'Invalid output buffer size';),
      (Err: ERR_INVALID_TAG_BUFFER_LEN; Msg: 'Invalid input buffer size';),
      (Err: ERR_GET_TAG_FAILED; Msg: 'Getting Tag failed (%s)';),
      (Err: ERR_SET_TAG_FAILED; Msg: 'Setting Tag failed (%s)';),
      (Err: ERR_TAG_VERIFY_FAILED; Msg: 'Tag verify failed (%s)';),
      (Err: ERR_INVALID_KEY_LEN; Msg: 'Invalid Key lenght';),
      (Err: ERR_FILE_HEDER_INVALID; Msg: 'Invalid file header';)
    );

type
  TOpenSSLGcmCrypto = class;

{$IFDEF BUFFERED_CRYPTO_STREAM_USED}
  // class for slow connections
 {$IFDEF FILE_STREAM_EX}
  TBufferedCryptoFileStreamEx = class(TFileStreamEx)
 {$ELSE}
  TBufferedCryptoFileStream = class(TFileStream)
 {$ENDIF}
  private
    FOwner: TOpenSSLGcmCrypto;
    FFilePos, FBufStartPos, FBufEndPos: Int64;
    FBuffer: PByte;
    FBufferSize: Integer;
    FModified: Boolean;
    FBuffered: Boolean;

    FBufferStep: Integer;
    FStepCount: Integer;
    FStatus: TCryptoStatusEvent;

    FLastError: Integer;
  protected
    procedure SetSize(const NewSize: Int64); override;
    /// <summary>
    ///    SyncBuffer writes buffered and not yet written data to the file.
    ///    When ReRead is True, then buffer will be repopulated from the file.
    ///    When ReRead is False, then buffer will be emptied, so next read or
    ///    write operation will repopulate buffer.
    /// </summary>
    function SyncBuffer(ReRead: Boolean): Boolean;  // returns True if process is not broken
  public
 {$IFDEF FILE_STREAM_EX}
    constructor Create(const AOwner: TOpenSSLGcmCrypto; const AFileName: string; Mode: DWORD;
      BufferSize: Integer = 32768; BufferStep: Integer = 32768; AStepCount: Integer = 50; const AStatus: TCryptoStatusEvent = nil); overload;
    constructor Create(const AOwner: TOpenSSLGcmCrypto; const AFileName: string; Mode: DWORD; Rights: Cardinal; BufferSize: Integer = 32768;
      BufferStep: Integer = 32768; AStepCount: Integer = 50; const AStatus: TCryptoStatusEvent = nil); overload;
    constructor Create(const AOwner: TOpenSSLGcmCrypto; const AHandle: THandle; BufferSize: Integer = 32768;
      BufferStep: Integer = 32768; AStepCount: Integer = 50; const AStatus: TCryptoStatusEvent = nil); overload;
 {$ELSE}
    constructor Create(const AOwner: TOpenSSLGcmCrypto; const AFileName: string; Mode: Word; BufferSize: Integer = 32768;
      BufferStep: Integer = 32768; AStepCount: Integer = 50; const AStatus: TCryptoStatusEvent = nil); overload;
    constructor Create(const AOwner: TOpenSSLGcmCrypto; const AFileName: string; Mode: Word; Rights: Cardinal; BufferSize: Integer = 32768;
      BufferStep: Integer = 32768; AStepCount: Integer = 50; const AStatus: TCryptoStatusEvent = nil); overload;
    constructor Create(const AOwner: TOpenSSLGcmCrypto; const AHandle: THandle; BufferSize: Integer = 32768;
      BufferStep: Integer = 32768; AStepCount: Integer = 50; const AStatus: TCryptoStatusEvent = nil); overload;
 {$ENDIF}
    destructor Destroy; override;
    /// <summary>
    ///    FlushBuffer writes buffered and not yet written data to the file.
    /// </summary>
    procedure FlushBuffer; inline;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(const Offset: Int64; Origin: TSeekOrigin): Int64; override;

    property Owner: TOpenSSLGcmCrypto read FOwner;
    property Status: TCryptoStatusEvent read FStatus;

    property LastError: Integer read FLastError write FLastError;
  end;
{$ENDIF}

  TOpenSSLGcmCrypto = class(TObject)
  strict private
    FIncludedErrNo: Boolean;
    FEncryptedFile, FDecryptedFile: string;
    FInputFile, FOutputFile: string;

    FLastPerc: Integer;
    FLastError: Integer;

    // property LastError: Integer read FLastError write FLastError;

    procedure FreeCtx(const Ctx: PEVP_CIPHER_CTX);
    function CheckKeyAndIv(const Recrypting: Boolean = False): Boolean;

    function DecryptInit(var Ctx: PEVP_CIPHER_CTX; const AIndex: Integer = 0): Integer;
    function DecryptText(const Ctx: PEVP_CIPHER_CTX; const CipherText, PlainText: PByte; CipherTextLen: Integer; var PlainTextLen: Integer): Integer;
    function DecryptFinal(const Ctx: PEVP_CIPHER_CTX; const PlainText: PByte; var PlainTextLen: Integer; const Tag: TBytes): Integer;

    function EncryptInit(var Ctx: PEVP_CIPHER_CTX; const AIndex: Integer = 0): Integer;
    function EncryptText(const Ctx: PEVP_CIPHER_CTX; const PlainText, CipherText: PByte; PlainTextLen: Integer; var CipherTextLen: Integer): Integer;
    function EncryptFinal(const Ctx: PEVP_CIPHER_CTX; const CipherText: PByte; var CipherTextLen: Integer; const Tag: TBytes): Integer;

    function GetAad(Index: Integer): string;
    function GetIv(Index: Integer): string;
    function GetKey(Index: Integer): string;
    procedure SetKey(Index: Integer; const Value: string);
    procedure SetIV(Index: Integer; const Value: string);
    procedure SetAad(Index: Integer; const Value: string);

    procedure HandleException(const ExceptObject: TObject);
    function GetLastErrorMsg: string;
    function GetIvBytes: TBytes;
    function GetKeyBytes: TBytes;

    procedure SetIvBytes(const Value: TBytes);
    procedure SetKeyBytes(const Value: TBytes);
    function GetIvStr: RawByteString;
    function GetKeyStr: RawByteString; protected
  strict protected
    FKey: array [0..1] of TBytes;
    FIv: array [0..1] of TBytes;
    FAad: array [0..1] of TBytes;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    function Encrypt(const InputFile, OutputFile: string; const ABufferSize: Integer = 32768; const AStatus: TCryptoStatusEvent = nil): Integer; overload;
    function Encrypt(const InputStream, OutputStream: TStream; const AStatus: TCryptoStatusEvent = nil): Integer; overload;

    function Decrypt(const InputFile, OutputFile: string; const ABufferSize: Integer = 32768; const AStatus: TCryptoStatusEvent = nil; const AEncrypted: Boolean = False): Integer; overload;
    function Decrypt(const InputStream, OutputStream: TStream; const AStatus: TCryptoStatusEvent = nil): Integer; overload;

    function Recrypt(const InputFile, OutputFile: string; const ABufferSize: Integer = 32768; const AStatus: TCryptoStatusEvent = nil): Integer; overload;
    function Recrypt(const InputStream, OutputStream: TStream; const AStatus: TCryptoStatusEvent = nil): Integer; overload;

    property Key: string index 0 read GetKey write SetKey;
    property Iv: string index 0 read GetIv write SetIv;
    property Aad: string index 0 read GetAad write SetAad;

    property KeyBytes: TBytes read GetKeyBytes write SetKeyBytes;
    property IvBytes: TBytes read GetIvBytes write SetIvBytes;

    property KeyStr: RawByteString read GetKeyStr;
    property IvStr: RawByteString  read GetIvStr;

    property Key2: string index 1 read GetKey write SetKey;
    property Iv2: string index 1 read GetIv write SetIv;
    property Aad2: string index 1 read GetAad write SetAad;

//    property LastError: Integer read FLastError;
    property LastErrorMsg: string read GetLastErrorMsg;
    property IncludedErrNo: Boolean read FIncludedErrNo write FIncludedErrNo;

    property LastPerc: Integer read FLastPerc write FLastPerc;
    property LastError: Integer read FLastError;
  end;

  TOpenSSLCbcCrypto = class(TOpenSSLGcmCrypto)
  public
    constructor Create; override;

    function ComputeIV(const InputStream: TStream): Integer;
    function ComputeKey(const ExportPassword, Salt: string; const Iterations: Integer): Boolean;

    function Decrypt(const Cipher: TBytes; var Plain: TBytes): Integer;

    function PrepareLastPass(const ExportPassword, Salt: string; const Iterations: Integer; const InputStream: TStream; out Cipher: TBytes): Integer;
  end;

function GetCryptoErrorMessage(Err: Integer; InclErr: Boolean; OSGC: TOpenSSLGcmCrypto = nil; const Inp: string = ''; const Outf: string = ''; const Enc: string = ''; const Dec: string = ''): string;

procedure InitOpenSslCrypt;

implementation

uses
  System.Types,
{$IFDEF OPEN_SSL_LOG}
  System.AnsiStrings,
{$ENDIF}
  System.NetEncoding;

const
//  MAX_INTER_STEP = 16 * 1024;
  MAX_INTER_STEP = 1 * 1024;  // for slow connections

  TAG_SIZE_BYTES = 16;
  KEY_SIZE_BYTES = 256 div 8;

  BUFFER_STEP = 32768;
  STEP_COUNT = 100; // 1024;

var
  OpenSslErrStr: string;

{$IFDEF OPEN_SSL_LOG}
function LogErrorToConsole(const str: PAnsiChar; Len: Integer; var Ctx): Integer; cdecl;
begin
  OpenSslErrStr := string(System.AnsiStrings.StrPas(str));
  // writeln(str);

  Result := 0;
end;

procedure LogOpenSslErrors();
var
  U: Pointer;
begin
  U := nil;
  ERR_print_errors_cb(@LogErrorToConsole, U);
end;
{$ENDIF}

{$IFDEF BUFFERED_CRYPTO_STREAM_USED}

{ TBufferedCryptoFileStream }

{$IFDEF FILE_STREAM_EX}
constructor TBufferedCryptoFileStreamEx.Create(const AOwner: TOpenSSLGcmCrypto; const AFileName: string; Mode: DWORD;
  BufferSize, BufferStep, AStepCount: Integer; const AStatus: TCryptoStatusEvent);
{$ELSE}
constructor TBufferedCryptoFileStream.Create(const AOwner: TOpenSSLGcmCrypto; const AFileName: string; Mode: Word; BufferSize,
  BufferStep: Integer; AStepCount: Integer; const AStatus: TCryptoStatusEvent);
{$ENDIF}
begin
  Create(AOwner, AFilename, Mode, 0, BufferSize, BufferStep, AStepCount, AStatus);
end;

{$IFDEF FILE_STREAM_EX}
constructor TBufferedCryptoFileStreamEx.Create(const AOwner: TOpenSSLGcmCrypto; const AFileName: string; Mode: DWORD; Rights: Cardinal;
  BufferSize, BufferStep, AStepCount: Integer; const AStatus: TCryptoStatusEvent);
{$ELSE}
constructor TBufferedCryptoFileStream.Create(const AOwner: TOpenSSLGcmCrypto; const AFileName: string; Mode: Word; Rights: Cardinal;
  BufferSize, BufferStep: Integer; AStepCount: Integer; const AStatus: TCryptoStatusEvent);
{$ENDIF}
begin
  inherited Create(AFileName, Mode);
  FOwner := AOwner;
  FBufferSize := BufferSize;
  GetMem(FBuffer, FBufferSize);
  FBuffered := True;
  FBufferStep := BufferStep;
  FStepCount := AStepCount;
  FStatus := AStatus;
  SyncBuffer(True);
end;

{$IFDEF FILE_STREAM_EX}
constructor TBufferedCryptoFileStreamEx.Create(const AOwner: TOpenSSLGcmCrypto; const AHandle: THandle; BufferSize,
  BufferStep, AStepCount: Integer; const AStatus: TCryptoStatusEvent);
{$ELSE}
constructor TBufferedCryptoFileStream.Create(const AOwner: TOpenSSLGcmCrypto; const AHandle: THandle;
  BufferSize, BufferStep, AStepCount: Integer;
  const AStatus: TCryptoStatusEvent);
{$ENDIF}
begin
{$IFDEF FILE_STREAM_EX}
  THandleStream.Create(AHandle);
{$ELSE}
  inherited Create(AHandle);
{$ENDIF}

  FOwner := AOwner;
  FBufferSize := BufferSize;
  GetMem(FBuffer, FBufferSize);
  FBuffered := True;
  FBufferStep := BufferStep;
  FStepCount := AStepCount;
  FStatus := AStatus;
  SyncBuffer(True);
end;

{$IFDEF FILE_STREAM_EX}
destructor TBufferedCryptoFileStreamEx.Destroy;
{$ELSE}
destructor TBufferedCryptoFileStream.Destroy;
{$ENDIF}
begin
  SyncBuffer(False);
  FreeMem(FBuffer, FBufferSize);
  inherited Destroy;
end;

{$IFDEF FILE_STREAM_EX}
procedure TBufferedCryptoFileStreamEx.FlushBuffer;
{$ELSE}
procedure TBufferedCryptoFileStream.FlushBuffer;
{$ENDIF}
begin
  SyncBuffer(False);
end;

{$IFDEF FILE_STREAM_EX}
function TBufferedCryptoFileStreamEx.Read(var Buffer; Count: Longint): Longint;
{$ELSE}
function TBufferedCryptoFileStream.Read(var Buffer; Count: Longint): Longint;
{$ENDIF}
var
  PSrc: PByte;
begin
  Result := 0;
  if Count >= FBufferSize then
  begin
    if not SyncBuffer(False) then
      Exit;
    Result := inherited Read(Buffer, Count)
  end else
  begin
    if (FBufStartPos > FFilePos) or (FFilePos + Count > FBufEndPos) then
      if not SyncBuffer(True) then
        Exit;
    if Count < FBufEndPos - FFilePos then
      Result := Count
    else
      Result := FBufEndPos - FFilePos;
    PSrc := FBuffer + (FFilePos - FBufStartPos);

    case Result of
      SizeOf(Byte):
        PByte(@Buffer)^ := PByte(PSrc)^;
      SizeOf(Word):
        PWord(@Buffer)^ := PWord(PSrc)^;
      SizeOf(Cardinal):
        PCardinal(@Buffer)^ := PCardinal(PSrc)^;
      SizeOf(UInt64):
        PUInt64(@Buffer)^ := PUInt64(PSrc)^;
    else
      Move(PSrc^, Buffer, Result);
    end;
  end;
  FFilePos := FFilePos + Result;
end;

{$IFDEF FILE_STREAM_EX}
function TBufferedCryptoFileStreamEx.Seek(const Offset: Int64; Origin: TSeekOrigin): Int64;
{$ELSE}
function TBufferedCryptoFileStream.Seek(const Offset: Int64; Origin: TSeekOrigin): Int64;
{$ENDIF}
begin
  if not FBuffered then
    FFilePos := inherited Seek(Offset, Origin)
  else
    case Origin of
      soBeginning:
        begin
          if (Offset < FBufStartPos) or (Offset > FBufEndPos) then
            if not SyncBuffer(False) then
              Exit(-1);
          FFilePos := Offset;
        end;
      soCurrent:
        begin
          if (FFilePos + Offset < FBufStartPos) or (FFilePos + Offset > FBufEndPos) then
            if not SyncBuffer(False) then
              Exit(-1);
          FFilePos := FFilePos + Offset;
        end;
      soEnd:
        begin
          if not SyncBuffer(False) then
            Exit(-1);
          FFilePos := inherited Seek(Offset, soEnd);
        end;
    end;
  Result := FFilePos;
end;

{$IFDEF FILE_STREAM_EX}
procedure TBufferedCryptoFileStreamEx.SetSize(const NewSize: Int64);
{$ELSE}
procedure TBufferedCryptoFileStream.SetSize(const NewSize: Int64);
{$ENDIF}
begin
  if NewSize < FBufEndPos then
    if not SyncBuffer(False) then
      Exit;
  FBuffered := False;
  try
    inherited SetSize(NewSize);
  finally
    FBuffered := True;
  end;
end;

{$IFDEF FILE_STREAM_EX}
function TBufferedCryptoFileStreamEx.SyncBuffer(ReRead: Boolean): Boolean;
{$ELSE}
function TBufferedCryptoFileStream.SyncBuffer(ReRead: Boolean): Boolean;
{$ENDIF}
var
  LLen: Longint;
  BufSize, BufStep, Step: Integer;
  BufPos, BufEnd: Int64;
  Buf: PByte;
begin
  Result := False;

  if FModified then
  begin
    if inherited Seek(FBufStartPos, soBeginning) <> FBufStartPos then
    begin
      LastError := System.GetLastError;
      raise EWriteError.CreateRes(@SWriteError);
    end;
    LLen := Longint(FBufEndPos - FBufStartPos);
    // if inherited Write(FBuffer^, LLen) <> LLen then raise EWriteError.CreateRes(@SWriteError);

    Step := 0;
    Buf := FBuffer;
    BufSize := LLen;
    BufPos := FBufStartPos;
    BufEnd := BufPos;
    while BufSize > FBufferStep do
    begin
      BufStep := inherited Write(Buf^, FBufferStep);
      if BufStep <> FBufferStep then
      begin
        FModified := False; // dont do any operations inside Free
        FLastError := System.GetLastError;
        raise EWriteError.CreateRes(@SWriteError);
      end;
      BufEnd := BufPos + BufStep;

      if Step mod FStepCount = 0 then
      begin
        Step := 0;
        if Assigned(Status) then
          if not FStatus(Owner.LastPerc) then
            Exit;
      end;

      Inc(Step);
      Dec(BufSize, FBufferStep);
      BufPos := BufEnd;
      Inc(Buf, BufStep);
    end;
    if BufSize > 0 then
    begin
      BufStep := inherited Write(Buf^, BufSize);
      if BufStep <> BufSize then
      begin
        FLastError := System.GetLastError;
        raise EWriteError.CreateRes(@SWriteError);
      end;
      BufEnd := BufPos + BufStep;
    end;
    FBufEndPos := BufEnd;

    if Assigned(Status) then
      if not FStatus(Owner.LastPerc) then
        Exit;

    FModified := False;
  end;

  if ReRead then
  begin
    FBufStartPos := inherited Seek(FFilePos, soBeginning);
    // FBufEndPos := FBufStartPos + inherited Read(FBuffer^, FBufferSize);

    Step := 0;
    Buf := FBuffer;
    BufSize := FBufferSize;
    BufPos := FBufStartPos;
    BufEnd := BufPos;
    while BufSize > FBufferStep do
    begin
      BufStep := inherited Read(Buf^, FBufferStep);
      BufEnd := BufPos + BufStep;
      if BufStep <= 0 then
      begin
        BufSize := 0;
        Break;
      end;

      if Step mod FStepCount = 0 then
      begin
        Step := 0;
        if Assigned(FStatus) then
          if not FStatus(Owner.LastPerc) then
            Exit;
      end;

      Inc(Step);
      Dec(BufSize, BufStep);
      BufPos := BufEnd;
      Inc(Buf, BufStep);
    end;
    if BufSize > 0 then
    begin
      BufStep := inherited Read(Buf^, BufSize);
      BufEnd := BufPos + BufStep;
    end;
    FBufEndPos := BufEnd;

    if Assigned(FStatus) then
      if not FStatus(Owner.LastPerc) then
        Exit;
  end else
  begin
    inherited Seek(FFilePos, soBeginning);
    FBufEndPos := FBufStartPos;
  end;
  Result := True;
end;

{$IFDEF FILE_STREAM_EX}
function TBufferedCryptoFileStreamEx.Write(const Buffer; Count: Longint): Longint;
{$ELSE}
function TBufferedCryptoFileStream.Write(const Buffer; Count: Longint): Longint;
{$ENDIF}
var
  PDest: PByte;
begin
  Result := -1;
  if Count >= FBufferSize then
  begin
    if not SyncBuffer(False) then
      Exit;
    Result := inherited Write(Buffer, Count);
    FFilePos := FFilePos + Result;
  end else
  begin
    if (FBufStartPos > FFilePos) or (FFilePos + Count > FBufStartPos + FBufferSize) then
      if not SyncBuffer(True) then
        Exit;
    Result := Count;
    PDest := FBuffer + (FFilePos - FBufStartPos);

    case Result of
      SizeOf(Byte):
        PByte(PDest)^ := PByte(@Buffer)^;
      SizeOf(Word):
        PWord(PDest)^ := PWord(@Buffer)^;
      SizeOf(Cardinal):
        PCardinal(PDest)^ := PCardinal(@Buffer)^;
      SizeOf(UInt64):
        PUInt64(PDest)^ := PUInt64(@Buffer)^;
    else
      Move(Buffer, PDest^, Result);
    end;
    FModified := True;
    FFilePos := FFilePos + Result;
    if FFilePos > FBufEndPos then
      FBufEndPos := FFilePos;
  end;
end;
{$ENDIF}

{ TOpenSSLGcmCrypto }

constructor TOpenSSLGcmCrypto.Create;
var
  I: Integer;
begin
  inherited Create;

  for I := Low(FKey) to High(FKey) do
  begin
    SetLength(FKey[I], 0);
    SetLength(FIV[I], 0);
    SetLength(FAad[I], 0);
  end;
end;

destructor TOpenSSLGcmCrypto.Destroy;
var
  I: Integer;
begin
  for I := Low(FKey) to High(FKey) do
  begin
    SetLength(FKey[I], 0);
    SetLength(FIV[I], 0);
    SetLength(FAad[I], 0);
  end;

  inherited;
end;

function TOpenSSLGcmCrypto.CheckKeyAndIv(const Recrypting: Boolean): Boolean;
begin
  FLastError := 0;
  Result := (Length(FKey[0]) = KEY_SIZE_BYTES) and (Length(FIv[0]) > 0);
  if Result and Recrypting then
    Result := (Length(FKey[1]) = KEY_SIZE_BYTES) and (Length(FIv[1]) > 0);
end;

function TOpenSSLGcmCrypto.Decrypt(const InputStream, OutputStream: TStream; const AStatus: TCryptoStatusEvent): Integer;
var
  Hdr: TExternalFileHeader;
  Ctx: PEVP_CIPHER_CTX;
  Remaining, InputLen: Int64;
  Ret, RdLen, WrLen, Written: Longint;
  InputBufSize, OutputBufSize: Integer;
  InputBuffer, OutputBuffer: PByte;
  Tag: TBytes;

  SrcSize: Int64;
  Perc, Step: Integer;
begin
  if not CheckKeyAndIv then
    Exit(ERR_INVALID_KEY_LEN);

  InputStream.ReadBuffer(Hdr, SizeOf(Hdr));
  if not Hdr.Valid then
    Exit(ERR_FILE_HEDER_INVALID);

  SrcSize := InputStream.Size;
  FLastPerc := 0;
  Step := 0;

  SetLength(Tag, TAG_SIZE_BYTES);

  InputBufSize := 32 * 1024;
  OutputBufSize := InputBufSize + Length(FKey[0]);

  Ctx := nil;
  InputBuffer := AllocMem(InputBufSize);
  OutputBuffer := AllocMem(OutputBufSize);
  try
    Result := DecryptInit(Ctx);

    if Result <> 0 then
      Exit;

    if Assigned(AStatus) then
      AStatus(0);

    Remaining := InputStream.Size - InputStream.Position;

    if Remaining < TAG_SIZE_BYTES then
      Exit(ERR_DECRYPT_INIT_FAILED);

    Dec(Remaining, TAG_SIZE_BYTES);

    while Remaining > 0 do
    begin
      if Remaining >= InputBufSize then
        InputLen := InputBufSize
      else
        InputLen := Remaining;

      RdLen := InputStream.Read(InputBuffer^, InputLen);
      Dec(Remaining, RdLen);

      WrLen := OutputBufSize;
      Ret := DecryptText(Ctx, InputBuffer, OutputBuffer, RdLen, WrLen);

      if Ret = 0 then
      begin
        Written := OutputStream.Write(OutputBuffer^, WrLen);
        if Written <> WrLen then
          // out of disk space!
          Break;

        if Assigned(AStatus) then
        begin
          Perc := (OutputStream.Position * 100) div SrcSize;
          if Perc > FLastPerc then
          begin
            FLastPerc := Perc;
            Step := 0;
            if not AStatus(FLastPerc) then Exit;
          end else
          begin
            Inc(Step);
            if Step > MAX_INTER_STEP then
            begin
              Step := 0;
              if not AStatus(FLastPerc) then Exit;
            end;
          end;
        end;

      end;
    end;

    RdLen := InputStream.Read(Tag, Length(Tag));
    if RdLen <> Length(Tag) then
      Exit(ERR_TAG_VERIFY_FAILED);

    WrLen := OutputBufSize;

    Result := DecryptFinal(Ctx, OutputBuffer, WrLen, Tag);
    if Result = 0 then
    begin
      if WrLen > 0 then // it shouldn't be > 0 in GCM mode (no padding), but just to be sure
      begin
        Written := OutputStream.Write(OutputBuffer^, WrLen);
        if Written <> WrLen then // out of disk space!
          Exit(IO_WRITE_ERROR);
      end;
    end;

    if Assigned(AStatus) then
      AStatus(100);
  finally
    FreeCtx(Ctx);

    FreeMem(InputBuffer);
    FreeMem(OutputBuffer);
  end;
end;

function TOpenSSLGcmCrypto.Decrypt(const InputFile, OutputFile: string; const ABufferSize: Integer; const AStatus: TCryptoStatusEvent; const AEncrypted: Boolean): Integer;
var
{$IFDEF FILE_STREAM_EX}
{$ELSE}
  Handle: THandle;
{$ENDIF}
{$IFDEF BUFFERED_CRYPTO_STREAM_USED}
 {$IFDEF FILE_STREAM_EX}
  InputStream: TBufferedCryptoFileStreamEx;
  OutputStream: TBufferedCryptoFileStreamEx;
 {$ELSE}
  InputStream: TBufferedCryptoFileStream;
  OutputStream: TBufferedCryptoFileStream;
 {$ENDIF}
{$ELSE}
  InputStream: TBufferedFileStream;
  OutputStream: TBufferedFileStream;
{$ENDIF}
begin
  Result := ERR_DECRYPT_INIT_FAILED;

  if not CheckKeyAndIv then
  begin
    FLastError := ERR_INVALID_KEY_LEN;
    Exit(ERR_INVALID_KEY_LEN);
  end;

  InputStream := nil;
  OutputStream := nil;
  try
{$IFDEF BUFFERED_CRYPTO_STREAM_USED}
  {$IFDEF FILE_STREAM_EX}
    InputStream := TBufferedCryptoFileStreamEx.Create(Self, InputFile, fmOpenRead + fmShareDenyWrite, 0,
      ABufferSize, BUFFER_STEP, STEP_COUNT, AStatus);
  {$ELSE}
    InputStream := TBufferedCryptoFileStream.Create(Self, InputFile, fmOpenRead + fmShareDenyWrite, ABufferSize, BUFFER_STEP, STEP_COUNT, AStatus);
  {$ENDIF}
    if AEncrypted then
    begin
    {$IFDEF FILE_STREAM_EX}
      OutputStream := TBufferedCryptoFileStreamEx.Create(Self, OutputFile, fmCreate + SAB.Classes.fmEncrypted + GetRemoteMode(OutputFile),
        ABufferSize, BUFFER_STEP, STEP_COUNT, AStatus);
    {$ELSE}
      Handle := CreateFileEnc(OutputFile);
      if Handle = INVALID_HANDLE_VALUE then
      begin
        FLastError := System.GetLastError;
        raise EFCreateError.CreateResFmt(@SFCreateErrorEx, [ExpandFileName(OutputFile), SysErrorMessage(FLastError)]);
      end;
      OutputStream := TBufferedCryptoFileStream.Create(Self, Handle, ABufferSize, BUFFER_STEP, STEP_COUNT, AStatus);
    {$ENDIF}
    end else
  {$IFDEF FILE_STREAM_EX}
      OutputStream := TBufferedCryptoFileStreamEx.Create(Self, OutputFile, fmCreate + GetRemoteMode(OutputFile),
        ABufferSize, BUFFER_STEP, STEP_COUNT, AStatus);
  {$ELSE}
      OutputStream := TBufferedCryptoFileStream.Create(Self, OutputFile, fmCreate, ABufferSize, BUFFER_STEP, STEP_COUNT, AStatus);
  {$ENDIF}
{$ELSE}
    InputStream := TBufferedFileStream.Create(InputFile, fmOpenRead + fmShareDenyWrite, ABufferSize);
    OutputStream := TBufferedFileStream.Create(OutputFile, fmCreate, ABufferSize);
{$ENDIF}
    try
      FInputFile := InputFile;
      FOutputFile := OutputFile;
      FEncryptedFile := InputFile;
      FDecryptedFile := OutputFile;
      Result := Decrypt(InputStream, OutputStream, AStatus);
      if Result <> 0 then
        FLastError := Result;
    except
      HandleException(System.ExceptObject);
    end;
  finally
    FreeAndNil(InputStream);
    FreeAndNil(OutputStream);
  end;
end;

// Ctx           - cipher context obtained from gcm_decrypt_init
// plaintext     - output buffer for decrypted data
// PlainTextLen  - should be set to plaintext buffer capacity when calling gcm_decrypt_final (must be at least KeyLen bytes)
// Tag           - buffer with authentication Tag
// TagLen        - authentication Tag length in bytes (should be 16)
//
// on success Exit(0 and sets *PlainTextLen to length of encrypted data in bytes (if any - for GCM this is 0) Tag will contain computed authentication Tag
// on failure returns ERR_* error

function TOpenSSLGcmCrypto.DecryptFinal(const Ctx: PEVP_CIPHER_CTX; const PlainText: PByte; var PlainTextLen: Integer; const Tag: TBytes): Integer;
var
  Len, KeyLen, Ret: Integer;
begin
  Len := 0;
  KeyLen := EVP_CIPHER_CTX_key_length(Ctx);

  if PlainTextLen < KeyLen then
  begin
    PlainTextLen := KeyLen;
    Exit(ERR_INVALID_OUT_BUFFER_LEN);
  end;

  if Length(Tag) <> 0 then
  begin
    if Length(Tag) <> TAG_SIZE_BYTES then
      Exit(ERR_INVALID_TAG_BUFFER_LEN);

    { Set expected Tag value. Works in OpenSSL 1.0.1d and later }
    if EVP_CIPHER_CTX_ctrl(Ctx, EVP_CTRL_GCM_SET_TAG, TAG_SIZE_BYTES, Tag) = 0 then
    begin
{$IFDEF OPEN_SSL_LOG}
      LogOpenSslErrors();
{$ENDIF}
      Exit(ERR_SET_TAG_FAILED);
    end;
    Ret := EVP_DecryptFinal_ex(Ctx, PAnsiChar(PlainText), Len);
  end else
    Ret := 1;

  PlainTextLen := Len;

  if Ret > 0 then
  begin
    { Success }
    PlainTextLen := Len;
    Result := 0;
  end else
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Result := ERR_TAG_VERIFY_FAILED;
  end;
end;

// key     - symmetric encryption key (256 bit AES = 256 / 8 bytes = 32 bytes)
// KeyLen - key length in bytes - just to make sure it's the correct key
// iv      - initialization vector - stored alongside the key (unencrypted) - must be identical to encryption iv
// iv_len  - initialization vector length in bytes - can be more than 12 bytes for AES, but it will be hashed to 12 bytes
// aad     - additional authentication date - must be identical to encryption aad
// aad_len - aad length in bytes
//
// on success returns 0 and sets cipher_ctx to initialized EVP_CIPHER_CTX
// on failure returns ERR_* error

function TOpenSSLGcmCrypto.DecryptInit(var Ctx: PEVP_CIPHER_CTX; const AIndex: Integer = 0): Integer;
var
  Len: Integer;
  SKey, SIv: RawByteString;
begin
  Len := 0;

  { Create and initialise the context }
  Ctx := EVP_CIPHER_CTX_new();
  if not Assigned(Ctx) then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_CIPHER_INIT_FAILED);
  end;

  { Initialise the decryption operation. }
  if EVP_DecryptInit_ex(Ctx, EVP_aes_256_gcm(), nil, nil, nil) = 0 then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_DECRYPT_INIT_FAILED);
  end;

  if Length(FKey[AIndex]) <> EVP_CIPHER_CTX_key_length(Ctx) then
    Exit(ERR_INVALID_KEY_LEN);

  { Set IV length. Not necessary if this is 12 bytes (96 bits) }
  if EVP_CIPHER_CTX_ctrl(Ctx, EVP_CTRL_GCM_SET_IVLEN, Length(FIv[AIndex]), nil) = 0 then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_IVLEN_INIT_FAILED);
  end;

  { Initialise key and IV }
  SKey := KeyStr;
  SIv := IvStr;
  if EVP_DecryptInit_ex(Ctx, nil, nil, PAnsiChar(SKey), PAnsiChar(SIv)) = 0 then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_KEY_INIT_FAILED);
  end;

  {
   * Provide any AAD data. This can be called zero or more times as
   * required
   }
  if EVP_DecryptUpdate(Ctx, nil, Len, @FAad[AIndex][0], Length(FAad[AIndex])) = 0 then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_AAD_INIT_FAILED);
  end;

  Result := 0;
end;

// Ctx            - cipher context obtained from gcm_encrypt_init
// CipherText     - output buffer with encrypted data
// CipherTextLen - length of encrypted data in bytes
// plaintext      - output buffer for decrypted data
// plaintext_len  - length of plaintext in bytes - should be set to plaintext buffer capacity when calling gcm_decrypt (must be at least CipherTextLen + KeyLen - 1 bytes)
//
// on success Exit(0 and sets *plaintext_len to length of decrypted data in bytes
// on failure returns ERR_* error and *plaintext_len is set to 0
// on ERR_INVALID_OUT_BUFFER_LEN sets *plaintext_len to minimum required plaintext capacity (computed as plaintext_len + KeyLen - 1)

function TOpenSSLGcmCrypto.DecryptText(const Ctx: PEVP_CIPHER_CTX; const CipherText, PlainText: PByte; CipherTextLen: Integer; var PlainTextLen: Integer): Integer;
var
  Len, KeyLen: Integer;
begin
  Len := 0;
  KeyLen := EVP_CIPHER_CTX_key_length(Ctx);

  if PlainTextLen < CipherTextLen + KeyLen - 1 then
  begin
    PlainTextLen := CipherTextLen + KeyLen - 1;
    Exit(ERR_INVALID_OUT_BUFFER_LEN);
  end;

  if EVP_DecryptUpdate(Ctx, PAnsiChar(PlainText), Len, PAnsiChar(CipherText), CipherTextLen) = 0 then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_DECRYPT_UPDATE_FAILED);
  end;

  PlainTextLen := Len;

  Result := 0;
end;

function TOpenSSLGcmCrypto.Encrypt(const InputFile, OutputFile: string; const ABufferSize: Integer; const AStatus: TCryptoStatusEvent): Integer;
var
  InputStream: TBufferedFileStream;
{$IFDEF BUFFERED_CRYPTO_STREAM_USED}
  {$IFDEF FILE_STREAM_EX}
  OutputStream: TBufferedCryptoFileStreamEx;
  {$ELSE}
  OutputStream: TBufferedCryptoFileStream;
  {$ENDIF}
{$ELSE}
  OutputStream: TBufferedFileStream;
{$ENDIF}
begin
  Result := ERR_ENCRYPT_INIT_FAILED;

  if not CheckKeyAndIv then
  begin
    FLastError := ERR_INVALID_KEY_LEN;
    Exit(ERR_INVALID_KEY_LEN);
  end;

  InputStream := nil;
  OutputStream := nil;
  try
    InputStream := TBufferedFileStream.Create(InputFile, fmOpenRead or fmShareDenyWrite, ABufferSize);
{$IFDEF BUFFERED_CRYPTO_STREAM_USED}
  {$IFDEF FILE_STREAM_EX}
    OutputStream := TBufferedCryptoFileStreamEx.Create(Self, OutputFile, fmCreate + GetRemoteMode(OutputFile),
      ABufferSize, BUFFER_STEP, STEP_COUNT, AStatus);
  {$ELSE}
    OutputStream := TBufferedCryptoFileStream.Create(Self, OutputFile, fmCreate, FILE_ATTRIBUTE_NORMAL + FILE_FLAG_NO_BUFFERING + FILE_FLAG_WRITE_THROUGH,
      ABufferSize, BUFFER_STEP, STEP_COUNT, AStatus);
  {$ENDIF}
{$ELSE}
    OutputStream := TBufferedFileStream.Create(OutputFile, fmCreate, ABufferSize);
{$ENDIF}
    try
      FInputFile := InputFile;
      FOutputFile := OutputFile;
      FEncryptedFile := OutputFile;
      FDecryptedFile := InputFile;
      Result := Encrypt(InputStream, OutputStream, AStatus);
      if Result <> 0 then
        FLastError := Result;
    except
      HandleException(System.ExceptObject);
    end;
  finally
    FreeAndNil(InputStream);
    FreeAndNil(OutputStream);
  end;
end;

function TOpenSSLGcmCrypto.Encrypt(const InputStream, OutputStream: TStream; const AStatus: TCryptoStatusEvent): Integer;
var
  Hdr: TExternalFileHeader;
  Ctx: PEVP_CIPHER_CTX;
  RdLen, WrLen, Written: Longint;
  InputBufSize, OutputBufSize: Integer;
  InputBuffer, OutputBuffer: PByte;
  Tag: TBytes;

  SrcSize: Int64;
  Perc, Step: Integer;
begin
  if not CheckKeyAndIv then
    Exit(ERR_INVALID_KEY_LEN);

  Hdr.Init;
  Written := OutputStream.write(Hdr, SizeOf(Hdr));
  if Written <> SizeOf(Hdr) then
    Exit(IO_WRITE_ERROR);

  SrcSize := InputStream.Size;
  FLastPerc := 0;
  Step := 0;

  SetLength(Tag, TAG_SIZE_BYTES);

  InputBufSize := 32 * 1024;
  OutputBufSize := InputBufSize + Length(FKey[0]);

  Ctx := nil;
  InputBuffer := AllocMem(InputBufSize);
  OutputBuffer := AllocMem(OutputBufSize);
  try
    Result := EncryptInit(Ctx);

    if Result <> 0 then
      Exit;

    if Assigned(AStatus) then
      AStatus(0);

    RdLen := InputStream.Read(InputBuffer^, InputBufSize);
    while RdLen > 0 do
    begin
      WrLen := OutputBufSize;

      Result := EncryptText(Ctx, InputBuffer, OutputBuffer, RdLen, WrLen);

      if Result = 0 then
      begin
        Written := OutputStream.Write(OutputBuffer^, WrLen);

        if Written <> WrLen then
          // out of disk space!
          Exit(IO_WRITE_ERROR); //  Break;

        if Assigned(AStatus) then
        begin
          Perc := (OutputStream.Position * 100) div SrcSize;
          if Perc > FLastPerc then
          begin
            FLastPerc := Perc;
            Step := 0;
            if not AStatus(FLastPerc) then
              Exit;
          end else
          begin
            Inc(Step);
            if Step > MAX_INTER_STEP then
            begin
              Step := 0;
              if not AStatus(FLastPerc) then
                Exit;
            end;
          end;
        end;
      end;
      RdLen := InputStream.Read(InputBuffer^, InputBufSize);
    end;

    WrLen := OutputBufSize;

    Result := EncryptFinal(Ctx, OutputBuffer, WrLen, Tag);
    if Result = 0 then
    begin
      if WrLen > 0 then // it shouldn't be > 0 in GCM mode (no padding), but just to be sure
      begin
        Written := OutputStream.Write(OutputBuffer^, WrLen);
        if Written <> WrLen then // out of disk space!
          Exit(IO_WRITE_ERROR);
      end;
    end;

    if Assigned(AStatus) then
      AStatus(100);

    Written := OutputStream.Write(Tag, Length(Tag));
    if Written <> Length(Tag) then // out of disk space!
      Exit(IO_WRITE_ERROR);
  finally
    FreeCtx(Ctx);

    FreeMem(InputBuffer);
    FreeMem(OutputBuffer);
  end;
end;

// Ctx            - cipher context obtained from gcm_encrypt_init
// ciphertext     - output buffer for encrypted data
// ciphertext_len - should be set to ciphertext buffer capacity when calling gcm_encrypt_final (must be at least key_len bytes)
// tag            - buffer for computed authentication tag of encrypted data
// tag_len        - authentication tag length in bytes (should be 16)
//
// on success return 0 and sets *ciphertext_len to length of encrypted data in bytes (if any - for GCM this is 0) tag will contain computed authentication tag
// on failure returns ERR_* error

function TOpenSSLGcmCrypto.EncryptFinal(const Ctx: PEVP_CIPHER_CTX; const CipherText: PByte; var CipherTextLen: Integer; const Tag: TBytes): Integer;
var
  Len, KeyLen: Integer;
begin
// Tag    - checksum tag (16 bytes)
// TagLen - checksum tag buffer lenght - should be 16

  Len := 0;
  KeyLen := EVP_CIPHER_CTX_key_length(Ctx);

  if CipherTextLen < KeyLen then
  begin
    CipherTextLen := KeyLen;
    Exit(ERR_INVALID_OUT_BUFFER_LEN);
  end;

  if EVP_EncryptFinal_ex(Ctx, PAnsiChar(CipherText), Len) <> 1 then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_ENCRYPT_FINAL_FAILED);
  end;

  if Length(Tag) <> TAG_SIZE_BYTES then
    Exit(ERR_INVALID_TAG_BUFFER_LEN);

  { Get the tag }
  if EVP_CIPHER_CTX_ctrl(Ctx, EVP_CTRL_GCM_GET_TAG, TAG_SIZE_BYTES, Tag) <> 1 then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_GET_TAG_FAILED);
  end;

  CipherTextLen := Len;
  Result := 0;
end;

// key     - symmetric encryption key (256 bit AES = 256 / 8 bytes = 32 bytes)
// key_len - key length in bytes - just to make sure it's the correct key
// iv      - initialization vector - random bytes - must be stored alongside the key (unencrypted) to later decrypt the file
// iv_len  - initialization vector length in bytes - can be more than 12 bytes for AES, but it will be hashed to 12 bytes
// aad     - additional authentication date - confused deputy protection :-) - must be protected and passed to decryption - can be username, file path, or something else
// aad_len - aad length in bytes
//
// on success returns 0 and sets cipher_ctx to initialized EVP_CIPHER_CTX
// on failure returns ERR_* error

function TOpenSSLGcmCrypto.EncryptInit(var Ctx: PEVP_CIPHER_CTX; const AIndex: Integer): Integer;
var
  Len: Integer;
begin
  Len := 0;

  { Create and initialise the context }
  Ctx := EVP_CIPHER_CTX_new();
  if not Assigned(Ctx) then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_CIPHER_INIT_FAILED);
  end;

  { Initialise the encryption operation. }
  if EVP_EncryptInit_ex(Ctx, EVP_aes_256_gcm(), nil, nil, nil) <> 1 then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_ENCRYPT_INIT_FAILED);
  end;

  if Length(FKey[AIndex]) <> EVP_CIPHER_CTX_key_length(Ctx) then
    Exit(ERR_INVALID_KEY_LEN);

  {
   * Set IV length if default 12 bytes (96 bits) is not appropriate
   }
  if EVP_CIPHER_CTX_ctrl(Ctx, EVP_CTRL_GCM_SET_IVLEN, Length(FIv[AIndex]), nil) <> 1 then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_IVLEN_INIT_FAILED);
  end;

  { Initialise key and IV }
  if EVP_EncryptInit_ex(Ctx, nil, nil, @FKey[AIndex][0], @FIv[AIndex][0]) <> 1 then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_KEY_INIT_FAILED);
  end;

  {
   * Provide any AAD data. This can be called zero or more times as
   * required
   }
  if EVP_EncryptUpdate(Ctx, nil, Len, @FAad[AIndex][0], Length(FAad[AIndex])) <> 1 then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_AAD_INIT_FAILED);
  end;

  Result := 0;
end;

// Ctx            - cipher context obtained from gcm_encrypt_init
// plaintext      - data to encrypt
// plaintext_len  - length of plaintext in bytes
// ciphertext     - output buffer for encrypted data
// ciphertext_len - should be set to ciphertext buffer capacity when calling gcm_encrypt (must be at least plaintext_len + key_len - 1 bytes)
//
// on success return 0 and sets *ciphertext_len to length of encrypted data in bytes
// on failure returns ERR_* error and *ciphertext_len is set to 0
// on ERR_INVALID_OUT_BUFFER_LEN sets *ciphertext_len to minimum required ciphertext capacity (computed as plaintext_len + key_len - 1)

function TOpenSSLGcmCrypto.EncryptText(const Ctx: PEVP_CIPHER_CTX; const PlainText, CipherText: PByte; PlainTextLen: Integer; var CipherTextLen: Integer): Integer;
var
  Len, KeyLen: Integer;
begin
  Len := 0;
  KeyLen := EVP_CIPHER_CTX_key_length(Ctx);

  if CipherTextLen < PlainTextLen + KeyLen - 1 then
  begin
    CipherTextLen := PlainTextLen + KeyLen - 1;
    Exit(ERR_INVALID_OUT_BUFFER_LEN);
  end;

  if EVP_EncryptUpdate(Ctx, PansiChar(CipherText), Len, PAnsiChar(PlainText), PlainTextLen) <> 1 then
  begin
{$IFDEF OPEN_SSL_LOG}
    LogOpenSslErrors();
{$ENDIF}
    Exit(ERR_AAD_INIT_FAILED);
  end;

  CipherTextLen := Len;
  Result := 0;
end;

// releases the cipher context (if Ctx == 0 does nothing)
// must be called if gcm_*_init call returned 0 and non-nil CipherCtx

procedure TOpenSSLGcmCrypto.FreeCtx(const Ctx: PEVP_CIPHER_CTX);
begin
  if Assigned(Ctx) then
    EVP_CIPHER_CTX_free(Ctx);
end;

function TOpenSSLGcmCrypto.GetAad(Index: Integer): string;
var
  I: Integer;
begin
  Result := '';
  for I := Low(FAad[Index]) to High(FAad[Index]) do
    Result := Result + IntToHex(FAad[Index][I], 2);
end;

function TOpenSSLGcmCrypto.GetIv(Index: Integer): string;
var
  I: Integer;
begin
  Result := '';
  for I := Low(FIv[Index]) to High(FIv[Index]) do
    Result := Result + IntToHex(FIv[Index][I], 2);
end;

function TOpenSSLGcmCrypto.GetIvBytes: TBytes;
begin
  Result := FIv[0];
end;

function TOpenSSLGcmCrypto.GetIvStr: RawByteString;
var
  I, N: Integer;
begin
  SetLength(Result, Length(FIv[0]) + 1);
  N := 1;
  for I := Low(FIv[0]) to High(FIv[0]) do
  begin
    Result[N] := AnsiChar(FIv[0, I]);
    Inc(N);
  end;
  Result[N] := #0;
end;

function TOpenSSLGcmCrypto.GetKey(Index: Integer): string;
var
  I: Integer;
begin
  Result := '';
  for I := Low(FKey[Index]) to High(FKey[Index]) do
    Result := Result + IntToHex(FKey[Index][I], 2);
end;

function TOpenSSLGcmCrypto.GetKeyBytes: TBytes;
begin
  Result := FKey[0];
end;

function TOpenSSLGcmCrypto.GetKeyStr: RawByteString;
var
  I, N: Integer;
begin
  SetLength(Result, Length(FKey[0]) + 1);
  N := 1;
  for I := Low(FKey[0]) to High(FKey[0]) do
  begin
    Result[N] := AnsiChar(FKey[0, I]);
    Inc(N);
  end;
  Result[N] := #0;
end;

function TOpenSSLGcmCrypto.GetLastErrorMsg: string;
begin
  Result := GetCryptoErrorMessage(FLastError, FIncludedErrNo, Self, FInputFile, FOutputFile, FEncryptedFile, FDecryptedFile);
end;

procedure TOpenSSLGcmCrypto.HandleException(const ExceptObject: TObject);
begin
  if ExceptObject is EWriteError then
    FLastError := IO_WRITE_ERROR
  else if ExceptObject is EReadError then
    FLastError := IO_READ_ERROR
  else if ExceptObject is EFOpenError then
    FLastError := IO_OPEN_ERROR
  else if ExceptObject is EFCreateError then
    FLastError := IO_CREATE_ERROR
  else
    FLastError := IO_ERROR;
end;

function TOpenSSLGcmCrypto.Recrypt(const InputStream, OutputStream: TStream; const AStatus: TCryptoStatusEvent): Integer;
var
  Hdr: TExternalFileHeader;
  SrcCtx, DstCtx: PEVP_CIPHER_CTX;
  Ret, RdLen, WrLen, InterLen, Written: Longint;
  InputBufSize, OutputBufSize, InterBufSize: Integer;
  InputLen, Remaining: Int64;
  InputBuffer, OutputBuffer, InterBuffer: PByte;
  Tag: TBytes;

  SrcSize: Int64;
  Perc, Step: Integer;
begin
  if not CheckKeyAndIv(True) then
    Exit(ERR_INVALID_KEY_LEN);

  InputStream.ReadBuffer(Hdr, SizeOf(Hdr));
  if not Hdr.Valid then
    Exit(ERR_FILE_HEDER_INVALID);

  Written := OutputStream.Write(Hdr, SizeOf(Hdr));
  if Written <> SizeOf(Hdr) then
    Exit(IO_WRITE_ERROR);

  SrcSize := InputStream.Size;
  FLastPerc := 0;
  Step := 0;

  SetLength(Tag, TAG_SIZE_BYTES);

  InputBufSize := 32 * 1024;
  OutputBufSize := InputBufSize + Length(FKey[1]);
  InterBufSize := InputBufSize + Length(FKey[1]);

  SrcCtx := nil;
  DstCtx := nil;
  InputBuffer := AllocMem(InputBufSize);
  InterBuffer := AllocMem(OutputBufSize);
  OutputBuffer := AllocMem(InterBufSize);
  try
    Result := DecryptInit(SrcCtx);
    if Result <> 0 then
      Exit;
    Result := EncryptInit(DstCtx, 1);
    if Result <> 0 then
      Exit;

    if Assigned(AStatus) then
      AStatus(0);

    Remaining := InputStream.Size - InputStream.Position;
    if Remaining < TAG_SIZE_BYTES then
      Exit(ERR_DECRYPT_INIT_FAILED);

    Dec(Remaining, TAG_SIZE_BYTES);

    while Remaining > 0 do
    begin
      if Remaining >= InputBufSize then
        InputLen := InputBufSize
      else
        InputLen := Remaining;

      RdLen := InputStream.Read(InputBuffer^, InputLen);
      Dec(Remaining, RdLen);

      InterLen := InterBufSize;
      Ret := DecryptText(SrcCtx, InputBuffer, InterBuffer, RdLen, InterLen);

      if Ret = 0 then
      begin
        WrLen := OutputBufSize;
        Ret := EncryptText(DstCtx, InterBuffer, OutputBuffer, InterLen, WrLen);

        if Ret = 0 then
        begin
          Written := OutputStream.Write(OutputBuffer^, WrLen);

          if Written <> WrLen then
            // out of disk space!
            Break;

          if Assigned(AStatus) then
          begin
            Perc := (OutputStream.Position * 100) div SrcSize;
            if Perc > FLastPerc then
            begin
              FLastPerc := Perc;
              Step := 0;
              if not AStatus(FLastPerc) then Exit;
            end else
            begin
              Inc(Step);
              if Step > MAX_INTER_STEP then
              begin
                Step := 0;
                if not AStatus(FLastPerc) then Exit;
              end;
            end;
          end;
        end;
      end else
        Exit(Ret);
    end;

    RdLen := InputStream.Read(Tag, Length(Tag));
    if RdLen <> Length(Tag) then
      Exit(ERR_TAG_VERIFY_FAILED);

    WrLen := OutputBufSize;

    Result := DecryptFinal(SrcCtx, OutputBuffer, WrLen, Tag);
    if Result = 0 then
    begin
      WrLen := OutputBufSize;
      Result := EncryptFinal(DstCtx, OutputBuffer, WrLen, Tag);
      if Result = 0 then
      begin
        if WrLen > 0 then // it shouldn't be > 0 in GCM mode (no padding), but just to be sure
        begin
          Written := OutputStream.Write(OutputBuffer^, WrLen);
          if Written <> WrLen then // out of disk space!
            Exit(IO_WRITE_ERROR);
        end;
      end;
    end;

    if Assigned(AStatus) then
      AStatus(100);

    Written := OutputStream.Write(Tag, Length(Tag));
    if Written <> Length(Tag) then // out of disk space!
      Exit(IO_WRITE_ERROR);
  finally
    FreeCtx(SrcCtx);
    FreeCtx(DstCtx);

    FreeMem(InputBuffer);
    FreeMem(InterBuffer);
    FreeMem(OutputBuffer);
  end;
end;

function TOpenSSLGcmCrypto.Recrypt(const InputFile, OutputFile: string; const ABufferSize: Integer; const AStatus: TCryptoStatusEvent): Integer;
var
{$IFDEF BUFFERED_CRYPTO_STREAM_USED}
  {$IFDEF FILE_STREAM_EX}
  InputStream, OutputStream: TBufferedCryptoFileStreamEx;
  {$ELSE}
  InputStream, OutputStream: TBufferedCryptoFileStream;
  {$ENDIF}
{$ELSE}
  InputStream, OutputStream: TBufferedFileStream;
{$ENDIF}
begin
  Result := ERR_DECRYPT_INIT_FAILED;

  if not CheckKeyAndIv(True) then
  begin
    FLastError := ERR_INVALID_KEY_LEN;
    Exit(ERR_INVALID_KEY_LEN);
  end;

  InputStream := nil;
  OutputStream := nil;
  try
{$IFDEF BUFFERED_CRYPTO_STREAM_USED}
  {$IFDEF FILE_STREAM_EX}
    InputStream := TBufferedCryptoFileStreamEx.Create(Self, InputFile, fmOpenRead or fmShareDenyWrite, 0, ABufferSize, BUFFER_STEP, STEP_COUNT, AStatus);
    OutputStream := TBufferedCryptoFileStreamEx.Create(Self, OutputFile, fmCreate + GetRemoteMode(OutputFile), ABufferSize, BUFFER_STEP, STEP_COUNT, AStatus);
  {$ELSE}
    InputStream := TBufferedCryptoFileStream.Create(Self, InputFile, fmOpenRead or fmShareDenyWrite, ABufferSize, BUFFER_STEP, STEP_COUNT, AStatus);
    OutputStream := TBufferedCryptoFileStream.Create(Self, OutputFile, fmCreate, ABufferSize, BUFFER_STEP, STEP_COUNT, AStatus);
  {$ENDIF}
{$ELSE}
    InputStream := TBufferedFileStream.Create(InputFile, fmOpenRead or fmShareDenyWrite, ABufferSize);
    OutputStream := TBufferedFileStream.Create(OutputFile, fmCreate, ABufferSize);
{$ENDIF}
    try
      FInputFile := InputFile;
      FOutputFile := OutputFile;
      FEncryptedFile := OutputFile;
      FDecryptedFile := InputFile;
      Result := Recrypt(InputStream, OutputStream, AStatus);
      if Result <> 0 then
        FLastError := Result;
    except
      HandleException(System.ExceptObject);
    end;
  finally
    FreeAndNil(InputStream);
    FreeAndNil(OutputStream);
  end;
end;

procedure TOpenSSLGcmCrypto.SetAad(Index: Integer; const Value: string);
begin
  SetLength(FAad[Index], Length(Value) div 2);
  System.Classes.HexToBin(PChar(Value), 0, FAad[Index], 0, Length(FAad[Index]));
end;

procedure TOpenSSLGcmCrypto.SetIV(Index: Integer; const Value: string);
begin
  SetLength(FIv[Index], Length(Value) div 2);
  System.Classes.HexToBin(PChar(Value), 0, FIv[Index], 0, Length(FIv[Index]));
end;

procedure TOpenSSLGcmCrypto.SetIvBytes(const Value: TBytes);
begin
  SetLength(FIv[0], Length(Value));
  System.Move(Value[0], FIv[0, 0], Length(Value));
end;

procedure TOpenSSLGcmCrypto.SetKey(Index: Integer; const Value: string);
begin
  SetLength(FKey[Index], Length(Value) div 2);
  System.Classes.HexToBin(PChar(Value), 0, FKey[Index], 0, Length(FKey[Index]));
end;

procedure TOpenSSLGcmCrypto.SetKeyBytes(const Value: TBytes);
begin
  SetLength(FKey[0], Length(Value));
  System.Move(Value[0], FKey[0, 0], Length(Value));
end;

{ TOpenSSLCbcCrypto }

constructor TOpenSSLCbcCrypto.Create;
begin
  inherited;
end;

// Prepare encrypting attributes from LastPass Encrypted File

function TOpenSSLCbcCrypto.ComputeIV(const InputStream: TStream): Integer;
const
  MAX_IV_ZONE_LEN = 128;
  IV_LEN = 16;
var
  Buf1: Byte;
  IvBuf: packed array [0..128 - 1] of Byte;
  IvDyn: TByteDynArray;
  Read, I: Integer;
  Net: TNetEncoding;
  IvT: TBytes;
begin
  Result := 3;

  InputStream.Position := 0;
  FillChar(IvBuf, Length(IvBuf), 0);

  Read := InputStream.Read(Buf1, SizeOf(Buf1));
  if Read = 0 then
    Exit;

  if AnsiChar(Buf1) <> '!' then
    Exit;

  I := 0;
  repeat
    if InputStream.Position >= MAX_IV_ZONE_LEN then
      Exit;

    Read := InputStream.Read(Buf1, SizeOf(Buf1));
    if Read = 0 then
      Exit;

    if I < Length(IvBuf) then
    begin
      IvBuf[I] := Buf1;
      Inc(I);
    end;
  until AnsiChar(Buf1) = '|';

  Net := TNetEncoding.Create;
  try
    SetLength(IvDyn, I - 1);
    for I := Low(IvDyn) to High(IvDyn) do
      IvDyn[I] := IvBuf[I];

    IvT := Net.Base64.Decode(IvDyn);

    if Length(IvT) > IV_LEN then
      SetLength(IvT, IV_LEN);

    FIv[0] := IvT;

  finally
    Net.Free;
  end;

  Result := 0;
end;

function TOpenSSLCbcCrypto.ComputeKey(const ExportPassword, Salt: string; const Iterations: Integer): Boolean;
const
  PROTECTION_KEY_SIZE = 256 div 8;
var
  ComputedKey: packed array [0..PROTECTION_KEY_SIZE - 1] of AnsiChar;
  Pswd, LocSalt: UTF8String;
begin
  FillChar(ComputedKey, SizeOf(ComputedKey), 0);
  Pswd := UTF8Encode(ExportPassword);
  LocSalt := UTF8Encode(Salt);

  Result := PKCS5_PBKDF2_HMAC(PAnsiChar(Pswd), Length(Pswd), PAnsiChar(LocSalt), Length(LocSalt),
    Iterations, EVP_sha256, PROTECTION_KEY_SIZE, @ComputedKey) = 1;

  if Result then
    KeyBytes := BytesOf(@ComputedKey, SizeOf(ComputedKey));
end;

// InputStream must by positioned at firab base64 byte
function TOpenSSLCbcCrypto.Decrypt(const Cipher: TBytes; var Plain: TBytes): Integer;
var
  Ctx: PEVP_CIPHER_CTX;
  OutLen: Integer;
  DecryptedLen: Integer;
  SKey, SIv: RawByteString;
begin
  Result := 0;

  Ctx := nil;
  OutLen := 0;
  DecryptedLen := 0;
  try
    SetLength(Plain, Length(Cipher) + AES_BLOCK_SIZE * 2 + 1);

    Ctx := EVP_CIPHER_CTX_new();
    if not Assigned(Ctx) then
      Exit(ERR_CIPHER_INIT_FAILED);

    SKey := KeyStr;
    SIv := IvStr;
    if EVP_DecryptInit_ex(Ctx, EVP_aes_256_cbc(), nil, PAnsiChar(SKey), PAnsiChar(SIv)) = 0 then
      Exit(ERR_DECRYPT_INIT_FAILED);

    if EVP_DecryptUpdate(Ctx, PAnsiChar(@Plain[0]), OutLen, PAnsiChar(@Cipher[0]), Length(Cipher)) = 0 then
      Exit(ERR_DECRYPT_UPDATE_FAILED);

    Inc(DecryptedLen, OutLen);
    if EVP_DecryptFinal_ex(Ctx, PAnsiChar(@Plain[DecryptedLen]), OutLen) = 0 then
      Exit(ERR_GET_TAG_FAILED);

    Inc(DecryptedLen, Outlen);
    SetLength(Plain, DecryptedLen);
  finally
    if Assigned(Ctx) then
      EVP_CIPHER_CTX_free(Ctx);
  end;
end;

function TOpenSSLCbcCrypto.PrepareLastPass(const ExportPassword, Salt: string;
  const Iterations: Integer; const InputStream: TStream; out Cipher: TBytes): Integer;
var
  Err: Integer;
  BufLen: Integer;
  Buf: TByteDynArray;
  Net: TNetEncoding;
begin
  Result := 0;

  InputStream.Position := 0;

  if not ComputeKey(ExportPassword, Salt, Iterations) then
    Exit(ERR_KEY_INIT_FAILED);

  // ERR_DECRYPT_INIT_FAILED = CRYPT_ERR_BASE + 3;

  Err := ComputeIV(InputStream);
  if Err <> 0 then
    Exit(ERR_IVLEN_INIT_FAILED);

  BufLen := InputStream.Size - InputStream.Position;
  SetLength(Buf, BufLen);
  InputStream.Read(Buf[0], BufLen);

  Net := TNetEncoding.Create;
  try
    Cipher := Net.Base64.Decode(Buf);
  finally
    Net.Free;
  end;
end;

{ GetErrorMessage }

function GetCryptoErrorMessage(Err: Integer; InclErr: Boolean; OSGC: TOpenSSLGcmCrypto; const Inp, Outf, Enc, Dec: string): string;
var
  I: Integer;
begin
  I := Low(ErrorMessagesCrypto);
  while I <= High(ErrorMessagesCrypto) do
  begin
    if ErrorMessagesCrypto[I].Err = Err then
    begin
      case Err of
        IO_READ_ERROR, IO_OPEN_ERROR, IO_ERROR:
          Result := System.SysUtils.Format(ErrorMessagesCrypto[I].Msg, [Inp]);
        IO_WRITE_ERROR, IO_CREATE_ERROR:
          Result := System.SysUtils.Format(ErrorMessagesCrypto[I].Msg, [Outf]);

        ERR_ENCRYPT_INIT_FAILED,
        ERR_ENCRYPT_UPDATE_FAILED,
        ERR_ENCRYPT_FINAL_FAILED,
        ERR_GET_TAG_FAILED:
          Result := System.SysUtils.Format(ErrorMessagesCrypto[I].Msg, [Enc]);

        ERR_DECRYPT_UPDATE_FAILED,
        ERR_DECRYPT_INIT_FAILED,
        ERR_SET_TAG_FAILED,
        ERR_TAG_VERIFY_FAILED:
          Result := System.SysUtils.Format(ErrorMessagesCrypto[I].Msg, [Dec]);
      else
        Result := ErrorMessagesCrypto[I].Msg;
      end;

      if InclErr then
        Result := System.SysUtils.Format('#%d: %s', [Err, Result]);

      Break;
    end;
    Inc(I);
  end;

  if I >= Length(ErrorMessagesCrypto) then
    Result := System.SysUtils.Format('#%d', [Err]) + Result;
end;

{ InitOpenSslCrypt }

procedure InitOpenSslCrypt;
begin
  OpenSSL_add_all_digests;
  OpenSSL_add_all_ciphers;
end;

end.
