
{*******************************************************}
{                                                       }
{       SAB Object Framework                            }
{       SAB Utils                                       }
{                                                       }
{       Copyright (c) 2017-23 TDP                       }
{                                                       }
{****************************************************tk*}

unit SAB.Utils;

interface

uses
  WinApi.Windows, System.SysUtils, System.Classes,
  SAB.Types, SAB.Consts, SAB.Classes;

function GetSabObjectClass(const AType: string): TSabObjectClass;
function GetSabObjectTypeName(const AType: TSabObjectType): string;
function GetSabObjectTypeHumanText(const AType: TSabObjectType): string;

function StringToBytes(const S: string): TBytes;
function StrToLang(const S: string): TSabLanguage;

function UnixTimeToDateTime(UMSec: Int64): TDateTime;
function DateTimeToFileTime(ADateTime: TDateTime; const UseLocalTimeZone: Boolean = False): TFileTime;
function FileTimeToDateTime(const AFileTime: TFileTime): TDateTime;

function GetPasswordQualityIndex(const AEntropy: Double): Integer;
function GetPasswordQualityPerc(const AEntropy: Double): Integer;
function GetAccessCodeQualityIndex(const AEntropy: Double): Integer;
function GetAccessCodeQualityPerc(const AEntropy: Double): Integer;

implementation

uses
  Math, DateUtils;

const
  SabTypeClasses: array [TSabObjectType] of TSabObjectClass = (
    TSabSafeBox,
    TSabSafeBoxGroup,
    TSabAccount,
    TSabSecAccount,
    TSabFile,
    TSabCertificate,
    TSabCreditCard,
    TSabStation,
    TSabConfiguration,
    TSabExternalFile
    );

{ GetSabObjectClass }

function GetSabObjectClass(const AType: string): TSabObjectClass;
var
  I: TSabObjectType;
begin
  for I := Low(SabTypeClasses) to High(SabTypeClasses) do
    if AnsiSameText(AType, SabTypeValues[I]) then
      Exit(SabTypeClasses[I]);

  raise SabException.CreateFmt('Unknown SAB object type/template ("%s")', [AType]);
end;

{ GetSabObjectTypeName }

function GetSabObjectTypeName(const AType: TSabObjectType): string;
begin
  if (AType >= Low(SabTypeValues)) and (AType <= High(SabTypeValues)) then
    Result := SabTypeValues[AType]
  else
    Result := '';
end;

function GetSabObjectTypeHumanText(const AType: TSabObjectType): string;
begin
  if (AType >= Low(SabTypeHumanTexts)) and (AType <= High(SabTypeHumanTexts)) then
    Result := SabTypeHumanTexts[AType]
  else
    Result := '';
end;

{ StringToBytes }
function StringToBytes(const S: string): TBytes;
var
  UTF8Str: UTF8String;
begin
  UTF8Str := Utf8Encode(S);
  SetLength(Result, Length(UTF8Str) + 1);
  if Length(Result) > 0 then
    Move(UTF8Str[1], Result[0], Length(UTF8Str));
  Result[High(Result)] := 0;
end;

{ UnixTimeToDateTime }

const
  UnixStartDate: TDateTime = 25569.0;

function UnixTimeToDateTime(UMSec: Int64): TDateTime;
begin
  Result := UMSec / (1000 * 86400) + UnixStartDate;
end;

function DateTimeToFileTime(ADateTime: TDateTime; const UseLocalTimeZone: Boolean): TFileTime;
var
  UFileTime: TFileTime;
  SysTime: TSystemTime;
begin
  Result.dwLowDateTime := 0;
  Result.dwLowDateTime := 0;
  DecodeDateTime(ADateTime, SysTime.wYear, SysTime.wMonth, SysTime.wDay,
    SysTime.wHour, SysTime.wMinute, SysTime.wSecond, SysTime.wMilliseconds);

  if SystemTimeToFileTime(SysTime, UFileTime) then
    if UseLocalTimeZone then
      LocalFileTimeToFileTime(UFileTime, Result)
    else
      Result := UFileTime;
end;

function FileTimeToDateTime(const AFileTime: TFileTime): TDateTime;
var
  SysTime: TSystemTime;
begin
  if FileTimeToSystemTime(AFileTime, SysTime) then
    Result := SystemTimeToDateTime(SysTime)
  else
    Result := 0;
end;

{ StrToLang }

function StrToLang(const S: string): TSabLanguage;
begin
  for Result := Low(SabLanguages) to High(SabLanguages) do
    if SameText(SabLanguages[Result], S) then
      Exit;
  Result := Low(SabLanguages);
end;

{ StrToLang }

function GetPasswordQualityIndex(const AEntropy: Double): Integer;
begin
  Result := Low(PasswordQualityByEntropy);
  while Result <= High(PasswordQualityByEntropy) do
  begin
    if AEntropy < PasswordQualityByEntropy[Result] then
      Break;
    Inc(Result);
  end;

  Dec(Result);
  Result := Max(Result, 0);
end;

function GetPasswordQualityPerc(const AEntropy: Double): Integer;
begin
  Result := GetPasswordQualityIndex(AEntropy);
  Result := Min(Round((Result / High(PasswordQualityByEntropy)) * 100), 100);
end;

function GetAccessCodeQualityIndex(const AEntropy: Double): Integer;
begin
  Result := GetPasswordQualityIndex(AEntropy);
end;

function GetAccessCodeQualityPerc(const AEntropy: Double): Integer;
begin
  Result := GetPasswordQualityPerc(AEntropy);
end;

end.
