
{*******************************************************}
{                                                       }
{       SAB Object Framework                            }
{       SAB Errors                                      }
{                                                       }
{       Copyright (c) 2017-22 TDP                       }
{                                                       }
{****************************************************tk*}

unit SAB.Errors;

interface

uses
  System.Types, System.Classes, System.SysUtils,
  SAB.Types;

type
  TSabErrorRec = record
    Key: string;
    Msg: string;
  end;

  TSabErrorTable = array of TSabErrorRec;

const
  SDuplicatePart                = '.duplicate';
  SFileDuplicateErr             = 'file' + SDuplicatePart;
  SAccountDuplicateErr          = 'account' + SDuplicatePart;
  SCertificateDuplErr           = 'cert' + SDuplicatePart;
  SStringTooLong                = 'stringtoolong';
  SDuplicateBoxKey              = 'safebox.duplicate';
  SResNotAvail                  = 'resource.not.available';
  SRecordNotFound               = 'record.not.found';

  SSabFileDuplicateKey          = 'file.duplicate';
  SSabExtFileDuplicateKey       = 'external_file.duplicate';
  SSabAccountDuplicateKey       = 'account.duplicate';
  SSabSafeBoxDuplicateKey       = 'safebox.duplicate';
  SErrSecondFailedRequested     = 'err.secondfactor.invalid';

  SSabLicense                   = 'license';
  SSabLicenseDot                = 'license.';
  SSabExceeded                  = 'exceeded';
  SSabExceededDot               = '.exceeded';
  SSabSafeBoxes                 = 'safeboxes';
  SSabSafeBoxGroups             = 'safeboxGroups';

  SSabLicenseSbExceeded         = SSabLicenseDot + SSabSafeBoxes + SSabExceededDot;
  SSabLicenseSbgExceeded        = SSabLicenseDot + SSabSafeBoxGroups + SSabExceededDot;

const
//* ----------------------- HTTP Status Codes  ------------------------- */

//* The size of the static array in http_protocol.c for storing
//* all of the potential response status-lines (a sparse table).
//* A future version should dynamically generate the table at startup.
//* /

  HTTP_RESPONSE_CODES                = 49;

  HTTP_CONTINUE                      = 100;
  HTTP_SWITCHING_PROTOCOLS           = 101;
  HTTP_PROCESSING                    = 102;
  HTTP_RESET_BY_PEER                 = 104;
  HTTP_OK                            = 200;
  HTTP_CREATED                       = 201;
  HTTP_ACCEPTED                      = 202;
  HTTP_NON_AUTHORITATIVE             = 203;
  HTTP_NO_CONTENT                    = 204;
  HTTP_RESET_CONTENT                 = 205;
  HTTP_PARTIAL_CONTENT               = 206;
  HTTP_MULTI_STATUS                  = 207;
  HTTP_MULTIPLE_CHOICES              = 300;
  HTTP_MOVED_PERMANENTLY             = 301;
  HTTP_MOVED_TEMPORARILY             = 302;
  HTTP_SEE_OTHER                     = 303;
  HTTP_NOT_MODIFIED                  = 304;
  HTTP_USE_PROXY                     = 305;
  HTTP_TEMPORARY_REDIRECT            = 307;
  HTTP_BAD_REQUEST                   = 400;
  HTTP_UNAUTHORIZED                  = 401;
  HTTP_PAYMENT_REQUIRED              = 402;
  HTTP_FORBIDDEN                     = 403;
  HTTP_NOT_FOUND                     = 404;
  HTTP_METHOD_NOT_ALLOWED            = 405;
  HTTP_NOT_ACCEPTABLE                = 406;
  HTTP_PROXY_AUTHENTICATION_REQUIRED = 407;
  HTTP_REQUEST_TIME_OUT              = 408;
  HTTP_CONFLICT                      = 409;
  HTTP_GONE                          = 410;
  HTTP_LENGTH_REQUIRED               = 411;
  HTTP_PRECONDITION_FAILED           = 412;
  HTTP_REQUEST_ENTITY_TOO_LARGE      = 413;
  HTTP_REQUEST_URI_TOO_LARGE         = 414;
  HTTP_UNSUPPORTED_MEDIA_TYPE        = 415;
  HTTP_RANGE_NOT_SATISFIABLE         = 416;
  HTTP_EXPECTATION_FAILED            = 417;
  HTTP_UNPROCESSABLE_ENTITY          = 422;
  HTTP_LOCKED                        = 423;
  HTTP_FAILED_DEPENDENCY             = 424;
  HTTP_INTERNAL_SERVER_ERROR         = 500;
  HTTP_NOT_IMPLEMENTED               = 501;
  HTTP_BAD_GATEWAY                   = 502;
  HTTP_SERVICE_UNAVAILABLE           = 503;
  HTTP_GATEWAY_TIME_OUT              = 504;
  HTTP_VERSION_NOT_SUPPORTED         = 505;
  HTTP_VARIANT_ALSO_VARIES           = 506;
  HTTP_INSUFFICIENT_STORAGE          = 507;
  HTTP_NOT_EXTENDED                  = 510;

  CONNECTION_TIMEOUT_ERR             = 10060;

const
  FSAB_ERROR_BASE        = 601;
  SAB_OPENSSL_ERROR      = 998;
  SAB_UNRECOGNIZED_ERROR = 999;

function IsSabErrorTableInitialized: Boolean;
procedure InitSabErrorTable(const ABytes: TBytesStream = nil);
procedure DoneSabErrorTable;

function IsLicenseError(const AErrKey: string): Boolean;

function GetSabError(const AKey: string; const AParams: TStringDynArray; out AResMsg: string; out ACode: Integer): Boolean;
function GetHttpError(const ACode: Integer; out AResMsg: string; ALanguage: TSabLanguage = sablngEnglish): Boolean;

function ParseErrorMessage(const AMsg: string; out Key: string; var AParams: TStringDynArray): Boolean;

implementation

uses
  System.StrUtils, System.JSON, System.Generics.Collections,
  TDP.IntList,
  SAB.Utils;

type
  THttpCode = record
    Code: Integer;
    Msg: string;
  end;
  THttpCodes = array [0..HTTP_RESPONSE_CODES - 1] of THttpCode;
  PHttpCodes = ^THttpCodes;

const
  HttpCodesCze: THttpCodes = (
    (Code: HTTP_CONTINUE                      ; Msg: 'Pokraovn'),  // 100
    (Code: HTTP_SWITCHING_PROTOCOLS           ; Msg: 'Zmna protokolu'),  // 101
    (Code: HTTP_PROCESSING                    ; Msg: 'Nen odpov'),  // 102
    (Code: HTTP_RESET_BY_PEER                 ; Msg: 'Reset spojen'),  // 104
    (Code: HTTP_OK                            ; Msg: 'OK'),  // 200
    (Code: HTTP_CREATED                       ; Msg: 'Nov objekt'),  // 201
    (Code: HTTP_ACCEPTED                      ; Msg: 'Nedokoneno'),  // 202
    (Code: HTTP_NON_AUTHORITATIVE             ; Msg: 'Jin zdroj'),  // 203
    (Code: HTTP_NO_CONTENT                    ; Msg: 'dn data'),  // 204
    (Code: HTTP_RESET_CONTENT                 ; Msg: 'Reset dat'),  // 205
    (Code: HTTP_PARTIAL_CONTENT               ; Msg: 'sten obsah'),  // 206
    (Code: HTTP_MULTI_STATUS                  ; Msg: 'Vce kd'),  // 207
    (Code: HTTP_MULTIPLE_CHOICES              ; Msg: 'Vce monost'),  // 300
    (Code: HTTP_MOVED_PERMANENTLY             ; Msg: 'Pesunuto'),  // 301
    (Code: HTTP_MOVED_TEMPORARILY             ; Msg: 'Doasn pesunuto'),  // 302
    (Code: HTTP_SEE_OTHER                     ; Msg: 'Pouijte jin poadavek'),  // 303
    (Code: HTTP_NOT_MODIFIED                  ; Msg: 'Nezmnno'),  // 304
    (Code: HTTP_USE_PROXY                     ; Msg: 'Pouijte proxy'),  // 305
    (Code: HTTP_TEMPORARY_REDIRECT            ; Msg: 'Doasn mimo'),  // 307
    (Code: HTTP_BAD_REQUEST                   ; Msg: 'Nesprvn poadavek'),  // 400
    (Code: HTTP_UNAUTHORIZED                  ; Msg: 'Neznm jmno nebo nesprvn heslo'), // 'Neautorizovno'),  // 401
    (Code: HTTP_PAYMENT_REQUIRED              ; Msg: 'Poadovna platba'),  // 402
    (Code: HTTP_FORBIDDEN                     ; Msg: 'Odmtnuto'),  // 403
    (Code: HTTP_NOT_FOUND                     ; Msg: 'Nenalezeno'),  // 404
    (Code: HTTP_METHOD_NOT_ALLOWED            ; Msg: 'Nepodporovno'),  // 405
    (Code: HTTP_NOT_ACCEPTABLE                ; Msg: 'Neakceptovno'),  // 406
    (Code: HTTP_PROXY_AUTHENTICATION_REQUIRED ; Msg: 'Proxy neautorizoval'),  // 407
    (Code: HTTP_REQUEST_TIME_OUT              ; Msg: 'Vyprel as'),  // 408
    (Code: HTTP_CONFLICT                      ; Msg: 'Konflikt'),  // 409
    (Code: HTTP_GONE                          ; Msg: 'Nedostupn'),  // 410
    (Code: HTTP_LENGTH_REQUIRED               ; Msg: 'Nedefinovan dlka'),  // 411
    (Code: HTTP_PRECONDITION_FAILED           ; Msg: 'Chybn pednastaven'),  // 412
    (Code: HTTP_REQUEST_ENTITY_TOO_LARGE      ; Msg: 'Pli velk poadavek'),  // 413
    (Code: HTTP_REQUEST_URI_TOO_LARGE         ; Msg: 'Pli dlouh URI'),  // 414
    (Code: HTTP_UNSUPPORTED_MEDIA_TYPE        ; Msg: 'Nepodoporovan mdium'),  // 415
    (Code: HTTP_RANGE_NOT_SATISFIABLE         ; Msg: 'Nevyhovujc rozsah'),  // 416
    (Code: HTTP_EXPECTATION_FAILED            ; Msg: 'Selhalo oekvn'),  // 417
    (Code: HTTP_UNPROCESSABLE_ENTITY          ; Msg: 'Nezpracovateln'),  // 422
    (Code: HTTP_LOCKED                        ; Msg: 'Uzameno'),  // 423
    (Code: HTTP_FAILED_DEPENDENCY             ; Msg: 'Nespch'),  // 424
    (Code: HTTP_INTERNAL_SERVER_ERROR         ; Msg: 'Chyba serveru'),  // 500
    (Code: HTTP_NOT_IMPLEMENTED               ; Msg: 'Neimplementovno'),  // 501
    (Code: HTTP_BAD_GATEWAY                   ; Msg: 'Chyba proxy'),  // 502
    (Code: HTTP_SERVICE_UNAVAILABLE           ; Msg: 'Nedostupn sluba'),  // 503
    (Code: HTTP_GATEWAY_TIME_OUT              ; Msg: 'Vyprel as proxy'),  // 504
    (Code: HTTP_VERSION_NOT_SUPPORTED         ; Msg: 'Nepodporovan protokol'),  // 505
    (Code: HTTP_VARIANT_ALSO_VARIES           ; Msg: 'Jin configurace'),  // 506
    (Code: HTTP_INSUFFICIENT_STORAGE          ; Msg: 'Nedostatek msta'),  // 507
    (Code: HTTP_NOT_EXTENDED                  ; Msg: 'Chyb rozen')   // 510
    );

  HttpCodesEng: THttpCodes = (
    (Code: HTTP_CONTINUE                      ; Msg: 'Continued'),  // 100
    (Code: HTTP_SWITCHING_PROTOCOLS           ; Msg: 'Switching protocols'),  // 101
    (Code: HTTP_PROCESSING                    ; Msg: 'Processing'),  // 102
    (Code: HTTP_RESET_BY_PEER                 ; Msg: 'Reset by peer'),  // 104
    (Code: HTTP_OK                            ; Msg: 'OK'),  // 200
    (Code: HTTP_CREATED                       ; Msg: 'New object'),  // 201
    (Code: HTTP_ACCEPTED                      ; Msg: 'Accepted'),  // 202
    (Code: HTTP_NON_AUTHORITATIVE             ; Msg: 'Other source'),  // 203
    (Code: HTTP_NO_CONTENT                    ; Msg: 'No contrent'),  // 204
    (Code: HTTP_RESET_CONTENT                 ; Msg: 'Reset content'),  // 205
    (Code: HTTP_PARTIAL_CONTENT               ; Msg: 'Partial content'),  // 206
    (Code: HTTP_MULTI_STATUS                  ; Msg: 'More codes'),  // 207
    (Code: HTTP_MULTIPLE_CHOICES              ; Msg: 'More choises'),  // 300
    (Code: HTTP_MOVED_PERMANENTLY             ; Msg: 'Moved'),  // 301
    (Code: HTTP_MOVED_TEMPORARILY             ; Msg: 'Temporarily moved'),  // 302
    (Code: HTTP_SEE_OTHER                     ; Msg: 'Use other request'),  // 303
    (Code: HTTP_NOT_MODIFIED                  ; Msg: 'Not modified'),  // 304
    (Code: HTTP_USE_PROXY                     ; Msg: 'Use proxy'),  // 305
    (Code: HTTP_TEMPORARY_REDIRECT            ; Msg: 'Temporary redirected'),  // 307
    (Code: HTTP_BAD_REQUEST                   ; Msg: 'Bad request'),  // 400
    (Code: HTTP_UNAUTHORIZED                  ; Msg: 'Unauthorized'),  // 401
    (Code: HTTP_PAYMENT_REQUIRED              ; Msg: 'Payment required'),  // 402
    (Code: HTTP_FORBIDDEN                     ; Msg: 'Forbidden'),  // 403
    (Code: HTTP_NOT_FOUND                     ; Msg: 'Not found'),  // 404
    (Code: HTTP_METHOD_NOT_ALLOWED            ; Msg: 'Not allowed'),  // 405
    (Code: HTTP_NOT_ACCEPTABLE                ; Msg: 'Not acceptable'),  // 406
    (Code: HTTP_PROXY_AUTHENTICATION_REQUIRED ; Msg: 'Unauthoriuzed proxy'),  // 407
    (Code: HTTP_REQUEST_TIME_OUT              ; Msg: 'Timeout'),  // 408
    (Code: HTTP_CONFLICT                      ; Msg: 'Conflict'),  // 409
    (Code: HTTP_GONE                          ; Msg: 'Gone'),  // 410
    (Code: HTTP_LENGTH_REQUIRED               ; Msg: 'Length required'),  // 411
    (Code: HTTP_PRECONDITION_FAILED           ; Msg: 'Failed'),  // 412
    (Code: HTTP_REQUEST_ENTITY_TOO_LARGE      ; Msg: 'Too large'),  // 413
    (Code: HTTP_REQUEST_URI_TOO_LARGE         ; Msg: 'Too large URI'),  // 414
    (Code: HTTP_UNSUPPORTED_MEDIA_TYPE        ; Msg: 'Unsupported media type'),  // 415
    (Code: HTTP_RANGE_NOT_SATISFIABLE         ; Msg: 'Not satisfiable'),  // 416
    (Code: HTTP_EXPECTATION_FAILED            ; Msg: 'Expectation failed'),  // 417
    (Code: HTTP_UNPROCESSABLE_ENTITY          ; Msg: 'Unprocessable'),  // 422
    (Code: HTTP_LOCKED                        ; Msg: 'Locked'),  // 423
    (Code: HTTP_FAILED_DEPENDENCY             ; Msg: 'Failed dependency'),  // 424
    (Code: HTTP_INTERNAL_SERVER_ERROR         ; Msg: 'Server error'),  // 500
    (Code: HTTP_NOT_IMPLEMENTED               ; Msg: 'Not implemented'),  // 501
    (Code: HTTP_BAD_GATEWAY                   ; Msg: 'Bad gateway'),  // 502
    (Code: HTTP_SERVICE_UNAVAILABLE           ; Msg: 'Service unavailable'),  // 503
    (Code: HTTP_GATEWAY_TIME_OUT              ; Msg: 'Gateway timeout'),  // 504
    (Code: HTTP_VERSION_NOT_SUPPORTED         ; Msg: 'Version not supported'),  // 505
    (Code: HTTP_VARIANT_ALSO_VARIES           ; Msg: 'Variant also varies'),  // 506
    (Code: HTTP_INSUFFICIENT_STORAGE          ; Msg: 'Insufficient storage'),  // 507
    (Code: HTTP_NOT_EXTENDED                  ; Msg: 'Not extended')   // 510
    );

const
  ERROR_LEADING_TAG  = 'm:';

var
  SabErrorTable: TSabErrorTable;
  SabErrorKeyList: TStringList;

  HttpCodesListCze: TIntegerList;
  HttpCodesListEng: TIntegerList;

function IsSabErrorTableInitialized: Boolean;
begin
  Result := Length(SabErrorTable) > 0;
end;

procedure DoneSabErrorTable;
begin
  SetLength(SabErrorTable, 0);
end;

procedure InitSabErrorTable(const ABytes: TBytesStream);
var
  I: Integer;
  Obj: TJSONObject;
  Rec: TSabErrorRec;
  Bytes: TBytes;
{$IFDEF DEBUG}
  FullTable: TStringList;
{$ENDIF}
begin
  if Assigned(ABytes) then
  begin
    Bytes := ABytes.Bytes;
    SetLength(Bytes, ABytes.Size);

    Obj := TJSONObject.Create;
    if Obj.Parse(Bytes, 0) > 0 then
    begin
      SetLength(SabErrorTable, Obj.Count);
      for I := 0 to Obj.Count - 1 do
      begin
        Rec.Key := Obj.Pairs[I].JsonString.Value;
        Rec.Msg := Obj.Pairs[I].JsonValue.Value;
        SabErrorTable[I] := Rec;
      end;

      for I := Low(SabErrorTable) to High(SabErrorTable) do
        SabErrorKeyList.AddObject(SabErrorTable[I].Key, TObject(I));

{$IFDEF DEBUG}
      FullTable := TStringList.Create(dupAccept, True, False);
      try
        for I := Low(SabErrorTable) to High(SabErrorTable) do
          FullTable.Add(SabErrorTable[I].Key + FullTable.NameValueSeparator + SabErrorTable[I].Msg);

        FullTable.Add('ZZZZZ' + FullTable.NameValueSeparator + 'AAAAA');
      finally
        FullTable.Free;
      end;
{$ENDIF}
    end;
  end else
    SetLength(SabErrorTable, 0);
end;

{ GetHttpError }
function GetHttpError(const ACode: Integer; out AResMsg: string; ALanguage: TSabLanguage): Boolean;
var
  I, Idx: Integer;
  List: TIntegerList;
  Codes: PHttpCodes;
begin
  case ALanguage of
    sablngEnglish,
    sablngGermany,
    sablngFrance:
      begin
        List := HttpCodesListEng;
        Codes := @HttpCodesEng;
      end
  else
    // sablngCzech
    List := HttpCodesListCze;
    Codes := @HttpCodesCze;
  end;
  Result := List.Find(ACode, I);
  if Result then
  begin
    Idx := List.DWords[I];
    AResMsg := System.SysUtils.Format('%d %s ', [ACode, Codes[Idx].Msg]);
  end;
end;

{ GetSabError }

function GetSabError(const AKey: string; const AParams: TStringDynArray; out AResMsg: string; out ACode: Integer): Boolean;
var
  I, Idx: Integer;
  Msg: string;
begin
  Result := False;
  if SabErrorKeyList.Find(AKey, I) then
  begin
    if I <= High(SabErrorTable) then
    begin
      Idx := Integer(SabErrorKeyList.Objects[I]);
      Msg := SabErrorTable[Idx].Msg;

      Idx := Pos('<a href', Msg);
      if Idx > 0 then
        Delete(Msg, Idx, Length(Msg));

      for I := Low(AParams) to High(AParams) do
        Msg := StringReplace(Msg, System.SysUtils.Format('{%d}', [I]), AParams[I], []);

      AResMsg := Msg;
      Result := True;
    end else
      AResMsg := AKey;
  end else
    AResMsg := AKey;
end;

function ParseErrorMessage(const AMsg: string; out Key: string; var AParams: TStringDynArray): Boolean;
var
  M, N: Integer;
  ParStr, Par: string;
begin
  Result := StartsText(ERROR_LEADING_TAG, AMsg);
  if Result then
  begin
    N := Pos('[', AMsg);
    if N > 0 then
    begin
      ParStr := Trim(Copy(AMsg, N + 1, Length(AMsg)));
      Key := Trim(Copy(AMsg, Length(ERROR_LEADING_TAG) + 1, N - Length(ERROR_LEADING_TAG) - 1));

      // for example: m:record.not.found["Agent Config", "Prague Office"]

      if ParStr[Length(ParStr)] = ']' then
        Delete(ParStr, Length(ParStr), 1);

      while ParStr <> '' do
      begin
        M := 1;
        repeat
          N := Pos(',', ParStr, M);
          if (N > 1) and (ParStr[N - 1] <> '\') then
            Break;
          M := N + 1;
        until N = 0;
        if N = 0 then
          N := Length(ParStr) + 1;

        Par := Copy(ParStr, 1, N - 1);
        Par := StringReplace(Par, '\\', '\', [rfReplaceAll]);
        Par := StringReplace(Par, '\"', '"', [rfReplaceAll]);
        Par := StringReplace(Par, '\,', ',', [rfReplaceAll]);

        SetLength(AParams, Length(AParams) + 1);
        AParams[High(AParams)] := AnsiDequotedStr(Par, '"');;

        ParStr := Trim(Copy(ParStr, N + 1, Length(ParStr)));
      end;
    end else
    begin
      SetLength(ParStr, 0);
      Key := Trim(Copy(AMsg, Length(ERROR_LEADING_TAG) + 1, Length(AMsg)));
    end;
  end;
end;

procedure InitSabErrorList;
begin
  SabErrorKeyList := TStringList.Create;
  SabErrorKeyList.CaseSensitive := False;
  SabErrorKeyList.Sorted := True;
  SabErrorKeyList.Duplicates := dupAccept;
end;

procedure InitHttpCodesList;
var
  I: Integer;
begin
  HttpCodesListCze := TIntegerList.Create;
  for I := Low(HttpCodesCze) to High(HttpCodesCze) do
    HttpCodesListCze.AddDWord(HttpCodesCze[I].Code, I);

  HttpCodesListEng := TIntegerList.Create;
  for I := Low(HttpCodesEng) to High(HttpCodesEng) do
    HttpCodesListEng.AddDWord(HttpCodesEng[I].Code, I);
end;

procedure Init;
begin
  InitSabErrorTable;
  InitSabErrorList;
  InitHttpCodesList;
end;

procedure Done;
begin
  SabErrorKeyList.Free;
  HttpCodesListCze.Free;
  HttpCodesListEng.Free;
  DoneSabErrorTable;
end;

function IsLicenseError(const AErrKey: string): Boolean;
begin
  Result := System.StrUtils.StartsText(SSabLicenseDot, AErrKey) and
    System.StrUtils.EndsText(SSabExceededDot, AErrKey);
end;

initialization
  Init;

finalization
  Done;

end.
