
{*******************************************************}
{                                                       }
{       Key Shield Project                              }
{                                                       }
{       OpenSSLHeadersExt.pas                           }
{                                                       }
{       Copyright (c) 2017-2022 TDP                     }
{       All Rights Reserved                             }
{                                                       }
{****************************************************tk*}

unit OpenSSLHeadersExt;

(*
  There is used built-in UNICODE directive here because it has been used since Delphi2009.
  D2009 and newer versions have more common units and types.
*)

{$INCLUDE DEFS.INC}

{$IFNDEF USE_SSL_0_9}
  {$DEFINE SSL_1_1_1}
{$ENDIF}

interface

uses
  System.SysUtils,
{$IFDEF UNICODE}
  IdCTypes,
{$ENDIF}
{$IFDEF INET_INDY}
  IdGlobal,
  IdSSLOpenSSLHeaders;
{$ELSE}
 {$IFDEF SSL_1_1_1}
  ssl_openssl11_lib;
 {$ELSE}
  ssl_openssl_lib;
 {$ENDIF}
{$ENDIF}

const
  AES_BLOCK_SIZE = 16;

{$IF NOT Declared(GEN_OTHERNAME)}
const
  GEN_OTHERNAME = 0;
  GEN_EMAIL = 1;
  GEN_DNS = 2;
  GEN_X400 = 3;
  GEN_DIRNAME = 4;
  GEN_EDIPARTY = 5;
  GEN_URI = 6;
  GEN_IPADD = 7;
  GEN_RID = 8;
{$IFEND}

{$IF NOT Declared(BIO_C_SET_FILENAME)}
const
  BIO_CLOSE = $01;
  BIO_FP_READ = $02;
  BIO_C_SET_FILENAME = 108;
{$IFEND}

{$IF NOT Declared(NID_subject_alt_name)}
const
  NID_subject_alt_name = 85;
{$IFEND}

{$IF NOT Declared(X509_PURPOSE_ANY)}
const
  X509_PURPOSE_ANY = 7;
{$IFEND}

{$IF NOT Declared(SHA_DIGEST_LENGTH)}
const
  SHA_DIGEST_LENGTH = 20;
{$IFEND}

{$IF NOT Declared(X509_V_FLAG_USE_CHECK_TIME)}
  X509_V_FLAG_CB_ISSUER_CHECK             = Cardinal($0);      // Deprecated
  X509_V_FLAG_USE_CHECK_TIME              = Cardinal($2);      // Use check time instead of current time
  X509_V_FLAG_CRL_CHECK                   = Cardinal($4);      // Lookup CRLs
  X509_V_FLAG_CRL_CHECK_ALL               = Cardinal($8);      // Lookup CRLs for whole chain
  X509_V_FLAG_IGNORE_CRITICAL             = Cardinal($10);     // Ignore unhandled critical extensions
  X509_V_FLAG_X509_STRICT                 = Cardinal($20);     // Disable workarounds for broken certificates
  X509_V_FLAG_ALLOW_PROXY_CERTS           = Cardinal($40);     // Enable proxy certificate validation
  X509_V_FLAG_POLICY_CHECK                = Cardinal($80);     // Enable policy checking
  X509_V_FLAG_EXPLICIT_POLICY             = Cardinal($100);    // Policy variable require-explicit-policy
  X509_V_FLAG_INHIBIT_ANY                 = Cardinal($200);    // Policy variable inhibit-any-policy
  X509_V_FLAG_INHIBIT_MAP                 = Cardinal($400);    // Policy variable inhibit-policy-mapping
  X509_V_FLAG_NOTIFY_POLICY               = Cardinal($800);    // Notify callback that policy is OK
  X509_V_FLAG_EXTENDED_CRL_SUPPORT        = Cardinal($1000);   // Extended CRL features such as indirect CRLs, alternate CRL signing keys
  X509_V_FLAG_USE_DELTAS                  = Cardinal($2000);   // Delta CRL support
  X509_V_FLAG_CHECK_SS_SIGNATURE          = Cardinal($4000);   // Check self-signed CA signature
  X509_V_FLAG_TRUSTED_FIRST               = Cardinal($8000);   // Use trusted store first
  X509_V_FLAG_SUITEB_128_LOS_ONLY         = Cardinal($10000);  // Suite B 128 bit only mode: not normally used
  X509_V_FLAG_SUITEB_192_LOS              = Cardinal($20000);  // Suite B 192 bit only mode
  X509_V_FLAG_SUITEB_128_LOS              = Cardinal($30000);  // Suite B 128 bit mode allowing 192 bit algorithms
  X509_V_FLAG_PARTIAL_CHAIN               = Cardinal($80000);  // Allow partial chains if at least one certificate is in trusted store 
  X509_V_FLAG_NO_ALT_CHAINS               = Cardinal($100000); (* If the initial chain is not trusted, do not attempt to build an alternative
                                                                  chain. Alternate chain checking was introduced in 1.1.0. Setting this flag
                                                                  will force the behaviour to match that of previous versions. *)
  X509_V_FLAG_NO_CHECK_TIME               = Cardinal($200000); // Do not check certificate/CRL validity against current time 
{$IFEND}

  SHA256_DIGEST_LENGTH = 32;
  KDF_HASH_LEN = SHA256_DIGEST_LENGTH;

  EVP_CTRL_INIT	            = $00;
  EVP_CTRL_SET_KEY_LENGTH   = $01;
  EVP_CTRL_GET_RC2_KEY_BITS = $02;
  EVP_CTRL_SET_RC2_KEY_BITS = $03;
  EVP_CTRL_GET_RC5_ROUNDS   = $04;
  EVP_CTRL_SET_RC5_ROUNDS   = $05;
  EVP_CTRL_RAND_KEY	    = $06;
  EVP_CTRL_PBE_PRF_NID	    = $07;
  EVP_CTRL_COPY	            = $08;
  EVP_CTRL_GCM_SET_IVLEN    = $09;
  EVP_CTRL_GCM_GET_TAG      = $10;
  EVP_CTRL_GCM_SET_TAG      = $11;
  EVP_CTRL_GCM_SET_IV_FIXED = $12;
  EVP_CTRL_GCM_IV_GEN       = $13;
  EVP_CTRL_CCM_SET_IVLEN    = EVP_CTRL_GCM_SET_IVLEN;
  EVP_CTRL_CCM_GET_TAG      = EVP_CTRL_GCM_GET_TAG;
  EVP_CTRL_CCM_SET_TAG      = EVP_CTRL_GCM_SET_TAG;
  EVP_CTRL_CCM_SET_L        = $14;
  EVP_CTRL_CCM_SET_MSGLEN   = $15;

{$IF NOT Declared(PPByte)}
type
  PPByte = ^PByte;
{$IFEND}

{$IF NOT Declared(pSTACK_OF_X509)}
type
  pSTACK_OF_X509 = PSTACK;
{$IFEND}

{$IF NOT Declared(PX509_POLICY_TREE)}
  PX509_POLICY_TREE = Pointer;
{$IFEND}

type
  EOpenSSLError = Exception;

  EOpenSSLLibError = class(EOpenSSLError)
  private
    FErrorCode: Integer;
  public
    constructor Create(ACode: Integer; const AMsg: string);
    property ErrorCode: Integer read FErrorCode;
  end;

{$IF NOT Declared(PASN1_STRING)}
type
  PASN1_STRING = ^ASN1_STRING;

  ASN1_STRING = record
    length: integer;
    asn1_type: integer;
    data: pointer;
    flags: longint;
	end;

  pASN1_IA5STRING = pASN1_STRING;
  pASN1_INTEGER = pASN1_STRING;
  pASN1_ENUMERATED = pASN1_STRING;
  pASN1_TIME = pASN1_STRING;
  pASN1_OCTET_STRING = pASN1_STRING;
{$IFEND}

{$IF NOT Declared(PASN1_OBJECT)}
type
  ASN1_OBJECT = record
    sn, ln : PAnsiChar;
    nid    : Integer;
    length : Integer;
    data   : PAnsiChar;
    flags  : Integer;
  end;
  PASN1_OBJECT = ^ASN1_OBJECT;
{$IFEND}

{$IF NOT Declared(OTHERNAME)}
type
  pOTHERNAME = ^OTHERNAME;
  OTHERNAME = record
    type_id: pASN1_OBJECT;
    value: pointer;  //pASN1_TYPE;
  end;
{$IFEND}

{$IF NOT Declared(GENERAL_NAME)}
type
  PGENERAL_NAME = ^GENERAL_NAME ;
  PGENERAL_NAMEDATA = record
    case integer of
      GEN_EMAIL: (ia5: pASN1_IA5STRING);
      GEN_IPADD: (ip: pASN1_OCTET_STRING);
      GEN_DIRNAME: (dirn: pX509_NAME);
      GEN_RID: (rid: pASN1_OBJECT);
      GEN_OTHERNAME: (otherName: pOTHERNAME);
      GEN_X400: (other: pointer);
  end;

  GENERAL_NAME = record
    _type: integer;
    d: PGENERAL_NAMEDATA;
  end;
{$IFEND}

{$IF NOT Declared(X509_NAME_ENTRY)}
type
  PX509_NAME_ENTRY = ^X509_NAME_ENTRY;
  X509_NAME_ENTRY = record
    _object : PASN1_OBJECT;
    value : PASN1_STRING;
    _set : Integer;
    size : Integer;
  end;
{$IFEND}

{$IF NOT Declared(ASN1_BOOLEAN)}
  ASN1_BOOLEAN = TIdC_INT;
{$IFEND}

{$IF NOT Declared(ASN1_BIT_STRING)}
  ASN1_BIT_STRING = ASN1_STRING;
  PASN1_BIT_STRING = ^ASN1_BIT_STRING;
{$IFEND}

{$IF NOT Declared(ASN1_PRINTABLESTRING)}
  ASN1_PRINTABLESTRING = ASN1_STRING;
  PASN1_PRINTABLESTRING = ^ASN1_PRINTABLESTRING;
{$IFEND}

{$IF NOT Declared(ASN1_T61STRING)}
  ASN1_T61STRING = ASN1_STRING;
  PASN1_T61STRING = ^ASN1_T61STRING;
{$IFEND}

{$IF NOT Declared(ASN1_GENERALSTRING)}
  ASN1_GENERALSTRING = ASN1_STRING;
  PASN1_GENERALSTRING = ^ASN1_GENERALSTRING;
{$IFEND}

{$IF NOT Declared(ASN1_BMPSTRING)}
  ASN1_BMPSTRING = ASN1_STRING;
  PASN1_BMPSTRING = ^ASN1_BMPSTRING;
{$IFEND}

{$IF NOT Declared(ASN1_UNIVERSALSTRING)}
  ASN1_UNIVERSALSTRING = ASN1_STRING;
  PASN1_UNIVERSALSTRING = ^ASN1_UNIVERSALSTRING;
  PPASN1_UNIVERSALSTRING = ^PASN1_UNIVERSALSTRING;
{$IFEND}

{$IF NOT Declared(ASN1_GENERALIZEDTIME)}
  ASN1_GENERALIZEDTIME = ASN1_STRING;
  PASN1_GENERALIZEDTIME = ^ASN1_GENERALIZEDTIME;
  PPASN1_GENERALIZEDTIME = ^PASN1_GENERALIZEDTIME;
{$IFEND}

{$IF NOT Declared(ASN1_VISIBLESTRING)}
  ASN1_VISIBLESTRING = ASN1_STRING;
  PASN1_VISIBLESTRING = ^ASN1_VISIBLESTRING;
  PPASN1_VISIBLESTRING = ^PASN1_VISIBLESTRING;
{$IFEND}

{$IF NOT Declared(ASN1_UTF8STRING)}
  ASN1_UTF8STRING = ASN1_STRING;
  PASN1_UTF8STRING = ^ASN1_UTF8STRING;
  PPASN1_UTF8STRING = ^PASN1_UTF8STRING;
{$IFEND}

{$IF NOT Declared(ASN1_ENCODING)}
  ASN1_ENCODING = record
    enc: PAnsiChar;
    len: TIdC_LONG;
    modified: TIdC_INT;
  end;
  PASN1_ENCODING = ^ASN1_ENCODING;
  PPASN1_ENCODING = ^PASN1_ENCODING;
{$IFEND}

{$IF NOT Declared(ASN1_TYPE)}
  ASN1_TYPE = record
    case Integer of
      0:  (ptr: PAnsiChar);
      1:  (boolean: ASN1_BOOLEAN);
      2:  (asn1_string: PASN1_STRING);
      3:  (_object: PASN1_OBJECT);
      4:  (integer: PASN1_INTEGER);
      5:  (enumerated: PASN1_ENUMERATED);
      6:  (bit_string: PASN1_BIT_STRING);
      7:  (octet_string: PASN1_OCTET_STRING);
      8:  (printablestring: PASN1_PRINTABLESTRING);
      9:  (t61string: PASN1_T61STRING);
      10: (ia5string: PASN1_IA5STRING);
      11: (generalstring: PASN1_GENERALSTRING);
      12: (bmpstring: PASN1_BMPSTRING);
      13: (universalstring: PASN1_UNIVERSALSTRING);
      14: (utctime: PASN1_UTCTIME);
      15: (generalizedtime: PASN1_GENERALIZEDTIME);
      16: (visiblestring: PASN1_VISIBLESTRING);
      17: (utf8string: PASN1_UTF8STRING);
      { set and sequence are left complete and still
      contain the set or sequence bytes }
      18: (_set: PASN1_STRING);
      19: (sequence: PASN1_STRING);
  end;
  PASN1_TYPE = ^ASN1_TYPE;
  PPASN1_TYPE = ^PASN1_TYPE;
{$IFEND}

{$IF NOT Declared(X509_ALGOR)}
  X509_ALGOR = record
    algorithm : PASN1_OBJECT;
    parameter : PASN1_TYPE;
  end;
  PX509_ALGOR  = ^X509_ALGOR;
  PPX509_ALGOR =^PX509_ALGOR;
{$IFEND}

{$IF NOT Declared(STACK)}
  STACK = record
    num : TIdC_INT; //int num;
    data : PAnsiChar;  //char **data;
    sorted : TIdC_INT;//int sorted;
    num_alloc : TIdC_INT; //int num_alloc;
    comp : function (_para1: PPAnsiChar; _para2: PPAnsiChar):  TIdC_INT; cdecl;
    //int (*comp)(const char * const *, const char * const *);
  end;
  PSTACK          = ^STACK;
  PPSTACK         = ^PSTACK;
  PSTACK_OF_POINTER = Pointer;
{$IFEND}

{$IF NOT Declared(time_t)}
  time_t = TIdC_LONG;
{$IFEND}

{$IF NOT Declared(STACK_OF_ASN1_OBJECT)}
  STACK_OF_ASN1_OBJECT = record
    stack: STACK;
  end;
  PSTACK_OF_ASN1_OBJECT = ^STACK_OF_ASN1_OBJECT;
{$IFEND}

{$IF NOT Declared(X509_VERIFY_PARAM)}
  X509_VERIFY_PARAM = record
    name : PAnsiChar;
    check_time : time_t;          // Time to use
    inh_flags : TIdC_ULONG;       // Inheritance flags
    flags : TIdC_ULONG;           // Various verify flags
    purpose : TIdC_INT;           // purpose to check untrusted certificates
    trust : TIdC_INT;             // trust setting to check
    depth : TIdC_INT;             // Verify depth
    policies : PSTACK_OF_ASN1_OBJECT; // Permissible policies
  end;
  PX509_VERIFY_PARAM = ^X509_VERIFY_PARAM;
{$IFEND}

{$IF NOT Declared(CRYPTO_EX_DATA)}
  CRYPTO_EX_DATA = record
    sk : PSTACK;
    dummy : TIdC_INT; // gcc is screwing up this data structure :-(
  end;
  PCRYPTO_EX_DATA = ^CRYPTO_EX_DATA;
{$IFEND}

{$IF NOT Declared(STACK_OF_X509_LOOKUP)}
  STACK_OF_X509_LOOKUP = record
    _stack: STACK;
  end;
  PSTACK_OF_X509_LOOKUP = ^STACK_OF_X509_LOOKUP;
{$IFEND}

{$IF NOT Declared(STACK_OF_X509_OBJECT)}
  STACK_OF_X509_OBJECT = record
    _stack: STACK;
  end;
  PSTACK_OF_X509_OBJECT = ^STACK_OF_X509_OBJECT;
{$IFEND}

{$IF NOT Declared(STACK_OF_X509_CRL)}
  STACK_OF_X509_CRL = record
    _stack: STACK;
  end;
  PSTACK_OF_X509_CRL = ^STACK_OF_X509_CRL;
{$IFEND}

{$IF NOT Declared(X509_STORE_CTX)}
  PX509_STORE_CTX = ^X509_STORE_CTX;
{$IFEND}

{$IF NOT Declared(PPX509)}
  PPX509 = ^PX509;
{$IFEND}

{$IF NOT Declared(PX509_CRL_METHOD)}
  PX509_CRL_METHOD = Pointer;
{$IFEND}

{$IF NOT Declared(STACK_OF_X509_REVOKED)}
  STACK_OF_X509_REVOKED = record
    _stack: stack;
  end;
  PSTACK_OF_X509_REVOKED = ^STACK_OF_X509_REVOKED;
{$IFEND}

{$IF NOT Declared(STACK_OF_X509_EXTENSION)}
  STACK_OF_X509_EXTENSION = record
    _stack: STACK;
  end;
  X509_EXTENSIONS = STACK_OF_X509_EXTENSION;
  PSTACK_OF_X509_EXTENSION = ^STACK_OF_X509_EXTENSION;
{$IFEND}

{$IF NOT Declared(X509_CRL_INFO)}
  X509_CRL_INFO = record
    version : PASN1_INTEGER;
    sig_alg : PX509_ALGOR;
    issuer : PX509_NAME;
    lastUpdate : PASN1_TIME;
    nextUpdate : PASN1_TIME;
    revoked : PSTACK_OF_X509_REVOKED;
    extensions : PSTACK_OF_X509_EXTENSION; // [0]
    enc : ASN1_ENCODING;
  end;
  PX509_CRL_INFO     = ^X509_CRL_INFO;
  PPX509_CRL_INFO    =^PX509_CRL_INFO;
{$IFEND}

{$IF NOT Declared(AUTHORITY_KEYID)}
  STACK_OF_GENERAL_NAME = record
    _stack: STACK;
  end;
  PSTACK_OF_GENERAL_NAME = ^STACK_OF_GENERAL_NAME;
  PGENERAL_NAMES = PSTACK_OF_GENERAL_NAME;
{$IFEND}

{$IF NOT Declared(AUTHORITY_KEYID)}
  AUTHORITY_KEYID = record
    keyid : PASN1_OCTET_STRING;
    issuer : PGENERAL_NAMES;
    serial : PASN1_INTEGER;
  end;
  PAUTHORITY_KEYID = ^AUTHORITY_KEYID;
{$IFEND}

{$IF NOT Declared(STACK_OF_X509_NAME_ENTRY)}
  STACK_OF_X509_NAME_ENTRY = record
    _stack: STACK;
  end;
  PSTACK_OF_X509_NAME_ENTRY = ^STACK_OF_X509_NAME_ENTRY;
{$IFEND}

{$IF NOT Declared(DIST_POINT_NAME_union)}
  DIST_POINT_NAME_union = record
    case byte of
    0 : (fullname : PGENERAL_NAMES);
    1 : (relativename : PSTACK_OF_X509_NAME_ENTRY);
  end;
{$IFEND}

{$IF NOT Declared(DIST_POINT_NAME)}
  DIST_POINT_NAME = record
    _type : TIdC_INT;
    name : DIST_POINT_NAME_union;
  end;
  PDIST_POINT_NAME = ^DIST_POINT_NAME;
{$IFEND}

{$IF NOT Declared(ISSUING_DIST_POINT)}
  ISSUING_DIST_POINT = record
    distpoint : PDIST_POINT_NAME;
    onlyuser : TIdC_INT;
    onlyCA : TIdC_INT;
    onlysomereasons : PASN1_BIT_STRING;
    indirectCRL : TIdC_INT;
    onlyattr : TIdC_INT;
  end;
  PISSUING_DIST_POINT = ^ISSUING_DIST_POINT;
{$IFEND}

{$IF NOT Declared(STACK_OF_GENERAL_NAMES)}
  STACK_OF_GENERAL_NAMES = record
    _stack: STACK;
  end;
  PSTACK_OF_GENERAL_NAMES = ^STACK_OF_GENERAL_NAMES;
{$IFEND}

{$IF NOT Declared(X509_CRL)}
  X509_CRL = record
    crl : PX509_CRL_INFO;
    sig_alg : PX509_ALGOR;
    signature : PASN1_BIT_STRING;
    references : TIdC_INT;
    flags : TIdC_INT;
    //* Copies of various extensions */
    akid : PAUTHORITY_KEYID;
    idp : PISSUING_DIST_POINT;
    //* Convenient breakdown of IDP */
    idp_flags : TIdC_INT;
    idp_reasons : TIdC_INT;
    //* CRL and base CRL numbers for delta processing */
    crl_number : PASN1_INTEGER;
    base_crl_number : PASN1_INTEGER;
{$ifndef OPENSSL_NO_SHA}
    sha1_hash : array [0..(SHA_DIGEST_LENGTH - 1)] of byte;
{$endif}
    issuers : PSTACK_OF_GENERAL_NAMES;
    meth : PX509_CRL_METHOD;
    meth_data : Pointer;
  end;
  PX509_CRL = ^X509_CRL;
  PPX509_CRL = ^PX509_CRL;
{$IFEND}

{$IF NOT Declared(X509_STORE)}
  X509_STORE = record
    // The following is a cache of trusted certs
    cache : TIdC_INT;               // if true, stash any hits
    objs : PSTACK_OF_X509_OBJECT;   // Cache of all objects
    // These are external lookup methods
    get_cert_methods : PSTACK_OF_X509_LOOKUP;
    param : PX509_VERIFY_PARAM;
    // Callbacks for various operations
    verify : function (ctx : PX509_STORE_CTX) : TIdC_INT; cdecl; // called to verify a certificate
    verify_cb : function (ok : TIdC_INT; ctx : PX509_STORE_CTX) : TIdC_INT; cdecl; // error callback
    get_issuer : function (issuer : PPX509; ctx : PX509_STORE_CTX; x : PX509) : TIdC_INT; cdecl; // get issuers cert from ctx
    check_issued : function (ctx : PX509_STORE_CTX; x : PX509; issuer : PX509) : TIdC_INT; cdecl; // check issued
    check_revocation : function (ctx : PX509_STORE_CTX) : TIdC_INT; cdecl; // Check revocation status of chain
    get_crl : function (ctx : PX509_STORE_CTX; crl : PPX509_CRL; x : PX509) : TIdC_INT; cdecl;// retrieve CRL
    check_crl : function(ctx : PX509_STORE_CTX; crl : PX509_CRL) : TIdC_INT; cdecl; // Check CRL validity
    cert_crl : function(ctx : PX509_STORE_CTX; crl : PX509_CRL; x : PX509) : TIdC_INT; cdecl; // Check certificate against CRL
    lookup_certs : function(ctx : PX509_STORE_CTX; nm : PX509_NAME) : PSTACK_OF_X509 cdecl;
    lookup_crls : function(ctx : PX509_STORE_CTX; nm : PX509_NAME) : PSTACK_OF_X509_CRL cdecl;
    cleanup : function(ctx : PX509_STORE_CTX) : TIdC_INT; cdecl;
    ex_data : CRYPTO_EX_DATA;
    references : TIdC_INT;
  end;
  PX509_STORE = ^X509_STORE;
{$IFEND}

{$IF NOT Declared(X509_STORE_CTX)}
  X509_STORE_CTX = record   // X509_STORE_CTX
    ctx : PX509_STORE;
    current_method : TIdC_INT;  // used when looking up certs
    // The following are set by the caller
    cert : PX509;               // The cert to check
    untrusted : PSTACK_OF_X509; // chain of X509s - untrusted - passed in
    crls : PSTACK_OF_X509_CRL;  // set of CRLs passed in
    param : PX509_VERIFY_PARAM;
    other_ctx : Pointer;  // Other info for use with get_issuer()
    // Callbacks for various operations
    verify : function (ctx : PX509_STORE_CTX) : TIdC_INT; cdecl;  // called to verify a certificate
    verify_cb : function (ok : TIdC_INT; ctx : PX509_STORE_CTX) : TIdC_INT; cdecl; // error callback
    get_issuer : function (var issuer : PX509; ctx : PX509_STORE_CTX; x : PX509) : TIdC_INT; cdecl; // get issuers cert from ctx
    check_issued : function(ctx : PX509_STORE_CTX; x, issuer : PX509) : TIdC_INT; cdecl; // check issued
    check_revocation : function (ctx : PX509_STORE_CTX): TIdC_INT; cdecl; // Check revocation status of chain
    get_crl : function (ctx : PX509_STORE_CTX; var crl : X509_CRL; x : PX509): TIdC_INT; cdecl; // retrieve CRL
    check_crl : function (ctx : PX509_STORE_CTX; var crl : X509_CRL) : TIdC_INT; cdecl; // Check CRL validity
    cert_crl : function (ctx : PX509_STORE_CTX; crl : PX509_CRL; x : PX509) : TIdC_INT; cdecl; // Check certificate against CRL
    check_policy : function (ctx : PX509_STORE_CTX) : TIdC_INT;  cdecl;
    cleanup : function (ctx : PX509_STORE_CTX) : TIdC_INT;  cdecl;
    //* The following is built up */
    valid : TIdC_INT;		//* if 0, rebuild chain */
    last_untrusted : TIdC_INT;	//* index of last untrusted cert */
    chain : PSTACK_OF_X509; 		//* chain of X509s - built up and trusted */
    tree : PX509_POLICY_TREE;	//* Valid policy tree */

    explicit_policy : TIdC_INT;	//* Require explicit policy value */

    //* When something goes wrong, this is why */
    error_depth : TIdC_INT;
    error : TIdC_INT;
    current_cert : PX509;
    current_issuer : PX509;	//* cert currently being tested as valid issuer */
    current_crl : PX509_CRL;	//* current CRL */

    current_crl_score : TIdC_INT;  //* score of current CRL */
    current_reasons : TIdC_UINT;  //* Reason mask */

    parent : PX509_STORE_CTX; //* For CRL path validation: parent context */

    ex_data : CRYPTO_EX_DATA;
  end;
{$IFEND}

{$IF NOT Declared(TPrintErrorCB)}
  TPrintErrorCB = function(Str: PAnsiChar; Len: Integer; var Ctx): Integer; cdecl;
{$IFEND}

{$IF NOT Declared(TSetVerify_cb)}
type
  TSetVerify_cb = function(Ok: Integer; StoreCtx: PX509_STORE_CTX): Integer; cdecl;
{$IFEND}

{$IF NOT Declared(PBIO)}
type
  PBIO = pointer;
{$IFEND}

{$IF NOT Declared(PENGINE)}
  PENGINE = pointer;
{$IFEND}

{$IF NOT Declared(PEVP_CIPHER_CTX)}
  PEVP_CIPHER_CTX = pointer;
{$IFEND}

{$IF NOT Declared(PEVP_CIPHER)}
  PEVP_CIPHER = pointer;
{$IFEND}

{$IF NOT Declared(sk_value)}
var
  sk_value: function (St: Pointer; I: Integer): Pointer; cdecl = nil;
{$IFEND}
{$IF NOT Declared(sk_num)}
var
  sk_num: function (St: Pointer): Integer; cdecl = nil;
{$IFEND}
{$IF NOT Declared(BIO_ctrl)}
var
  BIO_ctrl: function (BP: PBIO; cmd: Integer; larg: Longint; parg: Pointer): Longint; cdecl = nil;
{$IFEND}
{$IF NOT Declared(BIO_new_mem_buf)}
var
  BIO_new_mem_buf: function (buf: Pointer; len: Integer): PBIO; cdecl = nil;
{$IFEND}
{$IF NOT Declared(d2i_X509_bio)}
var
  d2i_X509_bio: function (BP: PBIO; X509: PX509): pX509; cdecl = nil;
{$IFEND}
{$IF NOT Declared(X509_STORE_CTX_init)}
var
  X509_STORE_CTX_init:function(ctx: PX509_STORE_CTX; store: pX509_STORE; x509: pX509; chain: pSTACK_OF_X509): Integer; cdecl = nil;
{$IFEND}
{$IF NOT Declared(X509_STORE_CTX_set_flags)}
var
  X509_STORE_CTX_set_flags: procedure (Ctx: PX509_STORE_CTX; Flags: Cardinal); cdecl = nil;
{$IFEND}
{$IF NOT Declared(X509_STORE_set_flags)}
var
  X509_STORE_set_flags : procedure (ctx: PX509_STORE; flags: Cardinal); cdecl = nil;
{$IFEND}

{$IFDEF CPUX64}
{$ELSE}
  {$IF NOT Declared(SSL_CTX_get0_param)}
  var
    SSL_CTX_get0_param: function(Ctx: PSsl_Ctx): PX509_VERIFY_PARAM; cdecl = nil;
  {$IFEND}
{$ENDIF}

{$IF NOT Declared(X509_VERIFY_PARAM_new)}
var
  X509_VERIFY_PARAM_new: function(): PX509_VERIFY_PARAM; cdecl = nil;
{$IFEND}
{$IF NOT Declared(X509_VERIFY_PARAM_set_flags)}
var
  X509_VERIFY_PARAM_set_flags: procedure(param: PX509_VERIFY_PARAM; flags: Cardinal); cdecl = nil;
{$IFEND}
{$IF NOT Declared(SSL_CTX_set1_param)}
var
  SSL_CTX_set1_param: function(Ctx: PSsl_Ctx; vpm: PX509_VERIFY_PARAM): Integer; cdecl = nil;
{$IFEND}
{$IF NOT Declared(X509_VERIFY_PARAM_free)}
var
  X509_VERIFY_PARAM_free: procedure(vpm: PX509_VERIFY_PARAM); cdecl = nil;
{$IFEND}

{$IF NOT Declared(X509_verify_cert)}
var
  X509_verify_cert: function(ctx: PSsl_Ctx {PX509_STORE_CTX}): Integer; cdecl = nil;
{$IFEND}
//{$IF NOT Declared(X509_STORE_CTX_verify)}
//var
//  X509_STORE_CTX_verify: function(ctx: PSsl_Ctx {PX509_STORE_CTX}): Integer; cdecl = nil;
//{$IFEND}

{$IF NOT Declared(X509_STORE_new)}
var
  X509_STORE_new: function: pX509_STORE; cdecl = nil;
{$IFEND}
{$IF NOT Declared(X509_STORE_free)}
var
  X509_STORE_free: procedure(v: pX509_STORE); cdecl = nil;
{$IFEND}
//{$IF NOT Declared(X509_STORE_add_cert)}
//var
//  X509_STORE_add_cert : function (Ctx: PX509_STORE; X: PX509): Integer cdecl = nil;
//{$IFEND}
{$IF NOT Declared(X509_STORE_CTX_new)}
var
  X509_STORE_CTX_new: function: PX509_STORE_CTX; cdecl = nil;
{$IFEND}
{$IF NOT Declared(X509_STORE_CTX_free)}
var
  X509_STORE_CTX_free: procedure (ctx: PX509_STORE_CTX); cdecl = nil;
{$IFEND}
{$IF NOT Declared(X509_STORE_CTX_set_cert)}
var
  X509_STORE_CTX_set_cert: procedure (ctx: PX509_STORE_CTX; X: PX509); cdecl = nil;
{$IFEND}
{$IF NOT Declared(X509_STORE_CTX_set_purpose)}
var
  X509_STORE_CTX_set_purpose: procedure (ctx: PX509_STORE_CTX; purpose: Integer); cdecl = nil;
{$IFEND}
{$IF NOT Declared(X509_STORE_CTX_cleanup)}
var
  X509_STORE_CTX_cleanup: procedure (ctx: PX509_STORE_CTX); cdecl = nil;
{$IFEND}
{$IF NOT Declared(OBJ_txt2obj)}
var
  OBJ_txt2obj: function(const _S: PAnsiChar; _no_name: Integer): PASN1_OBJECT; cdecl = nil;
{$IFEND}
{$IF NOT Declared(X509_get_ext_d2i)}
var
  X509_get_ext_d2i: function(X: PX509; _NID: Integer; var _crit: Integer; var _idx: Integer): Pointer; cdecl = nil;
{$IFEND}
{$IF NOT Declared(OBJ_cmp)}
var
  OBJ_cmp: function(const A: PASN1_OBJECT; const B: PASN1_OBJECT): Integer; cdecl = nil;
{$IFEND}
{$IF NOT Declared(BIO_free)}
var
  BIO_free : function(Bio: PBIO): Integer cdecl = nil;
{$IFEND}
{$IF NOT Declared(d2i_X509)}
var
  d2i_X509 : function(Pr: PX509; InFile: PPByte; Len: Integer):PX509 cdecl = nil;
{$IFEND}
{$IF NOT Declared(X509_free)}
var
  X509_free : procedure(C: PX509) cdecl = nil;
{$IFEND}
{$IF NOT Declared(ERR_get_error)}
var
  ERR_get_error : function: Integer cdecl = nil;
{$IFEND}
{$IF NOT Declared(ERR_error_string_n)}
var
  ERR_error_string_n: procedure(e: Cardinal; buf: PAnsiChar; len : Integer) cdecl = nil;
{$IFEND}
{$IF NOT Declared(ERR_error_string)}
var
  ERR_error_string: function(e: Cardinal; buf: PAnsiChar): PAnsiChar cdecl = nil;
{$IFEND}
{$IF NOT Declared(ERR_clear_error)}
var
  ERR_clear_error : procedure cdecl = nil;
{$IFEND}
{$IF NOT Declared(_SSLeay_version)}
var
  _SSLeay_version : function(_type : Integer) : PAnsiChar cdecl = nil;
{$IFEND}
{$IF NOT Declared(SSLeay)}
var
  SSLeay : function : Longint cdecl = nil;
{$IFEND}

{$IF NOT Declared(RAND_bytes)}
var
  RAND_bytes : function(buf: PAnsiChar; num: Integer): Integer cdecl = nil;
{$IFEND}
{$IF NOT Declared(RAND_pseudo_bytes)}
var
  RAND_pseudo_bytes : function (buf: PAnsiChar; num: Integer): Integer cdecl = nil;
{$IFEND}
{$IF NOT Declared(RAND_status)}
var
  RAND_status: function: Integer cdecl = nil;
{$IFEND}
{$IF NOT Declared(RAND_poll)}
var
  RAND_poll: function: Integer cdecl = nil;
{$IFEND}
{$IF NOT Declared(RAND_file_name)}
var
  RAND_file_name: function (buf: PAnsiChar; num: Integer): PAnsiChar cdecl = nil;
{$IFEND}
{$IF NOT Declared(RAND_load_file)}
var
  RAND_load_file: function (filename: PAnsiChar; max_bytes: Integer): Integer cdecl = nil;
{$IFEND}
{$IF NOT Declared(RAND_write_file)}
var
  RAND_write_file: function (filename: PAnsiChar): Integer cdecl = nil;
{$IFEND}
//{$IF NOT Declared(SSL_CTX_get_cert_store)}
//var
//  SSL_CTX_get_cert_store: function (ctx : PX509_STORE_CTX): PX509_STORE cdecl = nil;
//{$IFEND}
{$IF NOT Declared(X509_STORE_CTX_get_current_cert)}
var
  X509_STORE_CTX_get_current_cert : function(ctx: PX509_STORE_CTX): PX509 cdecl = nil;
{$IFEND}
{$IF NOT Declared(EVP_sha1)}
var
  EVP_sha1 : function: PEVP_MD cdecl = nil;
{$IFEND}

{$IF NOT Declared(X509_STORE_CTX_get_error_depth)}
var
  X509_STORE_CTX_get_error_depth : function(ctx: PX509_STORE_CTX): Integer cdecl = nil;
{$IFEND}

{$IF NOT Declared(X509_get_serialNumber)}
var
  X509_get_serialNumber : function(X: PX509): PASN1_INTEGER; cdecl = nil;
{$ENDIF}
{$IF NOT Declared(X509_check_ca)}
  X509_check_ca : function(cert: PX509): Integer; cdecl = nil;
{$ENDIF}
{$IF NOT Declared(X509_get1_email)}
  X509_get1_email : function(cert: PX509): PSTACK; cdecl = nil;
{$ENDIF}
{$IF NOT Declared(X509_get_subject_name)}
  X509_get_subject_name : function(cert: PX509): PX509_NAME; cdecl = nil;
{$ENDIF}

{$IF NOT Declared(SSL_CTX_load_verify_locations)}
  SSL_CTX_load_verify_locations : function(ctx: PSsl_Ctx; CAFile: PAnsiChar; CAPath: PAnsiChar): Integer; cdecl = nil;
{$ENDIF}

{$IF NOT Declared(ASN1_INTEGER_get)}
  ASN1_INTEGER_get : function (const A: PASN1_INTEGER): Int64; cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_CIPHER_CTX_new)}
  EVP_CIPHER_CTX_new : function : PEVP_CIPHER_CTX cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_CIPHER_CTX_free)}
  EVP_CIPHER_CTX_free: procedure(a : PEVP_CIPHER_CTX) cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_CIPHER_CTX_key_length)}
  EVP_CIPHER_CTX_key_length : function(ctx : PEVP_CIPHER_CTX) : TIdC_INT cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_CIPHER_CTX_ctrl)}
  EVP_CIPHER_CTX_ctrl : function(ctx : PEVP_CIPHER_CTX; _type : TIdC_INT; arg : TIdC_INT; ptr : Pointer) : TIdC_INT cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_EncryptInit_ex)}
  EVP_EncryptInit_ex : function(ctx : PEVP_CIPHER_CTX; cipher : PEVP_CIPHER; impl : PENGINE;
    key : PAnsiChar; iv : PAnsiChar) : TIdC_INT cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_aes_256_gcm)}
  EVP_aes_256_gcm: function : PEVP_CIPHER cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_aes_256_cbc)}
  EVP_aes_256_cbc: function : PEVP_CIPHER cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_EncryptUpdate)}
  EVP_EncryptUpdate : function(ctx : PEVP_CIPHER_CTX; _out : PAnsiChar;
          var outl : TIdC_INT; _in : PAnsiChar; inl : TIdC_INT) : TIdC_INT cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_EncryptFinal_ex)}
  EVP_EncryptFinal_ex : function(ctx : PEVP_CIPHER_CTX; _out : PAnsiChar; var outl : TIdC_INT) : TIdC_INT cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_DecryptInit_ex)}
  EVP_DecryptInit_ex : function(ctx : PEVP_CIPHER_CTX; cipher : PEVP_CIPHER; impl : PENGINE;
		key : PAnsiChar; iv : PAnsiChar) : TIdC_INT cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_DecryptUpdate)}
  EVP_DecryptUpdate : function(ctx : PEVP_CIPHER_CTX; _out : PAnsiChar;
		var outl : TIdC_INT; _in : PAnsiChar; inl : TIdC_INT) : TIdC_INT cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_DecryptFinal_ex)}
  EVP_DecryptFinal_ex : function(ctx : PEVP_CIPHER_CTX; outm: PAnsiChar; var outl : TIdC_INT) : TIdC_INT cdecl = nil;
{$ENDIF}

{$IF NOT Declared(ERR_print_errors_cb)}
  ERR_print_errors_cb : procedure(CB: TPrintErrorCB; var U); cdecl = nil;
{$ENDIF}

{$IF NOT Declared(OPENSSL_add_all_algorithms_conf)}
  OPENSSL_add_all_algorithms_conf: procedure cdecl = nil;
{$ENDIF}

{$IF NOT Declared(OPENSSL_add_all_algorithms_noconf)}
  OPENSSL_add_all_algorithms_noconf: procedure cdecl = nil;
{$ENDIF}

{$IF NOT Declared(OpenSSL_add_all_ciphers)}
  OpenSSL_add_all_ciphers : procedure cdecl = nil;
{$ENDIF}

{$IF NOT Declared(OpenSSL_add_all_digests)}
  OpenSSL_add_all_digests : procedure cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_CIPHER_CTX_init)}
  EVP_CIPHER_CTX_init : procedure(Ctx : PEVP_CIPHER_CTX) cdecl = nil;
{$ENDIF}

{$IF NOT Declared(PKCS5_PBKDF2_HMAC)}
  PKCS5_PBKDF2_HMAC : function(pass : PAnsiChar; passlen : TIdC_INT;
                        salt : PAnsiChar; saltlen : TIdC_INT; iter : TIdC_INT;
                        digest : PEVP_MD;
                        keylen : TIdC_INT; _out : PAnsiChar) : TIdC_INT cdecl = nil;
{$ENDIF}

{$IF NOT Declared(EVP_sha256)}
  EVP_sha256 : function : PEVP_MD cdecl = nil;
{$ENDIF}


{$IF NOT Declared(SSL_load_error_strings)}
  SSL_load_error_strings: procedure cdecl = nil;
{$ENDIF}

function sk_GENERAL_NAMES_value(Stack: PGENERAL_NAME; I: Integer): PGENERAL_NAME;
function sk_GENERAL_NAMES_num(const Stack: PGENERAL_NAME): Integer;
function BIO_read_filename(BP: PBIO; Filename: PAnsiChar): Integer;

procedure Load(SSL_111: Boolean = False);

procedure RaiseOpenSSLError(const AMessage: string = '');

implementation

uses
  Windows;

const
//{$IFDEF CPUX64}
//  SSLCLIB_DLL_name     = 'libeay64.dll';
//  SSLUTIL_DLL_name     = 'ssleay64.dll';
//{$ELSE}
  SSLCLIB_DLL_name     = 'libeay32.dll';
  SSLUTIL_DLL_name     = 'ssleay32.dll';
//{$ENDIF}

  fn_sk_value = 'sk_value';
  fn_sk_num = 'sk_num';
  fn_BIO_ctrl = 'BIO_ctrl';
  fn_BIO_new_mem_buf = 'BIO_new_mem_buf';
  fn_d2i_X509_bio = 'd2i_X509_bio';
  fn_X509_STORE_CTX_init = 'X509_STORE_CTX_init';
  fn_X509_STORE_CTX_set_flags = 'X509_STORE_CTX_set_flags';
  fn_X509_STORE_set_flags = 'X509_STORE_set_flags';

{$IFDEF CPUX64}
{$ELSE}
  fn_SSL_CTX_get0_param = 'SSL_CTX_get0_param';
{$ENDIF}
  fn_X509_VERIFY_PARAM_new = 'X509_VERIFY_PARAM_new';
  fn_X509_VERIFY_PARAM_set_flags = 'X509_VERIFY_PARAM_set_flags';
  fn_SSL_CTX_set1_param = 'SSL_CTX_set1_param';
  fn_X509_VERIFY_PARAM_free = 'X509_VERIFY_PARAM_free';

  fn_X509_verify_cert = 'X509_verify_cert';
//  fn_X509_STORE_CTX_verify = 'X509_STORE_CTX_verify';
  fn_X509_STORE_new = 'X509_STORE_new';
  fn_X509_STORE_free = 'X509_STORE_free';
  fn_X509_STORE_CTX_new = 'X509_STORE_CTX_new';
  fn_X509_STORE_CTX_free = 'X509_STORE_CTX_free';
  fn_X509_STORE_CTX_set_cert = 'X509_STORE_CTX_set_cert';
  fn_X509_STORE_CTX_set_purpose = 'X509_STORE_CTX_set_purpose';
  fn_X509_STORE_CTX_cleanup = 'X509_STORE_CTX_cleanup';
  fn_OBJ_txt2obj = 'OBJ_txt2obj';
  fn_X509_get_ext_d2i = 'X509_get_ext_d2i';
  fn_OBJ_cmp = 'OBJ_cmp';
//  fn_X509_STORE_add_cert = 'X509_STORE_add_cert';
  fn_BIO_free = 'BIO_free';
  fn_d2i_X509 = 'd2i_X509';
  fn_X509_free = 'X509_free';
  fn_ERR_get_error = 'ERR_get_error';
  fn_ERR_error_string_n = 'ERR_error_string_n';
  fn_ERR_error_string = 'ERR_error_string';
  fn_ERR_clear_error = 'ERR_clear_error';
  fn__SSLeay_version = 'SSLeay_version';
  fn_SSLeay = 'SSLeay';

  fn_RAND_bytes = 'RAND_bytes';
  fn_RAND_pseudo_bytes = 'RAND_pseudo_bytes';
  fn_RAND_status = 'RAND_status';
  fn_RAND_poll = 'RAND_poll';
  fn_RAND_file_name = 'RAND_file_name';
  fn_RAND_load_file = 'RAND_load_file';
  fn_RAND_write_file = 'RAND_write_file';

//  fn_SSL_CTX_get_cert_store = 'SSL_CTX_get_cert_store';
  fn_X509_STORE_CTX_get_current_cert = 'X509_STORE_CTX_get_current_cert';
  fn_EVP_sha1 = 'EVP_sha1';
  fn_X509_STORE_CTX_get_error_depth = 'X509_STORE_CTX_get_error_depth';
  fn_X509_get_serialNumber = 'X509_get_serialNumber';
  fn_X509_check_ca = 'X509_check_ca';
  fn_X509_get1_email = 'X509_get1_email';
  fn_X509_get_subject_name = 'X509_get_subject_name';
  fn_SSL_CTX_load_verify_locations = 'SSL_CTX_load_verify_locations';
  fn_ASN1_INTEGER_get = 'ASN1_INTEGER_get';

  fn_EVP_CIPHER_CTX_new = 'EVP_CIPHER_CTX_new';
  fn_EVP_CIPHER_CTX_free = 'EVP_CIPHER_CTX_free';
  fn_EVP_CIPHER_CTX_key_length = 'EVP_CIPHER_CTX_key_length';
  fn_EVP_CIPHER_CTX_ctrl = 'EVP_CIPHER_CTX_ctrl';
  fn_EVP_EncryptInit_ex = 'EVP_EncryptInit_ex';
  fn_EVP_aes_256_gcm = 'EVP_aes_256_gcm';
  fn_EVP_aes_256_cbc = 'EVP_aes_256_cbc';
  fn_EVP_EncryptUpdate = 'EVP_EncryptUpdate';
  fn_EVP_EncryptFinal_ex = 'EVP_EncryptFinal_ex';
  fn_EVP_DecryptInit_ex = 'EVP_DecryptInit_ex';
  fn_EVP_DecryptUpdate = 'EVP_DecryptUpdate';
  fn_EVP_DecryptFinal_ex = 'EVP_DecryptFinal_ex';
  fn_ERR_print_errors_cb = 'ERR_print_errors_cb';

  fn_OPENSSL_add_all_algorithms_noconf = 'OpenSSL_add_all_algorithms_noconf';

  fn_OpenSSL_add_all_ciphers = 'OpenSSL_add_all_ciphers';
  fn_OpenSSL_add_all_digests = 'OpenSSL_add_all_digests';

  fn_EVP_CIPHER_CTX_init = 'EVP_CIPHER_CTX_init';
  fn_PKCS5_PBKDF2_HMAC = 'PKCS5_PBKDF2_HMAC';
  fn_EVP_sha256 = 'EVP_sha256';

  fn_SSL_load_error_strings = 'SSL_load_error_strings';

function LoadFunction(Handle: THandle; const FceName: string; SSLeay: Boolean = False): Pointer;
begin
  Result := GetProcAddress(Handle, PChar(FceName));
  if Result = nil then
  begin
    if SSLeay then
      raise Exception.CreateFmt('Cannot load "%s" function from %s (err: #%d)', [FceName, SSLUTIL_DLL_name, GetLastError])
    else
      raise Exception.CreateFmt('Cannot load "%s" function from %s (err: #%d)', [FceName, SSLCLIB_DLL_name, GetLastError]);
  end;
end;

procedure Load(SSL_111: Boolean);
var
  hLibeayDLL, hSSLeayDLL: THandle;
begin
  hLibeayDLL := GetModuleHandle(SSLCLIB_DLL_name);
  if hLibeayDLL = 0 then
    raise Exception.CreateFmt('"%s" not loaded', [SSLCLIB_DLL_name]);

  hSSLeayDLL := GetModuleHandle(SSLUTIL_DLL_name);
  if hSSLeayDLL = 0 then
    raise Exception.CreateFmt('"%s" not loaded', [SSLUTIL_DLL_name]);

{$IF Declared(sk_value)}
  sk_value := LoadFunction(hLibeayDLL, fn_sk_value);
{$IFEND}
{$IF Declared(sk_num)}
  sk_num := LoadFunction(hLibeayDLL, fn_sk_num);
{$IFEND}
{$IF Declared(BIO_ctrl)}
  BIO_ctrl := LoadFunction(hLibeayDLL, fn_BIO_ctrl);
{$IFEND}
{$IF Declared(BIO_new_mem_buf)}
  BIO_new_mem_buf := LoadFunction(hLibeayDLL, fn_BIO_new_mem_buf);
{$IFEND}
{$IF Declared(d2i_X509_bio)}
  d2i_X509_bio := LoadFunction(hLibeayDLL, fn_d2i_X509_bio);
{$IFEND}
{$IF Declared(X509_STORE_CTX_init)}
  @X509_STORE_CTX_init := LoadFunction(hLibeayDLL, fn_X509_STORE_CTX_init);
{$IFEND}
{$IF Declared(X509_STORE_CTX_set_flags)}
  @X509_STORE_CTX_set_flags := LoadFunction(hLibeayDLL, fn_X509_STORE_CTX_set_flags);
{$IFEND}
{$IF Declared(X509_STORE_set_flags)}
  @X509_STORE_set_flags := LoadFunction(hLibeayDLL, fn_X509_STORE_set_flags);
{$IFEND}

{$IFDEF CPUX64}
{$ELSE}
  {$IF Declared(SSL_CTX_get0_param)}
    @SSL_CTX_get0_param := LoadFunction(hSSLeayDLL, fn_SSL_CTX_get0_param, True);
  {$IFEND}
{$ENDIF}

{$IF Declared(X509_verify_cert)}
  @X509_verify_cert := LoadFunction(hLibeayDLL, fn_X509_verify_cert);
{$IFEND}
//{$IF Declared(X509_STORE_CTX_verify)}
//  @X509_STORE_CTX_verify := LoadFunction(hLibeayDLL, fn_X509_STORE_CTX_verify);
//{$IFEND}

{$IF NOT Declared(X509_VERIFY_PARAM_new)}
var
  @X509_VERIFY_PARAM_new := LoadFunction(hLibeayDLL, fn_X509_VERIFY_PARAM_new);
{$IFEND}
{$IF NOT Declared(X509_VERIFY_PARAM_set_flags)}
var
  @X509_VERIFY_PARAM_set_flags := LoadFunction(hLibeayDLL, fn_X509_VERIFY_PARAM_set_flags);
{$IFEND}
{$IF NOT Declared(SSL_CTX_set1_param)}
var
  @SSL_CTX_set1_param := LoadFunction(hSSLeayDLL, fn_SSL_CTX_set1_param, True);
{$IFEND}
{$IF NOT Declared(X509_VERIFY_PARAM_free)}
var
  @X509_VERIFY_PARAM_free := LoadFunction(hLibeayDLL, fn_X509_VERIFY_PARAM_free);
{$IFEND}

{$IF Declared(X509_STORE_new)}
  @X509_STORE_new := LoadFunction(hLibeayDLL, fn_X509_STORE_new);
{$IFEND}
{$IF Declared(X509_STORE_free)}
  @X509_STORE_free := LoadFunction(hLibeayDLL, fn_X509_STORE_free);
{$IFEND}
{$IF Declared(X509_STORE_CTX_new)}
  @X509_STORE_CTX_new := LoadFunction(hLibeayDLL, fn_X509_STORE_CTX_new);
{$IFEND}
{$IF Declared(X509_STORE_CTX_free)}
  @X509_STORE_CTX_free := LoadFunction(hLibeayDLL, fn_X509_STORE_CTX_free);
{$IFEND}
{$IF Declared(X509_STORE_CTX_set_cert)}
  @X509_STORE_CTX_set_cert := LoadFunction(hLibeayDLL, fn_X509_STORE_CTX_set_cert);
{$IFEND}
{$IF Declared(X509_STORE_CTX_set_purpose)}
  @X509_STORE_CTX_set_purpose := LoadFunction(hLibeayDLL, fn_X509_STORE_CTX_set_purpose);
{$IFEND}
{$IF Declared(X509_STORE_CTX_cleanup)}
  @X509_STORE_CTX_cleanup := LoadFunction(hLibeayDLL, fn_X509_STORE_CTX_cleanup);
{$IFEND}
{$IF Declared(OBJ_txt2obj)}
  @OBJ_txt2obj := LoadFunction(hLibeayDLL, fn_OBJ_txt2obj);
{$IFEND}
{$IF Declared(X509_get_ext_d2i)}
  @X509_get_ext_d2i := LoadFunction(hLibeayDLL, fn_X509_get_ext_d2i);
{$IFEND}
{$IF Declared(OBJ_cmp)}
  @OBJ_cmp := LoadFunction(hLibeayDLL, fn_OBJ_cmp);
{$IFEND}
//{$IF Declared(X509_STORE_add_cert)}
//  @X509_STORE_add_cert := LoadFunction(hLibeayDLL, fn_X509_STORE_add_cert);
//{$IFEND}
{$IF Declared(BIO_free)}
  @BIO_free := LoadFunction(hLibeayDLL, fn_BIO_free);
{$IFEND}
{$IF Declared(d2i_X509)}
  @d2i_X509 := LoadFunction(hLibeayDLL, fn_d2i_X509);
{$IFEND}
{$IF Declared(X509_free)}
  @X509_free := LoadFunction(hLibeayDLL, fn_X509_free);
{$IFEND}
{$IF Declared(ERR_get_error)}
  @ERR_get_error := LoadFunction(hLibeayDLL, fn_ERR_get_error);
{$IFEND}
{$IF Declared(ERR_error_string_n)}
  @ERR_error_string_n := LoadFunction(hLibeayDLL, fn_ERR_error_string_n);
{$IFEND}
{$IF Declared(ERR_error_string)}
  @ERR_error_string := LoadFunction(hLibeayDLL, fn_ERR_error_string);
{$IFEND}
{$IF Declared(ERR_clear_error)}
  @ERR_clear_error := LoadFunction(hLibeayDLL, fn_ERR_clear_error);
{$IFEND}

{$IF Declared(RAND_bytes)}
  @RAND_bytes := LoadFunction(hLibeayDLL, fn_RAND_bytes);
{$IFEND}
{$IF Declared(RAND_pseudo_bytes)}
  @RAND_pseudo_bytes := LoadFunction(hLibeayDLL, fn_RAND_pseudo_bytes);
{$IFEND}
{$IF Declared(RAND_status)}
  @RAND_status := LoadFunction(hLibeayDLL, fn_RAND_status);
{$IFEND}
{$IF Declared(RAND_poll)}
  @RAND_poll := LoadFunction(hLibeayDLL, fn_RAND_poll);
{$IFEND}
{$IF Declared(RAND_file_name)}
  @RAND_file_name := LoadFunction(hLibeayDLL, fn_RAND_file_name);
{$IFEND}
{$IF Declared(RAND_load_file)}
  @RAND_load_file := LoadFunction(hLibeayDLL, fn_RAND_load_file);
{$IFEND}
{$IF Declared(RAND_write_file)}
  @RAND_write_file := LoadFunction(hLibeayDLL, fn_RAND_write_file);
{$IFEND}
//{$IF Declared(SSL_CTX_get_cert_store)}
//  @SSL_CTX_get_cert_store := LoadFunction(hSSLeayDLL, fn_SSL_CTX_get_cert_store, True);
//{$IFEND}
{$IF Declared(X509_STORE_CTX_get_current_cert)}
  @X509_STORE_CTX_get_current_cert := LoadFunction(hLibeayDLL, fn_X509_STORE_CTX_get_current_cert);
{$IFEND}
{$IF Declared(X509_STORE_CTX_get_error_depth)}
  @X509_STORE_CTX_get_error_depth := LoadFunction(hLibeayDLL, fn_X509_STORE_CTX_get_error_depth);
{$IFEND}
{$IF Declared(X509_get_serialNumber)}
  @X509_get_serialNumber := LoadFunction(hLibeayDLL, fn_X509_get_serialNumber);
{$IFEND}
{$IF Declared(X509_check_ca)}
  @X509_check_ca := LoadFunction(hLibeayDLL, fn_X509_check_ca);
{$IFEND}
{$IF Declared(X509_check_ca)}
  @X509_get1_email := LoadFunction(hLibeayDLL, fn_X509_get1_email);
{$IFEND}
{$IF Declared(X509_get_subject_name)}
  @X509_get_subject_name := LoadFunction(hLibeayDLL, fn_X509_get_subject_name);
{$IFEND}
{$IF Declared(SSL_CTX_load_verify_locations)}
  @SSL_CTX_load_verify_locations := LoadFunction(hSSLeayDLL, fn_SSL_CTX_load_verify_locations, True);
{$IFEND}
{$IF Declared(ASN1_INTEGER_get)}
  @ASN1_INTEGER_get := LoadFunction(hLibeayDLL, fn_ASN1_INTEGER_get);
{$IFEND}

{$IF Declared(EVP_CIPHER_CTX_new)}
  @EVP_CIPHER_CTX_new := LoadFunction(hLibeayDLL, fn_EVP_CIPHER_CTX_new);
{$ENDIF}

{$IF Declared(EVP_CIPHER_CTX_free)}
  @EVP_CIPHER_CTX_free := LoadFunction(hLibeayDLL, fn_EVP_CIPHER_CTX_free);
{$ENDIF}

{$IF Declared(EVP_CIPHER_CTX_key_length)}
  @EVP_CIPHER_CTX_key_length := LoadFunction(hLibeayDLL, fn_EVP_CIPHER_CTX_key_length);
{$ENDIF}

{$IF Declared(EVP_CIPHER_CTX_ctrl)}
  @EVP_CIPHER_CTX_ctrl := LoadFunction(hLibeayDLL, fn_EVP_CIPHER_CTX_ctrl);
{$ENDIF}

{$IF Declared(EVP_EncryptInit_ex)}
  @EVP_EncryptInit_ex := LoadFunction(hLibeayDLL, fn_EVP_EncryptInit_ex);
{$ENDIF}

{$IF Declared(EVP_aes_256_gcm)}
  @EVP_aes_256_gcm := LoadFunction(hLibeayDLL, fn_EVP_aes_256_gcm);
{$ENDIF}

{$IF Declared(EVP_aes_256_cbc)}
  @EVP_aes_256_cbc := LoadFunction(hLibeayDLL, fn_EVP_aes_256_cbc);
{$ENDIF}

{$IF Declared(EVP_EncryptUpdate)}
  @EVP_EncryptUpdate := LoadFunction(hLibeayDLL, fn_EVP_EncryptUpdate);
{$ENDIF}

{$IF Declared(EVP_EncryptFinal_ex)}
  @EVP_EncryptFinal_ex := LoadFunction(hLibeayDLL, fn_EVP_EncryptFinal_ex);
{$ENDIF}

{$IF Declared(EVP_DecryptInit_ex)}
  @EVP_DecryptInit_ex := LoadFunction(hLibeayDLL, fn_EVP_DecryptInit_ex);
{$ENDIF}

{$IF Declared(EVP_DecryptUpdate)}
  @EVP_DecryptUpdate := LoadFunction(hLibeayDLL, fn_EVP_DecryptUpdate);
{$ENDIF}

{$IF Declared(EVP_DecryptFinal_ex)}
  @EVP_DecryptFinal_ex := LoadFunction(hLibeayDLL, fn_EVP_DecryptFinal_ex);
{$ENDIF}

{$IF Declared(ERR_print_errors_cb)}
  @ERR_print_errors_cb := LoadFunction(hLibeayDLL, fn_ERR_print_errors_cb);
{$ENDIF}

{$IF Declared(OpenSSL_add_all_algorithms_noconf)}
//  @OpenSSL_add_all_algorithms_noconf := LoadFunction(hLibeayDLL, fn_OpenSSL_add_all_algorithms_noconf);
{$ENDIF}

{$IF Declared(OpenSSL_add_all_ciphers)}
  OpenSSL_add_all_ciphers := LoadFunction(hLibeayDLL, fn_OpenSSL_add_all_ciphers);
{$ENDIF}

{$IF Declared(OpenSSL_add_all_digests)}
  OpenSSL_add_all_digests := LoadFunction(hLibeayDLL, fn_OpenSSL_add_all_digests);
{$ENDIF}

{$IF Declared(EVP_CIPHER_CTX_init)}
  EVP_CIPHER_CTX_init := LoadFunction(hLibeayDLL, fn_EVP_CIPHER_CTX_init);
{$ENDIF}

{$IF Declared(PKCS5_PBKDF2_HMAC)}
  PKCS5_PBKDF2_HMAC := LoadFunction(hLibeayDLL, fn_PKCS5_PBKDF2_HMAC);
{$ENDIF}

{$IF Declared(EVP_sha256)}
  EVP_sha256 := LoadFunction(hLibeayDLL, fn_EVP_sha256);
{$ENDIF}

{$IF Declared(SSL_load_error_strings)}
  SSL_load_error_strings := LoadFunction(hSSLeayDLL, fn_SSL_load_error_strings);
{$ENDIF}

(*****************************************************************************)

{$IF Declared(EVP_sha1)}
  @EVP_sha1 := LoadFunction(hLibeayDLL, fn_EVP_sha1);
{$IFEND}

  if not SSL_111 then
  begin
{$IF Declared(SSLeay)}
    @SSLeay := LoadFunction(hLibeayDLL, fn_SSLeay);
{$IFEND}
{$IF Declared(_SSLeay_version)}
    @_SSLeay_version := LoadFunction(hLibeayDLL, fn__SSLeay_version);
{$IFEND}
  end;

  SSL_load_error_strings();
end;

function sk_GENERAL_NAMES_value(Stack: PGENERAL_NAME; I: Integer): PGENERAL_NAME;
begin
  Result := PGENERAL_NAME(sk_value(PSTACK(Stack), I));
end;

function sk_GENERAL_NAMES_num(const Stack: PGENERAL_NAME): Integer;
begin
  Result := sk_num(PSTACK(Stack));
end;

function BIO_read_filename(BP: PBIO; Filename: PAnsiChar): Integer;
begin
  Result := BIO_ctrl(BP, BIO_C_SET_FILENAME, BIO_CLOSE or BIO_FP_READ, Filename);
end;

{ EOpenSSLLibError }

constructor EOpenSSLLibError.Create(ACode: Integer; const AMsg: string);
begin
  FErrorCode := ACode;
  inherited Create(AMsg);
end;

{ RaiseOpenSSLError }
procedure RaiseOpenSSLError(const AMessage: string = '');
var
  ErrCode: Integer;
  ErrMsg, FullMsg: string;
begin
  ErrCode := ERR_get_error;
  ErrMsg := string(AnsiString(ERR_error_string(ErrCode, nil)));
  if AMessage = '' then
    FullMsg := ErrMsg
  else
    FullMsg := AMessage + ': ' + ErrMsg;

  raise EOpenSSLLibError.Create(ErrCode, FullMsg);
end;

end.
