
{*******************************************************}
{                                                       }
{       Rich Integer List                               }
{                                                       }
{   Copyright (c) 2001-2022 TDP, Tomas Korinek          }
{                                                       }
{****************************************************tk*}

{$DEFINE PURE64LIST}

unit TDP.IntList;

interface

uses
  WinApi.Windows, System.SysUtils, System.Classes;

{ TIntegerList class }

const
  MaxIntListSize = MaxInt div 16;

type
  { TIntegerList class }

  TIntegerList = class;

  PIntegerItem = ^TIntegerItem;
  TIntegerItem = record
    FInteger: Integer;
    FObject: TObject;
  end;

  PIntegerItemList = ^TIntegerItemList;
  TIntegerItemList = array[0..MaxIntListSize - 1] of TIntegerItem;
  TIntegerListSortCompare = function(List: TIntegerList; Index1, Index2: Integer): Integer;

  TIntegerList = class(TObject)
  private
    FList: PIntegerItemList;
    FCount: Integer;
    FCapacity: Integer;
    FSorted: Boolean;
    FOwnsObjects: Boolean;
    FDuplicates: TDuplicates;
    function GetIndexes(Value: Integer): Integer;

    procedure ExchangeItems(Index1, Index2: Integer);
    procedure FinalizeObjects(Start, Stop: Integer);
    procedure Grow;
    procedure QuickSort(L, R: Integer; ICompare: TIntegerListSortCompare);
    procedure SetSorted(Value: Boolean);
    function GetDWord(Index: Integer): DWORD;
    procedure PutDWord(Index: Integer; const Value: DWORD);
    function GetText: string;
    procedure SetText(const Value: string);
    function GetDwordText: string;
    function GetFmtText(Fmt: string): string;
  protected
    class procedure Error(const Msg: string; Data: Integer); overload;
    class procedure Error(Msg: PResStringRec; Data: Integer); overload;

    function Get(Index: Integer): Integer;
    function GetCapacity: Integer;
    function GetCount: Integer;
    function GetObject(Index: Integer): TObject;
    procedure Put(Index: Integer; const I: Integer);
    procedure PutObject(Index: Integer; AObject: TObject);
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetCount(NewCount: Integer);
    procedure SetUpdateState(Updating: Boolean);
    function CompareIntegers(const I1, I2: Integer; const Index1: Integer = -1; const Index2: Integer = -1): Integer; virtual;
    procedure InsertItem(Index: Integer; const I: Integer; AObject: TObject);
  public
    constructor Create(AOwnsObjects: Boolean = False);
    destructor Destroy; override;

    function Add(const I: Integer): Integer;
    function AddObject(const I: Integer; AObject: TObject): Integer;
    function AddDWord(const I: Integer; const Value: DWORD): Integer;
    procedure Assign(Source: TIntegerList; const DWords: Boolean = False);
    procedure Clear(AffectCapacity: Boolean = True);
    procedure Delete(Index: Integer);
    procedure Exchange(Index1, Index2: Integer);
    procedure Extract(Index: Integer);
    function Find(const N: Integer; var Index: Integer): Boolean;
    function IndexOf(const Value: Integer): Integer;
    procedure Insert(Index: Integer; const I: Integer);
    procedure InsertObject(Index: Integer; const I: Integer; AObject: TObject);
    procedure Merge(Source: TIntegerList; const DWords: Boolean = False);
    procedure Sort;
    procedure Reverse;
    procedure CustomSort(Compare: TIntegerListSortCompare);

    function SimpleText(const Fmt: string = '%d'): string;

    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
    property Capacity: Integer read GetCapacity write SetCapacity;
    property Count: Integer read GetCount;
    property Integers[Index: Integer]: Integer read Get write Put; default;
    property Objects[Index: Integer]: TObject read GetObject write PutObject;
    property DWords[Index: Integer]: DWORD read GetDWord write PutDWord;
    property OwnsObjects: Boolean read FOwnsObjects write FOwnsObjects;
    property Sorted: Boolean read FSorted write SetSorted;

    // properties added for debugging
    property Indexes[Value: Integer]: Integer read GetIndexes;
    property Text: string read GetText write SetText;
    property DText: string read GetDwordText;
    property FmtText[Fmt: string]: string read GetFmtText;
  end;

  { TInt64List class }

  TInt64List = class;

  TInt64ListSortCompare = function(List: TInt64List; Index1, Index2: Integer): Integer;

{$REGION ' Pure Int64List Interface '}

{$IFDEF PURE64LIST}

  PInt64Item = ^TInt64Item;
  TInt64Item = record
    FInt64: Int64;
    case Integer of
      0: (FObject: TObject);
      1: (FTetraWord: Int64);
  end;

  PInt64ItemList = ^TInt64ItemList;
  TInt64ItemList = array[0..MaxIntListSize div 2 - 1] of TInt64Item;

  TInt64List = class(TObject)
  private
    FList: PInt64ItemList;
    FCount: Integer;
    FCapacity: Integer;
    FSorted: Boolean;
    FOwnsObjects: Boolean;
    FDuplicates: TDuplicates;
    function GetIndexes(Value: Int64): Integer;

    procedure ExchangeItems(Index1, Index2: Integer);
    procedure FinalizeObjects(Start, Stop: Integer);
    procedure Grow;
    procedure QuickSort(L, R: Integer; ICompare: TInt64ListSortCompare);
    procedure SetSorted(Value: Boolean);
    function GetTetraWord(Index: Integer): Int64;
    procedure PutTetraWord(Index: Integer; const Value: Int64);
    function GetText: string;
    procedure SetText(const Value: string);
  protected
    class procedure Error(const Msg: string; Data: Integer); overload;
    class procedure Error(Msg: PResStringRec; Data: Integer); overload;

    function Get(Index: Integer): Int64;
    function GetCapacity: Integer;
    function GetCount: Integer;
    function GetObject(Index: Integer): TObject;
    procedure Put(Index: Integer; const I: Int64);
    procedure PutObject(Index: Integer; AObject: TObject);
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetCount(NewCount: Integer);
    procedure SetUpdateState(Updating: Boolean);
    function CompareInt64s(const I1, I2: Int64; const Index1: Integer = -1; const Index2: Integer = -1): Integer; virtual;
    procedure InsertItem(Index: Integer; const I: Int64; AObject: TObject); overload;
    procedure InsertItem(Index: Integer; const I: Int64; ATetraWord: Int64); overload;
    procedure InsertItemObj(Index: Integer; const I: Int64; ATetraWord: Int64);
  public
    constructor Create(AOwnsObjects: Boolean = False);
    destructor Destroy; override;

    function Add(const I: Int64): Integer;
    function AddObject(const I: Int64; AObject: TObject): Integer; overload;
    function AddObject(const I: Int64; ATetraWord: Int64): Integer; overload;
    procedure Assign(Source: TInt64List);
    procedure Clear(AffectCapacity: Boolean = True);
    procedure Delete(Index: Integer);
    procedure Exchange(Index1, Index2: Integer);
    procedure Extract(Index: Integer);
    function Find(const N: Int64; var Index: Integer): Boolean;
    function IndexOf(const Value: Int64): Integer;
    procedure Insert(Index: Integer; const I: Int64);
    procedure InsertObject(Index: Integer; const I: Int64; AObject: TObject); overload;
    procedure InsertObject(Index: Integer; const I: Int64; ATetraWord: Int64); overload;
    procedure Sort;
    procedure Reverse;
    procedure CustomSort(Compare: TInt64ListSortCompare);

    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
    property Capacity: Integer read GetCapacity write SetCapacity;
    property Count: Integer read GetCount;
    property Integers[Index: Integer]: Int64 read Get write Put; default;
    property Objects[Index: Integer]: TObject read GetObject write PutObject;
    property OwnsObjects: Boolean read FOwnsObjects write FOwnsObjects;
    property TetraWords[Index: Integer]: Int64 read GetTetraWord write PutTetraWord;
    property Sorted: Boolean read FSorted write SetSorted;

    // properties added for debugging
    property Indexes[Value: Int64]: Integer read GetIndexes;

    property Text: string read GetText write SetText;
  end;
{$ENDREGION}

{$ELSE}

{$REGION ' DoubleInt Int64List Interface '}

  TInt64List = class(TObject)
  private
    FList: PIntegerItemList;
    FCount: Integer;
    FCapacity: Integer;
    FSorted: Boolean;
    FOwnsObjects: Boolean;
    FDuplicates: TDuplicates;
    function GetIndexes(Value: Int64): Integer;

    procedure ExchangeItems(Index1, Index2: Integer);
    procedure FinalizeObjects(Start, Stop: Integer);
    procedure Grow;
    procedure QuickSort(L, R: Integer; ICompare: TInt64ListSortCompare);
    procedure SetSorted(Value: Boolean);
    function GetTetraWord(Index: Integer): Int64;
    procedure PutTetraWord(Index: Integer; const Value: Int64);
  protected
    class procedure Error(const Msg: string; Data: Integer); overload;
    class procedure Error(Msg: PResStringRec; Data: Integer); overload;

    function Get(Index: Integer): Int64;
    function GetCapacity: Integer;
    function GetCount: Integer;
    function GetObject(Index: Integer): TObject;
    procedure Put(Index: Integer; const I: Int64);
    procedure PutObject(Index: Integer; AObject: TObject);
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetUpdateState(Updating: Boolean);
    function CompareInt64s(const I1, I2: Int64; const Index1: Integer = -1; const Index2: Integer = -1): Integer; virtual;
    procedure InsertItem(Index: Integer; const I: Int64; AObject: TObject); overload;
    procedure InsertItem(Index: Integer; const I: Int64; ATetraWord: Int64); overload;
    procedure InsertItemObj(Index: Integer; const I: Int64; ATetraWord: Int64);
  public
    constructor Create(AOwnsObjects: Boolean = False);
    destructor Destroy; override;

    function Add(const I: Int64): Integer;
    function AddObject(const I: Int64; AObject: TObject): Integer;
    procedure Clear;
    procedure Delete(Index: Integer);
    procedure Exchange(Index1, Index2: Integer);
    procedure Extract(Index: Integer);
    function Find(const N: Int64; var Index: Integer): Boolean;
    function IndexOf(const Value: Int64): Integer;
    procedure Insert(Index: Integer; const I: Int64);
    procedure InsertObject(Index: Integer; const I: Int64; AObject: TObject); overload;
    procedure InsertObject(Index: Integer; const I: Int64; ATetraWord: Int64); overload;
    procedure Sort;
    procedure CustomSort(Compare: TInt64ListSortCompare);

    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
    property Capacity: Integer read GetCapacity write SetCapacity;
    property Count: Integer read GetCount;
    property Integers[Index: Integer]: Int64 read Get write Put; default;
    property Objects[Index: Integer]: TObject read GetObject write PutObject;
    property OwnsObjects: Boolean read FOwnsObjects write FOwnsObjects;
    property TetraWords[Index: Integer]: Int64 read GetTetraWord write PutTetraWord;
    property Sorted: Boolean read FSorted write SetSorted;

    // property added for debugging
    property Indexes[Value: Int64]: Integer read GetIndexes;
  end;
{$ENDIF}

{$ENDREGION}

function MakeInt64(const ALo, AHi: Integer): Int64; inline;

implementation

uses
  System.RTLConsts, System.StrUtils,
  TDP.CommonCode;

const
  SListOwnsObjects = 'TetraWords not allowed on list with owns objects';

{ TIntegerList }

constructor TIntegerList.Create(AOwnsObjects: Boolean = False);
begin
  inherited Create;
  FOwnsObjects := AOwnsObjects;
  FSorted := True;
end;

destructor TIntegerList.Destroy;
begin
  inherited Destroy;
  FinalizeObjects(0, FCount);
  FCount := 0;
  SetCapacity(0);
end;

function TIntegerList.Add(const I: Integer): Integer;
begin
  Result := AddObject(I, nil);
end;

function TIntegerList.AddDWord(const I: Integer; const Value: DWORD): Integer;
begin
  Result := AddObject(I, TObject(Value));
end;

function TIntegerList.AddObject(const I: Integer; AObject: TObject): Integer;
begin
  if not Sorted then
    Result := FCount
  else
    if Find(I, Result) then
      case Duplicates of
        dupIgnore: Exit;
        dupError: raise EListError.Create('List does not allow duplicates');
      end;
  InsertItem(Result, I, AObject);
end;

procedure TIntegerList.Assign(Source: TIntegerList; const DWords: Boolean);
var
  I: Integer;
begin
  Clear;
  for I := 0 to Source.Count - 1 do
    if DWords then
      AddDWord(Source[I], Source.DWords[I])
    else
      Add(Source[I]);
end;

procedure TIntegerList.Clear(AffectCapacity: Boolean);
begin
  if FCount <> 0 then
  begin
    // FinalizeObjects(0, FCount);
    SetCount(0);
    if AffectCapacity then
      SetCapacity(0);
  end;
end;

procedure TIntegerList.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then
    Error(@SListIndexError, Index);
  FinalizeObjects(Index, Index + 1);
  Dec(FCount);
  if Index < FCount then
    System.Move(FList^[Index + 1], FList^[Index],
      (FCount - Index) * SizeOf(TIntegerItem));
end;

class procedure TIntegerList.Error(const Msg: string; Data: Integer);

  function ReturnAddr: Pointer;
  asm
          MOV     EAX,[EBP+4]
  end;

begin
  raise EListError.CreateFmt(Msg, [Data]) at ReturnAddr;
end;

class procedure TIntegerList.Error(Msg: PResStringRec; Data: Integer);
begin
  TList.Error(LoadResString(Msg), Data);
end;

procedure TIntegerList.Exchange(Index1, Index2: Integer);
begin
  if Sorted then Error(@SSortedListError, 0);
  if (Index1 < 0) or (Index1 >= FCount) then Error(@SListIndexError, Index1);
  if (Index2 < 0) or (Index2 >= FCount) then Error(@SListIndexError, Index2);
  ExchangeItems(Index1, Index2);
end;

procedure TIntegerList.ExchangeItems(Index1, Index2: Integer);
var
  TempInt: Integer;
  Temp: NativeInt;
  Item1, Item2: PIntegerItem;
begin
  Item1 := @FList^[Index1];
  Item2 := @FList^[Index2];
  TempInt := Integer(Item1^.FInteger);
  Item1^.FInteger := Item2^.FInteger;
  Item2^.FInteger := TempInt;
  Temp := NativeInt(Item1^.FObject);
  NativeInt(Item1^.FObject) := NativeInt(Item2^.FObject);
  NativeInt(Item2^.FObject) := Temp;
end;

procedure TIntegerList.Extract(Index: Integer);
var
  SaveOwn: Boolean;
begin
  SaveOwn := FOwnsObjects;
  FOwnsObjects := True;
  try
    Delete(Index);
  finally
    FOwnsObjects := SaveOwn;
  end;
end;

procedure TIntegerList.FinalizeObjects(Start, Stop: Integer);
begin
  if FOwnsObjects then
    while Start < Stop do
    begin
      FList^[Start].FObject.Free;
      Inc(Start);
    end;
end;

function TIntegerList.Find(const N: Integer; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;

  if not Sorted then
  begin
    for I := 0 to Count - 1 do
      if FList^[I].FInteger = N then
      begin
        Index := I;
        Result := True;
        Break;
      end;
  end else
  begin
    L := 0;
    H := FCount - 1;
    while L <= H do
    begin
      I := (L + H) shr 1;
      C := CompareIntegers(FList^[I].FInteger, N);
      if C < 0 then L := I + 1 else
      begin
        H := I - 1;
        if C = 0 then
        begin
          Result := True;
          if Duplicates <> dupAccept then L := I;
        end;
      end;
    end;
    Index := L;
  end;
end;

function TIntegerList.Get(Index: Integer): Integer;
begin
  if (Index < 0) or (Index >= FCount) then
    Error(@SListIndexError, Index);
  Result := FList^[Index].FInteger;
end;

function TIntegerList.GetCapacity: Integer;
begin
  Result := FCapacity;
end;

function TIntegerList.GetCount: Integer;
begin
  Result := FCount;
end;

function TIntegerList.GetDWord(Index: Integer): DWORD;
begin
  if (Index < 0) or (Index >= FCount) then
    Error(@SListIndexError, Index);
  Result := DWORD(FList^[Index].FObject);
end;

function TIntegerList.GetDwordText: string;
var
  I: Integer;
begin
  Result := '';
  for I := 0 to Count - 1 do
    Result := Result + IfThen(I > 0, ' ', '') + Format('[%d: %d]', [Self[I], DWords[I]]);
end;

function TIntegerList.GetFmtText(Fmt: string): string;
var
  I: Integer;
  N, Cnt, Store: Integer;
begin
  Result := '';
  if Fmt = '' then Fmt := '%d';

  if Sorted and (Duplicates = dupAccept) and (Count > 0) then
  begin
//    N := Integers[0];
//    Result := SysUtils.Format(Fmt, [N]);
    Result := '';
    Store := Integers[0];
    Cnt := 0;
    for I := 0 to Count - 1 do
    begin
      N := Integers[I];
      if N = Store then
      begin
        Inc(Cnt);
      end else
      begin
        if Cnt > 1 then
          Result := Result + System.SysUtils.Format(' ' + Fmt + '{%d}', [Store, Cnt])
        else
          Result := Result + System.SysUtils.Format(' ' + Fmt, [Store]);
        Cnt := 1;
      end;
      Store := N;
    end;
    if Cnt > 1 then
      Result := Result + System.SysUtils.Format(' ' + Fmt + '{%d}', [Store, Cnt])
    else
      Result := Result + System.SysUtils.Format(' ' + Fmt, [Store]);
  end else
    for I := 0 to Count - 1 do
      Result := Result + System.SysUtils.Format('%s' + Fmt, [IfThen(I > 0, ' ', ''), Integers[I]]);
end;

function TIntegerList.GetIndexes(Value: Integer): Integer;
begin
  Result := IndexOf(Value);
end;

function TIntegerList.GetObject(Index: Integer): TObject;
begin
  if (Index < 0) or (Index >= FCount) then
    Error(@SListIndexError, Index);
  Result := FList^[Index].FObject;
end;

function TIntegerList.GetText: string;
begin
  Result := GetFmtText('');
end;

procedure TIntegerList.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 64 then Delta := FCapacity div 4 else
    if FCapacity > 8 then Delta := 16 else
      Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

function TIntegerList.IndexOf(const Value: Integer): Integer;
begin
  if not Sorted then
  begin
    for Result := 0 to GetCount - 1 do
      if CompareIntegers(Get(Result), Value) = 0 then Exit;
    Result := -1;
  end else
    if not Find(Value, Result) then Result := -1;
end;

procedure TIntegerList.Insert(Index: Integer; const I: Integer);
begin
  InsertObject(Index, I, nil);
end;

procedure TIntegerList.InsertObject(Index: Integer; const I: Integer;
  AObject: TObject);
begin
  if Sorted then Error(@SSortedListError, 0);
  if (Index < 0) or (Index > FCount) then
    Error(@SListIndexError, Index);
  InsertItem(Index, I, AObject);
end;

procedure TIntegerList.Merge(Source: TIntegerList; const DWords: Boolean);
var
  I: Integer;
begin
  for I := 0 to Source.Count - 1 do
    if DWords then
      AddDWord(Source[I], Source.DWords[I])
    else
      Add(Source[I]);
end;

procedure TIntegerList.InsertItem(Index: Integer; const I: Integer; AObject: TObject);
begin
  if FCount = FCapacity then Grow;
  if Index < FCount then
    System.Move(FList^[Index], FList^[Index + 1],
      (FCount - Index) * SizeOf(TIntegerItem));
  with FList^[Index] do
  begin
    FInteger := 0;
    FObject := AObject;
    FInteger := I;
  end;
  Inc(FCount);
end;

procedure TIntegerList.Put(Index: Integer; const I: Integer);
begin
  if Sorted then Error(@SSortedListError, 0);
  if (Index < 0) or (Index >= FCount) then
    Error(@SListIndexError, Index);
  FList^[Index].FInteger := I;
end;

procedure TIntegerList.PutDWord(Index: Integer; const Value: DWORD);
begin
  if (Index < 0) or (Index >= FCount) then
    Error(@SListIndexError, Index);
  FList^[Index].FObject := TObject(Value);
end;

procedure TIntegerList.PutObject(Index: Integer; AObject: TObject);
begin
  if (Index < 0) or (Index >= FCount) then
    Error(@SListIndexError, Index);
  FList^[Index].FObject := AObject;
end;

procedure TIntegerList.QuickSort(L, R: Integer; ICompare: TIntegerListSortCompare);
var
  I, J, P: Integer;
begin
  repeat
    I := L;
    J := R;
    P := (L + R) shr 1;
    repeat
      while ICompare(Self, I, P) < 0 do Inc(I);
      while ICompare(Self, J, P) > 0 do Dec(J);
      if I <= J then
      begin
        ExchangeItems(I, J);
        if P = I then
          P := J
        else if P = J then
          P := I;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J, ICompare);
    L := I;
  until I >= R;
end;

procedure TIntegerList.Reverse;
var
  Tmp: TIntegerItem;
  B, E: Integer;
begin
  B := 0;
  E := Count - 1;
  while B < E do
  begin
    Tmp := FList^[B];
    FList^[B] := FList^[E];
    FList^[E] := Tmp;
    Inc(B);
    Dec(E);
  end;
end;

procedure TIntegerList.SetCapacity(NewCapacity: Integer);
begin
  if NewCapacity < FCount then
    FinalizeObjects(NewCapacity, FCount);

  ReallocMem(FList, NewCapacity * SizeOf(TIntegerItem));
  FCapacity := NewCapacity;

  if FCount > FCapacity then FCount := FCapacity;
end;

procedure TIntegerList.SetCount(NewCount: Integer);
var
  I: Integer;
begin
  if (NewCount < 0) or (NewCount > MaxIntListSize) then
    Error(@SListCountError, NewCount);
  if NewCount > FCapacity then
    SetCapacity(NewCount);
  if NewCount > FCount then
    FillChar(FList^[FCount], (NewCount - FCount) * SizeOf(TIntegerItem), 0)
  else
    for I := FCount - 1 downto NewCount do
      Delete(I);
  FCount := NewCount;
end;

procedure TIntegerList.SetSorted(Value: Boolean);
begin
  if FSorted <> Value then
  begin
    if Value then Sort;
    FSorted := Value;
  end;
end;

procedure TIntegerList.SetText(const Value: string);
var
  Tkn: TStringTokenizer;
  I: Integer;
begin
  Tkn := TStringTokenizer.Create(' ');
  try
    Clear;
    Tkn.Str := Value;
    for I := 0 to Tkn.Count - 1 do
      Add(Tkn.Numbers[I]);
  finally
    Tkn.Free;
  end;
end;

procedure TIntegerList.SetUpdateState(Updating: Boolean);
begin
  // if Updating then Changing else Changed;
end;

function TIntegerList.SimpleText(const Fmt: string): string;
var
  I: Integer;
begin
  Result := '';
  for I := 0 to Count - 1 do
    Result := Result + System.SysUtils.Format('%s' + Fmt, [IfThen(I > 0, ' ', ''), Integers[I]]);
end;

function IntegerListCompareIntegers(List: TIntegerList; Index1, Index2: Integer): Integer;
begin
  Result := List.CompareIntegers(List.FList^[Index1].FInteger,
    List.FList^[Index2].FInteger, Index1, Index2);
end;

procedure TIntegerList.Sort;
begin
  CustomSort(IntegerListCompareIntegers);
end;

procedure TIntegerList.CustomSort(Compare: TIntegerListSortCompare);
begin
  if not Sorted and (FCount > 1) then
  begin
    QuickSort(0, FCount - 1, Compare);
  end;
end;

function TIntegerList.CompareIntegers(const I1, I2, Index1, Index2: Integer): Integer;
begin
  Result := I1 - I2;
end;

{ TInt64List }

{$REGION ' Pure Int64List Implementation '}

{$IFDEF PURE64LIST}

constructor TInt64List.Create(AOwnsObjects: Boolean = False);
begin
  inherited Create;
  FOwnsObjects := AOwnsObjects;
  FSorted := True;
end;

destructor TInt64List.Destroy;
begin
  inherited Destroy;
  FinalizeObjects(0, FCount);
  FCount := 0;
  SetCapacity(0);
end;

function TInt64List.Add(const I: Int64): Integer;
begin
  Result := AddObject(I, nil);
end;

function TInt64List.AddObject(const I: Int64; AObject: TObject): Integer;
begin
  if not Sorted then
    Result := FCount
  else
    if Find(I, Result) then
      case Duplicates of
        dupIgnore: Exit;
        dupError: raise EListError.Create('List does not allow duplicates');
      end;
  InsertItem(Result, I, AObject);
end;

function TInt64List.AddObject(const I: Int64; ATetraWord: Int64): Integer;
begin
  if not Sorted then
    Result := FCount
  else
    if Find(I, Result) then
      case Duplicates of
        dupIgnore: Exit;
        dupError: raise EListError.Create('List does not allow duplicates');
      end;
  InsertItem(Result, I, ATetraWord);
end;

procedure TInt64List.Assign(Source: TInt64List);
var
  I: Integer;
begin
  Clear;
  for I := 0 to Source.Count - 1 do
    Add(Source[I]);
end;

procedure TInt64List.Clear(AffectCapacity: Boolean = True);
begin
  if FCount <> 0 then
  begin
    // FinalizeObjects(0, FCount);
    SetCount(0);
    if AffectCapacity then
      SetCapacity(0);
  end;
end;

procedure TInt64List.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then
    Error(@SListIndexError, Index);
  FinalizeObjects(Index, Index + 1);
  Dec(FCount);
  if Index < FCount then
    System.Move(FList^[Index + 1], FList^[Index], (FCount - Index) * SizeOf(TInt64Item));
end;

class procedure TInt64List.Error(const Msg: string; Data: Integer);

  function ReturnAddr: Pointer;
  asm
          MOV     EAX,[EBP+4]
  end;

begin
  raise EListError.CreateFmt(Msg, [Data]) at ReturnAddr;
end;

class procedure TInt64List.Error(Msg: PResStringRec; Data: Integer);
begin
  TList.Error(LoadResString(Msg), Data);
end;

procedure TInt64List.Exchange(Index1, Index2: Integer);
begin
  if (Index1 < 0) or (Index1 >= FCount) then Error(@SListIndexError, Index1);
  if (Index2 < 0) or (Index2 >= FCount) then Error(@SListIndexError, Index2);
  ExchangeItems(Index1, Index2);
end;

procedure TInt64List.ExchangeItems(Index1, Index2: Integer);
var
  Temp: Int64;
  Item1, Item2: PInt64Item;
begin
  Item1 := @FList^[Index1];
  Item2 := @FList^[Index2];
  Temp := Item1^.FInt64;
  Item1^.FInt64 := Item2^.FInt64;
  Item2^.FInt64 := Temp;
{$IFDEF VER200}
  Temp := Integer(Item1^.FObject);
  Integer(Item1^.FObject) := Integer(Item2^.FObject);
  Integer(Item2^.FObject) := Temp;
{$ELSE}
{$IFDEF CPUX86}
  Temp := Integer(Item1^.FObject);
  Integer(Item1^.FObject) := Integer(Item2^.FObject);
  Integer(Item2^.FObject) := Temp;
{$ELSE}
  Temp := Int64(Item1^.FObject);
  Item1^.FObject := Item2^.FObject;
  Int64(Item2^.FObject) := Temp;
{$ENDIF}
{$ENDIF}
end;

procedure TInt64List.Extract(Index: Integer);
var
  SaveOwn: Boolean;
begin
  SaveOwn := FOwnsObjects;
  FOwnsObjects := True;
  try
    Delete(Index);
  finally
    FOwnsObjects := SaveOwn;
  end;
end;

procedure TInt64List.FinalizeObjects(Start, Stop: Integer);
begin
  if FOwnsObjects then
    while Start < Stop do
    begin
      FList^[Start].FObject.Free;
      Inc(Start);
    end;
end;

function TInt64List.Find(const N: Int64; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;

  if not Sorted then
  begin
    for I := 0 to Count - 1 do
      if FList^[I].FInt64 = N then
      begin
        Index := I;
        Result := True;
        Break;
      end;
  end else
  begin
    L := 0;
    H := FCount - 1;
    while L <= H do
    begin
      I := (L + H) shr 1;
      C := CompareInt64s(FList^[I].FInt64, N);
      if C < 0 then L := I + 1 else
      begin
        H := I - 1;
        if C = 0 then
        begin
          Result := True;
          if Duplicates <> dupAccept then L := I;
        end;
      end;
    end;
    Index := L;
  end;
end;

function TInt64List.Get(Index: Integer): Int64;
begin
  if (Index < 0) or (Index >= FCount) then
    Error(@SListIndexError, Index);
  Result := FList^[Index].FInt64;
end;

function TInt64List.GetCapacity: Integer;
begin
  Result := FCapacity;
end;

function TInt64List.GetCount: Integer;
begin
  Result := FCount;
end;

function TInt64List.GetIndexes(Value: Int64): Integer;
begin
  Result := IndexOf(Value);
end;

function TInt64List.GetObject(Index: Integer): TObject;
begin
  if (Index < 0) or (Index >= FCount) then
    Error(@SListIndexError, Index);
  Result := FList^[Index].FObject;
end;

function TInt64List.GetTetraWord(Index: Integer): Int64;
begin
  if FOwnsObjects then Error(SListOwnsObjects, Index);
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);

  Result := FList^[Index].FTetraWord;
end;

function TInt64List.GetText: string;
var
  I: Integer;
  N, Store: Int64;
  Cnt: Integer;
begin
  Result := '';
  if Sorted and (Duplicates = dupAccept) and (Count > 0) then
  begin
    N := Integers[0];
    Result := IntToStr(N);
    Store := N;
    Cnt := 1;
    for I := 1 to Count - 1 do
    begin
      N := Integers[I];
      if N = Store then
      begin
        Inc(Cnt);
      end else
      begin
        if Cnt > 1 then
          Result := Result + System.SysUtils.Format(' %d (%d)', [N, Cnt])
        else
          Result := Result + ' ' + IntToStr(N);
        Cnt := 1;
      end;
      Store := N;
    end;
    if Cnt > 1 then
      Result := Result + System.SysUtils.Format(' %d (%d)', [N, Cnt]);
  end else
    for I := 0 to Count - 1 do
      Result := Result + IfThen(I > 0, ' ', '') + IntToStr(Integers[I]);
end;

procedure TInt64List.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 64 then Delta := FCapacity div 4 else
    if FCapacity > 8 then Delta := 16 else
      Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

function TInt64List.IndexOf(const Value: Int64): Integer;
begin
  if not Sorted then
  begin
    for Result := 0 to GetCount - 1 do
      if CompareInt64s(Get(Result), Value) = 0 then Exit;
    Result := -1;
  end else
    if not Find(Value, Result) then Result := -1;
end;

procedure TInt64List.Insert(Index: Integer; const I: Int64);
begin
  InsertObject(Index, I, nil);
end;

procedure TInt64List.InsertItem(Index: Integer; const I: Int64; AObject: TObject);
begin
  InsertItemObj(Index, I, NativeInt(AObject));
end;

procedure TInt64List.InsertItem(Index: Integer; const I: Int64; ATetraWord: Int64);
begin
  InsertItemObj(Index, I, ATetraWord);
end;

procedure TInt64List.InsertItemObj(Index: Integer; const I: Int64; ATetraWord: Int64);
begin
  if FCount = FCapacity then Grow;
  if Index < FCount then
    System.Move(FList^[Index], FList^[Index + 1],
      (FCount - Index) * SizeOf(TInt64Item));
  with FList^[Index] do
  begin
    FInt64 := I;
    FTetraWord := ATetraWord;
  end;
  Inc(FCount);
end;

procedure TInt64List.InsertObject(Index: Integer; const I: Int64; AObject: TObject);
begin
  if Sorted then Error(@SSortedListError, 0);
  if (Index < 0) or (Index > FCount) then Error(@SListIndexError, Index);
  InsertItem(Index, I, AObject);
end;

procedure TInt64List.InsertObject(Index: Integer; const I: Int64; ATetraWord: Int64);
begin
  if FOwnsObjects then Error(SListOwnsObjects, Index);
  if Sorted then Error(@SSortedListError, 0);
  if (Index < 0) or (Index > FCount) then Error(@SListIndexError, Index);
  InsertItem(Index, I, ATetraWord);
end;

procedure TInt64List.Put(Index: Integer; const I: Int64);
begin
  if Sorted then Error(@SSortedListError, 0);
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);
  FList^[Index].FInt64 := I;
end;

procedure TInt64List.PutObject(Index: Integer; AObject: TObject);
begin
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);
  FList^[Index].FObject := AObject;
end;

procedure TInt64List.PutTetraWord(Index: Integer; const Value: Int64);
begin
  if FOwnsObjects then Error(SListOwnsObjects, Index);
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);
  FList^[Index].FTetraWord := Value;
end;

procedure TInt64List.QuickSort(L, R: Integer; ICompare: TInt64ListSortCompare);
var
  I, J, P: Integer;
begin
  repeat
    I := L;
    J := R;
    P := (L + R) shr 1;
    repeat
      while ICompare(Self, I, P) < 0 do Inc(I);
      while ICompare(Self, J, P) > 0 do Dec(J);
      if I <= J then
      begin
        ExchangeItems(I, J);
        if P = I then
          P := J
        else if P = J then
          P := I;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J, ICompare);
    L := I;
  until I >= R;
end;

procedure TInt64List.Reverse;
var
  Tmp: TInt64Item;
  B, E: Integer;
begin
  B := 0;
  E := Count - 1;
  while B < E do
  begin
    Tmp := FList^[B];
    FList^[B] := FLIst^[E];
    FList^[E] := Tmp;
    Inc(B);
    Dec(E);
  end;
end;

procedure TInt64List.SetCapacity(NewCapacity: Integer);
begin
  if NewCapacity < FCount then
    FinalizeObjects(NewCapacity, FCount);

  ReallocMem(FList, NewCapacity * SizeOf(TInt64Item));
  FCapacity := NewCapacity;

  if FCount > FCapacity then FCount := FCapacity; 
end;

procedure TInt64List.SetCount(NewCount: Integer);
var
  I: Integer;
begin
  if (NewCount < 0) or (NewCount > MaxIntListSize) then
    Error(@SListCountError, NewCount);
  if NewCount > FCapacity then
    SetCapacity(NewCount);
  if NewCount > FCount then
    FillChar(FList^[FCount], (NewCount - FCount) * SizeOf(TInt64Item), 0)
  else
    for I := FCount - 1 downto NewCount do
      Delete(I);
  FCount := NewCount;
end;

procedure TInt64List.SetSorted(Value: Boolean);
begin
  if FSorted <> Value then
  begin
    if Value then Sort;
    FSorted := Value;
  end;
end;

procedure TInt64List.SetText(const Value: string);
var
  Tkn: TStringTokenizer;
  I: Integer;
begin
  Tkn := TStringTokenizer.Create(' ');
  try
    Clear;
    Tkn.Str := Value;
    for I := 0 to Tkn.Count - 1 do
      Add(Tkn.Longs[I]);
  finally
    Tkn.Free;
  end;
end;

procedure TInt64List.SetUpdateState(Updating: Boolean);
begin
  // if Updating then Changing else Changed;
end;

function Int64ListCompareInt64s(List: TInt64List; Index1, Index2: Integer): Integer;
begin
  Result := List.CompareInt64s(List.FList^[Index1].FInt64,
    List.FList^[Index2].FInt64, Index1, Index2);
end;

procedure TInt64List.Sort;
begin
  CustomSort(Int64ListCompareInt64s);
end;

procedure TInt64List.CustomSort(Compare: TInt64ListSortCompare);
begin
  if not Sorted and (FCount > 1) then
  begin
    QuickSort(0, FCount - 1, Compare);
  end;
end;

function TInt64List.CompareInt64s(const I1, I2: Int64; const Index1, Index2: Integer): Integer;
var
  Res: Int64;
begin
  Res := I1 - I2;
  if Res < 0 then      Result := -1
  else if Res > 0 then Result := 1
                  else Result := 0;
end;

{$ENDREGION}

{$ELSE}

{$REGION ' DoubleInt Int64List Implementation '}

type
  TInt64Item = array[0..1] of TIntegerItem;

constructor TInt64List.Create(AOwnsObjects: Boolean = False);
begin
  inherited Create;
  FOwnsObjects := AOwnsObjects;
  FSorted := True;
end;

destructor TInt64List.Destroy;
begin
  inherited Destroy;
  FinalizeObjects(0, FCount);
  FCount := 0;
  SetCapacity(0);
end;

function TInt64List.Add(const I: Int64): Integer;
begin
  Result := AddObject(I, nil);
end;

function TInt64List.AddObject(const I: Int64; AObject: TObject): Integer;
begin
  if not Sorted then
    Result := FCount
  else
    if Find(I, Result) then
      case Duplicates of
        dupIgnore: Exit;
        dupError: raise EListError.Create('List does not allow duplicates');
      end;
  InsertItem(Result, I, AObject);
end;

procedure TInt64List.Clear;
begin
  if FCount <> 0 then
  begin
    FinalizeObjects(0, FCount);
    FCount := 0;
    SetCapacity(0);
  end;
end;

procedure TInt64List.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);
  FinalizeObjects(Index, Index + 1);
  Dec(FCount);
  if Index < FCount then
    System.Move(FList^[(Index + 1) * 2], FList^[Index * 2],
      (FCount - Index) * SizeOf(TIntegerItem) * 2);
end;

class procedure TInt64List.Error(const Msg: string; Data: Integer);

  function ReturnAddr: Pointer;
  asm
          MOV     EAX,[EBP+4]
  end;

begin
  raise EListError.CreateFmt(Msg, [Data]) at ReturnAddr;
end;

class procedure TInt64List.Error(Msg: PResStringRec; Data: Integer);
begin
  TList.Error(LoadResString(Msg), Data);
end;

procedure TInt64List.Exchange(Index1, Index2: Integer);
begin
  if (Index1 < 0) or (Index1 >= FCount) then Error(@SListIndexError, Index1);
  if (Index2 < 0) or (Index2 >= FCount) then Error(@SListIndexError, Index2);
  ExchangeItems(Index1, Index2);
end;

procedure TInt64List.ExchangeItems(Index1, Index2: Integer);
var
  Item1, Item2: TInt64Item;
begin
  Move(FList^[Index1 * 2], Item1, SizeOf(TInt64Item));
  Move(FList^[Index2 * 2], Item2, SizeOf(TInt64Item));

  Move(Item2, FList^[Index1 * 2], SizeOf(TInt64Item));
  Move(Item1, FList^[Index2 * 2], SizeOf(TInt64Item));
end;

procedure TInt64List.Extract(Index: Integer);
var
  SaveOwn: Boolean;
begin
  SaveOwn := FOwnsObjects;
  FOwnsObjects := True;
  try
    Delete(Index);
  finally
    FOwnsObjects := SaveOwn;
  end;
end;

procedure TInt64List.FinalizeObjects(Start, Stop: Integer);
begin
  if FOwnsObjects then
  begin
    Inc(Start, Start);              // Start *= 2
    Inc(Stop, Stop);                // Stop *= 2
    while Start < Stop do
    begin
      FList^[Start].FObject.Free;
      Inc(Start, 2);
    end;
  end;
end;

function TInt64List.Find(const N: Int64; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;

  if not Sorted then
  begin
    for I := 0 to Count - 1 do
      if FList^[I].FInt64 = N then
      begin
        Index := I;
        Result := True;
        Break;
      end;
  end else
  begin
    L := 0;
    H := FCount - 1;
    while L <= H do
    begin
      I := (L + H) shr 1;

      C := CompareInt64s(MakeInt64(FList^[I * 2].FInteger, FList^[I * 2 + 1].FInteger), N);
      if C < 0 then L := I + 1 else
      begin
        H := I - 1;
        if C = 0 then
        begin
          Result := True;
          if Duplicates <> dupAccept then L := I;
        end;
      end;
    end;
    Index := L;
  end;
end;

function TInt64List.Get(Index: Integer): Int64;
begin
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);
  Result := MakeInt64(FList^[Index * 2].FInteger, FList^[Index * 2 + 1].FInteger);
end;

function TInt64List.GetCapacity: Integer;
begin
  Result := FCapacity;
end;

function TInt64List.GetCount: Integer;
begin
  Result := FCount;
end;

function TInt64List.GetIndexes(Value: Int64): Integer;
begin
  Result := IndexOf(Value);
end;

function TInt64List.GetObject(Index: Integer): TObject;
begin
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);
  Result := FList^[Index * 2].FObject;
end;

function TInt64List.GetTetraWord(Index: Integer): Int64;
begin
  if FOwnsObjects then Error(SListOwnsObjects, Index);
  Int64Rec(Result).Lo := Cardinal(FList^[Index * 2].FObject);
  Int64Rec(Result).Hi := Cardinal(FList^[Index * 2 + 1].FObject);
end;

procedure TInt64List.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 64 then Delta := FCapacity div 4 else
    if FCapacity > 8 then Delta := 16 else
      Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

function TInt64List.IndexOf(const Value: Int64): Integer;
begin
  if not Sorted then
  begin
    for Result := 0 to GetCount - 1 do
      if CompareInt64s(Get(Result), Value) = 0 then Exit;
    Result := -1;
  end else
    if not Find(Value, Result) then Result := -1;
end;

procedure TInt64List.Insert(Index: Integer; const I: Int64);
begin
  InsertObject(Index, I, nil);
end;

procedure TInt64List.InsertItem(Index: Integer; const I: Int64; AObject: TObject);
begin
  InsertItemObj(Index, I, NativeInt(AObject));
end;

procedure TInt64List.InsertItem(Index: Integer; const I: Int64; ATetraWord: Int64);
begin
  InsertItemObj(Index, I, ATetraWord);
end;

procedure TInt64List.InsertItemObj(Index: Integer; const I: Int64; ATetraWord: Int64);
begin
  if FCount = FCapacity then Grow;
  if Index < FCount then
    System.Move(FList^[Index * 2], FList^[(Index + 1) * 2],
      (FCount - Index) * SizeOf(TIntegerItem) * 2);

  with FList^[Index * 2] do
  begin
    FObject := Pointer(Int64Rec(ATetraWord).Lo);
    FInteger := Integer(Int64Rec(I).Lo);
  end;

  with FList^[Index * 2 + 1] do
  begin
    FObject := Pointer(Int64Rec(ATetraWord).Hi);
    FInteger := Integer(Int64Rec(I).Hi);
  end;

  Inc(FCount);
end;

procedure TInt64List.InsertObject(Index: Integer; const I: Int64; AObject: TObject);
begin
  if Sorted then Error(@SSortedListError, 0);
  if (Index < 0) or (Index > FCount) then Error(@SListIndexError, Index);
  InsertItem(Index, I, AObject);
end;

procedure TInt64List.InsertObject(Index: Integer; const I: Int64; ATetraWord: Int64);
begin
  if FOwnsObjects then Error(SListOwnsObjects, Index);
  if Sorted then Error(@SSortedListError, 0);
  if (Index < 0) or (Index > FCount) then Error(@SListIndexError, Index);
  InsertItem(Index, I, ATetraWord);
end;

procedure TInt64List.Put(Index: Integer; const I: Int64);
begin
  if Sorted then Error(@SSortedListError, 0);
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);

  FList^[Index * 2].FInteger := Integer(Int64Rec(I).Lo);
  FList^[Index * 2 + 1].FInteger := Integer(Int64Rec(I).Hi);
end;

procedure TInt64List.PutObject(Index: Integer; AObject: TObject);
begin
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);
  FList^[Index * 2].FObject := AObject;
  FList^[Index * 2 + 1].FObject := nil;
end;

procedure TInt64List.PutTetraWord(Index: Integer; const Value: Int64);
begin
  if FOwnsObjects then Error(SListOwnsObjects, Index);
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);
  FList^[Index * 2].FObject := Pointer(Int64Rec(Value).Lo);
  FList^[Index * 2 + 1].FObject := Pointer(Int64Rec(Value).Hi);
end;

procedure TInt64List.QuickSort(L, R: Integer; ICompare: TInt64ListSortCompare);
var
  I, J, P: Integer;
begin
  repeat
    I := L;
    J := R;
    P := (L + R) shr 1;
    repeat
      while ICompare(Self, I, P) < 0 do Inc(I);
      while ICompare(Self, J, P) > 0 do Dec(J);
      if I <= J then
      begin
        ExchangeItems(I, J);
        if P = I then
          P := J
        else if P = J then
          P := I;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J, ICompare);
    L := I;
  until I >= R;
end;

procedure TInt64List.SetCapacity(NewCapacity: Integer);
begin
  if NewCapacity < FCount then
    FinalizeObjects(NewCapacity, FCount);

  ReallocMem(FList, NewCapacity * SizeOf(TIntegerItem) * 2);
  FCapacity := NewCapacity;

  if FCount > FCapacity then FCount := FCapacity; 
end;

procedure TInt64List.SetSorted(Value: Boolean);
begin
  if FSorted <> Value then
  begin
    if Value then Sort;
    FSorted := Value;
  end;
end;

procedure TInt64List.SetUpdateState(Updating: Boolean);
begin
  // if Updating then Changing else Changed;
end;

function Int64ListCompareInt64s(List: TInt64List; Index1, Index2: Integer): Integer;
var
  I1, I2: Int64;
begin
  Int64Rec(I1).Lo := Cardinal(List.FList^[Index1 * 2].FInteger);
  Int64Rec(I1).Hi := Cardinal(List.FList^[Index1 * 2 + 1].FInteger);

  Int64Rec(I2).Lo := Cardinal(List.FList^[Index2 * 2].FInteger);
  Int64Rec(I2).Hi := Cardinal(List.FList^[Index2 * 2 + 1].FInteger);

  Result := List.CompareInt64s(I1, I2, Index1, Index2);
end;

procedure TInt64List.Sort;
begin
  CustomSort(Int64ListCompareInt64s);
end;

procedure TInt64List.CustomSort(Compare: TInt64ListSortCompare);
begin
  if not Sorted and (FCount > 1) then
  begin
    QuickSort(0, FCount - 1, Compare);
  end;
end;

function TInt64List.CompareInt64s(const I1, I2: Int64; const Index1, Index2: Integer): Integer;
var
  Res: Int64;
begin
  Res := I1 - I2;
  if Res < 0 then      Result := -1
  else if Res > 0 then Result := 1
                  else Result := 0;
end;

{$ENDIF}

{$ENDREGION}

function MakeInt64(const ALo, AHi: Integer): Int64; inline;
begin
  Int64Rec(Result).Lo := Cardinal(ALo);
  Int64Rec(Result).Hi := Cardinal(AHi);
end;

end.
